// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// General purpose math routines

#pragma once

#include "constants.h"

#include <numeric>
#include <spdlog/spdlog.h>
#include <vector>

namespace spex {

// Compute the median of an unsorted array. Return either the middle
// element (array has odd number of values) or the average of the
// middle two (even number of values). Note that this function
// modifies the input array.
template <typename T>
[[nodiscard]] auto median(std::vector<T>& values) -> auto
{
    const int mid_idx { static_cast<int>(values.size()) / 2 };
    auto mid_it { values.begin() + mid_idx };
    std::nth_element(values.begin(), mid_it, values.end());
    auto median { *mid_it };
    // If With an even number of elements, mid_it corresponds to the
    // first value in the second half of the array.
    if (values.size() % 2 == 0) {
        const int prev_idx { mid_idx - 1 };
        auto prev_it { values.begin() + prev_idx };
        std::nth_element(values.begin(), prev_it, values.end());
        median = (median + *prev_it) / 2;
    }
    return median;
}

// Do a binary search to locate an index n such that x falls in the
// range list[n]...list[n+1].
[[nodiscard]] auto binaryFindIdx(const std::vector<double>& list,
                                 const double x) -> int;

// Compare array values against a threshold and set mask values to
// true for values that are outside the threshold. Default comparison
// function is 'less', meaning if data[i] < threshold then mask[i] =
// true. For testing the upper threshold call with the template
// argument std::greater instead. This function reports the number of
// *new* bad values. If a value was already masked bad it is not
// included in the count.
template <typename comparator = std::less<double>>
auto updateMask(const std::vector<double>& data,
                const double threshold,
                const std::string& info_str,
                std::vector<bool>& mask) -> void
{
    int n_bad {};
    for (int i {}; i < static_cast<int>(mask.size()); ++i) {
        if (!mask[i] && comparator {}(data[i], threshold)) {
            mask[i] = true;
            ++n_bad;
        }
    }
    if (n_bad > 0) {
        spdlog::info("{} {} {}", n_bad, info_str, threshold);
    }
}

// Shorthands for comparing if the data values are less than (lt) or
// greater than (gt) a threshold
auto updateMaskLt(const std::vector<double>& data,
                  const double threshold,
                  const std::string& info_str,
                  std::vector<bool>& mask) -> void;

auto updateMaskGt(const std::vector<double>& data,
                  const double threshold,
                  const std::string& info_str,
                  std::vector<bool>& mask) -> void;

template <typename T>
auto vectorNorm(const T& v) -> double
{
    return std::sqrt(std::accumulate(
      v.begin(), v.end(), 0.0, [](const auto& s, const auto& x) {
          return s + x * x;
      }));
}

auto normalizeVector(std::array<double, dims::vec>& v) -> void;

[[nodiscard]] auto dotProduct(const std::vector<double>::iterator& a,
                              const std::vector<double>::iterator& b) -> double;

[[nodiscard]] auto dotProduct(const std::array<double, dims::vec>& a,
                              const std::array<double, dims::vec>& b) -> double;

template <typename T>
auto crossProduct(const T a, const T b, std::array<double, dims::vec>& c)
  -> void
{
    c[0] = a[1] * b[2] - a[2] * b[1];
    c[1] = a[2] * b[0] - a[0] * b[2];
    c[2] = a[0] * b[1] - a[1] * b[0];
}

} // namespace spex
