// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "arb_special_functions.h"

#include <cassert>
#include <acb_hypgeom.h>
#include <complex>

namespace spex {

// We'll be working with double precision throughout
constexpr int arb_precision { 64 };

// Gamma function
[[nodiscard]] static auto gamma(const std::complex<double> x)
  -> std::complex<double>
{
    acb_t _x {};
    acb_init(_x);
    acb_set_d_d(_x, x.real(), x.imag());
    acb_hypgeom_gamma(_x, _x, arb_precision);
    return { arf_get_d(&_x->real.mid, ARF_RND_NEAR),
             arf_get_d(&_x->imag.mid, ARF_RND_NEAR) };
}

// Hypergeometric function 2F1(a, b, c, d)
[[nodiscard]] static auto hypgeom_2f1(const std::complex<double> a,
                                      const std::complex<double> b,
                                      const std::complex<double> c,
                                      const std::complex<double> d)
  -> std::complex<double>
{
    acb_t _a {};
    acb_t _b {};
    acb_t _c {};
    acb_t _d {};
    acb_init(_a);
    acb_init(_b);
    acb_init(_c);
    acb_init(_d);
    acb_set_d_d(_a, a.real(), a.imag());
    acb_set_d_d(_b, b.real(), b.imag());
    acb_set_d_d(_c, c.real(), c.imag());
    acb_set_d_d(_d, d.real(), d.imag());
    acb_t res_2f1 {};
    acb_init(res_2f1);
    acb_hypgeom_2f1(
      res_2f1, _a, _b, _c, _d, ACB_HYPGEOM_2F1_REGULARIZED, arb_precision);
    acb_t res_gamma {};
    acb_init(res_gamma);
    acb_hypgeom_gamma(res_gamma, _c, arb_precision);
    acb_mul(res_2f1, res_2f1, res_gamma, arb_precision);
    return { arf_get_d(&res_2f1->real.mid, ARF_RND_NEAR),
             arf_get_d(&res_2f1->imag.mid, ARF_RND_NEAR) };
}

// H(z) = 2F1(1, 2-2m, 2-m+i(1/2)s, 1/2+i(1/2)(z/a))
//        x exp(-s atan(z/a))(1+(a/z)^2)^(1-m)
[[nodiscard]] static auto H(const double z,
                            const double a,
                            const double m,
                            const double s) -> std::complex<double>
{
    const double za { z / a };
    return hypgeom_2f1({ 1.0, 0.0 },
                       { 2 - 2 * m, 0.0 },
                       { 2 - m, 0.5 * s },
                       { 0.5, 0.5 * za })
           * std::exp(-s * std::atan(za)) * std::pow(1 + za * za, 1 - m);
}

[[nodiscard]] auto pearsonIV(const double x,
                             const double a,
                             const double m,
                             const double s,
                             const double r) -> double
{
    assert(a > 0.0);
    assert(m > 0.5);
    assert(r > 0.0);
    const auto gamma1 { gamma({ m, 0.5 * s }) };
    const auto gamma2 { gamma({ 2 * m - 1, 0.0 }) };
    const auto gamma1_mod { std::abs(gamma1) };
    const auto term1 { std::pow(2.0, 2 * m - 3) * gamma1_mod * gamma1_mod
                       / (std::numbers::pi * r * std::abs(gamma2)) };
    const auto term2 { (H(x + r, a, m, s) - H(x - r, a, m, s))
                       / std::complex<double> { -s, 2 * (1 - m) } };
    return (term1 * term2).real();
}

} // namespace spex
