// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "binning_table.h"

#include <algorithm>
#include <iomanip>

namespace spex {

BinningTable::BinningTable(const int detector_n_rows,
                           const int detector_n_cols,
                           const std::string binning_table,
                           const int table_id,
                           const bool spectral_binning)
  : do_spectral_binning { spectral_binning }
{
    if (table_id == 0) {
        // Default binning table is the same size as the unbinned detector
        n_rows = detector_n_rows;
        n_cols = detector_n_cols;
        n_bins = n_rows * n_cols;
        bin_indices.resize(n_bins);
        std::iota(bin_indices.begin(), bin_indices.end(), 0);
        count_table.resize(n_bins, 1);
        lineskip_arr.resize(n_rows, 1);
    } else {
        netCDF::NcFile nc { binning_table, netCDF::NcFile::read };
        // Determine group name
        std::stringstream group_id {};
        group_id << std::setfill('0') << std::setw(3) << table_id;
        const std::string group_name { "Table_" + group_id.str() };
        // Get dimensions
        n_rows = static_cast<int>(nc.getDim("row").getSize());
        n_cols = static_cast<int>(nc.getDim("column").getSize());
        n_bins =
          static_cast<int>(nc.getGroup(group_name).getDim("bins").getSize());
        // Get variables
        bin_indices.resize(n_rows * n_cols);
        nc.getGroup(group_name)
          .getVar("binning_table")
          .getVar(bin_indices.data());
        count_table.resize(n_bins);
        nc.getGroup(group_name)
          .getVar("count_table")
          .getVar(count_table.data());
        lineskip_arr.resize(n_rows);
        nc.getGroup(group_name)
          .getVar("lineskip_arr")
          .getVar(lineskip_arr.data());
        // Linearly unbin the binning table
        const int linear_bin_r { detector_n_rows / n_rows };
        const int linear_bin_c { detector_n_cols / n_cols };
        std::vector<uint32_t> bin_indices_unbinned(detector_n_rows
                                                   * detector_n_cols);
        for (int i {}; i < n_rows; ++i) {
            for (int j {}; j < n_cols; ++j) {
                for (int m {}; m < linear_bin_r; ++m) {
                    for (int n {}; n < linear_bin_c; ++n) {
                        bin_indices_unbinned[(i * linear_bin_r + m)
                                               * detector_n_cols
                                             + j * linear_bin_c + n] =
                          bin_indices[i * n_cols + j];
                    }
                }
            }
        }
        bin_indices = std::move(bin_indices_unbinned);
        // Linearly unbin the lineskip array
        std::vector<uint8_t> lineskip_arr_unbinned(detector_n_cols, 0);
        for (int i {}; i < n_rows; ++i) {
            for (int k {}; k < linear_bin_r; ++k) {
                lineskip_arr_unbinned[i * linear_bin_r + k] = lineskip_arr[i];
            }
        }
        lineskip_arr = std::move(lineskip_arr_unbinned);
        // The count table already refers to binned data but after
        // linearly unbinning the counts must by updated
        for (auto& val : count_table) {
            val *= static_cast<uint16_t>(linear_bin_r * linear_bin_c);
        }
        // Finally, reset the binning table dimensions
        n_rows = detector_n_rows;
        n_cols = detector_n_cols;
    }
}

auto BinningTable::binCKD(CKD& ckd) -> void
{
    spdlog::info("Binning the CKD");
    // Pixel mask
    binPixelMask(ckd.pixel_mask);
    // Dark
    binDetector(ckd.dark.offset_short);
    binDetector(ckd.dark.offset_long);
    binDetector(ckd.dark.current);
    // Noise
    binDetector(ckd.noise.g);
    binDetector(ckd.noise.n);
    // Nonlinearity
    binDetector(ckd.nonlin_signal_scale);
    binDetector(ckd.nonlin_fit, ckd.dim_nonlin_spline);
    binDetector(ckd.nonlin.linear_spline.y,
                static_cast<int>(ckd.nonlin.linear_spline.y.size() / ckd.npix));
    binDetector(ckd.nonlin.control_points, ckd.nonlin.order);
    binDetector(ckd.nonlin.control_points_deriv, ckd.nonlin.order);
    // PRNU
    binDetector(ckd.prnu.prnu);
    // FOV
    transformFOVPixels(ckd);
    if (!do_spectral_binning) {
        return;
    }
    generateSpectralBinningTable(ckd);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        // Spectral
        if (!ckd.wave_full.empty()) {
            binSpectral(ckd, ckd.wave_full.at(i_pol));
        }
        // Radiometric
        if (!ckd.rad_spectra.empty()) {
            binSpectral(ckd, ckd.rad_spectra.at(i_pol));
        }
    }
    if (!ckd.wave_common.empty()) {
        binSpectral(ckd, ckd.wave_common);
    }
    // Polarimetric
    if (!ckd.pol.m_q.empty()) {
        binSpectral(ckd, ckd.pol.m_q);
        binSpectral(ckd, ckd.pol.m_u);
        binSpectral(ckd, ckd.pol.m_q_solar);
        binSpectral(ckd, ckd.pol.m_u_solar);
        binSpectral(ckd, ckd.pol.m_t);
    }
}

auto BinningTable::binPixelMask(std::vector<bool>& pixel_mask) -> void
{
    std::vector<bool> pixel_mask_binned(n_bins, false);
    for (int i {}; i < n_rows; ++i) {
        if (lineskip_arr[i] == 1) {
            for (int j {}; j < n_cols; ++j) {
                const int idx { i * n_cols + j };
                pixel_mask_binned[bin_indices[idx]] =
                  pixel_mask_binned[bin_indices[idx]] || pixel_mask[idx];
            }
        }
    }
    pixel_mask = std::move(pixel_mask_binned);
}

auto BinningTable::binDetector(std::vector<double>& data, const int batch_size)
  -> void
{
    // Skip if there is no data or the number of bins is the same as
    // the number of data points, i.e. a trivial binning table is in
    // use.
    if (data.empty() || static_cast<int>(data.size()) == n_bins * batch_size) {
        return;
    }
    std::vector<double> data_binned(n_bins * batch_size, 0.0);
    for (int i {}; i < n_rows; ++i) {
        if (lineskip_arr[i] == 1) {
            for (int j {}; j < n_cols; ++j) {
                const int idx { i * n_cols + j };
                for (int i_val {}; i_val < batch_size; ++i_val) {
                    data_binned[bin_indices[idx] * batch_size + i_val] +=
                      data[idx * batch_size + i_val];
                }
            }
        }
    }
    for (int i {}; i < n_bins; ++i) {
        for (int i_val {}; i_val < batch_size; ++i_val) {
            data_binned[i * batch_size + i_val] /= count_table[i];
        }
    }
    data = std::move(data_binned);
}

auto BinningTable::transformFOVPixels(CKD& ckd) -> void
{
    if (ckd.fov.pix_indices.front().empty()) {
        return;
    }
    // The FOV indices must be transformed from those corresponding to
    // an unbinned image to a binned image.
    std::vector<int> row_indices(dims::pol * ckd.fov.n_indices);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        for (int i_spec {}; i_spec < ckd.n_detector_cols; ++i_spec) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                for (int i_ind {}; i_ind < ckd.fov.n_indices; ++i_ind) {
                    row_indices[i_pol * ckd.fov.n_indices + i_ind] =
                      ckd.fov.pix_indices[i_pol][i_fov]
                                         [i_spec * ckd.fov.n_indices + i_ind]
                      / ckd.n_detector_cols;
                }
            }
            // If any of those is a bad value (index), all indices for
            // this spectral element (both polarization channels) are
            // marked bad. One might object to throwing out good
            // indices this way but i) it should be rare to begin with
            // and ii) it significantly simplifies the math and
            // implementation if all indices are treated the same way
            // at every step.
            if (std::ranges::any_of(row_indices, [this](const int val) {
                    return lineskip_arr[val] == 0;
                })) {
                for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                    for (int i_ind {}; i_ind < ckd.fov.n_indices; ++i_ind) {
                        ckd.fov
                          .pix_indices[i_pol][i_fov][i_spec * ckd.fov.n_indices
                                                     + i_ind] = def::idx;
                    }
                }
            } else {
                for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                    for (int i_ind {}; i_ind < ckd.fov.n_indices; ++i_ind) {
                        const int idx { i_spec * ckd.fov.n_indices + i_ind };
                        ckd.fov.pix_indices[i_pol][i_fov][idx] =
                          bin_indices[ckd.fov.pix_indices[i_pol][i_fov][idx]];
                    }
                }
            }
        }
    }
}

auto BinningTable::generateSpectralBinningTable(CKD& ckd) -> void
{
    if (ckd.fov.pix_indices.front().empty()) {
        return;
    }
    spectral_bin_indices.resize(ckd.fov.n_act,
                                std::vector<uint32_t>(ckd.n_detector_cols, 0));
    spectral_count_table.resize(ckd.fov.n_act, std::vector<uint16_t>(1, 1));
    // The size of a spectral bin is determined by the number of
    // indices that are the same within a segment of the FOV CKD. Here
    // we are iterating through several FOV CKD arrays simultaneously
    // (spectrum width in pixels times number of polarization
    // states). As long as any of them remain constant we continue to
    // expand the superpixel size for the given spectrum.
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        for (int i_spec { 1 }; i_spec < ckd.n_detector_cols; ++i_spec) {
            bool is_same { false };
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                const auto& indices { ckd.fov.pix_indices[i_pol][i_fov] };
                for (int i_ind {}; i_ind < ckd.fov.n_indices; ++i_ind) {
                    is_same =
                      is_same
                      || indices[i_spec * ckd.fov.n_indices + i_ind]
                           == indices[(i_spec - 1) * ckd.fov.n_indices + i_ind];
                }
            }
            if (is_same) {
                spectral_bin_indices[i_fov][i_spec] =
                  spectral_bin_indices[i_fov][i_spec - 1];
                ++spectral_count_table[i_fov].back();
            } else {
                spectral_bin_indices[i_fov][i_spec] =
                  spectral_bin_indices[i_fov][i_spec - 1] + 1;
                // The next count table value starts at 1 which is the
                // minimum superpixel size.
                spectral_count_table[i_fov].push_back(1);
            }
        }
    }
    // Size of each spectrum
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        ckd.fov.dims_spec[i_fov] = spectral_bin_indices[i_fov].back() + 1;
    }
}

auto BinningTable::binSpectralCore(const int i_fov,
                                   const std::vector<double>& data,
                                   std::vector<double>& data_binned) const
  -> void
{
    data_binned.resize(spectral_count_table[i_fov].size(), 0.0);
    for (int i {}; i < static_cast<int>(spectral_bin_indices[i_fov].size());
         ++i) {
        data_binned[spectral_bin_indices[i_fov][i]] += data[i];
    }
    for (int i {}; i < static_cast<int>(data_binned.size()); ++i) {
        data_binned[i] /= spectral_count_table[i_fov][i];
    }
}

auto BinningTable::binSpectralCore(const int i_fov,
                                   const std::vector<bool>& data,
                                   std::vector<bool>& data_binned) const -> void
{
    data_binned.resize(spectral_count_table[i_fov].size(), false);
    for (int i {}; i < static_cast<int>(spectral_bin_indices[i_fov].size());
         ++i) {
        data_binned[spectral_bin_indices[i_fov][i]] =
          data_binned[spectral_bin_indices[i_fov][i]] || data[i];
    }
}

auto BinningTable::binSpectral(const CKD& ckd,
                               std::vector<std::vector<double>>& data) -> void
{
    std::vector<std::vector<double>> data_binned(ckd.fov.n_act);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        binSpectralCore(i_fov, data[i_fov], data_binned[i_fov]);
    }
    data = std::move(data_binned);
}

auto BinningTable::binImage(const std::vector<double>& data_unbinned,
                            std::vector<double>& data_binned) const -> void
{
    std::ranges::fill(data_binned, 0.0);
    for (int i {}; i < static_cast<int>(data_unbinned.size()); ++i) {
        data_binned[bin_indices[i]] += data_unbinned[i];
    }
    for (int i {}; i < static_cast<int>(data_binned.size()); ++i) {
        data_binned[i] /= count_table[i];
    }
}

} // namespace spex
