// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for storing the binning table, other related quantities such
// as the lineskip array, and operations on data and the CKD using the
// binning table (e.g. binning and unbinning).

#pragma once

#include "ckd.h"

namespace spex {

class BinningTable
{
private:
    // Number of rows and columns of the binning table (dimensions of
    // bin_indices)
    int n_rows {};
    int n_cols {};
    // Number of pixels in binned data (dimension of count_table)
    int n_bins {};
    // bin_indices[i] is the index of the ith element of the unbinned
    // data. The data could represent, e.g., an image or part of the
    // CKD.
    std::vector<uint32_t> bin_indices {};
    // count_table[i] is the superpixel size of the ith element of
    // binned data
    std::vector<uint16_t> count_table {};
    // Value 0 means skip that line
    std::vector<uint8_t> lineskip_arr {};
    // spectral_bin_indices[i_fov][i] is the ith element of the binned
    // spectrum of the i_fov'th field of view. Each spectrum has a
    // separate 1D binning table.
    std::vector<std::vector<uint32_t>> spectral_bin_indices {};
    // Like count_table but used on binned spectra
    std::vector<std::vector<uint16_t>> spectral_count_table {};
    // If false, the detector CKDs are binned and the FOV indices are
    // transformed but the number of spectral elements extracted from
    // an image remains unaffected by the binning table. Normally we
    // want to do spectral binning.
    const bool do_spectral_binning {};

public:
    // Construct a binning table from the binning table name given by
    // user input and the binning table ID. The latter is stored in an
    // L1A file and corresponds to a group name in the binning table
    // file. For example, ID 24 means use the binning table from group
    // Table_024 of the binning table file.
    BinningTable(const int detector_n_rows,
                 const int detector_n_cols,
                 const std::string binning_table,
                 const int table_id,
                 const bool spectral_binning);
    // Size of binned data
    [[nodiscard]] auto nBins() const -> int { return n_bins; }
    // Return a binned index
    [[nodiscard]] auto binIndex(const int idx) const -> int
    {
        return static_cast<int>(bin_indices[idx]);
    }
    // Return a bin (superpixel) size
    [[nodiscard]] auto binSize(const int idx) const -> int
    {
        return static_cast<int>(count_table[idx]);
    }
    // Whether the row i_row should be skipped
    [[nodiscard]] auto skipRow(const int i_row) const -> bool
    {
        return lineskip_arr[i_row] == 0;
    }
    // Bin a detector image
    auto binImage(const std::vector<double>& data_unbinned,
                  std::vector<double>& data_binned) const -> void;
    // Bin a spectrum corresponding to FOV i_fov. This function
    // modifies 'data'. It is a wrapper around the core algorithm of
    // binSpectralCore.
    template <typename T>
    auto binSpectral(const int i_fov, std::vector<T>& data) const -> void
    {
        if (spectral_bin_indices.empty()) {
            return;
        }
        std::vector<T> data_binned {};
        binSpectralCore(i_fov, data, data_binned);
        data = std::move(data_binned);
    }
    // Bin everything that can be binned in the CKD (stray light
    // kernel is not binned)
    auto binCKD(CKD& ckd) -> void;
    // Unbin data, e.g. a binned detector image, using the binning
    // table. Output array is the same size as the binning table,
    // bin_indices.size().
    template <typename T>
    auto unbin(const std::vector<T>& data, std::vector<T>& data_unbinned) const
      -> void
    {
        for (int i {}; i < static_cast<int>(bin_indices.size()); ++i) {
            data_unbinned[i] = data[bin_indices[i]];
        }
        const int n_cols { static_cast<int>(bin_indices.size()
                                            / lineskip_arr.size()) };
        for (int i {}; i < static_cast<int>(lineskip_arr.size()); ++i) {
            if (lineskip_arr[i] == 0) {
                for (int j {}; j < n_cols; ++j) {
                    if constexpr (std::is_same_v<T, bool>) {
                        data_unbinned[i * n_cols + j] = false;
                    } else {
                        data_unbinned[i * n_cols + j] = 0.0;
                    }
                }
            }
        }
    }

private:
    // Bin the mask array
    auto binPixelMask(std::vector<bool>& mask) -> void;
    // Bin all detector steps, e.g. the PRNU values. batch_size is the
    // number of values associated with each pixel.
    auto binDetector(std::vector<double>& data, const int batch_size = 1)
      -> void;
    // Transform the FOV indices for up and down pixels according to
    // the binning table.
    auto transformFOVPixels(CKD& ckd) -> void;
    // Generate the spectral binning and count tables
    auto generateSpectralBinningTable(CKD& ckd) -> void;
    // Spectrally bin some quantity. We use separate algorithms for
    // binning numbers and booleans. i_fov is the index of the
    // extracted spectrum. The result data_binned should be allocated
    // and swapped with data (if that is desired) separately. See
    // binSpectral above for example.
    auto binSpectralCore(const int i_fov,
                         const std::vector<double>& data,
                         std::vector<double>& data_binned) const -> void;
    auto binSpectralCore(const int i_fov,
                         const std::vector<bool>& data,
                         std::vector<bool>& data_binned) const -> void;
    // Spectrally bin an array of arrays, for instance the radiometric
    // CKD where each FOV has its own set of calibration constants.
    auto binSpectral(const CKD& ckd, std::vector<std::vector<double>>& data)
      -> void;
};

} // namespace spex
