// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "bspline.h"

namespace spex {

BSpline::BSpline(const int order, const std::vector<double>& knots)
  : order { order }
{
    this->knots.resize(knots.size() + 2 * order);
    // Copy the B-spline knots and pad them from both ends by n
    // elements where n is the spline order.
    for (int i {}; i < static_cast<int>(knots.size()); ++i) {
        this->knots[i + order] = knots[i];
    }
    for (int i {}; i < order; ++i) {
        this->knots[i] = knots.front();
        this->knots[this->knots.size() - 1 - i] = knots.back();
    }
    control_points_tmp.resize(this->knots.size(), 0.0);
}

[[nodiscard]] auto BSpline::nStates() const -> int
{
    return static_cast<int>(knots.size()) - order - 1;
}

auto BSpline::findInterval(const double x) const -> int
{
    switch (knot_spacing) {
    case KnotSpacing::irregular:
    default:
        return binaryFindIdx(knots, x);
    }
}

auto BSpline::deBoor(const std::vector<double>& control_points,
                     const double x,
                     const int i_shift) const -> double
{
    const int p { findInterval(x) };
    if (p == def::idx) {
        return 0.0;
    }
    const auto control_points_begin { control_points.cbegin()
                                      + i_shift * nStates() };
    std::vector<double> d_cur { control_points_begin + p - order,
                                control_points_begin + p + 1 };
    for (int r { 1 }; r <= order; ++r) {
        for (int j { order }; j >= r; --j) {
            const double alpha { (x - knots[j + p - order])
                                 / (knots[j + 1 + p - r]
                                    - knots[j + p - order]) };
            d_cur[j] = (1 - alpha) * d_cur[j - 1] + alpha * d_cur[j];
        }
    }
    return d_cur[order];
}

auto BSpline::evalBasis(const double x,
                        const int i_row,
                        PackedMatrix& packed_matrix) -> void
{
    const int i_x { findInterval(x) };
    if (i_x == def::idx) {
        return;
    }
    std::vector<double> control_points(knots.size(), 0.0);
    std::vector<double> values(order + 1);
    for (int k {}; k <= order; ++k) {
        control_points[i_x - k] = 1.0;
        values[order - k] = deBoor(control_points, x);
        control_points[i_x - k] = 0.0;
    }
    packed_matrix.row_sizes[i_row] = order + 1;
    packed_matrix.global_col_indices[i_row] = i_x - order;
    packed_matrix.row_locations[i_row] =
      static_cast<int>(packed_matrix.data.size());
    for (auto& x : values) {
        packed_matrix.data.push_back(x);
    }
}

auto BSpline::evalBasis(const std::vector<double>& x_data,
                        std::vector<double>& B) -> void
{
    const int n_states { nStates() };
    B.resize(x_data.size() * n_states);
    for (int i_point {}; i_point < static_cast<int>(x_data.size());
         ++i_point) {
        const int i_x { findInterval(x_data[i_point]) };
        if (i_x == def::idx) {
            return;
        }
        for (int k {}; k <= order; ++k) {
            control_points_tmp[i_x - k] = 1.0;
            B[i_point * n_states + i_x - k] =
              deBoor(control_points_tmp, x_data[i_point]);
            control_points_tmp[i_x - k] = 0.0;
        }
    }
}

auto BSpline::genDerivControlPoints(
  const std::vector<double>& control_points,
  std::vector<double>& control_points_deriv) const -> void
{
    // The state vector (number of control points) of the
    // derivative B-spline is one less.
    const int deriv_n_states { nStates() - 1 };
    // If control_points contains several sets of control points,
    // iterate over all of them.
    for (int i_set {};
         i_set < static_cast<int>(control_points.size() / nStates());
         ++i_set) {
        const int shift { i_set * nStates() };
        const int shift_deriv { i_set * deriv_n_states };
        for (int i {}; i < deriv_n_states; ++i) {
            control_points_deriv[shift_deriv + i] =
              order
              * (control_points[shift + i + 1] - control_points[shift + i])
              / (knots[order + i + 1] - knots[i + 1]);
        }
    }
}

Bspline::Bspline( // {{{
    int a_order, // Order of the B-spline in Wikipedia perspective (n).
    int a_nknot, // Number of knot points.
    double const *a_knots // Knot points in ascending order.
)
{
    // 1. Copy the arguments.
    order = a_order;
    nknot = a_nknot;
    knots.resize(nknot);
    memcpy(knots.data(),a_knots,nknot*sizeof(double));

    error = false;
    // Assert that you at least have two knots.
    if (nknot < 2) {error = true; return;}

    // Assert that the knots are ascending.
    for (int iknot=1 ; iknot<nknot ; iknot++) {
        if (knots[iknot] < knots[iknot-1]) {error = true; return;}
    }
    nspline = nknot + order - 2;

    polyfactors.resize((nknot-1)*order*order);
    std::vector<double> other_polyfactors((nknot-1)*(order-1)*(order-1)); // Large enough.
    double *poly_new;
    double *poly_old;
    if ((order & 1) == 1) {
        poly_new = polyfactors.data();
        poly_old = other_polyfactors.data();
    } else {
        poly_old = polyfactors.data();
        poly_new = other_polyfactors.data();
    }
    for (int ipoly=0 ; ipoly<nknot-1 ; ipoly++) poly_new[ipoly] = 1.0;
    for (int iorder=1 ; iorder<order ; iorder++) {
        double *sw = poly_old;
        poly_old = poly_new;
        poly_new = sw;
        for (int ipoly=0 ; ipoly<(nknot-1)*(iorder+1)*(iorder+1) ; ipoly++) poly_new[ipoly] = 0.0;
        double *poly_new_run = poly_new;
        double *poly_old_run = poly_old;
        for (int iint=0 ; iint<nknot-1 ; iint++) {
            for (int ispline=0 ; ispline<iorder ; ispline++) {
                // Our interval starts at iint and ends at iint+1.
                // For iorder=1, that is the interval. For higher iorders, the leftmost
                // interval expands to the left, so the left knot of the ramp is.
                // iint+1-iorder and the right knot is iint+1.
                // But then, we move ispline to the right, so we are at iint+1-iorder+ispline
                // to iint+1+ispline.
                int ileft = iint+1+ispline<iorder?0:iint+1-iorder+ispline;
                int iright = nknot-1<iint+1+ispline?nknot-1:iint+1+ispline;
                double left_x = 1.0 / (knots[ileft]-knots[iright]);
                double right_x = -left_x;
                double left_c = knots[iright] * right_x;
                double right_c = knots[ileft] * left_x;
                for (int iterm=0 ; iterm<iorder ; iterm++) {
                    poly_new_run[iterm] += left_c * poly_old_run[iterm];
                    poly_new_run[iterm+1] += left_x * poly_old_run[iterm];
                }
                poly_new_run += iorder+1;
                for (int iterm=0 ; iterm<iorder ; iterm++) {
                    poly_new_run[iterm] += right_c * poly_old_run[iterm];
                    poly_new_run[iterm+1] += right_x * poly_old_run[iterm];
                }
                poly_old_run += iorder;
            }
            poly_new_run += iorder+1;
        }
    }

} // }}}

// Full Jacobian formula.
int Bspline::jaccalc( // {{{
    int nx, // Number of evaluation points.
    double *x, // Evaluation points, sorted in ascending order, or at least sorted in ascending knot intervals.
    int *ideriv, // Array that indicates for each point which derivative to take, zero is just the function itself.
    double* mat // Resulting Jacobian matrix (if valid pointer is given).
)
{
    for (int imat=0 ; imat<nx*nspline ; imat++) mat[imat] = 0.0;

    // Pre-calculate factorial terms.
    std::vector<double> facterm(order*order);
    double* fac_cur = facterm.data();
    std::vector<double> facs(order);
    facs[0] = 1.0;
    for (int ifac=1 ; ifac<order ; ifac++) facs[ifac] = ifac*facs[ifac-1];
    for (int iterm=0 ; iterm<order ; iterm++) {
        for (int ider=0 ; ider<=iterm ; ider++) {
            fac_cur[ider] = facs[iterm] / facs[iterm-ider];
        }
        fac_cur += order;
    }

    // Evaluate per interval, this assumes ascending x.
    int iint = 0;
    double *polyfactors_cur = polyfactors.data();
    for (int ix=0 ; ix<nx ; ix++) {
        while (x[ix] > knots[iint+1] && iint+2 != nknot) {
            iint++;
            mat += nx;
            polyfactors_cur += order*order;
        }
        while (x[ix] < knots[iint] && iint != 0) {
            iint--;
            mat -= nx;
            polyfactors_cur -= order*order;
        }
        // Evaluate x.
        for (int iterm=ideriv[ix] ; iterm<order ; iterm++) {
            double preterm = facterm[iterm*order+ideriv[ix]] * std::pow(x[ix],iterm-ideriv[ix]);
            for (int ispline=0 ; ispline<order ; ispline++) mat[ispline*nx+ix] += preterm * polyfactors_cur[ispline*order+iterm];
        }
    }

    return 0;

} // }}}

// Jacobian formula without ideriv.
int Bspline::jaccalc( // {{{
    int nx, // Number of evaluation points.
    double *x, // Evaluation points, sorted in ascending order, or at least sorted in ascending knot intervals.
    double *mat // Resulting Jacobian matrix (if valid pointer is given).
)
{
    for (int imat=0 ; imat<nx*nspline ; imat++) mat[imat] = 0.0;

    // Evaluate per interval, this assumes ascending x.
    int iint = 0;
    double *polyfactors_cur = polyfactors.data();
    for (int ix=0 ; ix<nx ; ix++) {
        while (iint+2 != nknot && x[ix] > knots[iint+1]) {
            iint++;
            mat += nx;
            polyfactors_cur += order*order;
        }
        while (iint != 0 && x[ix] < knots[iint]) {
            iint--;
            mat -= nx;
            polyfactors_cur -= order*order;
        }
        // Evaluate x.
        for (int iterm=0 ; iterm<order ; iterm++) {
            double preterm = pow(x[ix],iterm);
            for (int ispline=0 ; ispline<order ; ispline++) mat[ispline*nx+ix] += preterm * polyfactors_cur[ispline*order+iterm];
        }
    }

    return 0;

} // }}}

// Application of full Jacobian formula.
int Bspline::jacapply( // {{{
    int nx, // Number of evaluation points.
    double *x, // Evaluation points, sorted in ascending order, or at least sorted in ascending knot intervals.
    int *ideriv, // Array that indicates for each point which derivative to take, zero is just the function itself.
    double const* terms, // Terms with which to apply the Jacobian, saving the effort to save the Jacobian. You will get res = mat*terms then.
    double *res // Array where the result is written if you apply the Jacobian with res = mat*terms.
)
{
    for (int ix=0 ; ix<nx ; ix++) res[ix] = 0.0;

    // Matrix multiply the terms with the polyfactors over the spline dimension.
    std::vector<double> polyterms((nknot-1)*order);
    for (int imat=0 ; imat<(nknot-1)*order ; imat++) polyterms[imat] = 0.0;
    for (int iint=0 ; iint<nknot-1 ; iint++) {
        for (int ispline=0 ; ispline<order ; ispline++) {
            for (int iterm=0 ; iterm<order ; iterm++) {
                polyterms[iint*order+iterm] += terms[ispline+iint] * polyfactors[iint*order*order+ispline*order+iterm];
            }
        }
    }

    // Pre-calculate factorial terms.
    std::vector<double> facterm(order*order);
    double* fac_cur = facterm.data();
    std::vector<double> facs(order);
    facs[0] = 1.0;
    for (int ifac=1 ; ifac<order ; ifac++) facs[ifac] = ifac*facs[ifac-1];
    for (int iterm=0 ; iterm<order ; iterm++) {
        for (int ider=0 ; ider<=iterm ; ider++) {
            fac_cur[ider] = facs[iterm] / facs[iterm-ider];
        }
        fac_cur += order;
    }

    // Evaluate per interval, this assumes ascending x.
    double *polyterms_cur = polyterms.data();
    int iint = 0;
    for (int ix=0 ; ix<nx ; ix++) {
        while (x[ix] > knots[iint+1] && iint+2 != nknot) {
            iint++;
            polyterms_cur += order;
        }
        while (x[ix] < knots[iint] && iint != 0) {
            iint--;
            polyterms_cur -= order;
        }
        // Evaluate x.
        for (int iterm=ideriv[ix] ; iterm<order ; iterm++) {
            res[ix] += facterm[iterm*order+ideriv[ix]] * polyterms_cur[iterm] * pow(x[ix],iterm-ideriv[ix]);
        }
    }

    return 0;

} // }}}

// Application without ideriv.
int Bspline::jacapply( // {{{
    int nx, // Number of evaluation points.
    double *x, // Evaluation points, sorted in ascending order, or at least sorted in ascending knot intervals.
    double *terms, // Terms with which to apply the Jacobian, saving the effort to save the Jacobian. You will get res = mat*terms then.
    double *res // Array where the result is written if you apply the Jacobian with res = mat*terms.
)
{
    for (int ix=0 ; ix<nx ; ix++) res[ix] = 0.0;

    // Matrix multiply the terms with the polyfactors over the spline dimension.
    std::vector<double> polyterms((nknot-1)*order);
    for (int imat=0 ; imat<(nknot-1)*order ; imat++) polyterms[imat] = 0.0;
    for (int iint=0 ; iint<nknot-1 ; iint++) {
        for (int ispline=0 ; ispline<order ; ispline++) {
            for (int iterm=0 ; iterm<order ; iterm++) {
                polyterms[iint*order+iterm] += terms[ispline+iint] * polyfactors[iint*order*order+ispline*order+iterm];
            }
        }
    }

    // Evaluate per interval, this assumes ascending x.
    int iint = 0;
    double *polyterms_cur = polyterms.data();
    for (int ix=0 ; ix<nx ; ix++) {
        while (x[ix] > knots[iint+1] && iint+2 != nknot) {
            iint++;
            polyterms_cur += order;
        }
        while (x[ix] < knots[iint] && iint != 0) {
            iint--;
            polyterms_cur -= order;
        }
        // Evaluate x.
        for (int iterm=0 ; iterm<order ; iterm++) {
            res[ix] += polyterms_cur[iterm] * pow(x[ix],iterm);
        }
    }

    return 0;

} // }}}

} // namespace spex
