// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "algorithm.h"
#include "packed_matrix.h"

namespace spex {

class BSpline
{
private:
    enum class KnotSpacing
    {
        // Knots are not systematically spaced
        irregular,
        // Knots are equidistantly spaced
        equal,
    };
    KnotSpacing knot_spacing { KnotSpacing::irregular };
    std::vector<double> knots {};
    // B-spline order such that n means nth order polynomial
    const int order {};
    // Work array for controlling which splines are evaluated. This is
    // used in the construction of the B-spline matrix.
    std::vector<double> control_points_tmp {};

public:
    BSpline(const int order, const std::vector<double>& knots);
    // Return the size of the state vector for linear inversion. It is
    // (order+1) less than the number of knots because the endpoint
    // and the added knots should not be included.
    [[nodiscard]] auto nStates() const -> int;
    // Return index i such that x is in knots[i]..knots[i+1]
    auto findInterval(const double x) const -> int;
    // Return the B-spline function using the de Boor algorithm for a
    // given set of control points and argument x. If control_points
    // contains more than one set of control points, e.g. for all
    // pixels, then i_shift specifies which set to use. The absolute
    // shift is computed inside the function and the user only needs
    // to specify the set number, for i_shift = 0 for pixel 0, 1 for
    // pixel 1, etc.
    auto deBoor(const std::vector<double>& control_points,
                const double x,
                const int i_shift = 0) const -> double;
    // Generate B-spline matrix elements for the point x and store
    // them in row i_row in the packed format.
    auto evalBasis(const double x,
                   const int i_row,
                   PackedMatrix& packed_matrix) -> void;
    // Same but the result is stored in a normal matrix.
    auto evalBasis(const std::vector<double>& x_data, std::vector<double>& B)
      -> void;
    // Given a set of control points, generate the control points for
    // the B-spline derivative. The derivative array needs to be
    // allocated outside.
    auto genDerivControlPoints(const std::vector<double>& control_points,
                               std::vector<double>& control_points_deriv) const
      -> void;
    ~BSpline() = default;
};

// Old implementation, removed soon
class Bspline {

    private:
    // Passed in constructor.
    int order {}; // Order in Wikipedia perspective (n), so n=1 is order-zero polynomial.
    int nknot {}; // Number of knots.
    std::vector<double> knots {}; // Position of the knots in ascending order.

    bool error = true; // Flag for proper construction.

    public:
    std::vector<double> polyfactors {}; // Pre-calculated prefactors for the Jacobian.

    public:
    Bspline(int a_order, int a_nknot, double const *knots);
    ~Bspline() = default;
    int nspline; // Number of B-splines (N+n-2). This is the number of state parameters if you do an inversion.
    // Routines to directly apply a Jacobian. This does not return the Jacobian
    // matrix.
    int jacapply(int nx, double *x, double *terms, double *res); // Without ideriv, so ideriv is considered 0 everywhere.
    int jacapply(int nx, double *x, int *ideriv, double const* terms, double* res); // Full implementation.
    // Routines for returning the Jacobian, for instance, for inversion.
    int jaccalc(int nx, double *x, double *mat); // Without ideriv, so ideriv is considered 0 everywhere.
    int jaccalc(int nx, double *x, int *ideriv, double* mat); // Full implementation.

};

} // namespace spex
