// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "ckd.h"

#include "io.h"
#include "nc_object.h"
#include "spex_settings.h"

namespace spex {

// Retrieve values from a NetCDF variable into a buffer and reshape
// for destination.
static auto getAndReshape(const netCDF::NcVar& nc_var,
                          std::vector<std::vector<double>>& dest) -> void
{
    std::vector<double> buffer(dest.size() * dest.front().size());
    nc_var.getVar(buffer.data());
    for (int i {}; i < static_cast<int>(dest.size()); ++i) {
        for (int j {}; j < static_cast<int>(dest.at(i).size()); ++j) {
            dest[i][j] = buffer[i * dest.front().size() + j];
        }
    }
}

// Similar but for a 3-dimensional array
static auto getAndReshape(
  const netCDF::NcVar& nc_var,
  std::array<std::vector<std::vector<double>>, dims::pol>& dest) -> void
{
    const std::array<int, 3> dims {
        static_cast<int>(dest.size()),
        static_cast<int>(dest.front().size()),
        static_cast<int>(dest.front().front().size()),
    };
    // Sometimes the first dimension is dims::pol in the destination
    // array but second dimension in the NetCDF variable.
    const bool flip_dim { static_cast<int>(nc_var.getDim(0).getSize())
                          != dims[0] };
    std::vector<double> buffer(
      std::accumulate(dims.cbegin(), dims.cend(), 1, std::multiplies<>()));
    nc_var.getVar(buffer.data());
    // Later idx + k is the full index of the source array
    int idx {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int j {}; j < dims[1]; ++j) {
            if (flip_dim) {
                idx = (j * dims::pol + i_pol) * dims[2];
            } else {
                idx = (i_pol * dims[1] + j) * dims[2];
            }
            for (int k {}; k < dims[2]; ++k) {
                dest[i_pol][j][k] = buffer[idx + k];
            }
        }
    }
}

auto CKD::readNonlinDebugData(const netCDF::NcGroup& grp) -> void
{
    const int npix_d { static_cast<int>(grp.getDim("debug_pixels").getSize()) };
    const int n_signal_rates { static_cast<int>(
      grp.getDim("signal_rates").getSize()) };
    const int n_S_max { static_cast<int>(grp.getDim("S_max").getSize()) };
    nonlin.debug_pixels.resize(npix_d);
    grp.getVar("debug_pixels").getVar(nonlin.debug_pixels.data());
    nonlin.f1.resize(npix);
    nonlin.f2.resize(npix);
    nonlin.m0.resize(npix);
    nonlin.S_max_data.resize(n_S_max);
    nonlin.S_max_fit.resize(n_S_max);
    nonlin.I_c_med.resize(n_S_max);
    if (nonlin.model == NonlinModel::sigmoidal) {
        nonlin.r.resize(nonlin_n_r, std::vector<double>(npix));
        nonlin.c.resize(npix);
    }
    grp.getVar("f1").getVar(nonlin.f1.data());
    grp.getVar("f2").getVar(nonlin.f2.data());
    grp.getVar("m0").getVar(nonlin.m0.data());
    if (nonlin.model == NonlinModel::sigmoidal) {
        grp.getVar("r0").getVar(nonlin.r[0].data());
        grp.getVar("r1").getVar(nonlin.r[1].data());
        grp.getVar("r2").getVar(nonlin.r[2].data());
        grp.getVar("r3").getVar(nonlin.r[3].data());
        grp.getVar("r4").getVar(nonlin.r[4].data());
        grp.getVar("c").getVar(nonlin.c.data());
    }
    grp.getVar("m1").getVar(&nonlin.m1);
    grp.getVar("m2").getVar(&nonlin.m2);
    grp.getVar("S_max_data").getVar(nonlin.S_max_data.data());
    grp.getVar("S_max_fit").getVar(nonlin.S_max_fit.data());
    grp.getVar("I_c_med").getVar(nonlin.I_c_med.data());
    if (npix_d > 0) {
        const int model_fit_points { static_cast<int>(
          grp.getDim("model_fit_points").getSize()) };
        nonlin.signal_rates.resize(npix_d, std::vector<double>(n_signal_rates));
        nonlin.reference_rates.resize(npix_d,
                                      std::vector<double>(n_signal_rates));
        nonlin.S_lin.resize(npix_d, std::vector<double>(model_fit_points));
        nonlin.S_meas.resize(npix_d, std::vector<double>(model_fit_points));
        nonlin.nonlinearity_fits.resize(npix_d,
                                        std::vector<double>(model_fit_points));
        getAndReshape(grp.getVar("signal_rates"), nonlin.signal_rates);
        getAndReshape(grp.getVar("reference_rates"), nonlin.reference_rates);
        getAndReshape(grp.getVar("S_lin"), nonlin.S_lin);
        getAndReshape(grp.getVar("S_meas"), nonlin.S_meas);
        getAndReshape(grp.getVar("nonlinearity_fit"), nonlin.nonlinearity_fits);
    }
}

auto CKD::readCKD(
  const std::string& ckd_file_name,
  const ProcLevel target_level,
  std::array<bool, static_cast<int>(ProcLevel::n_levels)>& debug,
  std::array<bool, static_cast<int>(ProcLevel::n_levels)>& skip,
  const std::array<std::string, static_cast<int>(ProcLevel::n_levels)>&
    alternative_ckd_files,
  const double fov_spectrum_width) -> void
{
    const NcObject nc { netCDF::NcFile::read, ckd_file_name };
    n_detector_rows =
      static_cast<int>(nc.getDim("spatial_detector_pixels").getSize());
    n_detector_cols =
      static_cast<int>(nc.getDim("spectral_detector_pixels").getSize());
    n_vp = static_cast<int>(nc.getDim("viewports").getSize());
    // Number of pixels is the product of length and width. This is
    // unbinned CKD.
    npix = n_detector_cols * n_detector_rows;
    // Print the basic dimensions which are fixed from now on
    spdlog::info("Using an input CKD: {}", ckd_file_name);
    spdlog::info("Detector dimension  (spatial): {}", n_detector_rows);
    spdlog::info("Detector dimension (spectral): {}", n_detector_cols);
    spdlog::info("Number of viewports: {}", n_vp);
    // Read the pixel mask which is always there, possibly not yet in
    // its final state. It may be updated by one or more detector
    // calibration steps. When reading into shared arrays make sure to
    // call the root process only.
    std::vector<uint8_t> pixel_mask_u8(npix);
    nc.getVar("mask").getVar(pixel_mask_u8.data());
    pixel_mask.resize(npix);
    for (int i {}; i < npix; ++i) {
        pixel_mask[i] = static_cast<bool>(pixel_mask_u8[i]);
    }
    // Only attempt to read data if the target level is equal or
    // higher than the process. The reusable variable grp will then
    // hold a group read from the input CKD file. If the group is not
    // present in the input CKD, grp is a null group.
    netCDF::NcGroup grp {};
    // Similar to grp, grp_d will point to the debug group of a given CKD step
    netCDF::NcGroup grp_d {};
    // For experimental purposes, one may provide an alternative CKD
    // file for reading the that section of the CKD. If so then things
    // like the detector dimensions have to be consistent between the
    // two CKDs. No check are made for that.
    NcObject alternative_nc { netCDF::NcFile::read };
    // Function to test at each step whether the group is present and
    // should be read from the input CKD. Additionally, if the step is
    // considered active, the debug information might or might not be
    // read depending on the debug flag and whether the relevant data
    // is present in the input CKD.
    const auto activateStep { [this,
                               target_level,
                               &skip,
                               &debug,
                               &nc,
                               &grp,
                               &grp_d,
                               &alternative_ckd_files,
                               &alternative_nc](const ProcLevel proc_level,
                                                const std::string& group_name) {
        const int i_proc_level { static_cast<int>(proc_level) };
        if (alternative_ckd_files.at(i_proc_level).empty()) {
            grp = nc.getGroup(group_name);
        } else {
            alternative_nc.open(alternative_ckd_files.at(i_proc_level));
            grp = alternative_nc.getGroup(group_name);
        }
        grp_d = {};
        // Active means that data for a given CKD step have been
        // previously generated and will be used to calibrate
        // images during this processor run.
        const bool active { target_level >= proc_level && !grp.isNull() };
        if (active) {
            level = proc_level;
            skip.at(i_proc_level) = false;
            if (debug.at(i_proc_level)) {
                if (alternative_ckd_files.at(i_proc_level).empty()) {
                    grp_d = nc.getGroup("DEBUG/" + group_name);
                } else {
                    grp_d = alternative_nc.getGroup("DEBUG/" + group_name);
                }
            }
            if (grp_d.isNull()) {
                debug.at(i_proc_level) = false;
            }
        }
        return active;
    } };
    if (activateStep(ProcLevel::dark, "DARK")) {
        spdlog::info("Reading the dark CKD");
        dark.offset_short.resize(npix);
        dark.offset_long.resize(npix);
        dark.current.resize(npix);
        grp.getVar("offset_short").getVar(dark.offset_short.data());
        grp.getVar("offset_long").getVar(dark.offset_long.data());
        grp.getVar("dark_current").getVar(dark.current.data());
        if (!grp_d.isNull()) {
            dark.chi2.resize(npix);
            dark.sigma_beta.resize(npix);
            dark.sigma_alpha.resize(npix);
            dark.last_exposure_times.resize(npix);
            grp_d.getVar("chi2").getVar(dark.chi2.data());
            grp_d.getVar("sigma_beta").getVar(dark.sigma_beta.data());
            grp_d.getVar("sigma_alpha").getVar(dark.sigma_alpha.data());
            grp_d.getVar("last_exposure_times")
              .getVar(dark.last_exposure_times.data());
        }
    }
    if (activateStep(ProcLevel::noise, "NOISE")) {
        spdlog::info("Reading the noise CKD");
        noise.g.resize(npix);
        noise.n.resize(npix);
        grp.getVar("g").getVar(noise.g.data());
        grp.getVar("n").getVar(noise.n.data());
        if (!grp_d.isNull()) {
            noise.std_dev_g.resize(npix);
            noise.std_dev_n.resize(npix);
            grp_d.getVar("std_dev_g").getVar(noise.std_dev_g.data());
            grp_d.getVar("std_dev_n").getVar(noise.std_dev_n.data());
        }
    }
    if (activateStep(ProcLevel::nonlin, "NON_LINEARITY")) {
        spdlog::info("Reading the nonlinearity CKD");
        uint8_t model_u8 {};
        grp.getVar("model").getVar(&model_u8);
        nonlin.model = static_cast<NonlinModel>(model_u8);
        if (nonlin.model == NonlinModel::bspline) {
            const int n_knots { static_cast<int>(
              grp.getDim("knots").getSize()) };
            grp.getVar("order").getVar(&nonlin.order);
            nonlin.b_spline_knots.resize(n_knots);
            grp.getVar("knots").getVar(nonlin.b_spline_knots.data());
            nonlin.b_spline =
              std::make_unique<BSpline>(nonlin.order, nonlin.b_spline_knots);
            nonlin.b_spline_deriv = std::make_unique<BSpline>(
              nonlin.order - 1, nonlin.b_spline_knots);
            nonlin.control_points.resize(npix * nonlin.b_spline->nStates());
            nonlin.control_points_deriv.resize(
              npix * nonlin.b_spline_deriv->nStates());
            grp.getVar("control_points").getVar(nonlin.control_points.data());
            nonlin.b_spline->genDerivControlPoints(nonlin.control_points,
                                                   nonlin.control_points_deriv);
            if (!grp_d.isNull()) {
                readNonlinDebugData(grp_d);
            }
        } else if (nonlin.model == NonlinModel::sigmoidal) {
            const int n_knots { static_cast<int>(
              grp.getDim("knots").getSize()) };
            nonlin.linear_spline = { npix, n_knots };
            grp.getVar("knots").getVar(nonlin.linear_spline.knots.data());
            grp.getVar("y").getVar(nonlin.linear_spline.y.data());
            nonlin.linear_spline.setLookupIdxScale();
            if (!grp_d.isNull()) {
                readNonlinDebugData(grp_d);
            }
        } else {
            dim_nonlin_spline = static_cast<int>(
              grp.getDim("nonlin_number_of_coefficients").getSize());
            dim_nonlin_knot =
              static_cast<int>(grp.getDim("nonlin_number_of_knots").getSize());
            nonlin.b_spline_knots.resize(dim_nonlin_knot);
            nonlin_fit.resize(npix * dim_nonlin_spline);
            nonlin_signal_scale.resize(npix);
            int intorder {}; // For integer type conversion
            grp.getVar("nonlin_order").getVar(&intorder);
            nonlin.order = intorder;
            grp.getVar("nonlin_knots").getVar(nonlin.b_spline_knots.data());
            grp.getVar("nonlin_signal_scale")
              .getVar(nonlin_signal_scale.data());
            grp.getVar("nonlin_fit").getVar(nonlin_fit.data());
        }
    }
    if (activateStep(ProcLevel::prnu, "PRNU")) {
        spdlog::info("Reading the PRNU CKD");
        prnu.prnu.resize(npix);
        grp.getVar("prnu").getVar(prnu.prnu.data());
        if (!grp_d.isNull()) {
            prnu.sensitivity_fit.resize(npix);
            grp_d.getVar("sensitivity").getVar(prnu.sensitivity_fit.data());
        }
    }
    if (activateStep(ProcLevel::stray, "STRAYLIGHT")) {
        spdlog::info("Reading stray light CKD");
        stray.n_kernels = static_cast<int>(grp.getDim("kernels").getSize());
        uint8_t reduced_kernels_u8 {};
        grp.getVar("reduced_kernels").getVar(&reduced_kernels_u8);
        stray.reduced_kernels = static_cast<bool>(reduced_kernels_u8);
        stray.n_spatial = static_cast<int>(grp.getDim("spatial").getSize());
        stray.n_spectral = static_cast<int>(grp.getDim("spectral").getSize());
        stray.kernel_rows.resize(stray.n_kernels);
        grp.getVar("kernel_rows").getVar(stray.kernel_rows.data());
        stray.kernel_cols.resize(stray.n_kernels);
        grp.getVar("kernel_cols").getVar(stray.kernel_cols.data());
        stray.kernel_fft_sizes.resize(stray.n_kernels);
        grp.getVar("kernel_fft_sizes").getVar(stray.kernel_fft_sizes.data());
        stray.kernels_fft.resize(stray.n_kernels);
        for (int i {}; i < stray.n_kernels; ++i) {
            stray.kernels_fft[i].resize(stray.kernel_fft_sizes[i]);
        }
        std::vector<size_t> start { 0 };
        for (int i_kernel {}; i_kernel < stray.n_kernels; ++i_kernel) {
            // Need to multiply by 2 because of complex numbers
            std::vector<size_t> count {
                2 * static_cast<size_t>(stray.kernel_fft_sizes.at(i_kernel))
            };
            grp.getVar("kernels_fft")
              .getVar(start, count, stray.kernels_fft.at(i_kernel).data());
            start.front() += count.front();
        }
        stray.eta.resize(stray.n_spatial * stray.n_spectral);
        grp.getVar("eta").getVar(stray.eta.data());
        stray.weights.resize(
          stray.n_kernels,
          std::vector<double>(stray.n_spatial * stray.n_spectral));
        getAndReshape(grp.getVar("weights"), stray.weights);
        stray.edges.resize(4 * stray.n_kernels);
        grp.getVar("edges").getVar(stray.edges.data());
        if (!grp_d.isNull()) {
            stray.kernel_n_rows_full =
              static_cast<int>(grp_d.getDim("kernel_rows").getSize());
            stray.kernel_n_cols_full =
              static_cast<int>(grp_d.getDim("kernel_cols").getSize());
            stray.kernels_real.resize(
              stray.n_kernels,
              std::vector<double>(stray.kernel_n_rows_full
                                  * stray.kernel_n_cols_full));
            getAndReshape(grp_d.getVar("kernels_real"), stray.kernels_real);
            stray.all_kernels.resize(npix);
            grp_d.getVar("all_kernels").getVar(stray.all_kernels.data());
            stray.etas.resize(
              stray.n_kernels,
              std::vector<double>(stray.n_spatial * stray.n_spectral));
            getAndReshape(grp_d.getVar("etas"), stray.etas);
        }
    }
    if (activateStep(ProcLevel::fov, "FIELD_OF_VIEW")) {
        spdlog::info("Reading the FOV CKD");
        if (alternative_ckd_files.at(static_cast<int>(ProcLevel::fov))
              .empty()) {
            fov.n_act = static_cast<int>(
              nc.getDim("spatial_samples_per_image").getSize());
        } else {
            fov.n_act = static_cast<int>(
              alternative_nc.getDim("spatial_samples_per_image").getSize());
        }
        fov.n_act_vp.resize(n_vp);
        fov.start_vp.resize(n_vp);
        fov.act_angles.resize(fov.n_act);
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            fov.row_indices[i_pol].resize(fov.n_act,
                                          std::vector<double>(n_detector_cols));
        }
        fov.dims_spec.resize(fov.n_act, n_detector_cols);
        grp.getVar("fov_nfov_vp").getVar(fov.n_act_vp.data());
        grp.getVar("fov_ifov_start_vp").getVar(fov.start_vp.data());
        grp.getVar("fov_act_angles").getVar(fov.act_angles.data());
        getAndReshape(grp.getVar("fov_ispat"), fov.row_indices);
        if (!grp_d.isNull()) {
            fov.n_l1a_total =
              static_cast<int>(grp_d.getDim("measurements").getSize());
            fov.n_columns =
              static_cast<int>(grp_d.getDim("peak_positions").getSize());
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                fov.peak_pos.at(i_pol).resize(
                  fov.n_l1a_total, std::vector<double>(fov.n_columns));
                fov.peak_pos_b.at(i_pol).resize(
                  fov.n_l1a_total, std::vector<double>(n_detector_cols));
            }
            getAndReshape(grp_d.getVar("peak_positions"), fov.peak_pos);
            getAndReshape(grp_d.getVar("peak_positions_b"), fov.peak_pos_b);
        }
        genFovIndices(fov_spectrum_width);
    }
    if (activateStep(ProcLevel::swath, "SWATH")) {
        spdlog::info("Reading swath vectors");
        swath.pointings.resize(fov.n_act * dims::vec);
        grp.getVar("swath_swathvectors").getVar(swath.pointings.data());
        swath.normals.resize(n_vp * dims::vec);
        grp.getVar("swath_vectorplane_normals").getVar(swath.normals.data());
    }
    if (activateStep(ProcLevel::wave, "WAVELENGTH")) {
        spdlog::info("Reading the wavelength CKD");
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            wave_full.at(i_pol).resize(fov.n_act,
                                       std::vector<double>(n_detector_cols));
        }
        wave_common.resize(fov.n_act, std::vector<double>(n_detector_cols));
        std::vector<double> buffer(fov.n_act * dims::pol * n_detector_cols);
        grp.getVar("wave_full").getVar(buffer.data());
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            for (int i_act {}; i_act < fov.n_act; ++i_act) {
                for (int i {}; i < n_detector_cols; ++i) {
                    wave_full[i_pol][i_act][i] =
                      buffer[(i_pol * fov.n_act + i_act) * n_detector_cols + i];
                }
            }
        }
        getAndReshape(grp.getVar("wave_common"), wave_common);
    }
    if (activateStep(ProcLevel::rad, "RADIOMETRIC")) {
        spdlog::info("Reading the radiometric CKD");
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            rad_spectra.at(i_pol).resize(fov.n_act,
                                         std::vector<double>(n_detector_cols));
        }
        std::vector<double> buffer(fov.n_act * dims::pol * n_detector_cols);
        grp.getVar("rad_spectra").getVar(buffer.data());
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            for (int i_act {}; i_act < fov.n_act; ++i_act) {
                for (int i {}; i < n_detector_cols; ++i) {
                    rad_spectra[i_pol][i_act][i] =
                      buffer[(i_act * dims::pol + i_pol) * n_detector_cols + i];
                }
            }
        }
    }
    if (activateStep(ProcLevel::pol, "POLARIMETRIC")) {
        spdlog::info("Reading the polarimetric CKD");
        pol.m_q.resize(fov.n_act, std::vector<double>(n_detector_cols));
        pol.m_u.resize(fov.n_act, std::vector<double>(n_detector_cols));
        pol.m_q_solar.resize(fov.n_act, std::vector<double>(n_detector_cols));
        pol.m_u_solar.resize(fov.n_act, std::vector<double>(n_detector_cols));
        pol.m_t.resize(fov.n_act, std::vector<double>(n_detector_cols));
        getAndReshape(grp.getVar("pol_m_q"), pol.m_q);
        getAndReshape(grp.getVar("pol_m_u"), pol.m_u);
        getAndReshape(grp.getVar("solar_m_q"), pol.m_q_solar);
        getAndReshape(grp.getVar("solar_m_u"), pol.m_u_solar);
        getAndReshape(grp.getVar("pol_m_t"), pol.m_t);
        if (!grp_d.isNull()) {
            pol.eps.resize(fov.n_act, std::vector<double>(n_detector_cols));
            pol.delta.resize(fov.n_act, std::vector<double>(n_detector_cols));
            getAndReshape(grp_d.getVar("eps"), pol.eps);
            getAndReshape(grp_d.getVar("delta"), pol.delta);
        }
    }
}

CKD::CKD(const std::string ckd_file_name,
         std::array<bool, static_cast<int>(ProcLevel::n_levels)>& skip,
         const std::array<std::string, static_cast<int>(ProcLevel::n_levels)>
           alternative_ckd_file_names,
         const double fov_spectrum_width)
{
    // Read debug information if present
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> debug {};
    std::ranges::fill(debug, true);
    readCKD(ckd_file_name,
            ProcLevel::n_levels,
            debug,
            skip,
            alternative_ckd_file_names,
            fov_spectrum_width);
}

CKD::CKD(const ProcLevel target_level, SPEXSettings& settings)
  : n_detector_rows { settings.main.detector_spat }
  , n_detector_cols { settings.main.detector_spec }
  , n_detector_cols_full { settings.main.detector_spec_full.value() }
  , n_detector_rows_full { settings.main.detector_spat_full.value() }
  , npix { settings.main.detector_spec * settings.main.detector_spat }
  , n_vp { settings.main.n_viewports }
{
    if (settings.main.processes.front() < ProcLevel::l1c) {
        printHeading("CKD initialization");
    }
    // Star with an empty pixel mask (all pixels are live)
    pixel_mask.resize(npix, false);
    if (settings.main.ckd_file_in.empty()) {
        if (settings.main.processes.front() < ProcLevel::l1b) {
            spdlog::info("Starting with an empty CKD: {}",
                         settings.main.ckd_file_out);
            spdlog::info("Detector dimension  (spatial): {}", n_detector_rows);
            spdlog::info("Detector dimension (spectral): {}", n_detector_cols);
            spdlog::info("Number of viewports: {}", n_vp);
        }
        for (int& bad_row : settings.main.bad_rows) {
            std::fill(pixel_mask.begin() + bad_row * n_detector_cols,
                      pixel_mask.begin() + (bad_row + 1) * n_detector_cols,
                      true);
        }
        return;
    }
    // Load debug flags into a temporary variable due to readCKD interface
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> debug {};
    for (int i {}; i < static_cast<int>(ProcLevel::n_levels); ++i) {
        debug.at(i) = settings.debug.at(i).value();
    }
    // Similarly load debug alternative CKD file names into a temporary variable
    std::array<std::string, static_cast<int>(ProcLevel::n_levels)>
      alternative_ckd_files {};
    for (int i {}; i < static_cast<int>(ProcLevel::n_levels); ++i) {
        alternative_ckd_files.at(i) = settings.ckd_file_in.at(i);
    }
    readCKD(settings.main.ckd_file_in,
            target_level,
            debug,
            settings.skip,
            alternative_ckd_files,
            settings.main.spectrum_width);
    // Read debug flags back into the original setting because some of
    // them might have changed in readCKD.
    for (int i {}; i < static_cast<int>(ProcLevel::n_levels); ++i) {
        settings.debug.at(i) = debug.at(i);
    }
    // If the user defined detector dimensions to not match those from
    // the CKD it is undefined behavior. Better to stop here.
    if (settings.main.detector_spat != 0
        && settings.main.detector_spat != n_detector_rows) {
        throw std::runtime_error { "[main][detector_spat] does not match the "
                                   "detector dimension from the CKD" };
    }
    if (settings.main.detector_spec != 0
        && settings.main.detector_spec != n_detector_cols) {
        throw std::runtime_error { "[main][detector_spec] does not match the "
                                   "detector dimension from the CKD" };
    }
    // Since more functions use the settings than the CKD copy the
    // dimensions to the settings so that they are easily available
    // (they might already be set).
    settings.main.detector_spec = n_detector_cols;
    settings.main.detector_spat = n_detector_rows;
}

// Like getAndReshape but in reverse
static auto reshapeAndPut(const std::vector<std::vector<double>>& source,
                          netCDF::NcVar& nc_var) -> void
{
    std::vector<double> buffer(source.size() * source.front().size());
    for (int i {}; i < static_cast<int>(source.size()); ++i) {
        for (int j {}; j < static_cast<int>(source.at(i).size()); ++j) {
            buffer[i * source.front().size() + j] = source[i][j];
        }
    }
    nc_var.putVar(buffer.data());
}

// Similar but for a 3-dimensional array
static auto reshapeAndPut(
  const std::array<std::vector<std::vector<double>>, dims::pol>& source,
  netCDF::NcVar& nc_var) -> void
{
    const std::array<int, 3> dims {
        static_cast<int>(source.size()),
        static_cast<int>(source.front().size()),
        static_cast<int>(source.front().front().size()),
    };
    // Same comments as above in getAndReshape
    const bool flip_dim { static_cast<int>(nc_var.getDim(0).getSize())
                          != dims[0] };
    std::vector<double> buffer(
      std::accumulate(dims.cbegin(), dims.cend(), 1, std::multiplies<>()));
    int idx {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int j {}; j < dims[1]; ++j) {
            if (flip_dim) {
                idx = (j * dims::pol + i_pol) * dims[2];
            } else {
                idx = (i_pol * dims[1] + j) * dims[2];
            }
            for (int k {}; k < dims[2]; ++k) {
                buffer[idx + k] = source[i_pol][j][k];
            }
        }
    }
    nc_var.putVar(buffer.data());
}

// Create a debug group with the same name as 'group' in the DEBUG
// section. Create the DEBUG group first if not already present.
static auto getOrCreateDebugGroup(netCDF::NcGroup& group) -> netCDF::NcGroup
{
    auto root { group.getParentGroup() };
    if (auto group_d { root.getGroup("DEBUG") }; group_d.isNull()) {
        return root.addGroup("DEBUG").addGroup(group.getName());
    } else {
        return group_d.addGroup(group.getName());
    }
}

auto CKD::write(const SPEXSettings& settings) -> void
{
    spdlog::info("");
    spdlog::info("Writing CKD file");
    // Open an output CKD file and write some global attributes and a
    // reduced YAML configuration into it.
    NcObject nc { netCDF::NcFile::replace, settings.main.ckd_file_out };
    nc.writeReducedConfig(settings);
    nc.putAtt("title", "PACE SPEXone calibration key data (CKD)");
    // Define basic dimensions and pixel mask
    const netCDF::NcDim polarization_states {
        nc.addDim("polarization_directions", dims::pol),
    };
    const netCDF::NcVar polarization_states_str {
        nc.addVar(
          "polarization_directions", netCDF::ncString, polarization_states),
    };
    const char* polarization_states_map[dims::pol] { "-", "+" };
    polarization_states_str.putVar(&polarization_states_map);
    const netCDF::NcDim vector_elements {
        nc.addDim("vector_elements", dims::vec),
    };
    const netCDF::NcDim spatial_detector_pixels {
        nc.addDim("spatial_detector_pixels", n_detector_rows),
    };
    const netCDF::NcDim spectral_detector_pixels {
        nc.addDim("spectral_detector_pixels", n_detector_cols),
    };
    const std::vector<netCDF::NcDim> detector_shape {
        spatial_detector_pixels, spectral_detector_pixels
    };
    const netCDF::NcDim viewports {
        nc.addDim("viewports", n_vp),
    };
    netCDF::NcVar viewports_str {
        nc.addVar("viewports", netCDF::ncString, viewports),
    };
    viewports_str.putAtt("units", "degrees");
    const char* viewports_map[n_vp] { "-50", "-20", "0", "20", "50" };
    viewports_str.putVar(&viewports_map);
    // Mask is the only global variable
    const std::vector<netCDF::NcDim> dims_mask { detector_shape };
    netCDF::NcVar mask_nc { nc.addVar("mask", netCDF::ncUbyte, dims_mask) };
    mask_nc.putAtt("long_name", "detector pixel mask");
    constexpr std::array<uint8_t, 2> mask_values { 0, 1 };
    mask_nc.putAtt(
      "flag_values", netCDF::ncUbyte, mask_values.size(), mask_values.data());
    mask_nc.putAtt("flag_meanings", "good bad");
    std::vector<uint8_t> pixel_mask_u8(npix);
    for (int i {}; i < npix; ++i) {
        pixel_mask_u8[i] = static_cast<uint8_t>(pixel_mask[i]);
    }
    mask_nc.putVar(pixel_mask_u8.data());
    // Work variable for NetCDF variables
    netCDF::NcVar var {};
    if (!settings.skip.at(static_cast<int>(ProcLevel::dark))) {
        netCDF::NcGroup grp { nc.addGroup("DARK") };
        var = grp.addVar("offset_short", netCDF::ncDouble, detector_shape);
        var.putAtt("long_name", "offset signal for short exposure times");
        var.putAtt("units", "counts");
        var.putVar(dark.offset_short.data());
        var = grp.addVar("offset_long", netCDF::ncDouble, detector_shape);
        var.putAtt("long_name", "offset signal for long exposure times");
        var.putAtt("units", "counts");
        var.putVar(dark.offset_long.data());
        var = grp.addVar("dark_current", netCDF::ncDouble, detector_shape);
        var.putAtt("long_name", "dark current");
        var.putAtt("comment",
                   "basis set continues as polynomials outside knot domain");
        var.putAtt("units", "counts/s");
        var.putVar(dark.current.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::dark)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            var = grp_d.addVar("chi2", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "sum of least squares of the linear fit");
            var.putAtt("units", "1");
            var.putVar(dark.chi2.data());
            var = grp_d.addVar("sigma_beta", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name",
                       "standard deviation of the long exposure time offset");
            var.putAtt("units", "1");
            var.putVar(dark.sigma_beta.data());
            var = grp_d.addVar("sigma_alpha", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "standard deviation of the dark current");
            var.putAtt("units", "1");
            var.putVar(dark.sigma_alpha.data());
            var = grp_d.addVar(
              "last_exposure_times", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name",
                       "highest exposure time used in the linear fits");
            var.putAtt("units", "s");
            var.putVar(dark.last_exposure_times.data());
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::noise))) {
        netCDF::NcGroup grp { nc.addGroup("NOISE") };
        var = grp.addVar("g", netCDF::ncDouble, detector_shape);
        var.putAtt("long_name", "noise conversion gain");
        var.putAtt("comment", "noise model: sigma = sqrt(g*signal + n)");
        var.putAtt("units", "counts/e");
        var.putVar(noise.g.data());
        var = grp.addVar("n", netCDF::ncDouble, detector_shape);
        var.putAtt("long_name", "scaled read noise");
        var.putAtt("comment", "noise model: sigma = sqrt(g*signal + n)");
        var.putAtt("units", "counts^2/e");
        var.putVar(noise.n.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::noise)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            var = grp_d.addVar("std_dev_g", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name",
                       "standard deviation of noise conversion gain");
            var.putAtt("units", "counts");
            var.putVar(noise.std_dev_g.data());
            var = grp_d.addVar("std_dev_n", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "standard deviation scaled dark noise");
            var.putAtt("units", "counts^2");
            var.putVar(noise.std_dev_n.data());
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::nonlin))) {
        netCDF::NcGroup grp { nc.addGroup("NON_LINEARITY") };
        var = grp.addVar("model", netCDF::ncUbyte);
        var.putAtt("long_name", "nonlinearity model");
        constexpr std::array<uint8_t, 3> model_values {
            static_cast<uint8_t>(NonlinModel::bspline),
            static_cast<uint8_t>(NonlinModel::sigmoidal),
            static_cast<uint8_t>(NonlinModel::single)
        };
        var.putAtt("flag_values",
                   netCDF::ncUbyte,
                   model_values.size(),
                   model_values.data());
        var.putAtt("flag_meanings", "bspline sigmoidal single");
        const uint8_t model_u8 { static_cast<uint8_t>(nonlin.model) };
        var.putVar(&model_u8);
        if (nonlin.model == NonlinModel::bspline) {
            const netCDF::NcDim n_states { grp.addDim(
              "states", nonlin.b_spline->nStates()) };
            const netCDF::NcDim n_knots { grp.addDim(
              "knots", nonlin.b_spline_knots.size()) };
            var = grp.addVar("order", netCDF::ncInt);
            var.putAtt("long_name", "B-spline order");
            var.putAtt("units", "1");
            var.putVar(&nonlin.order);
            var = grp.addVar("knots", netCDF::ncDouble, n_knots);
            var.putAtt("long_name", "B-spline knots of the measured signal");
            var.putAtt("units", "counts");
            var.putVar(nonlin.b_spline_knots.data());
            var = grp.addVar(
              "control_points",
              netCDF::ncDouble,
              { spatial_detector_pixels, spectral_detector_pixels, n_states });
            var.putAtt("long_name", "B-spline control points");
            var.putAtt("units", "counts");
            var.putVar(nonlin.control_points.data());
        } else if (nonlin.model == NonlinModel::sigmoidal) {
            const netCDF::NcDim n_knots { grp.addDim(
              "knots", nonlin.linear_spline.knots.size()) };
            var = grp.addVar("spline_type", netCDF::ncByte);
            var.putAtt("long_name", "nonlinearity spline type");
            var.putAtt("flag_values",
                       netCDF::ncUbyte,
                       mask_values.size(),
                       mask_values.data());
            var.putAtt("flag_meanings", "linear cubic");
            var.putAtt("units", "1");
            var = grp.addVar("knots", netCDF::ncDouble, n_knots);
            var.putAtt("long_name", "spline knots of the measured signal");
            var.putAtt("units", "counts");
            var.putVar(nonlin.linear_spline.knots.data());
            var = grp.addVar(
              "y",
              netCDF::ncDouble,
              { spatial_detector_pixels, spectral_detector_pixels, n_knots });
            var.putAtt("long_name", "function values as knots");
            var.putAtt("units", "1");
            var.putVar(nonlin.linear_spline.y.data());
        }
        if ((nonlin.model == NonlinModel::bspline
             || nonlin.model == NonlinModel::sigmoidal)
            && settings.debug.at(static_cast<int>(ProcLevel::nonlin)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            const int npix_d { static_cast<int>(nonlin.debug_pixels.size()) };
            const netCDF::NcDim n_debug_pixels { grp_d.addDim("debug_pixels",
                                                              npix_d) };
            const int n_rates { nonlin.signal_rates.empty()
                                  ? 0
                                  : static_cast<int>(
                                    nonlin.signal_rates.front().size()) };
            const netCDF::NcDim n_signal_rates { grp_d.addDim("signal_rates",
                                                              n_rates) };
            var = grp_d.addVar("debug_pixels", netCDF::ncInt, n_debug_pixels);
            var.putAtt(
              "long_name",
              "debug pixels for which additional information is provided");
            var.putAtt("units", "1");
            var.putVar(nonlin.debug_pixels.data());
            var = grp_d.addVar("signal_rates",
                               netCDF::ncDouble,
                               { n_debug_pixels, n_signal_rates });
            var.putAtt(
              "long_name",
              "signal rates used in the determining the reference current");
            var.putAtt("comment", "I_c in Eq. (81)");
            var.putAtt("units", "1");
            reshapeAndPut(nonlin.signal_rates, var);
            var = grp_d.addVar("reference_rates",
                               netCDF::ncDouble,
                               { n_debug_pixels, n_signal_rates });
            var.putAtt("long_name",
                       "reference diode rates used in the determining the "
                       "reference current");
            var.putAtt("comment", "I_ref in Eq. (81)");
            var.putAtt("units", "1");
            reshapeAndPut(nonlin.reference_rates, var);
            var = grp_d.addVar("f1", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "reference diode conversion factor f1");
            var.putAtt("comment", "f1 in Eq. (81)");
            var.putAtt("units", "1");
            var.putVar(nonlin.f1.data());
            var = grp_d.addVar("f2", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "reference diode conversion factor f2");
            var.putAtt("comment", "f2 in Eq. (81)");
            var.putAtt("units", "1");
            var.putVar(nonlin.f2.data());
            var = grp_d.addVar("m0", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "saturation signal m0 parameter");
            var.putAtt("comment", "m0 in Eq. (83)");
            var.putAtt("units", "1");
            var.putVar(nonlin.m0.data());
            var = grp_d.addVar("m1", netCDF::ncDouble);
            var.putAtt("long_name", "saturation signal m1 parameter");
            var.putAtt("comment", "m1 in Eq. (83)");
            var.putAtt("units", "1");
            var.putVar(&nonlin.m1);
            var = grp_d.addVar("m2", netCDF::ncDouble);
            var.putAtt("long_name", "saturation signal m2 parameter");
            var.putAtt("comment", "m2 in Eq. (83)");
            var.putAtt("units", "1");
            var.putVar(&nonlin.m2);
            const netCDF::NcDim n_S_max { grp_d.addDim(
              "S_max", nonlin.S_max_data.size()) };
            var = grp_d.addVar("I_c_med", netCDF::ncDouble, n_S_max);
            var.putAtt("long_name",
                       "detector medians of the linearized current");
            var.putAtt("comment", "Ic values in Eq. (83)");
            var.putAtt("units", "1");
            var.putVar(nonlin.I_c_med.data());
            var = grp_d.addVar("S_max_data", netCDF::ncDouble, n_S_max);
            var.putAtt(
              "long_name",
              "data points used for fitting the saturation parameters");
            var.putAtt("comment", "S_max values in Eq. (83)");
            var.putAtt("units", "1");
            var.putVar(nonlin.S_max_data.data());
            var = grp_d.addVar("S_max_fit", netCDF::ncDouble, n_S_max);
            var.putAtt("long_name", "fitting of the saturation parameters");
            var.putAtt("comment", "result of fitting Eq. (83)");
            var.putAtt("units", "1");
            var.putVar(nonlin.S_max_fit.data());
            if (nonlin.model == NonlinModel::sigmoidal) {
                for (int i_r {}; i_r < static_cast<int>(nonlin.r.size());
                     ++i_r) {
                    const auto i_r_str { std::to_string(i_r) };
                    var = grp_d.addVar(
                      "r" + i_r_str, netCDF::ncDouble, detector_shape);
                    var.putAtt("long_name", "r" + i_r_str + " in Eq. (73)");
                    var.putAtt("units", "1");
                    var.putVar(nonlin.r[i_r].data());
                }
                var = grp_d.addVar("c", netCDF::ncDouble, detector_shape);
                var.putAtt("long_name", "c in Eq. (72)");
                var.putAtt("units", "1");
                var.putVar(nonlin.c.data());
            }
            if (npix_d > 0) {
                const netCDF::NcDim model_fit_points { grp_d.addDim(
                  "model_fit_points", nonlin.S_lin.front().size()) };
                var = grp_d.addVar("S_lin",
                                   netCDF::ncDouble,
                                   { n_debug_pixels, model_fit_points });
                var.putAtt("long_name",
                           "linear signal for the nonlinearity model fit");
                var.putAtt("units", "1");
                reshapeAndPut(nonlin.S_lin, var);
                var = grp_d.addVar("S_meas",
                                   netCDF::ncDouble,
                                   { n_debug_pixels, model_fit_points });
                var.putAtt("long_name",
                           "expected signal for the nonlinearity model fit");
                var.putAtt("units", "1");
                reshapeAndPut(nonlin.S_meas, var);
                var = grp_d.addVar("nonlinearity_fit",
                                   netCDF::ncDouble,
                                   { n_debug_pixels, model_fit_points });
                var.putAtt("long_name", "nonlinearity model fit");
                var.putAtt("units", "1");
                reshapeAndPut(nonlin.nonlinearity_fits, var);
            }
        }
        if (nonlin.model == NonlinModel::single) {
            const netCDF::NcDim dimid_nonlin_spline { grp.addDim(
              "nonlin_number_of_coefficients", dim_nonlin_spline) };
            const netCDF::NcDim dimid_nonlin_knot { grp.addDim(
              "nonlin_number_of_knots", dim_nonlin_knot) };
            // Order (scalar)
            var = grp.addVar("nonlin_order", netCDF::ncInt);
            var.putAtt("long_name", "b-spline order of non-linearity fit");
            var.putAtt("definition",
                       "parameter n, so polynomial order is one lower");
            var.putVar(&nonlin.order);
            // Knots
            var =
              grp.addVar("nonlin_knots", netCDF::ncDouble, dimid_nonlin_knot);
            var.putAtt("long_name",
                       "scaled b-spline knots of non-linearity fit");
            var.putVar(nonlin.b_spline_knots.data());
            // Signal scales
            var =
              grp.addVar("nonlin_signal_scale",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
            var.putAtt("long_name", "scaling factor of the b-spline knots");
            var.putAtt("units", "counts");
            var.putVar(nonlin_signal_scale.data());
            // Fit results
            var = grp.addVar("nonlin_fit",
                             netCDF::ncDouble,
                             { spatial_detector_pixels,
                               spectral_detector_pixels,
                               dimid_nonlin_spline });
            var.putAtt("long_name",
                       "b-spline coefficients for non-linearity curve");
            var.putAtt("units", "counts");
            var.putVar(nonlin_fit.data());
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::prnu))) {
        netCDF::NcGroup grp { nc.addGroup("PRNU") };
        var = grp.addVar("prnu", netCDF::ncDouble, detector_shape);
        var.putAtt("long_name", "pixel response non-uniformity");
        var.putAtt("units", "1");
        var.putVar(prnu.prnu.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::prnu)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            var = grp_d.addVar("sensitivity", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name", "polynomial fit of the calibration data");
            var.putAtt("units", "1");
            var.putVar(prnu.sensitivity_fit.data());
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::stray))) {
        netCDF::NcGroup grp { nc.addGroup("STRAYLIGHT") };
        // Dimensions
        const netCDF::NcDim n_kernels {
            grp.addDim("kernels", stray.n_kernels),
        };
        const auto reduced_kernels_u8 { static_cast<uint8_t>(
          stray.reduced_kernels) };
        grp.addVar("reduced_kernels", netCDF::ncUbyte)
          .putVar(&reduced_kernels_u8);
        const netCDF::NcDim n_spatial { grp.addDim("spatial",
                                                   stray.n_spatial) };
        const netCDF::NcDim n_spectral { grp.addDim("spectral",
                                                    stray.n_spectral) };
        const netCDF::NcDim kernels_fft_size {
            // Need to multiply by 2 because of complex numbers
            grp.addDim("kernels_fft_size",
                       2
                         * std::accumulate(stray.kernel_fft_sizes.cbegin(),
                                           stray.kernel_fft_sizes.cend(),
                                           0)),
        };
        var = grp.addVar("kernel_rows", netCDF::ncInt, n_kernels);
        var.putAtt("long_name", "number of rows in each kernel");
        var.putAtt("units", "1");
        var.putVar(stray.kernel_rows.data());
        var = grp.addVar("kernel_cols", netCDF::ncInt, n_kernels);
        var.putAtt("long_name", "number of columns in each kernel");
        var.putAtt("units", "1");
        var.putVar(stray.kernel_cols.data());
        var = grp.addVar("kernel_fft_sizes", netCDF::ncInt, n_kernels);
        var.putAtt("long_name", "sizes of kernel FFTs");
        var.putAtt("units", "size of complex double");
        var.putVar(stray.kernel_fft_sizes.data());
        // Variables
        var = grp.addVar("kernels_fft", netCDF::ncDouble, kernels_fft_size);
        var.putAtt("long_name", "Fourier transforms of the kernels");
        var.putAtt("units", "1");
        std::vector<size_t> start { 0 };
        for (int i_kernel {}; i_kernel < stray.n_kernels; ++i_kernel) {
            std::vector<size_t> count {
                2 * static_cast<size_t>(stray.kernel_fft_sizes.at(i_kernel))
            };
            var.putVar(start, count, stray.kernels_fft.at(i_kernel).data());
            start.front() += count.front();
        }
        var = grp.addVar("eta", netCDF::ncDouble, { n_spatial, n_spectral });
        var.putAtt("long_name", "internal scattering factor");
        var.putAtt("units", "1");
        var.putVar(stray.eta.data());
        var = grp.addVar(
          "weights", netCDF::ncDouble, { n_kernels, n_spatial, n_spectral });
        var.putAtt("long_name", "kernel weights");
        var.putAtt("units", "1");
        reshapeAndPut(stray.weights, var);
        const netCDF::NcDim edges_dim { grp.addDim("edges",
                                                   stray.edges.size()) };
        var = grp.addVar("edges", netCDF::ncInt, edges_dim);
        var.putAtt("long_name", "subimage boundaries");
        var.putAtt("description",
                   "bottom and top row and left and right column defining an "
                   "image that a given kernel is convolved with");
        var.putAtt("units", "1");
        var.putVar(stray.edges.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::stray)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            const netCDF::NcDim n_kernels {
                grp_d.addDim("kernels", stray.n_kernels),
            };
            const netCDF::NcDim kernel_n_rows {
                grp_d.addDim("kernel_rows", stray.kernel_n_rows_full),
            };
            const netCDF::NcDim kernel_n_cols {
                grp_d.addDim("kernel_cols", stray.kernel_n_cols_full),
            };
            const netCDF::NcDim n_spatial { grp_d.addDim("spatial",
                                                         stray.n_spatial) };
            const netCDF::NcDim n_spectral { grp_d.addDim("spectral",
                                                          stray.n_spectral) };
            var = grp_d.addVar("kernels_real",
                               netCDF::ncDouble,
                               { n_kernels, kernel_n_rows, kernel_n_cols });
            var.putAtt("long_name", "real space kernels");
            var.putAtt("units", "1");
            reshapeAndPut(stray.kernels_real, var);
            var = grp_d.addVar(
              "etas", netCDF::ncDouble, { n_kernels, n_spatial, n_spectral });
            reshapeAndPut(stray.etas, var);
            var.putAtt("long_name", "individual internal scattering factors");
            var.putAtt("units", "1");
            var = grp_d.addVar("all_kernels", netCDF::ncDouble, detector_shape);
            var.putAtt("long_name",
                       "sum of all kernels shifted to their centers (for "
                       "illustration only)");
            var.putAtt("units", "1");
            var.putVar(stray.all_kernels.data());
        }
    }
    // Define this dimension outside the FOV scope because it is also
    // used by other steps.
    netCDF::NcDim spatial_samples_per_image {};
    if (fov.n_act > 0) {
        spatial_samples_per_image =
          nc.addDim("spatial_samples_per_image", fov.n_act);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::fov))) {
        netCDF::NcGroup grp { nc.addGroup("FIELD_OF_VIEW") };
        var = grp.addVar("fov_nfov_vp", netCDF::ncInt, viewports);
        var.putAtt("long_name", "number of spatial samples per viewport");
        var.putVar(fov.n_act_vp.data());
        var = grp.addVar("fov_ifov_start_vp", netCDF::ncInt, viewports);
        var.putAtt("long_name", "first spatial sample index in each viewport");
        var.putVar(fov.start_vp.data());
        var = grp.addVar(
          "fov_act_angles", netCDF::ncDouble, spatial_samples_per_image);
        var.putAtt("long_name",
                   "across track rotation stage position per spatial sample");
        var.putAtt("units", "radians");
        var.putVar(fov.act_angles.data());
        var = grp.addVar("fov_ispat",
                         netCDF::ncDouble,
                         { spatial_samples_per_image,
                           polarization_states,
                           spectral_detector_pixels });
        var.putAtt(
          "long_name",
          "floating point spatial detector pixel indices of the spectra");
        reshapeAndPut(fov.row_indices, var);
        if (settings.debug.at(static_cast<int>(ProcLevel::fov)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            const auto n_l1a { grp_d.addDim("measurements", fov.n_l1a_total) };
            const auto n_columns { grp_d.addDim("peak_positions",
                                                fov.n_columns) };
            var = grp_d.addVar("peak_positions",
                               netCDF::ncDouble,
                               { n_l1a, polarization_states, n_columns });
            var.putAtt(
              "long_name",
              "peak positions determined from Gaussian fits across columns");
            var.putAtt("units", "1");
            reshapeAndPut(fov.peak_pos, var);
            var = grp_d.addVar(
              "peak_positions_b",
              netCDF::ncDouble,
              { n_l1a, polarization_states, spectral_detector_pixels });
            var.putAtt("long_name", "B-spline fits of peak positions");
            var.putAtt("units", "1");
            reshapeAndPut(fov.peak_pos_b, var);
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::swath))) {
        netCDF::NcGroup grp { nc.addGroup("SWATH") };
        // Swath vectors
        const std::vector<netCDF::NcDim> dims_swathvectors {
            spatial_samples_per_image, vector_elements
        };
        var =
          grp.addVar("swath_swathvectors", netCDF::ncDouble, dims_swathvectors);
        var.putAtt("long_name",
                   "normalized pointing vectors in spacecraft coordinates");
        var.putAtt("axes", "x = flying direction, y = to the right, z = down");
        var.putVar(swath.pointings.data());
        const std::vector<netCDF::NcDim> dims_normals { viewports,
                                                        vector_elements };
        var = grp.addVar(
          "swath_vectorplane_normals", netCDF::ncDouble, dims_normals);
        var.putAtt(
          "long_name",
          "normal vectors to plane through swath vectors of viewport.");
        var.putAtt("axes", "x = flying direction, y = to the right, z = down");
        var.putVar(swath.normals.data());
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::wave))) {
        netCDF::NcGroup grp { nc.addGroup("WAVELENGTH") };
        // Wavelengths per spectrum
        const std::vector<netCDF::NcDim> dims_wave_full {
            polarization_states,
            spatial_samples_per_image,
            spectral_detector_pixels
        };
        var = grp.addVar("wave_full", netCDF::ncDouble, dims_wave_full);
        var.putAtt("long_name", "wavelengths of S and P spectra");
        var.putAtt("units", "nm");
        reshapeAndPut(wave_full, var);
        const std::vector<netCDF::NcDim> dims_wave_common {
            spatial_samples_per_image, spectral_detector_pixels
        };
        var = grp.addVar("wave_common", netCDF::ncDouble, dims_wave_common);
        var.putAtt("long_name",
                   "wavelengths of modulation spectra (common to both "
                   "polarization states)");
        var.putAtt("units", "nm");
        reshapeAndPut(wave_common, var);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::rad))) {
        netCDF::NcGroup grp { nc.addGroup("RADIOMETRIC") };
        // Radiometric calibration data on spectrum perspective
        const std::vector<netCDF::NcDim> dims_rad_spectra {
            spatial_samples_per_image,
            polarization_states,
            spectral_detector_pixels
        };
        var = grp.addVar("rad_spectra", netCDF::ncDouble, dims_rad_spectra);
        var.putAtt("long_name", "radiometric calibration constants");
        var.putAtt("units", "W sr-1 um-1 counts-1");
        reshapeAndPut(rad_spectra, var);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::pol))) {
        netCDF::NcGroup grp { nc.addGroup("POLARIMETRIC") };
        // Polarimetric calibration sampled on spectral pixels
        const std::vector<netCDF::NcDim> dims_pol { spatial_samples_per_image,
                                                    spectral_detector_pixels };
        // Mueller matrix ratio for Q
        var = grp.addVar("pol_m_q", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "modulation sensitivity to q-polarization");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_q, var);
        // Mueller matrix ratio for U
        var = grp.addVar("pol_m_u", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "modulation sensitivity to u-polarization");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_u, var);
        // Solar correction factor for mq
        var = grp.addVar("solar_m_q", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "solar spectrum correction factor for mq");
        var.putAtt("comment", "<mq I>_0 / (<I>_0 <mq>_0)");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_q_solar, var);
        // Solar correction factor for mu
        var = grp.addVar("solar_m_u", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "solar spectrum correction factor for mu");
        var.putAtt("comment", "<mu I>_0 / (<I>_0 <mu>_0)");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_u_solar, var);
        // Mueller element for telescope polarization
        var = grp.addVar("pol_m_t", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "telescope polarization");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_t, var);
        if (settings.debug.at(static_cast<int>(ProcLevel::pol)).value()) {
            auto grp_d { getOrCreateDebugGroup(grp) };
            var = grp_d.addVar("eps", netCDF::ncDouble, dims_pol);
            var.putAtt("long_name", "polarimetric efficiency");
            var.putAtt("units", "1");
            reshapeAndPut(pol.eps, var);
            var = grp_d.addVar("delta", netCDF::ncDouble, dims_pol);
            var.putAtt("long_name", "polarimetric retardance");
            var.putAtt("units", "nm");
            reshapeAndPut(pol.delta, var);
        }
    }
}

auto CKD::genFovIndices(const double spectrum_width) -> void
{
    // Convert fractional detector row coordinates into pixel indices
    // and weight factors.
    const double spectrum_width_half { spectrum_width / 2 };
    fov.n_indices = static_cast<int>(std::ceil(spectrum_width) + 1);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        fov.pix_indices[i_pol].resize(
          fov.n_act, std::vector<int>(n_detector_cols * fov.n_indices));
        fov.weights[i_pol].resize(
          fov.n_act, std::vector<double>(n_detector_cols * fov.n_indices));
    }
    // Based on the argument spectrum_width, each spectral element
    // spans N rows on the detector. The weight of the first and last
    // rows are determined by where exactly the edge of the spectrum
    // falls. All middle rows have the same constant weight (1 before
    // normalization).
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i_fov {}; i_fov < fov.n_act; ++i_fov) {
            for (int i {}; i < n_detector_cols; ++i) {
                const double row_first_d { fov.row_indices[i_pol][i_fov][i]
                                           + 0.5 - spectrum_width_half };
                const double row_last_d { fov.row_indices[i_pol][i_fov][i] + 0.5
                                          + spectrum_width_half };
                const int row_first_i { static_cast<int>(row_first_d) };
                const int row_last_i { static_cast<int>(row_last_d) };
                auto& indices { fov.pix_indices[i_pol][i_fov] };
                auto& weights { fov.weights[i_pol][i_fov] };
                indices[i * fov.n_indices] = row_first_i * n_detector_cols + i;
                indices[(i + 1) * fov.n_indices - 1] =
                  row_last_i * n_detector_cols + i;
                weights[i * fov.n_indices] =
                  std::ceil(row_first_d) - row_first_d;
                weights[(i + 1) * fov.n_indices - 1] =
                  row_last_d - std::floor(row_last_d);
                for (int i_ind { 1 }; i_ind < fov.n_indices - 1; ++i_ind) {
                    const int idx { i * fov.n_indices + i_ind };
                    indices[idx] = (row_first_i + i_ind) * n_detector_cols + i;
                    weights[idx] = 1.0;
                }
                const int idx { i * fov.n_indices };
                if (indices[idx] == indices[idx + 1]) {
                    weights[idx + 1] = 0.0;
                }
                if (indices[idx + fov.n_indices - 1]
                      == indices[idx + fov.n_indices - 2]
                    && fov.n_indices > 2) {
                    weights[idx + fov.n_indices - 2] = 0.0;
                }
                double weights_norm {};
                for (int i_ind {}; i_ind < fov.n_indices; ++i_ind) {
                    weights_norm += weights[i * fov.n_indices + i_ind];
                }
                for (int i_ind {}; i_ind < fov.n_indices; ++i_ind) {
                    weights[i * fov.n_indices + i_ind] /= weights_norm;
                }
            }
        }
    }
}

} // namespace spex
