// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include <cstdint>
#include <map>
#include <string>

namespace spex {

// Dimensions of mathematical spaces
namespace dims {

// Number of polarization states
constexpr int pol { 2 };
// Number of vector components
constexpr int vec { 3 };
// Number of quaternion components
constexpr int quat { 4 };

} // namespace dims

// Detector related
namespace detector {

// Number of data channels is 2 in full-frame mode and 8 in binning
// mode.
constexpr int n_channels_full_frame { 2 };
constexpr int n_channels_binning_mode { 8 };

} // namespace detector

// Whenever an array dimension is associated with polarization states,
// the pol_plus index refers to S+ spectra and pol_minus to S-
// spectra. For instance, rad_spectra[pol_plus][i_fov][i] is the
// radiometric calibration constant of S+ corresponding to field of
// view i_fov and spectral element i.
constexpr int pol_minus {};
constexpr int pol_plus { 1 };

namespace math {

// Multiply with this factor to convert from degrees to radians
constexpr double deg_to_rad { 0.017453292519943295 };

} // namespace math

// Default values for array indices and variables
namespace def {

// Index value for an array element that does not exist or is flagged
// bad for some reason
constexpr int idx { -1 };

// This fill value is used for all NetCDF variables in L1B and L1C products
constexpr double l1bc_fill { -32767.0 };

} // namespace def

// Geolocation related
namespace earth {

constexpr double a { 6378137.0 };
constexpr double b { 6356752.314245179499 };
constexpr double f_inv { 298.257223563 };
constexpr double f { 3.3528106647474804385e-3 };
constexpr double e2 { 6.694379990141316435e-3 };
constexpr double e { 8.1819190842621490908e-2 };

} // namespace earth

// Process ladder, an ordered list of possible states of data. A
// process level can be used to identify the process that is currently
// running or the maturity of the data (e.g. the CKD or an image).
enum class ProcLevel
{
    l1a,
    raw, // Like L1A but data stored as floating point numbers
    dark,
    noise,
    nonlin,
    prnu,
    stray,
    fov,
    swath,
    wave,
    rad,
    pol,
    l1b,
    l1c,
    n_levels,
};

enum class L1AType
{
    science,
    diagnostic,
    n_types,
};

// Nonlinearity models
enum class NonlinModel
{
    bspline,
    sigmoidal,
    single,
};

// Mapping between the viewport index and angle. Whenever we group L1A
// products into batches we follow this ordering. It does not reflect
// the actual ordering of viewports on a detector image but as long as
// this ordering is used consistently in the code it's fine. The true
// mapping of viewport angles to detector regions is relevant for
// swath vectors and is determined by [swath][viewport_alt_angles].
const std::map<uint8_t, double> viewport_idx_to_angle {
    { 1, -50.0 }, { 2, -20.0 }, { 4, 0.0 }, { 8, 20.0 }, { 16, 50.0 },
};

// Mapping between a process name and its corresponding entry in ProcLevel
const std::map<std::string, spex::ProcLevel> proc_level_str_to_enum {
    { "l1a", spex::ProcLevel::l1a },       { "raw", spex::ProcLevel::raw },
    { "dark", spex::ProcLevel::dark },     { "noise", spex::ProcLevel::noise },
    { "nonlin", spex::ProcLevel::nonlin }, { "prnu", spex::ProcLevel::prnu },
    { "stray", spex::ProcLevel::stray },   { "fov", spex::ProcLevel::fov },
    { "swath", spex::ProcLevel::swath },   { "wave", spex::ProcLevel::wave },
    { "rad", spex::ProcLevel::rad },       { "pol", spex::ProcLevel::pol },
    { "l1b", spex::ProcLevel::l1b },       { "l1c", spex::ProcLevel::l1c },
};

// Namespace for whenever we deal with rectangles and need to identify
// their bottom, top, left, and right sides
namespace box {

constexpr int b { 0 }; // bottom
constexpr int t { 1 }; // top
constexpr int l { 2 }; // left
constexpr int r { 3 }; // right
constexpr int n { 4 }; // number of sides a box has

} // namespace box

} // namespace spex
