// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "dark_gen.h"

#include <ranges>

namespace spex {

// Generate offset correponding to short exposure times
static auto genOffsetShort(const SPEXSettings& settings,
                           const std::vector<L1A>& l1a_products,
                           CKD& ckd) -> void
{
    ckd.dark.offset_short.resize(ckd.npix, 0.0);
    int n_l1a_short {};
    for (const auto& l1a :
         l1a_products
           | std::views::filter(
             [short_exposure_time =
                settings.dark.short_exposure_time](const auto& l1a) {
                 return l1a.exposure_time < short_exposure_time;
             })) {
        spdlog::info(
          "Generating short exposure offset from exposure time {:0.5f} ms",
          1e3 * l1a.exposure_time);
        for (int i {}; i < ckd.npix; ++i) {
            ckd.dark.offset_short[i] += l1a.image[i];
        }
        ++n_l1a_short;
    }
    if (n_l1a_short == 0) {
        throw std::runtime_error { "make sure that at least one measurement "
                                   "has exposure time shorter than "
                                   + std::to_string(
                                     settings.dark.short_exposure_time)
                                   + " s [dark][short_exposure_time]" };
    }
    for (int i {}; i < ckd.npix; ++i) {
        ckd.dark.offset_short[i] /= n_l1a_short;
    }
}

// Generate offset correponding to long exposure times and dark current
static auto genOffsetLongAndCurrent(const SPEXSettings& settings,
                                    const std::vector<L1A>& l1a_products,
                                    CKD& ckd) -> void
{
    ckd.dark.offset_long.resize(ckd.npix);
    ckd.dark.current.resize(ckd.npix);
    ckd.dark.chi2.resize(ckd.npix);
    ckd.dark.sigma_beta.resize(ckd.npix);
    ckd.dark.sigma_alpha.resize(ckd.npix);
    ckd.dark.last_exposure_times.resize(ckd.npix);
    // Only use measurements larger than the long exposure time
    // parameter and shorter than the maximum allowed exposure time
    // Beyond which the signal becomes saturated.
    const std::vector<std::reference_wrapper<const L1A>> l1a_refs {
        std::ranges::find_if(
          l1a_products,
          [exposure_time = settings.dark.long_exposure_time](const auto& l1a) {
              return l1a.exposure_time > exposure_time;
          }),
        l1a_products.end()
    };
    spdlog::info(
      "Generating long exposure offsets and dark current from exposure times:");
    for (const auto& l1a : l1a_refs) {
        spdlog::info("  {:2.5f} s", l1a.get().exposure_time);
    }
    // Default is to include all long exposure times in the set
    int n_l1a { static_cast<int>(l1a_refs.size()) };
    // Optionally reduce the number of files by an exposure time cutoff
    if (settings.dark.last_exposure_time) {
        n_l1a = std::ranges::distance(
          l1a_refs.begin(),
          std::ranges::find_if(
            l1a_refs,
            [exposure_time =
               settings.dark.last_exposure_time](const auto& l1a) {
                return l1a.get().exposure_time > exposure_time;
            }));
    }
    // At least this many exposure times are always includes in the linear fit
    const int n_l1a_min { static_cast<int>(settings.dark.data_start_fraction
                                           * n_l1a) };
    std::vector<double> exposure_times(n_l1a);
#pragma omp parallel for
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        // Fit for S = offset_long + tau * dark_current where tau is
        // the exposure time using least squares percentage
        // regression.
        double Q_00 {};
        double Q_01 {};
        double Q_02 {};
        double Q_11 {};
        double Q_12 {};
        double Q_22 {};
        for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
            exposure_times[i_l1a] = l1a_refs[i_l1a].get().exposure_time;
        }
        // Starting uncertainty of the dark current
        ckd.dark.sigma_alpha[i_pix] = 1e100;
        for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
            // Use weights to effectively spread the exposure times
            // evenly over the whole range, making the result less
            // sensitive to the actual spacing.
            double w;
            if (i_l1a == 0) {
                w = 2 * (exposure_times[1] - exposure_times.front());
            } else if (i_l1a == n_l1a - 1) {
                w = 2 * (exposure_times.back() - exposure_times[n_l1a - 2]);
            } else {
                w = exposure_times[i_l1a + 1] - exposure_times[i_l1a - 1];
            }
            const double w_over_exp { w / exposure_times[i_l1a] };
            const double w_over_exp2 { w_over_exp / exposure_times[i_l1a] };
            Q_00 += w;
            Q_01 += w_over_exp;
            Q_02 += w_over_exp2;
            const double& counts { l1a_refs[i_l1a].get().image[i_pix] };
            Q_11 += w_over_exp * counts;
            Q_12 += w_over_exp2 * counts;
            Q_22 += w_over_exp2 * counts * counts;
            const double Z_1 { Q_00 * Q_02 - Q_01 * Q_01 };
            const double Z_2 { Q_00 * Q_12 - Q_01 * Q_11 };
            const double Z_3 { Q_02 * Q_11 - Q_01 * Q_12 };
            const double offset_long { Z_2 / Z_1 };
            const double current { Z_3 / Z_1 };
            const double chi2 { Q_22 - Q_12 * offset_long - Q_11 * current };
            const double sigma_beta { std::sqrt(chi2 / (n_l1a - 1) * Q_00
                                                / Z_1) };
            const double sigma_alpha { std::sqrt(chi2 / (n_l1a - 1) * Q_02
                                                 / Z_1) };
            // The final values of the dark offset, current, and
            // related parameters are determined by either the last
            // exposure time specified by the user or the set of
            // exposure times yielding the smaller dark current
            // uncertainty. In the latter case, if the current
            // uncertainty is smaller than the smallest one yet
            // encountered for this pixel, stored in
            // ckd.dark.sigma_alpha, then update that and other CKD
            // variables.
            if (i_l1a > n_l1a_min
                && (settings.dark.last_exposure_time
                    || sigma_alpha < ckd.dark.sigma_alpha[i_pix])) {
                ckd.dark.offset_long[i_pix] = offset_long;
                ckd.dark.current[i_pix] = current;
                ckd.dark.chi2[i_pix] = chi2;
                ckd.dark.sigma_beta[i_pix] = sigma_beta;
                ckd.dark.sigma_alpha[i_pix] = sigma_alpha;
                ckd.dark.last_exposure_times[i_pix] = exposure_times[i_l1a];
            }
        }
    }
}

static auto updateMask(const SPEXSettings& settings, CKD& ckd) -> void
{
    if (settings.dark.mask_offset_min) {
        updateMaskLt(
          ckd.dark.offset_short,
          settings.dark.mask_offset_min.value(),
          "pixels masked bad because their offset (short) is less than",
          ckd.pixel_mask);
        updateMaskLt(
          ckd.dark.offset_long,
          settings.dark.mask_offset_min.value(),
          "pixels masked bad because their offset (long) is less than",
          ckd.pixel_mask);
    }
    if (settings.dark.mask_offset_max) {
        updateMaskGt(
          ckd.dark.offset_short,
          settings.dark.mask_offset_max.value(),
          "pixels masked bad because their offset (short) is greater than",
          ckd.pixel_mask);
        updateMaskGt(
          ckd.dark.offset_long,
          settings.dark.mask_offset_max.value(),
          "pixels masked bad because their offset (long) is greater than",
          ckd.pixel_mask);
    }
    if (settings.dark.mask_current_min) {
        updateMaskLt(
          ckd.dark.current,
          settings.dark.mask_current_min.value(),
          "pixels masked bad because their dark current is less than",
          ckd.pixel_mask);
    }
    if (settings.dark.mask_current_max) {
        updateMaskGt(
          ckd.dark.current,
          settings.dark.mask_current_max.value(),
          "pixels masked bad because their dark current is greater than",
          ckd.pixel_mask);
    }
}

auto DarkGen::processBatch(const int /*i_batch*/) -> void
{
    // Sort all measurements by exposure time because that is the main
    // criterion for which ones to select
    std::ranges::sort(batches.front().l1a_products,
                      [](const auto& l1a_1, const auto& l1a_2) {
                          return l1a_1.exposure_time < l1a_2.exposure_time;
                      });
    genOffsetShort(settings, batches.front().l1a_products, ckd);
    genOffsetLongAndCurrent(settings, batches.front().l1a_products, ckd);
    // Update pixel mask
    updateMask(settings, ckd);
}

} // namespace spex
