// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "dem.h"

#include "constants.h"

#include <algorithm>
#include <numbers>
#include <numeric>
#include <spdlog/spdlog.h>

namespace spex {

DEM::Bounds::Bounds(const std::vector<double>& latitudes,
                    const std::vector<double>& longitudes,
                    const double lat_min,
                    const double lat_max,
                    const double lon_min,
                    const double lon_max)
{
    // Helper functions for finding the index of the first or last
    // element from a list that is smaller or larger than a value.
    const auto idxMin { [](const std::vector<double>& values,
                           const double min_value) {
        // If all elements are smaller than the target value then
        // return the index of the first element outside the range.
        if (min_value >= values.back()) {
            return values.size();
        }
        return static_cast<size_t>(std::distance(
          values.begin(),
          std::ranges::find_if(
            values, [min_value](const double x) { return x > min_value; })));
    } };
    const auto idxMax { [](const std::vector<double>& values,
                           const double max_value) -> size_t {
        // If all elements are larger than the target value then
        // return the index of the first element outside the range in
        // the reverse direction.
        if (max_value <= values.front()) {
            return -1;
        }
        return std::distance(
                 values.begin(),
                 std::ranges::find_if_not(
                   values,
                   [max_value](const double x) { return x < max_value; }))
               - 1;
    } };
    idx_lat_beg = idxMin(latitudes, lat_min);
    idx_lat_end = idxMax(latitudes, lat_max);
    idx_lon_beg = idxMin(longitudes, lon_min);
    idx_lon_end = idxMax(longitudes, lon_max);
    lat_count = idx_lat_end - idx_lat_beg + 1;
    lon_count = idx_lon_end - idx_lon_beg + 1;
}

DEM::DEM(const std::string& filename, const double resolution)
  : resolution { resolution }
{
    nc.open(filename);
    const size_t n_lat { nc.getDim("lat").getSize() };
    const size_t n_lon { nc.getDim("lon").getSize() };
    latitudes.resize(n_lat);
    longitudes.resize(n_lon);
    nc.getVar("lat").getVar(latitudes.data());
    nc.getVar("lon").getVar(longitudes.data());
    nc_height = nc.getVar("height");
    nc_water_surface_height = nc.getVar("water_surface_height");
    nc_watermask = nc.getVar("watermask");
    for (double& val : latitudes) {
        val *= math::deg_to_rad;
    }
    for (double& val : longitudes) {
        val *= math::deg_to_rad;
    }
    if (!nc_water_surface_height.isNull()) {
        for (const auto& [name, att] : nc_water_surface_height.getAtts()) {
            if (name == "_FillValue") {
                att.getValues(&water_surface_height_fill);
            }
        }
    }
}

auto DEM::loadToMemory(const float lat_min,
                       const float lat_max,
                       const float lon_min,
                       const float lon_max) -> void
{
    spdlog::info("Loading DEM file ...");
    // Define a bounding box and retrieve the corresponding
    // heights. After that, two corner cases need to be considered.
    Bounds bounds { latitudes, longitudes, lat_min, lat_max, lon_min, lon_max };
    const std::vector<size_t> starts { bounds.idx_lat_beg, bounds.idx_lon_beg };
    const std::vector<size_t> counts { bounds.lat_count, bounds.lon_count };
    all_heights.resize(bounds.lat_count * bounds.lon_count);
    nc_height.getVar(starts, counts, all_heights.data());
    if (!nc_water_surface_height.isNull()) {
        std::vector<int16_t> water_surface_height(all_heights.size());
        std::vector<uint8_t> watermask(all_heights.size());
        nc_water_surface_height.getVar(
          starts, counts, water_surface_height.data());
        nc_watermask.getVar(starts, counts, watermask.data());
#pragma omp parallel for
        for (size_t i = 0; i < all_heights.size(); ++i) {
            if (watermask[i] == 1) {
                constexpr int16_t fill { -32767 };
                if (water_surface_height[i] == fill) {
                    all_heights[i] = 0.0;
                } else {
                    all_heights[i] = water_surface_height[i];
                }
            }
        }
    }
    latitudes = { latitudes.begin() + bounds.idx_lat_beg,
                  latitudes.begin() + bounds.idx_lat_beg + bounds.lat_count };
    longitudes = { longitudes.begin() + bounds.idx_lon_beg,
                   longitudes.begin() + bounds.idx_lon_beg + bounds.lon_count };
    nc.close();
    spdlog::info("Loading DEM file ... done");
}

auto DEM::extractHeights(const Bounds& bounds,
                         std::vector<int16_t>& heights) const -> void
{
    const auto lat_count { bounds.idx_lat_end - bounds.idx_lat_beg + 1 };
    const auto lon_count { bounds.idx_lon_end - bounds.idx_lon_beg + 1 };
    heights.resize(lat_count * lon_count);
    if (!all_heights.empty()) {
        auto heights_it { heights.begin() };
        for (int i {}; i < lat_count; ++i) {
            const auto lat_idx { (bounds.idx_lat_beg + i) * longitudes.size() };
            for (int j {}; j < lon_count; ++j) {
                *heights_it = all_heights[lat_idx + bounds.idx_lon_beg + j];
                ++heights_it;
            }
        }
        return;
    }
    const std::vector<size_t> starts { bounds.idx_lat_beg, bounds.idx_lon_beg };
    const std::vector<size_t> counts { lat_count, lon_count };
    std::vector<int16_t> water_surface_height {};
    std::vector<uint8_t> watermask {};
    if (!nc_water_surface_height.isNull()) {
        water_surface_height.resize(heights.size());
        watermask.resize(heights.size());
    }
#pragma omp critical
    {
        nc_height.getVar(starts, counts, heights.data());
        if (!nc_water_surface_height.isNull()) {
            nc_water_surface_height.getVar(
              starts, counts, water_surface_height.data());
            nc_watermask.getVar(starts, counts, watermask.data());
        }
    }
    if (!nc_water_surface_height.isNull()) {
        for (size_t i {}; i < heights.size(); ++i) {
            if (watermask[i] == 1) {
                if (water_surface_height[i] == water_surface_height_fill) {
                    heights[i] = 0.0;
                } else {
                    heights[i] = water_surface_height[i];
                }
            }
        }
    }
}

[[nodiscard]] auto DEM::getHeight(const double lat, const double lon) const
  -> std::array<double, 2>
{
    // Compute local Earth radius
    const double sin_phi { std::sin(lat) };
    const double normal_distance {
        earth::a / (std::sqrt(1.0 - earth::e2 * sin_phi * sin_phi))
    };
    const double local_r { normal_distance * std::cos(lat) };
    // Based on the given resolution, define the angular side length
    // of the bounding box as 2 * theta.
    const double theta { resolution / local_r / 2 };
    // In the longitude direction the angular distance is scaled by
    // the latitude cosine.
    const double theta_cos { theta * std::cos(lat) };
    // The minimum and maximum target latitudes are defined by +-theta.
    const double lat_min { lat - theta };
    const double lat_max { lat + theta };
    const double lon_min { lon - theta_cos };
    const double lon_max { lon + theta_cos };
    // We assume the bounding box never crosses the pole. If so, give
    // a warning but proceed. Effectively, some values will be
    // ignored.
    if (lat_min < latitudes.front() || lat_max > latitudes.back()) {
        spdlog::warn("Spacecraft is too close to one of the poles");
    }
    // Define the bounding and retrieve the corresponding
    // heights. After that, two corner cases need to be considered.
    Bounds bounds { latitudes, longitudes, lat_min, lat_max, lon_min, lon_max };
    std::vector<int16_t> height {};
    extractHeights(bounds, height);
    // First corner case is if the bounding box crosses the -180/180
    // longitude boundary from the right (target longitude is
    // negative).
    if (bounds.idx_lon_beg == 0) {
        // Define a new bounding box such that the maximum longitude
        // is close to 180.
        const double lon_min_new { lon_min + 2 * std::numbers::pi };
        bounds = {
            latitudes, longitudes, lat_min, lat_max, lon_min_new, lon_max
        };
        bounds.idx_lon_end = longitudes.size() - 1;
        std::vector<int16_t> height_extra {};
        extractHeights(bounds, height_extra);
        // Reallocation when adding new values to the list of heights
        // could be more efficient but this corner case is rarely
        // relevant so just use push_back here.
        for (const auto h : height_extra) {
            height.push_back(h);
        }
    }
    // Second corner case is if the bounding box crosses the boundary
    // from the left (target longitude is positive).
    if (bounds.idx_lon_end == longitudes.size() - 1) {
        const double lon_max_new { lon_max - 2 * std::numbers::pi };
        bounds = {
            latitudes, longitudes, lat_min, lat_max, lon_min, lon_max_new
        };
        bounds.idx_lon_beg = 0;
        std::vector<int16_t> height_extra {};
        extractHeights(bounds, height_extra);
        for (const auto h : height_extra) {
            height.push_back(h);
        }
    }
    // Finally, returned the averaged height and its variability
    const double height_ave { std::accumulate(
                                height.cbegin(), height.cend(), 0.0)
                              / static_cast<double>(height.size()) };
    double rms {};
    for (const auto& val : height) {
        const double diff { val - height_ave };
        rms += diff * diff;
    }
    rms = std::sqrt(rms
                    / static_cast<double>(
                      std::max(static_cast<size_t>(1), height.size() - 1)));
    return { height_ave, rms };
}

} // namespace spex
