// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Digital elevation model (DEM). Used for computing the surface
// height and its variability for target latitudes and longitudes.

#pragma once

#include "nc_object.h"

#include <string>
#include <vector>

namespace spex {

class DEM
{
private:
    // Minimum and maximum indices of latitudes and longitudes for
    // which to extract height data
    struct Bounds
    {
        size_t idx_lat_beg {};
        size_t idx_lat_end {}; // Inclusive
        size_t idx_lon_beg {};
        size_t idx_lon_end {}; // Inclusive
        size_t lat_count {};
        size_t lon_count {};
        Bounds(const std::vector<double>& latitudes,
               const std::vector<double>& longitudes,
               const double lat_min,
               const double lat_max,
               const double lon_min,
               const double lon_max);
    };

    // For a target latitude and longitude, height data is retrieved
    // for a box of size resolution * resolution km^2 where the target
    // point is in the middle of the box.
    const double resolution {};
    // DEM file grid
    std::vector<double> latitudes {};
    std::vector<double> longitudes {};
    // DEM file
    NcObject nc { netCDF::NcFile::read };
    // We don't read the DEM file contents into memory. Instead, store
    // their NetCDF variable handles.
    netCDF::NcVar nc_height {};
    netCDF::NcVar nc_water_surface_height {};
    netCDF::NcVar nc_watermask {};
    // Water surface height fill value. Used for recognizing when a
    // pixel is masked as water its surface height equals the fill
    // value. Then the height is set to 0.
    int16_t water_surface_height_fill {};
    // If load to memory was requested, this buffer contains all
    // heights from the DEM file which makes processing faster than
    // constantly them reading from file.
    std::vector<int16_t> all_heights {};

    // Extract a list of heights within a bounding box. The box bounds
    // correspond to the minimum and maximum indices of the latitude
    // and longitude variables read from the DEM file.
    auto extractHeights(const Bounds& bounds,
                        std::vector<int16_t>& heights) const -> void;

    // Latitude/longitude value in radians higher than any real value
    static constexpr float latlon_fill { 100.0 };

public:
    // Initialize using the DEM file name (in NetCDF format) and a
    // target resolution.
    DEM(const std::string& filename, const double resolution);
    // Call this to load all or a subset of the height values into
    // memory for faster processing. Calling with default arguments
    // load all values.
    auto loadToMemory(const float lat_min = -latlon_fill,
                      const float lat_max = latlon_fill,
                      const float lon_min = -latlon_fill,
                      const float lon_max = latlon_fill) -> void;

    // Return the height and its RMS (height variability) for a target
    // grid point and resolution which was set in the constructor.
    [[nodiscard]] auto getHeight(const double lat, const double lon) const
      -> std::array<double, 2>;

    ~DEM() = default;
};

} // namespace spex
