// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "fov_gen.h"

#include "io.h"
#include "l1a_metadata.h"
#include "linalg.h"

#include <gadfit/lm_solver.h>

namespace spex {

auto FovGen::processInit() -> void
{
    ckd.fov.n_columns = ckd.n_detector_cols - settings.fov.exclude_columns_right
                        - settings.fov.exclude_columns_left;
    ckd.fov.start_vp.resize(ckd.n_vp, 0);
    ckd.fov.n_act_vp.resize(ckd.n_vp, 0);
    ckd.fov.dims_spec.resize(ckd.n_detector_rows, ckd.n_detector_cols);
    // Allocate maximum size for the CKD and shrink in the end when we know the
    // number of FOVs (ACT angles used)
    ckd.fov.act_angles.resize(ckd.n_detector_rows);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        ckd.fov.row_indices[i_pol].resize(
          ckd.n_detector_rows, std::vector<double>(ckd.n_detector_cols));
        ckd.fov.peak_pos[i_pol].resize(ckd.n_detector_rows,
                                       std::vector<double>(ckd.fov.n_columns));
        ckd.fov.peak_pos_b[i_pol].resize(
          ckd.n_detector_rows, std::vector<double>(ckd.n_detector_cols));
    }
}

auto FovGen::prepareBatches() -> void
{
    batchViewport();
}

// Remove those calibration measurements from consideration that are
// not visible on the detector. This is determined by comparing light
// sums (sum over all pixel values) with the highest light sum of all
// measurements of the current batch.
static auto filterByLightSums(std::vector<L1A>& l1a_products) -> void
{
    const int n_l1a { static_cast<int>(l1a_products.size()) };
    const int npix { static_cast<int>(l1a_products.front().image.size()) };
    std::vector<double> light_sums(n_l1a, 0.0);
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        for (int i {}; i < npix; ++i) {
            if (!l1a_products[i_l1a].pixel_mask[i]) {
                light_sums[i_l1a] += l1a_products[i_l1a].image[i];
            }
        }
    }
    const double max_light_sum { *std::ranges::max_element(light_sums) };
    constexpr double light_sum_thres { 0.1 };
    auto l1a_first { l1a_products.begin() };
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        if (light_sums.at(i_l1a) > light_sum_thres * max_light_sum) {
            l1a_first = l1a_products.begin() + i_l1a;
            break;
        } else {
            spdlog::debug(
              "Discarding {}",
              shortenFilename(l1a_products.at(i_l1a).meta->filename));
        }
    }
    auto l1a_last { l1a_products.end() };
    for (int i_l1a { n_l1a - 1 }; i_l1a >= 0; --i_l1a) {
        if (light_sums.at(i_l1a) > light_sum_thres * max_light_sum) {
            l1a_last = l1a_products.begin() + i_l1a + 1;
            break;
        } else {
            spdlog::debug(
              "Discarding {}",
              shortenFilename(l1a_products.at(i_l1a).meta->filename));
        }
    }
    l1a_products = { l1a_first, l1a_last };
}

auto FovGen::processBatch(const int i_vp) -> void
{
    // Shorthand for all L1A products of this viewport
    auto& l1a { batches.at(i_vp).l1a_products };
    spdlog::info("Generating field of view CKD for viewport: {} deg",
                 viewport_idx_to_angle.at(l1a.front().meta->viewport));
    // Sort all L1A instances of this batch by ACT angle
    std::ranges::sort(l1a, [](const auto& a, const auto& b) {
        return a.meta->act_angle < b.meta->act_angle;
    });
    // Remove measurements that are not visible in this viewport
    filterByLightSums(l1a);
    spdlog::info("  Most negative ACT angle used: {} deg",
                 l1a.front().meta->act_angle / math::deg_to_rad);
    spdlog::info("  Most positive ACT angle used:  {} deg",
                 l1a.back().meta->act_angle / math::deg_to_rad);
    // Total number of measurements used from this batch after filtering
    const int n_l1a { static_cast<int>(l1a.size()) };
    // Prepare B-spline
    std::vector<double> knots(settings.fov.n_knots, 0.0);
    for (int i {}; i < settings.fov.n_knots; ++i) {
        knots[i] = static_cast<double>(i) * ckd.n_detector_cols
                   / (settings.fov.n_knots - 1);
    }
    // The knots should be slightly outside the data range
    knots.front() -= 1e-10;
    knots.back() += 1e-10;
    BSpline b_spline { settings.fov.bspline_order, knots };
    // x-values for constructing the B-spline matrix
    std::vector<double> x_data(ckd.fov.n_columns);
    std::iota(x_data.begin(),
              x_data.end(),
              static_cast<double>(settings.fov.exclude_columns_left));
    // B-spline basis matrix
    std::vector<double> B_mat {};
    b_spline.evalBasis(x_data, B_mat);
    // Standard deviations of the peak positions. This array store one
    // value per column for a given detector row.
    std::array<std::vector<double>, dims::pol> std_devs {};
    for (auto& std_dev : std_devs) {
        std_dev.resize(ckd.fov.n_columns);
    }
    // Polarization states are handled separately
    const int n_detector_rows_half { ckd.n_detector_rows / 2 };
#pragma omp parallel for firstprivate(b_spline, B_mat, std_devs)
    for (int i_l1a = 0; i_l1a < n_l1a; ++i_l1a) {
        if (const double p { getPercentage(i_l1a, n_l1a) }; p >= 0.0) {
            spdlog::info("  Fitting peaks along detector columns: {:5.1f}%", p);
        }
        // For the Gaussian fit, x-data are row (spatial) indices and
        // y-data are the pixel intensities.
        std::vector<double> x_data {};
        std::vector<double> y_data {};
        std::vector<double> noise {};
        x_data.reserve(n_detector_rows_half);
        y_data.reserve(n_detector_rows_half);
        noise.reserve(n_detector_rows_half);
        for (int i_col {}; i_col < ckd.fov.n_columns; ++i_col) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                x_data.clear();
                y_data.clear();
                noise.clear();
                for (int i_row { i_pol * n_detector_rows_half };
                     i_row < (i_pol + 1) * n_detector_rows_half;
                     ++i_row) {
                    const int idx { i_row * ckd.n_detector_cols + i_col
                                    + settings.fov.exclude_columns_left };
                    if (!l1a[i_l1a].pixel_mask[idx]) {
                        x_data.push_back(i_row);
                        y_data.push_back(l1a[i_l1a].image[idx]);
                        noise.push_back(l1a[i_l1a].noise[idx]);
                    }
                }
                if (x_data.empty()) {
                    throw std::runtime_error { "FOV has nothing to work with "
                                               "because all pixels are bad" };
                }
                gadfit::LMsolver solver {
                    [](const std::vector<gadfit::AdVar>& parameters,
                       const double x) {
                        // a * exp(-(x0-x)^2/sigma) + bgr
                        gadfit::AdVar diff { parameters[0] - x };
                        return parameters[2] * exp(-diff * diff / parameters[1])
                               + parameters[3];
                    }
                };
                solver.addDataset(x_data, y_data, noise);
                // Maximum pixel intensity across the column is set as
                // the initial guess for the Levenberg-Marquardt
                // solver.
                const auto max_loc { std::ranges::distance(
                  y_data.cbegin(), std::ranges::max_element(y_data)) };
                solver.setPar(0, x_data[max_loc], true); // x0
                solver.setPar(1, 1.0, true);             // sigma
                solver.setPar(2, y_data[max_loc], true); // a
                solver.setPar(3, 0.0, true);             // bgr
                solver.settings.iteration_limit =
                  settings.fov.fitting.iteration_limit;
                solver.settings.verbosity = gadfit::io::hide_all;
                solver.fit();
                ckd.fov.peak_pos[i_pol][ckd.fov.n_l1a_total + i_l1a][i_col] =
                  solver.getParValue(0);
                std_devs[i_pol][i_col] = solver.chi2()
                                         / solver.degreesOfFreedom()
                                         * solver.getInvJTJ().front();
                // Filter bad values. For instance, some columns close
                // to the edge of the detector produce unusually low
                // standard deviations.
                if (std::isnan(std_devs[i_pol][i_col])
                    || std_devs[i_pol][i_col] < 1e-10) {
                    std_devs[i_pol][i_col] = 1e30;
                }
            }
        }
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            std::vector<double> B_mat_copy { B_mat };
            std::vector<double> state_vector {};
            dppsv(ckd.fov.peak_pos[i_pol][ckd.fov.n_l1a_total + i_l1a],
                  std_devs[i_pol],
                  B_mat_copy,
                  state_vector);
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                ckd.fov.peak_pos_b[i_pol][ckd.fov.n_l1a_total + i_l1a][i] =
                  b_spline.deBoor(state_vector, static_cast<double>(i));
            }
        }
    }
    // Number of negative L1B ACT angles determined by the lowest ACT
    // angle in the measurement set (after filtering by light sums
    // above) and the user supplied ACT sampling.
    const int n_act_neg { static_cast<int>(
      -l1a.front().meta->act_angle / settings.fov.act_sampling.at(i_vp)) };
    // Number of positive L1B ACT angles
    const int n_act_pos { static_cast<int>(
      l1a.back().meta->act_angle / settings.fov.act_sampling.at(i_vp)) };
    // Total number is the sum plus one in the middle corresponding to ACT = 0
    ckd.fov.n_act_vp.at(i_vp) = n_act_neg + n_act_pos + 1;
    if (i_vp > 0) {
        ckd.fov.start_vp.at(i_vp) =
          ckd.fov.start_vp.at(i_vp - 1) + ckd.fov.n_act_vp.at(i_vp - 1);
    }
    spdlog::info("  Interpolating to the target L1B ACT angles ({})",
                 ckd.fov.n_act_vp.at(i_vp));
    // Generate a list of ACT angle for this viewport starting with
    // the most negative one.
    ckd.fov.act_angles.at(ckd.fov.start_vp.at(i_vp)) =
      -n_act_neg * settings.fov.act_sampling.at(i_vp);
    for (int i_act { 1 }; i_act < ckd.fov.n_act_vp.at(i_vp); ++i_act) {
        ckd.fov.act_angles.at(ckd.fov.start_vp.at(i_vp) + i_act) =
          ckd.fov.act_angles.at(ckd.fov.start_vp.at(i_vp))
          + i_act * settings.fov.act_sampling.at(i_vp);
    }
    // For the cubic spline fit, x-data are the ACT angles and y-data
    // are the peak positions along a given detector column.
    x_data.resize(n_l1a);
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        x_data[i_l1a] = l1a[i_l1a].meta->act_angle;
    }
#pragma omp parallel
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        std::vector<double> y_data(n_l1a);
#pragma omp for
        for (int i_col = 0; i_col < ckd.n_detector_cols; ++i_col) {
            for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                y_data[i_l1a] =
                  ckd.fov.peak_pos_b[i_pol][ckd.fov.n_l1a_total + i_l1a][i_col];
            }
            const CubicSpline spline { x_data, y_data };
            for (int i_act {}; i_act < ckd.fov.n_act_vp[i_vp]; ++i_act) {
                const int fov_idx { ckd.fov.start_vp[i_vp] + i_act };
                // This is where the primary FOV CKD quantity is constructed
                ckd.fov.row_indices[i_pol][fov_idx][i_col] =
                  spline.eval(ckd.fov.act_angles[fov_idx]);
            }
        }
    }
    ckd.fov.n_l1a_total += n_l1a;
}

auto FovGen::processFinalize() -> void
{
    // Now that we know the number of ACT angles of all viewports,
    // shrink all relevant arrays to their optimal sizes.
    ckd.fov.n_act =
      std::accumulate(ckd.fov.n_act_vp.cbegin(), ckd.fov.n_act_vp.cend(), 0);
    ckd.fov.dims_spec.resize(ckd.fov.n_act);
    ckd.fov.act_angles.resize(ckd.fov.n_act);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        ckd.fov.row_indices.at(i_pol).resize(ckd.fov.n_act);
        if (!ckd.fov.peak_pos[i_pol].empty()) {
            ckd.fov.peak_pos[i_pol].resize(ckd.fov.n_l1a_total);
            ckd.fov.peak_pos_b[i_pol].resize(ckd.fov.n_l1a_total);
        }
    }
    ckd.genFovIndices(settings.main.spectrum_width);
}

} // namespace spex
