// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "geometry.h"

#include "algorithm.h"
#include "dem.h"

#include <algorithm>
#include <cmath>

namespace spex {

auto fixAzimuthValues(const int row_beg,
                      const int row_end,
                      const int n_cols,
                      std::vector<float>& data) -> void
{
    const double min_val { *std::min_element(data.begin() + row_beg * n_cols,
                                             data.begin() + row_end * n_cols) };
    const double max_val { *std::max_element(data.begin() + row_beg * n_cols,
                                             data.begin() + row_end * n_cols) };
    if ((min_val < 0.0 ^ max_val < 0.0)
        && std::abs(max_val - min_val) > std::numbers::pi) {
        for (auto& val : data) {
            if (val < 0.0) {
                val += static_cast<float>(2 * std::numbers::pi);
            }
        }
    }
}

auto moduloAndConvertAngles(std::vector<float>& angles) -> void
{
    for (auto& angle : angles) {
        if (angle < 0.8 * def::l1bc_fill) {
            continue;
        }
        // This does nothing for latitudes
        if (angle > std::numbers::pi) {
            angle -= static_cast<float>(2 * std::numbers::pi);
        }
        if (angle < -std::numbers::pi) {
            angle += static_cast<float>(2 * std::numbers::pi);
        }
        angle /= static_cast<float>(math::deg_to_rad);
    }
}

auto removeQuaternionDiscontinuities(std::vector<Quaternion>& quaternions)
  -> void
{
    const int sc_records { static_cast<int>(quaternions.size()) };
    // Find the average difference of the first quaternion
    // component between spacecraft records
    double quat_diff_ave {};
    for (int i { 1 }; i < sc_records; ++i) {
        quat_diff_ave += std::abs(quaternions[i][0] - quaternions[i - 1][0]);
    }
    if (quaternions.size() > 1) {
        quat_diff_ave /= static_cast<double>(quaternions.size() - 1);
    }
    // If at any point the difference between records i and i-1 is
    // significantly greater than the average difference we have a
    // discontinuity. The discontinuity is not physical and is
    // fixed by flipping all four quaternion components.
    for (int i { 1 }; i < sc_records; ++i) {
        const double quat_diff { quaternions[i][0] - quaternions[i - 1][0] };
        if (std::abs(quat_diff) > 1e2 * quat_diff_ave) {
            // Flip the quaternion components for all remaining
            // records and continue with the main loop.
            for (int j { i }; j < sc_records; ++j) {
                quaternions[j][0] = -quaternions[j][0];
                quaternions[j][1] = -quaternions[j][1];
                quaternions[j][2] = -quaternions[j][2];
                quaternions[j].real() = -quaternions[j].real();
            }
        }
    }
}

auto interpolateQuaternion(const std::vector<double>& att_times,
                           const std::vector<Quaternion>& att_quaternions,
                           const double image_time,
                           Quaternion& quat) -> void
{
    // Find the first attitude time higher than the current time stamp
    const auto first_t_higher { std::ranges::find_if(
      att_times, [image_time](const double t) { return t > image_time; }) };
    // Index of the first attitude time lower than current time stamp
    const auto idx { std::distance(att_times.begin(), first_t_higher) - 1 };
    const double idx_delta { (image_time - att_times[idx])
                             / (att_times[idx + 1] - att_times[idx]) };
    quat = att_quaternions[idx];
    quat.slerp(att_quaternions[idx + 1], idx_delta);
}

auto cart2geo(const std::array<double, dims::vec>& xyz,
              double& lat,
              double& lon,
              double& alt) -> void
{
    // Longitude is easy
    lon = std::atan2(xyz[1], xyz[0]);
    // Latitude and height are computed using the algorithm of
    // Fukushima "Transformation from Cartesian to Geodetic
    // Coordinates Accelerated by Halley's Method", J. Geodesy 79, 689
    // (2006).
    static constexpr double ec { std::sqrt(1 - earth::e2) };
    static constexpr double c { earth::a * earth::e2 };
    const double p { std::sqrt(xyz[0] * xyz[0] + xyz[1] * xyz[1]) };
    const double zc { ec * std::abs(xyz[2]) };
    double Sn { std::abs(xyz[2]) };
    double Cn { ec * p };
    const double Sn_prev { Sn };
    // This algorithm is only run for one iteration
    const double SnCn { (Sn * Sn + Cn * Cn) };
    const double An { std::sqrt(SnCn) };
    const double cSnCn { c * Sn * Cn };
    const double Bn { 1.5 * cSnCn * ((p * Sn - zc * Cn) * An - cSnCn) };
    const double An3 { SnCn * An };
    const double Sn3 { Sn * Sn * Sn };
    const double Cn3 { Cn * Cn * Cn };
    Sn = (zc * An3 + c * Sn3) * An3 - Bn * Sn;
    Cn = (p * An3 - c * Cn3) * An3 - Bn * Cn;
    // End of iteration one
    const double Cc { ec * Cn };
    lat = std::atan(Sn / Cc);
    alt = (p * Cc + Sn_prev * Sn
           - earth::a * std::sqrt(ec * ec * Sn * Sn + Cc * Cc))
          / (std::sqrt(Sn * Sn + Cc * Cc));
    if (xyz[2] < 0.0) {
        lat = -lat;
    }
}

auto intersectEllipsoid(std::array<double, dims::vec> pointing,
                        std::array<double, dims::vec> sat,
                        std::array<double, dims::vec>& pos) -> void
{
    pointing[0] /= earth::a;
    pointing[1] /= earth::a;
    pointing[2] /= earth::b;
    sat[0] /= earth::a;
    sat[1] /= earth::a;
    sat[2] /= earth::b;
    const double ps { dotProduct(pointing, sat) };
    const double pp { dotProduct(pointing, pointing) };
    const double ss { dotProduct(sat, sat) };
    const double discriminant { ps * ps - pp * (ss - 1.0) };
    if (discriminant < 0.0) {
        throw std::runtime_error { "no intersection with the ellipsoid, the "
                                   "satellite is looking past Earth" };
    }
    const double d { (-ps - std::sqrt(discriminant)) / pp };
    for (int i {}; i < dims::vec; ++i) {
        pos[i] = sat[i] + d * pointing[i];
    }
    pos[0] *= earth::a;
    pos[1] *= earth::a;
    pos[2] *= earth::b;
}

auto intersectTerrain(const DEM& dem,
                      const std::array<double, dims::vec>& pointing,
                      std::array<double, dims::vec>& pos,
                      double& lat,
                      double& lon,
                      double& alt) -> void
{
    cart2geo(pos, lat, lon, alt);
    // Normal vector n at the target point
    const std::array<double, dims::vec> normal { std::cos(lat) * std::cos(lon),
                                                 std::cos(lat) * std::sin(lon),
                                                 std::sin(lat) };
    const double cosn { -dotProduct(pointing, normal) };
    double sign { -1.0 };
    // Highest point on Earth
    constexpr double max_jump { 10000.0 };
    double jump_size { max_jump / cosn };
    // Stop the algorithm if the altitude changes by less than this
    // distance in m.
    constexpr double target_accuracy { 1.0 };
    // Ideally the algorithm stops when the target accuracy is
    // reached. The maximum number of iterations is here to prevent an
    // infinite loop if something goes wrong. At this iteration level
    // the error is on the order of 1 cm.
    constexpr int max_iter { 20 };
    for (int iter {}; iter < max_iter; ++iter) {
        for (int i {}; i < dims::vec; ++i) {
            pos[i] += sign * jump_size * pointing[i];
        }
        const double alt_prev { alt };
        cart2geo(pos, lat, lon, alt);
        if (std::abs(alt_prev - alt) < target_accuracy) {
            break;
        }
        // The next jump direction is determined by whether the
        // current point is below or above the height given by the DEM
        // file.
        sign = (dem.getHeight(lat, lon).front() < alt ? 1.0 : -1.0);
        jump_size /= 2;
    }
}

auto solarAndViewingGeometry(const std::array<double, dims::vec>& sun,
                             const std::array<double, dims::vec>& pointing,
                             const double lat,
                             const double lon,
                             float& sza,
                             float& saa,
                             float& vza,
                             float& vaa) -> void
{
    // Normal vector of the target point
    const std::array<double, dims::vec> zenith { std::cos(lat) * std::cos(lon),
                                                 std::cos(lat) * std::sin(lon),
                                                 std::sin(lat) };
    // Normal vector in the east direction
    const std::array<double, dims::vec> east { -std::sin(lon),
                                               std::cos(lon),
                                               0 };
    // Normal vector in the north direction
    std::array<double, dims::vec> north {};
    crossProduct(zenith, east, north);
    // Solar geometry
    sza = static_cast<float>(std::acos(dotProduct(zenith, sun)));
    saa = static_cast<float>(
      std::atan2(dotProduct(east, sun), dotProduct(north, sun)));
    // Viewing geometry
    vza = static_cast<float>(std::acos(-dotProduct(zenith, pointing)));
    vaa = static_cast<float>(
      std::atan2(-dotProduct(east, pointing), -dotProduct(north, pointing)));
}

auto aolpShift(const Quaternion& q_sc,
               const std::array<double, dims::vec>& pointing,
               const std::array<double, dims::vec>& zenith,
               std::array<double, dims::vec> inst_plane_normal) -> double
{
    // The local meridian plane is defined as the plane containing the
    // local zenith and the spacecraft viewing direction.
    std::array<double, dims::vec> lmp_normal {};
    crossProduct(pointing, zenith, lmp_normal);
    normalizeVector(lmp_normal);
    // Transform the instrument reference plane normal to ECR
    q_sc.rotate(inst_plane_normal);
    std::array<double, dims::vec> lmp_x {};
    crossProduct(lmp_normal, pointing, lmp_x);
    std::array<double, dims::vec> inst_plane_x {};
    crossProduct(inst_plane_normal, pointing, inst_plane_x);
    normalizeVector(inst_plane_x);
    std::array<double, dims::vec> cross {};
    crossProduct(lmp_x, inst_plane_x, cross);
    const double dot_point_cross { dotProduct(pointing, cross) };
    const double dot_lmp_inst { dotProduct(lmp_normal, inst_plane_normal) };
    double sign { std::atan2(dot_point_cross, dot_lmp_inst) > 0.0 ? 1.0
                                                                  : -1.0 };
    return sign * std::acos(dot_lmp_inst)
           / std::sqrt(dotProduct(lmp_normal, lmp_normal))
           / std::sqrt(dotProduct(inst_plane_normal, inst_plane_normal));
}

} // namespace spex
