// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Algorithms related to geometrical operations

#pragma once

#include "constants.h"
#include "quaternion.h"

#include <numbers>

namespace spex {

class DEM;

// If the azimuth values cross the -180/180 boundary then add 2*pi to
// all negative values.
auto fixAzimuthValues(const int row_beg,
                      const int row_end,
                      const int n_cols,
                      std::vector<float>& data) -> void;

// Convert angles from radians to degrees. If dealing with longitudes,
// then first bring them to the -180/180 range.
auto moduloAndConvertAngles(std::vector<float>& angles) -> void;

// If the quaternions are set up such that the real part is always
// positive there could be a jump in the vector components the moment
// the real part touches or gets close to zero. This is handled by
// identifying where the jump is and flipping all quaternions beyond
// that point.
auto removeQuaternionDiscontinuities(std::vector<Quaternion>& quaternions)
  -> void;

// Given a series of timestamps and quaternions, find the quaternion
// corresponding to an arbitrary timestamp.
auto interpolateQuaternion(const std::vector<double>& att_times,
                           const std::vector<Quaternion>& att_quaternions,
                           const double image_time,
                           Quaternion& quat) -> void;

// Convert from cartesian to geodetic coordinates. Input
auto cart2geo(const std::array<double, dims::vec>& xyz,
              double& lat,
              double& lon,
              double& alt) -> void;

// Given a pointing vector (swath vector) and the satellite position
// in cartesian coordinates, return the point of intersection with the
// Earth ellipsoid in cartesian coordinates. Note that the pointing
// and satellite vector need to be rescaled which is why they are
// passed by value. Algorithm taken from Joint Polar Satellite System
// (JPPS) VIIRS Geolocation Algorithm Theoretical Basis Document
// (ATBD) section 3.3.2.2.1.
auto intersectEllipsoid(std::array<double, dims::vec> pointing,
                        std::array<double, dims::vec> sat,
                        std::array<double, dims::vec>& pos) -> void;

// Given a digital elevation model, a pointing vector (swath vector),
// the satellite position in cartesian coordinates, and the ellipsoid
// intersection point computed by intersectEllipsoid, return the
// latitude, longitude, and height (altitude) of the target
// point. Note that the intersection is modified by this function.
auto intersectTerrain(const DEM& dem,
                      const std::array<double, dims::vec>& pointing,
                      std::array<double, dims::vec>& pos,
                      double& lat,
                      double& lon,
                      double& alt) -> void;

// Given the sun and pointing (swath) vectors in ECR coordinates and
// the target point latitude and longitude, return the solar zenith
// and azimuth angles and the sensor viewing zenith and azimuth
// angles.
auto solarAndViewingGeometry(const std::array<double, dims::vec>& sun,
                             const std::array<double, dims::vec>& pointing,
                             const double lat,
                             const double lon,
                             float& sza,
                             float& saa,
                             float& vza,
                             float& vaa) -> void;

// Return the AoLP shift in radians needed to convert the polarization
// information from the instrument reference frame to the local
// meridian plane. Inputs are the spacecraft to ECR rotation
// quaternion, the pointing vector (direction of measurement), the
// local zenith vector, and the instrument reference plane normal.
auto aolpShift(const Quaternion& q_sc,
               const std::array<double, dims::vec>& pointing,
               const std::array<double, dims::vec>& zenith,
               std::array<double, dims::vec> inst_plane_normal) -> double;

} // namespace spex
