// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "inversion.h"

#include "lininv.h"

#include <cmath>
#include <cstring>
#include <spdlog/spdlog.h>

namespace spex {

Inversion::Inversion(size_t a_nstate, size_t a_nmeas)
{
    // Copy dimension size
    nstate = a_nstate;
    nmeas = a_nmeas;

    // The mask is the only one that has a default
    own_mask.resize(nmeas,false);
    mask = &own_mask; // Do not use data() for vector<bool>
}

// Setters
void Inversion::setApriori(
    double* x,
    bool copy
)
{
    if (copy) {
        own_apriori.resize(nstate);
        apriori = own_apriori.data();
        // Copy array with the size that is already known
        memcpy(apriori,x,nstate*sizeof(double));
    } else apriori = x;

}
void Inversion::setMeas(
    double* y,
    bool copy
)
{

    if (copy) {
        own_meas.resize(nmeas);
        meas = own_meas.data();
        // Copy array with the size that is already known
        memcpy(meas,y,nmeas*sizeof(double));
    } else meas = y;

}
void Inversion::setNoise(
    double* y_noise,
    bool s
)
{

    s_y.resize(nmeas);
    if (s) {
        for (size_t imeas=0 ; imeas<nmeas ; imeas++) s_y[imeas] = y_noise[imeas];
    } else {
        for (size_t imeas=0 ; imeas<nmeas ; imeas++) s_y[imeas] = std::pow(y_noise[imeas],2.0);
    }

}
void Inversion::removeNoise(
)
{
    s_y.resize(0);
}
void Inversion::setMask(
    std::vector<bool> *a_mask,
    bool copy
)
{

    if (copy) {
        own_mask = *a_mask; // Copy of the vector<bool>
        mask = &own_mask;
    } else mask = a_mask;

}
void Inversion::removeMask(
)
{
    for (size_t imeas=0 ; imeas<nmeas ; imeas++) own_mask[imeas] = false;
    mask = &own_mask;
}
void Inversion::setTol(
    double* t,
    bool copy
)
{

    if (copy) {
        own_tol.resize(nstate);
        tol = own_tol.data();
        // Copy array with the size that is already known
        memcpy(tol,t,nstate*sizeof(double));
    } else tol = t;

}

// Perform the inversion
auto Inversion::inversion_calculate(double* res) -> int
{
    // Construct local variables
    std::vector<double> mod(nmeas); // Modelled signal
    std::vector<double> residual(nmeas); // Measured minus model
    std::vector<double> jac(nmeas*nstate); // Jacobian
    std::vector<double> state(nstate); // Current state
    std::vector<double> state_des(nstate); // Desired state change

    // There may be an outlier protection mechanism in the linear inverison.
    // In that case, we keep track of all the measurement points that are
    // excluded on the way. It starts with the input pixel mask.
    std::vector<bool> excluded(*mask);

    // Go to a-priori state
    memcpy(state.data(),apriori,nstate*sizeof(double));

    // Run forward model
    fwd(state.data(),mod.data(),jac.data());
    for (size_t imeas=0 ; imeas<nmeas ; imeas++) residual[imeas] = meas[imeas]-mod[imeas];

    int conv = 1; // Not converged
    double step_reducer = initial_step_reducer;
    for (size_t iter=0 ; iter < maxiter ; iter++) {

        // Perform linear inversion algorithm
        int ierr;
        execute_linear_invert(jac.data(),residual.data(),&excluded,state_des.data());
        // The state_des is now a movement

        // Calculate initial chi squared using the mask provided by the inversion.
        double chi2_before = 0.0;
        if (s_y.size() == 0) {
            for (size_t imeas=0 ; imeas<nmeas ; imeas++) {
                if (!excluded[imeas]) chi2_before += pow(residual[imeas],2.0);
            }
        } else {
            for (size_t imeas=0 ; imeas<nmeas ; imeas++) {
                // In this inversion, the s_y can only be diagonal
                if (!excluded[imeas]) {
                    chi2_before += pow(residual[imeas],2.0)/s_y[imeas];
                }
            }
        }
        // No need to reduce

        // Apply retrieval to state
        for (size_t istate=0 ; istate<nstate ; istate++) state[istate] += (1.0/(1.0+step_reducer)) * state_des[istate];

        bool acc = true;
        int i_step {};
        while (true) { // Step control loop
            ++i_step;
            constexpr int step_control_iter_limit { 10000 };
            if (i_step > step_control_iter_limit) {
                spdlog::warn("Inversion step control loop not converged after "
                             "{} iterations",
                             step_control_iter_limit);
                break;
            }

            // Rerun forward model
            fwd(state.data(),mod.data(),jac.data());
            for (size_t imeas=0 ; imeas<nmeas ; imeas++) residual[imeas] = meas[imeas]-mod[imeas];

            // Calculate new chi squared using the same mask as during the inversion
            double chi2_after = 0.0;
            if (s_y.size() == 0) {
                for (size_t imeas=0 ; imeas<nmeas ; imeas++) {
                    if (!excluded[imeas]) chi2_after += pow(residual[imeas],2.0);
                }
            } else {
                for (size_t imeas=0 ; imeas<nmeas ; imeas++) {
                    if (!excluded[imeas]) chi2_after += pow(residual[imeas],2.0)/s_y[imeas];
                }
            }
            if (std::isnan(chi2_after)) return 2; // Numeric error occurred. Not a lethal program error.
            if (steptolerance == -999.0 || chi2_after <= chi2_before*steptolerance) {
                step_reducer *= reducerfactor_success;
                break;
            }

            acc = false; // First step was not accepted. No convergence can be triggered at this iteration.
            for (size_t istate=0 ; istate<nstate ; istate++) {
                // Undo step
                state[istate] -= (1.0/(1.0+step_reducer)) * state_des[istate];
            }
            // Update reducer
            step_reducer *= reducerfactor_fail;
            for (size_t istate=0 ; istate<nstate ; istate++) {
                state[istate] += (1.0/(1.0+step_reducer)) * state_des[istate];
            }
        }
        // Check for convergence
        // Direct accept
        if (acc && (reducer_limit == -999.0 || step_reducer < reducer_limit)) {
            double mag_diff_sq = 0.0;
            // The change of the state is (still) in state_des. It can be reduced by
            // a step reduction factor, but then the boolean acc had been false and
            // we would not be here.
            for (size_t istate=0 ; istate<nstate ; istate++) mag_diff_sq += pow(state_des[istate]/tol[istate],2.0);
            if (mag_diff_sq < 1.0) {
                conv = 0;
                break;
            }
        }

    }

    // Copy state to result
    for (size_t istate=0 ; istate<nstate ; istate++) res[istate] = state[istate];

    // This is zero if the loop is terminated by convergence, and one if the loop
    // was terminated by iterations.
    return conv;

}

// Default linear inversion. Can be overwritten by child classes.
auto Inversion::execute_linear_invert(double* jac,
                                      double* residual,
                                      std::vector<bool> *excluded,
                                      double* state_des) -> void
{
    squareoption_t opt_s_y = s_y.size() == 0?OPT_NONE:OPT_DIAG;
    linear_invert(nstate,nmeas,jac,opt_s_y,s_y.data(),excluded,residual,state_des);
}

} // namespace spex
