// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include <cstddef>
#include <vector>

namespace spex {

// A class for a iterative non-linear inversion.
// TODO: Expand this with more features like in Sicor or RemoTeC.
// This class is to be inherited from. Each inversion system has to define
// her own forward model and the desired additional parameters.
// So, they inherit from this system.
class Inversion
{
protected:
    size_t nstate = 0; // Number of state parameters
    size_t nmeas = 0; // Number of elements in the measurement vector
    std::vector<double> own_apriori {}; // A-priori state when transferring with copy
    double* apriori {}; // Pointer to apriori state
    std::vector<double> own_meas {}; // Measurement vector when transferring with copy
    double* meas {}; // Measurement vector
    std::vector<double> s_y {}; // Error covariance matrix (if used)
    std::vector<bool> own_mask {}; // Pixel mask, created with copy or by default
    std::vector<bool> *mask {}; // Pixel mask, eventually a pointer to its own mask, or to an external one
    std::vector<double> own_tol {}; // Tolerance for convergence in state vector when transfoerring with copy
    double* tol {}; // Tolerance for convergence in state vector
    size_t maxiter {}; // Option: Maximum number of iterations to reach convergence
    double steptolerance {}; // Maximum increment factor of chi square to accept a step
    double initial_step_reducer {}; // Starting value of the Levenberg-Marquardt step control parameter
    double reducerfactor_fail {}; // Multiplication of step reducer for a failed step
    double reducerfactor_success {}; // Multiplication of step reducer for a successful step
    double reducer_limit = -999.0; // Maximum step control reducer required for convergence

public:
    Inversion(size_t a_nstate, // Number of state parameters
              size_t a_nmeas); // Number of measurement
    virtual ~Inversion() {} // Do-nothing virtual destructor

    // Simple setters
    auto setMaxiter(size_t a_maxiter) -> void {maxiter = a_maxiter;}
    auto setSteptolerance(double a_steptolerance) -> void {steptolerance = a_steptolerance;}
    auto setInitialStepReducer(double a_initial_step_reducer) -> void {initial_step_reducer = a_initial_step_reducer;}
    auto setReducerfactorFail(double a_reducerfactor_fail) -> void {reducerfactor_fail = a_reducerfactor_fail;}
    auto setReducerfactorSuccess(double a_reducerfactor_success) -> void {reducerfactor_success = a_reducerfactor_success;}
    auto setReducerLimit(double a_reducer_limit) -> void {reducer_limit = a_reducer_limit;}

    // Less simple setters (memcpy)
    auto setApriori(double* x, bool copy=false) -> void;
    auto setMeas(double* y, bool copy=false) -> void;

    // s is flag for providing S-elements (no need to square anymore)
    auto setNoise(double* y_noise,bool s=false) -> void;
    auto removeNoise() -> void;
    auto setMask(std::vector<bool> *a_mask,bool copy=false) -> void;
    auto removeMask() -> void;
    auto setTol(double* t, bool copy=false) -> void;

protected:
    virtual auto fwd(double* state, // State vector
                     double* signal, // Modelled signal (outout)
                     double* jacobian) -> void = 0; // Modelled jacobian (outout)

    // The main function. Perform the inversion.
public:
    auto inversion_calculate(double* res) -> int; // Retrieved state (output)

    // Execution of linear inverison
    virtual auto execute_linear_invert(double* jac,
                                       double* residual,
                                       std::vector<bool> *excluded,
                                       double* state_des) -> void;

    // Simple getter
    auto getNstate() -> int {return nstate;}
};

} // namespace spex
