// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "io.h"

#include "constants.h"

#include <filesystem>
#include <fstream>
#include <numeric>
#include <omp.h>
#include <spdlog/sinks/stdout_color_sinks.h>
#include <spdlog/spdlog.h>
#include <unistd.h>

namespace spex {

auto spex_formatter_flag::format(const spdlog::details::log_msg& log_msg,
                                 const std::tm& /* tm_time */,
                                 spdlog::memory_buf_t& dest) -> void
{
    std::string text {};
    switch (log_msg.level) {
    case spdlog::level::info:
        break;
    case spdlog::level::warn:
        text = " [warning]";
        break;
    case spdlog::level::err:
        text = " [error]";
        break;
    case spdlog::level::debug:
        text = " [debug]";
        break;
    default:
        text = " [unknown]";
    }
    // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-pointer-arithmetic)
    dest.append(text.data(), text.data() + text.size());
}

auto spex_formatter_flag::clone() const
  -> std::unique_ptr<custom_flag_formatter>
{
    return spdlog::details::make_unique<spex_formatter_flag>();
}

auto initLogging() -> void
{
    // Only if the logger does not already exist
    if (!spdlog::get("plain")) {
        auto formatter { std::make_unique<spdlog::pattern_formatter>() };
        formatter->add_flag<spex_formatter_flag>('*').set_pattern(
          "[%H:%M:%S]%* %v");
        spdlog::set_formatter(std::move(formatter));
        spdlog::stdout_color_mt("plain");
        spdlog::get("plain")->set_pattern("%v");
    }
    setLoggerLevel(false);
}

auto setLoggerLevel(const bool use_debug_level) -> void
{
    if (use_debug_level) {
        spdlog::set_level(spdlog::level::debug);
    } else {
        spdlog::set_level(spdlog::level::info);
    }
}

auto printHeading(const std::string& heading, const bool incl_empty_line)
  -> void
{
    if (incl_empty_line) {
        spdlog::get("plain")->info("");
    }
    std::string hash_line(heading.size() + 4, '#');
    spdlog::get("plain")->info(hash_line);
    spdlog::get("plain")->info("# " + heading + " #");
    spdlog::get("plain")->info(hash_line);
}

[[nodiscard]] auto getPercentage(const int iteration,
                                 const size_t work_size,
                                 const double step_size_perc) -> double
{
    if (omp_get_thread_num() != 0) {
        return -1.0;
    }
    const int step_size { static_cast<int>(
      std::ceil(1e-2 * step_size_perc * work_size / omp_get_num_threads())) };
    if (step_size == 0 || iteration % step_size == 0) {
        return 1e2 * iteration * omp_get_num_threads() / work_size;
    }
    return -1.0;
}

auto shortenFilename(const std::string& filename) -> std::string
{
    constexpr int substring_length { 70 };
    if (static_cast<int>(filename.size()) <= substring_length) {
        return filename;
    }
    return ".."
           + filename.substr(filename.size() - substring_length,
                             filename.size() - 1);
}

auto checkPresenceOfFile(const std::string& filename, const bool required)
  -> void
{
    if (!required && filename.empty()) {
        return;
    }
    try {
        std::ifstream file { filename };
        file.exceptions(std::ifstream::failbit);
    } catch (const std::ifstream::failure& e) {
        const std::string msg { "\nCould not open file: " + filename };
        throw std::ifstream::failure { e.what() + msg, e.code() };
    }
}

auto checkFileWritable(const std::string& filename) -> void
{
    if (!filename.empty()) {
        std::string dir { std::filesystem::path { filename }.parent_path() };
        if (!dir.empty() && static_cast<bool>(access(dir.c_str(), W_OK))) {
            throw std::runtime_error { filename + " is not writable" };
        }
    }
}

auto extractFilesFromList(const std::string& filename,
                          std::vector<std::string>& list,
                          const bool required) -> void
{
    checkPresenceOfFile(filename, required);
    std::ifstream file { filename };
    std::string line {};
    while (std::getline(file, line)) {
        list.push_back(line);
    }
}

auto splitString(const std::string& list, const char delimiter)
  -> std::vector<std::string>
{
    std::stringstream ss { list };
    std::string name {};
    std::vector<std::string> strings {};
    while (getline(ss, name, delimiter)) {
        strings.push_back(name);
    }
    return strings;
}

auto copyNavigationData(NcObject& nc_l1a, NcObject& nc_l1b) -> void
{
    nc_l1a.reOpen();
    netCDF::NcGroup nav_l1a { nc_l1a.getGroup("navigation_data") };
    auto nav_l1b { nc_l1b.addGroup("navigation_data") };
    // Dimensions
    size_t SC_records {};
    try {
        SC_records = nc_l1a.getDim("SC_records").getSize();
    } catch (const netCDF::exceptions::NcBadId&) {
        SC_records = nav_l1a.getDim("att_time").getSize();
    }
    const auto nc_SC_records { nc_l1b.addDim("SC_records", SC_records) };
    const auto nc_quat { nc_l1b.addDim("quaternion_elements", dims::quat) };
    const auto nc_vec { nc_l1b.addDim("vector_elements", dims::vec) };
    // Work variables for NetCDF variables
    netCDF::NcVar var_in {};
    netCDF::NcVar var_out {};
    std::vector<double> buf {};
    // Global attributes of orbit data
    for (const auto& [name, att] : nc_l1a.getAtts()) {
        std::string buffer {};
        if (name == "startDirection") {
            att.getValues(buffer);
            nc_l1b.putAtt(name, buffer);
        } else if (name == "endDirection") {
            att.getValues(buffer);
            nc_l1b.putAtt(name, buffer);
        } else if (name == "time_coverage_start") {
            att.getValues(buffer);
            nc_l1b.putAtt(name, buffer);
        } else if (name == "time_coverage_stop") {
            att.getValues(buffer);
            nc_l1b.putAtt(name, buffer);
        } else if (name == "reference") {
            // Reference on which image time is based
            att.getValues(buffer);
            nc_l1b.putAtt(name, buffer);
        } else if (name == "orbit_number") {
            long int orbit_number {};
            att.getValues(&orbit_number);
            nc_l1b.putAtt(name, netCDF::ncInt64, orbit_number);
        }
    }

    const auto copy { [&](const netCDF::NcVar& var_in, netCDF::NcVar& var_out) {
        std::vector<size_t> dims {};
        for (const auto& dim : var_in.getDims()) {
            dims.push_back(dim.getSize());
        }
        size_t total_size { std::accumulate(dims.cbegin(),
                                            dims.cend(),
                                            static_cast<size_t>(1),
                                            std::multiplies<size_t>()) };
        std::vector<double> buf(total_size);
        var_in.getVar(buf.data());
        var_out.putVar(buf.data());
    } };

    var_in = nav_l1a.getVar("att_time");
    var_out = NcObject::copyVar(var_in, nav_l1b, { nc_SC_records });
    copy(var_in, var_out);

    var_in = nav_l1a.getVar("att_quat");
    var_out = NcObject::copyVar(var_in, nav_l1b, { nc_SC_records, nc_quat });
    copy(var_in, var_out);

    var_in = nav_l1a.getVar("orb_time");
    var_out = NcObject::copyVar(var_in, nav_l1b, { nc_SC_records });
    copy(var_in, var_out);

    var_in = nav_l1a.getVar("orb_pos");
    var_out = NcObject::copyVar(var_in, nav_l1b, { nc_SC_records, nc_vec });
    copy(var_in, var_out);

    var_in = nav_l1a.getVar("orb_vel");
    var_out = NcObject::copyVar(var_in, nav_l1b, { nc_SC_records, nc_vec });
    copy(var_in, var_out);

    nc_l1a.close();
}

auto copyImageAttributes(const int image_start,
                         const int image_end,
                         NcObject& nc_l1a,
                         NcObject& nc_l1b) -> void
{
    nc_l1a.reOpen();
    const auto grp_l1a { nc_l1a.getGroup("image_attributes") };
    auto grp_l1b { nc_l1b.addGroup("image_attributes") };
    const int number_of_images { image_end - image_start + 1 };
    const auto nc_number_of_images { grp_l1b.addDim("number_of_images",
                                                    number_of_images) };
    // Work variables for NetCDF variables
    netCDF::NcVar var_in {};
    netCDF::NcVar var_out {};
    const std::vector<size_t> start { static_cast<size_t>(image_start) };
    const std::vector<size_t> count { static_cast<size_t>(number_of_images) };
    std::vector<uint32_t> buf_u32(number_of_images);
    var_in = grp_l1a.getVar("icu_time_sec");
    var_out = NcObject::copyVar(var_in, grp_l1b, { nc_number_of_images });
    var_in.getVar(start, count, buf_u32.data());
    var_out.putVar(buf_u32.data());
    std::vector<uint16_t> buf_u16(number_of_images);
    var_in = grp_l1a.getVar("icu_time_subsec");
    var_out = NcObject::copyVar(var_in, grp_l1b, { nc_number_of_images });
    var_in.getVar(start, count, buf_u16.data());
    var_out.putVar(buf_u16.data());
    nc_l1a.close();
}

} // namespace spex
