// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Functions for formatting the output to stdout, for manipulating
// input/output files, and tests for things like whether a file is
// readable/writable.

#pragma once

#include "nc_object.h"

#include <spdlog/pattern_formatter.h>
#include <string>
#include <vector>

namespace spex {

// Define a new spdlog formatter flag. The primary purpose is to show
// labels such as [warning] for warnings but no label for regular
// (info) messages.
class spex_formatter_flag : public spdlog::custom_flag_formatter
{
public:
    auto format(const spdlog::details::log_msg& log_msg,
                const std::tm&,
                spdlog::memory_buf_t& dest) -> void override;
    auto clone() const -> std::unique_ptr<custom_flag_formatter> override;
};

// Set up two loggers: one with a verbose pattern (mostly used
// throughout the code) and a plain one (clean pattern, i.e. prints
// just the message text).
auto initLogging() -> void;

// Whether to set logging level to debug (else info level)
auto setLoggerLevel(const bool use_debug_level) -> void;

// Print the name of a major section at its beginning. For example,
// the CKD initialization would start with
//
// ######################
// # CKD initialization #
// ######################
auto printHeading(const std::string& heading, const bool incl_empty_line = true)
  -> void;

// Return the percentage of work done by the master process.
// 'iteration' is the iteration number of the total loop (for instance
// an omp parallel for) and work_size is the total number of
// iterations. 'step_size_perc', if given, returns a positive number
// only if step_size_perc percent has passed since the last call.
[[nodiscard]] auto getPercentage(const int iteration,
                                 const size_t work_size,
                                 const double step_size_perc = 0.0) -> double;

// Helper function for dealing with long filenames in the output
auto shortenFilename(const std::string& filename) -> std::string;

// Check if filename exists. If required == true and filename is an
// empty string raise an error, otherwise return without checking.
auto checkPresenceOfFile(const std::string& filename, const bool required)
  -> void;

auto checkFileWritable(const std::string& filename) -> void;

// This function is useful for things like a list of L1A files which
// can be stored in a separate file to avoid clutter in the main
// configuration file.
auto extractFilesFromList(const std::string& filename,
                          std::vector<std::string>& list,
                          const bool required) -> void;

auto splitString(const std::string& list, const char delimiter)
  -> std::vector<std::string>;

// Copy the navigation data over from an L1A file
auto copyNavigationData(NcObject& nc_l1a, NcObject& nc_l1b) -> void;

// Copy selected image attributes over from an L1A file
auto copyImageAttributes(const int image_start,
                         const int image_end,
                         NcObject& nc_l1a,
                         NcObject& nc_l1b) -> void;

} // namespace spex
