// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "isrf.h"

#include "arb_special_functions.h"

#include <netcdf>

namespace spex {

ISRF::ISRF(const std::string& filename)
{
    netCDF::NcFile nc_isrf { filename, netCDF::NcFile::read };
    const auto n_fov { nc_isrf.getDim("spatial_samples_per_image").getSize() };
    const auto n_spec { nc_isrf.getDim("spectral_detector_pixels").getSize() };
    std::vector<double> buf(n_fov * n_spec);
    // Read and reshape each of the four ISRF parameters
    for (auto& [name, var] : std::vector<
           std::pair<std::string, std::vector<std::vector<double>>&>> {
           { { "scale_a", scale_a },
             { "shape_m", shape_m },
             { "skew", skew },
             { "radius_r", radius_r } } }) {
        var.resize(n_fov, std::vector<double>(n_spec));
        nc_isrf.getVar(name).getVar(buf.data());
        for (int i_fov {}; i_fov < n_fov; ++i_fov) {
            std::copy(buf.begin() + i_fov * n_spec,
                      buf.begin() + (i_fov + 1) * n_spec,
                      var[i_fov].begin());
        }
    }
}

[[nodiscard]] auto ISRF::eval(const int i_fov,
                              const int i_spec,
                              const double arg) const -> double
{
    return pearsonIV(arg,
                     scale_a[i_fov][i_spec],
                     shape_m[i_fov][i_spec],
                     skew[i_fov][i_spec],
                     radius_r[i_fov][i_spec]);
}

} // namespace spex
