// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a.h"

#include "fourier.h"
#include "geometry.h"
#include "l1a_metadata.h"
#include "l1x.h"
#include "spex_settings.h"
#include "timer.h"

#include <span>

namespace spex {

L1A::L1A(const bool is_background) : is_background { is_background }
{
    l1x.resize(static_cast<int>(ProcLevel::n_levels));
}

auto L1A::initImageAttributes(const CKD& ckd) -> void
{
    // In diagnostic mode, image_offset is 0 and number_of_images is 1
    // or more. In science mode, image_offset is 0 or higher and
    // number_of_images is 1. That is not the number of images in the
    // full L1A product but the number of frames that make up one
    // flight image.
    const std::vector<double> image_times {
        meta->image_times.cbegin() + image_offset,
        meta->image_times.cbegin() + image_offset + meta->number_of_images
    };
    const std::vector<uint32_t> icu_times {
        meta->icu_times.cbegin() + image_offset,
        meta->icu_times.cbegin() + image_offset + meta->number_of_images
    };
    const std::vector<uint16_t> icu_times_subsec {
        meta->icu_times_subsec.cbegin() + image_offset,
        meta->icu_times_subsec.cbegin() + image_offset + meta->number_of_images
    };
    const std::vector<uint8_t> binning_tables {
        meta->binning_tables.cbegin() + image_offset,
        meta->binning_tables.cbegin() + image_offset + meta->number_of_images
    };
    const std::vector<int16_t> digital_offsets {
        meta->digital_offsets.cbegin() + image_offset,
        meta->digital_offsets.cbegin() + image_offset + meta->number_of_images
    };
    const std::vector<uint16_t> nr_coadditions_values {
        meta->nr_coadditions.cbegin() + image_offset,
        meta->nr_coadditions.cbegin() + image_offset + meta->number_of_images
    };
    const std::vector<double> exposure_times {
        meta->exposure_times.cbegin() + image_offset,
        meta->exposure_times.cbegin() + image_offset + meta->number_of_images
    };
    // The binning table should be consistent within an image
    binning_table_id = meta->binning_tables.front();
    // Coadd image attributes
    const auto coadd { [&nr_coadditions_values](const auto& arr) {
        return std::transform_reduce(arr.begin(),
                                     arr.end(),
                                     nr_coadditions_values.begin(),
                                     0.0,
                                     std::plus<>(),
                                     std::multiplies<>());
    } };
    exposure_time = coadd(exposure_times);
    digital_offset_single = digital_offsets.front();
    digital_offset = static_cast<double>(coadd(digital_offsets));
    nr_coadditions = std::accumulate(
      nr_coadditions_values.begin(), nr_coadditions_values.end(), 0);
    // Some of the coadded quantities should be averaged
    exposure_time /= nr_coadditions;
    // Time is defined as the beginning of the first exposure
    image_time = image_times.front();
    seconds = static_cast<uint32_t>((static_cast<long int>(icu_times.front())
                                     + static_cast<long int>(icu_times.back()))
                                    / 2);
    second_fraction = (static_cast<double>(icu_times_subsec.front())
                       + static_cast<double>(icu_times_subsec.back()))
                      / 2 / 65535;
    // Add a coaddition correction to frames where the coaddition
    // factor is greater than 1. This is only necessary when not using
    // background (dark) measurements.
    //  SRON-SPEX-TN-2020-001, issue 0.5, Table 10
    constexpr double c_coadd { 1.13 };
    for (const auto& coadd_factor : nr_coadditions_values) {
        if (coadd_factor > 1) {
            // Since it is an additive factor like the digital offset,
            // use the same variable for both.
            digital_offset += c_coadd * coadd_factor;
        }
    }
    // Calculate the jump pixel for applying dark offset when not using
    // background (dark) measurements
    // Time for reading up to the jump row in seconds
    constexpr double t_jump { 0.2124 };
    // Master clock period is seconds
    constexpr double t_mcp { 0.1e-6 };
    // FOT register value
    constexpr int r_fot { 20 };
    jump_row = static_cast<int>(std::round(
      meta->n_ch / 16.0 * ((t_jump - exposure_time) / (129.0 * t_mcp) - r_fot)
      - 3));
    jump_row = std::min(ckd.n_detector_cols - 1, std::max(0, jump_row));
}

auto L1A::processNavigationData() -> void
{
    if (meta->l1x_level > ProcLevel::l1a) {
        att_quat = meta->nav_att_quat[image_offset];
        satellite_position = meta->nav_orb_pos[image_offset];
        return;
    }
    if (image_time < meta->nav_att_time.front()
        || image_time > meta->nav_att_time.back()) {
        throw std::runtime_error {
            "Image time is not within the range of the attitude time axis"
        };
    }
    if (meta->sc_records == 1) {
        att_quat = meta->nav_att_quat.front();
        satellite_position = meta->nav_orb_pos.front();
        return;
    }
    interpolateQuaternion(
      meta->nav_att_time, meta->nav_att_quat, image_time, att_quat);
    for (int i {}; i < dims::vec; ++i) {
        satellite_position[i] = meta->pos_splines[i]->eval(image_time);
    }
}

// Bin raw data according to the ratio [main][detector_spec] /
// [main][detector_spec_reduced] in the spectral dimension and
// similarly in the spatial dimension. Here the data are almost raw -
// only the digital offset is applied after reading in images.
static auto binImage(const CKD& ckd,
                     const std::span<uint16_t>& image,
                     std::vector<double>& image_binned) -> void
{
    const int bin_spec { ckd.n_detector_cols_full / ckd.n_detector_cols };
    const int bin_spat { ckd.n_detector_rows_full / ckd.n_detector_rows };
    std::ranges::fill(image_binned, 0.0);
    for (int i {}; i < ckd.n_detector_rows; ++i) {
        for (int j {}; j < ckd.n_detector_cols; ++j) {
            for (int k {}; k < bin_spat; ++k) {
                for (int l {}; l < bin_spec; ++l) {
                    image_binned[i * ckd.n_detector_cols + j] +=
                      static_cast<double>(image[(i * bin_spat + k) * bin_spec
                                                  * ckd.n_detector_cols
                                                + j * bin_spec + l]);
                }
            }
        }
    }
}

auto L1A::readImage(const CKD& ckd) -> void
{
    // In L1B we are reading from the same file over an over again and
    // it is expected to be open at all times. For other processes
    // light swath vectors there are too many files to be open at
    // once. They need to be opened one by one when it is reading
    // time. If L1A files were requested to be held in memory
    // (frames_u16) then do not reopen the file.
    if (!meta->nc->isOpen() && frames_u16.empty()) {
        meta->nc->reOpen();
    }
    // Allocate the main work variables
    image.resize(binning_table->nBins());
    noise.resize(binning_table->nBins());
    pixel_mask.resize(binning_table->nBins());
    for (int i {}; i < binning_table->nBins(); ++i) {
        pixel_mask[i] = ckd.pixel_mask[i];
    }
    // Whether images should be immediately binned after reading. This
    // is done only if the user defined one of both of the reduced
    // dimensions in the [main] section.
    const bool bin_images { ckd.n_detector_cols_full > 0
                              && ckd.n_detector_cols != ckd.n_detector_cols_full
                            || ckd.n_detector_rows_full > 0
                                 && ckd.n_detector_rows
                                      != ckd.n_detector_rows_full };
    if (meta->l1x_level == ProcLevel::l1a) {
        // Need size_t here because of a potentially large allocation
        const size_t npix { static_cast<size_t>(meta->npix) };
        if (frames_u16.empty()) {
            const std::vector<size_t> start { static_cast<size_t>(image_offset),
                                              0 };
            const std::vector<size_t> count {
                static_cast<size_t>(meta->number_of_images),
                npix,
            };
            frames_u16.resize(meta->number_of_images * npix);
#pragma omp critical
            meta->nc->getGroup("science_data")
              .getVar("detector_images")
              .getVar(start, count, frames_u16.data());
        }
        std::vector<std::span<uint16_t>> frames(meta->number_of_images);
        for (int i {}; i < meta->number_of_images; ++i) {
            frames[i] = { frames_u16.begin() + i * npix,
                          frames_u16.begin() + (i + 1) * npix };
        }
        // Start with minus the total digital offset which is already
        // summed over images but not yet over binned pixels. Digital
        // offset can vary between different images (usually not).
        if (!meta->keep_frames) {
            for (int i {}; i < binning_table->nBins(); ++i) {
                image[i] = -digital_offset * binning_table->binSize(i);
            }
            for (int i_img {}; i_img < meta->number_of_images; ++i_img) {
                if (bin_images) {
                    std::vector<double> image_binned(ckd.npix);
                    binImage(ckd, frames[i_img], image_binned);
                    for (int i {}; i < ckd.npix; ++i) {
                        image[i] += image_binned[i];
                    }
                } else {
                    for (int i {}; i < binning_table->nBins(); ++i) {
                        image[i] += frames[i_img][i];
                    }
                }
            }
            for (int i {}; i < binning_table->nBins(); ++i) {
                image[i] /= nr_coadditions * binning_table->binSize(i);
            }
        } else {
            this->frames.resize(meta->number_of_images,
                                std::vector<double>(binning_table->nBins()));
            // Digital offset per single frame
            digital_offset /= meta->number_of_images;
            for (int i_img {}; i_img < meta->number_of_images; ++i_img) {
                if (bin_images) {
                    std::vector<double> frame_binned(ckd.npix);
                    binImage(ckd, frames[i_img], frame_binned);
                    for (int i {}; i < binning_table->nBins(); ++i) {
                        this->frames[i_img][i] =
                          frame_binned[i] - digital_offset;
                    }
                } else {
                    for (int i {}; i < ckd.npix; ++i) {
                        this->frames[i_img][i] =
                          static_cast<double>(frames[i_img][i])
                          - digital_offset;
                    }
                }
            }
        }
        frames_u16 = std::vector<uint16_t> {};
    } else if (meta->l1x_level <= ProcLevel::stray) {
        const std::vector<size_t> start { static_cast<size_t>(image_offset),
                                          0 };
        const std::vector<size_t> count { 1, image.size() };
        std::vector<uint8_t> pixel_mask_u8(binning_table->nBins());
#pragma omp critical
        {
            const auto grp { meta->nc->getGroup("science_data") };
            grp.getVar("detector_images").getVar(start, count, image.data());
            grp.getVar("detector_noise").getVar(start, count, noise.data());
            grp.getVar("pixel_mask").getVar(start, count, pixel_mask_u8.data());
        }
        for (int i {}; i < ckd.npix; ++i) {
            pixel_mask[i] =
              pixel_mask[i] || static_cast<bool>(pixel_mask_u8[i]);
        }
    } else if (meta->l1x_level < ProcLevel::l1b) {
        l1x_spectra.resize(ckd.fov.n_act);
        std::vector<int> spectra_sizes(ckd.fov.n_act);
#pragma omp critical
        meta->nc->getGroup("science_data")
          .getVar("spectra_sizes")
          .getVar(spectra_sizes.data());
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                const std::vector<size_t> start { l1x_image_offset,
                                                  static_cast<size_t>(i_fov),
                                                  static_cast<size_t>(i_pol),
                                                  0 };
                const std::vector<size_t> count {
                    1, 1, 1, static_cast<size_t>(spectra_sizes[i_fov])
                };
                l1x_spectra[i_fov].signal[i_pol].resize(spectra_sizes[i_fov]);
                l1x_spectra[i_fov].noise[i_pol].resize(spectra_sizes[i_fov]);
                l1x_spectra[i_fov].mask[i_pol].resize(spectra_sizes[i_fov]);
                std::vector<uint8_t> mask_u8(spectra_sizes[i_fov]);
#pragma omp critical
                {
                    const auto grp { meta->nc->getGroup("science_data") };
                    grp.getVar("spectra").getVar(
                      start, count, l1x_spectra[i_fov].signal[i_pol].data());
                    grp.getVar("spectra_noise")
                      .getVar(
                        start, count, l1x_spectra[i_fov].noise[i_pol].data());
                    grp.getVar("spectra_mask")
                      .getVar(start, count, mask_u8.data());
                }
                for (int i {}; i < spectra_sizes[i_fov]; ++i) {
                    l1x_spectra[i_fov].mask[i_pol][i] =
                      static_cast<bool>(mask_u8[i]);
                }
            }
        }
    } else {
        throw std::runtime_error { "attempting to read an L1A file with an "
                                   "unknown L1X level" };
    }
}

auto L1A::rowDependentOffset(const SPEXSettings& settings,
                             const ProcLevel proc_level,
                             const CKD& ckd) -> void
{
    const int& dark_cols_left { settings.dark_cols_left.at(
      static_cast<int>(proc_level)) };
    const int& dark_cols_right { settings.dark_cols_right.at(
      static_cast<int>(proc_level)) };
    if (dark_cols_left + dark_cols_right == 0
        || meta->l1x_level >= ProcLevel::nonlin) {
        return;
    }
    image_buf.assign(image.size(), 0.0);
    std::vector<double> edge_pixels(dark_cols_left + dark_cols_right);
    for (int i_row {}; i_row < ckd.n_detector_rows; ++i_row) {
        if (binning_table->skipRow(i_row)) {
            continue;
        }
        const int row_idx { i_row * ckd.n_detector_cols };
        int edge_pixel_idx {};
        for (int i_col {}; i_col < dark_cols_left; ++i_col) {
            const int idx { binning_table->binIndex(row_idx + i_col) };
            if (!pixel_mask[idx]) {
                edge_pixels[edge_pixel_idx] = image[idx];
                ++edge_pixel_idx;
            }
        }
        for (int i_col {}; i_col < dark_cols_right; ++i_col) {
            const int idx { binning_table->binIndex(
              row_idx + ckd.n_detector_cols - 1 - i_col) };
            if (!pixel_mask[idx]) {
                edge_pixels[edge_pixel_idx] = image[idx];
                ++edge_pixel_idx;
            }
        }
        const double med { median(edge_pixels) };
        for (int i_col {}; i_col < ckd.n_detector_cols; ++i_col) {
            const int idx { binning_table->binIndex(row_idx + i_col) };
            image_buf[idx] += (image[idx] - med) / binning_table->binSize(idx);
        }
    }
    image = std::move(image_buf);
}

// Smooth over bad values. This is necessary for stray light which
// uses all the image pixels.
static auto fillHoles(const std::vector<bool>& pixel_mask,
                      std::vector<double>& image) -> void
{
    for (int i {}; i < static_cast<int>(image.size()); ++i) {
        if (pixel_mask[i]) {
            // Unless we are at either end of the image array, take
            // the average of the neighboring values.
            if (i == 0) {
                image.front() = image[1];
            } else if (i == static_cast<int>(image.size() - 1)) {
                image.back() = image[static_cast<int>(image.size()) - 2];
            } else {
                image[i] = (image[i - 1] + image[i + 1]) / 2;
            }
            // If there are two or more bad pixels side by side we
            // could still end up with a bad value after
            // averaging. Then we just set it to zero. To test it we
            // use a large value that is close to the NetCDF float
            // fill value.
            constexpr double bad_value { 1e30 };
            if (std::isnan(image[i]) || std::abs(image[i]) > bad_value) {
                image[i] = 0.0;
            }
        }
    }
}

auto L1A::calibrateStrayLight(const CKD& ckd, const int n_van_cittert) -> void
{
    // Number of pixels in an unbinned image. If the kernels were
    // reduced then this is lower than the actual detector dimensions.
    const int unbin_npix { ckd.stray.n_spatial * ckd.stray.n_spectral };
    // Handle large (bad) values of the binned image.
    fillHoles(pixel_mask, image);
    // Unbin the image using the stray light binning table
    std::vector<double> image_unbin(unbin_npix);
    binning_table_stray->unbin(image, image_unbin);
    // "Ideal" image, i.e. the one without stray light
    std::vector<double> image_ideal { image_unbin };
    // Part of the unbinned detector image
    std::vector<double> sub_image {};
    // Result of convoling one kernel with a subimage
    std::vector<double> conv_result_sub {};
    // Allocate the image FFT once to save time
    std::vector<std::complex<double>> image_fft(
      *std::ranges::max_element(ckd.stray.kernel_fft_sizes));
    // Result of taking the whole convolution
    std::vector<double> conv_result(unbin_npix);
    // Van Cittert algorithm
    for (int i_vc {}; i_vc < n_van_cittert; ++i_vc) {
        std::ranges::fill(conv_result, 0.0);
        for (int i_kernel {}; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
            // Number of rows and column in this subimage
            const int image_n_rows {
                ckd.stray.edges[i_kernel * box::n + box::t]
                - ckd.stray.edges[i_kernel * box::n + box::b]
            };
            const int image_n_cols {
                ckd.stray.edges[i_kernel * box::n + box::r]
                - ckd.stray.edges[i_kernel * box::n + box::l]
            };
            // Each iteration starts with image_ideal as the best
            // current estimate.
            sub_image.resize(image_n_rows * image_n_cols);
            for (int i {}; i < image_n_rows; ++i) {
                for (int j {}; j < image_n_cols; ++j) {
                    const int idx {
                        (i + ckd.stray.edges[i_kernel * box::n + box::b])
                          * ckd.stray.n_spectral
                        + j + ckd.stray.edges[i_kernel * box::n + box::l]
                    };
                    sub_image[i * image_n_cols + j] =
                      image_ideal[idx] * ckd.stray.weights[i_kernel][idx];
                }
            }
            // Result of taking a convolution using one of the
            // subimages and kernels
            convolve(image_n_rows,
                     image_n_cols,
                     sub_image,
                     ckd.stray.kernel_rows[i_kernel],
                     ckd.stray.kernel_cols[i_kernel],
                     ckd.stray.kernels_fft[i_kernel],
                     image_fft,
                     conv_result_sub);
            // The full convolution is a sum over all convolutions
            for (int i {}; i < image_n_rows; ++i) {
                for (int j {}; j < image_n_cols; ++j) {
                    conv_result
                      [(i + ckd.stray.edges[i_kernel * box::n + box::b])
                         * ckd.stray.n_spectral
                       + j + ckd.stray.edges[i_kernel * box::n + box::l]] +=
                      conv_result_sub[i * image_n_cols + j];
                }
            }
        }
        for (int i {}; i < unbin_npix; ++i) {
            image_ideal[i] =
              (image_unbin[i] - conv_result[i]) / (1 - ckd.stray.eta[i]);
        }
    }
    binning_table_stray->binImage(image_ideal, image);
}

auto L1A::calibrateImage(const SPEXSettings& settings,
                         const ProcLevel proc_level,
                         const CKD& ckd) -> void
{
    // A calibration step is run if two conditions are satisfied:
    // * The L1X level is less than the corresponding process. For
    //   instance, if the L1X level is ProcLevel::prnu then the PRNU
    //   correction is not applied because the data is already
    //   corrected for PRNU.
    // * The CKD level is equal or higher than the given process. If
    //   the CKD level is not high enough then the calibration step
    //   cannot be applied because that section is missing from the
    //   CKD.
    // * If the process is not skipped

    // If requested, write the data without any calibration. This
    // differs from the original L1A contents in that the data has
    // been coadded, binning table counts have been applied, and the
    // detector image is saved in double precisions as opposed to
    // integers.
    if (meta->l1x_level < ProcLevel::raw
        && l1x[static_cast<int>(ProcLevel::raw)]) {
        l1x[static_cast<int>(ProcLevel::raw)]->writeImage(*this);
    }

    // Dark calibration: dark offset
    if (meta->l1x_level < ProcLevel::dark && ckd.level >= ProcLevel::dark
        && !settings.skip.at(static_cast<int>(ProcLevel::dark))) {
        timers[static_cast<int>(ProcLevel::dark)]->start();
        // If use of the jump row was not requested we set its value
        // to 0. Then the long exposure dark offset is used for the
        // entire detector and the short exposure offset is ignored.
        if (!settings.main.use_jump_row) {
            jump_row = 0;
        }
        const int jump_pixel { binning_table->binIndex(std::max(0, jump_row)
                                                       * ckd.n_detector_cols) };
        for (int i {}; i < jump_pixel; ++i) {
            if (!pixel_mask[i]) {
                image[i] -= ckd.dark.offset_short[i];
            }
        }
        for (int i { jump_pixel }; i < binning_table->nBins(); ++i) {
            if (!pixel_mask[i]) {
                image[i] -= ckd.dark.offset_long[i];
            }
        }
        if (meta->keep_frames) {
            for (int i_img {}; i_img < meta->number_of_images; ++i_img) {
                for (int i {}; i < jump_pixel; ++i) {
                    if (!pixel_mask[i]) {
                        frames[i_img][i] -= ckd.dark.offset_short[i];
                    }
                }
                for (int i { jump_pixel }; i < binning_table->nBins(); ++i) {
                    if (!pixel_mask[i]) {
                        frames[i_img][i] -= ckd.dark.offset_long[i];
                    }
                }
            }
        }
        timers[static_cast<int>(ProcLevel::dark)]->stop();
    }

    // Noise estimation
    if (meta->l1x_level < ProcLevel::noise && ckd.level >= ProcLevel::noise) {
        if (!settings.skip.at(static_cast<int>(ProcLevel::noise))) {
            timers[static_cast<int>(ProcLevel::noise)]->start();
            for (int i {}; i < binning_table->nBins(); ++i) {
                if (pixel_mask[i]) {
                    continue;
                }
                const double var { ckd.noise.g[i] * image[i] + ckd.noise.n[i] };
                if (var <= 0.0) {
                    pixel_mask[i] = true;
                } else {
                    noise[i] = std::sqrt(
                      var / (nr_coadditions * binning_table->binSize(i)));
                }
            }
            timers[static_cast<int>(ProcLevel::noise)]->stop();
        } else {
            std::ranges::fill(noise, 1.0);
        }
    }

    // Dark calibration: dark current
    if (meta->l1x_level < ProcLevel::dark && ckd.level >= ProcLevel::dark
        && !settings.skip.at(static_cast<int>(ProcLevel::dark))) {
        timers[static_cast<int>(ProcLevel::dark)]->start();
        for (int i {}; i < binning_table->nBins(); ++i) {
            image[i] -= ckd.dark.current[i] * exposure_time;
        }
        timers[static_cast<int>(ProcLevel::dark)]->stop();
        if (l1x[static_cast<int>(ProcLevel::dark)]) {
            l1x[static_cast<int>(ProcLevel::dark)]->writeImage(*this);
        }
    }

    // Subtract background image
    if (bgr_image && meta->l1x_level < ProcLevel::nonlin) {
        for (int i {}; i < binning_table->nBins(); ++i) {
            image[i] -= (*bgr_image)[i];
        }
    }

    // Apply row-dependent offset to the corrected image but not the
    // calibration image and background image separately.
    if (!is_background) {
        rowDependentOffset(settings, proc_level, ckd);
    }

    // Nonlinearity correction
    if (meta->l1x_level < ProcLevel::nonlin && ckd.level >= ProcLevel::nonlin
        && !settings.skip.at(static_cast<int>(ProcLevel::nonlin))) {
        timers[static_cast<int>(ProcLevel::nonlin)]->start();
        if (ckd.nonlin.model == NonlinModel::bspline) {
            for (int i {}; i < binning_table->nBins(); ++i) {
                if (pixel_mask[i]) {
                    continue;
                }
                const double dark { ckd.dark.current[i] * exposure_time };
                image[i] = ckd.nonlin.b_spline->deBoor(
                             ckd.nonlin.control_points, image[i] + dark, i)
                           - ckd.nonlin.b_spline->deBoor(
                             ckd.nonlin.control_points, dark, i);
                noise[i] *= ckd.nonlin.b_spline_deriv->deBoor(
                  ckd.nonlin.control_points_deriv, image[i] + dark, i);
            }
        } else if (ckd.nonlin.model == NonlinModel::sigmoidal) {
            for (int i {}; i < binning_table->nBins(); ++i) {
                if (pixel_mask[i]) {
                    continue;
                }
                const double dark { ckd.dark.current[i] * exposure_time };
                noise[i] *= ckd.nonlin.linear_spline.deriv(i, image[i] + dark);
                image[i] = ckd.nonlin.linear_spline.eval(i, image[i] + dark)
                           - ckd.nonlin.linear_spline.eval(i, dark);
            }
        } else {
            // Use the B-spline to evaluate the fit on the desired input
            // (the signal)
            Bspline b(ckd.nonlin.order + 1,
                      ckd.dim_nonlin_knot,
                      ckd.nonlin.b_spline_knots.data());
            std::vector<int> ideriv { 0, 1 };
            for (int i_pix {}; i_pix < binning_table->nBins(); ++i_pix) {
                if (pixel_mask[i_pix]) {
                    continue;
                }
                double signal_cur = image[i_pix];
                bool converged = false;
                for (size_t iter {}; iter < 100; ++iter) {
                    std::vector<double> model(2);
                    std::vector<double> abscissa(2);
                    abscissa[0] = signal_cur / ckd.nonlin_signal_scale[i_pix];
                    abscissa[1] = abscissa[0];
                    b.jacapply(2,
                               abscissa.data(),
                               ideriv.data(),
                               &ckd.nonlin_fit[i_pix * ckd.dim_nonlin_spline],
                               // nonlin_fit_cur,
                               model.data());
                    double gain = ckd.nonlin_signal_scale[i_pix] / model[1];
                    double addition = (image[i_pix] - model[0]) * gain;
                    signal_cur += addition;
                    if (abs(addition) < 1e-3) {
                        image[i_pix] = signal_cur;
                        noise[i_pix] *= abs(gain);
                        converged = true;
                        break;
                    }
                }
                if (!converged) {
                    pixel_mask[i_pix] = true;
                }
            }
        }
        timers[static_cast<int>(ProcLevel::nonlin)]->stop();
        if (l1x[static_cast<int>(ProcLevel::nonlin)]) {
            l1x[static_cast<int>(ProcLevel::nonlin)]->writeImage(*this);
        }
    }

    // Pixel response nonuniformity (PRNU)
    if (meta->l1x_level < ProcLevel::prnu && ckd.level >= ProcLevel::prnu
        && !settings.skip.at(static_cast<int>(ProcLevel::prnu))) {
        timers[static_cast<int>(ProcLevel::prnu)]->start();
        for (int i {}; i < binning_table->nBins(); ++i) {
            image[i] /= ckd.prnu.prnu[i];
            noise[i] /= ckd.prnu.prnu[i];
        }
        timers[static_cast<int>(ProcLevel::prnu)]->stop();
        if (l1x[static_cast<int>(ProcLevel::prnu)]) {
            l1x[static_cast<int>(ProcLevel::prnu)]->writeImage(*this);
        }
    }

    // Stray light
    const int n_van_cittert_steps {
        settings.van_cittert_steps.at(static_cast<int>(proc_level)),
    };
    if (ckd.level >= ProcLevel::stray && meta->l1x_level < ProcLevel::stray
        && !settings.skip.at(static_cast<int>(ProcLevel::stray))
        && n_van_cittert_steps > 0) {
        timers[static_cast<int>(ProcLevel::stray)]->start();
        calibrateStrayLight(ckd, n_van_cittert_steps);
        timers[static_cast<int>(ProcLevel::stray)]->stop();
        if (l1x[static_cast<int>(ProcLevel::stray)]) {
            l1x[static_cast<int>(ProcLevel::stray)]->writeImage(*this);
        }
    }
}

// Spectrally bin the signal, noise, and mask of both polarization
// channels of FOV i_fov. If the spectral binning table does not exist
// this function has no affect. That check is done inside binSpectral.
static auto binSpectra(const BinningTable& binning_table,
                       const int i_fov,
                       Spectra& spectra) -> void
{
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        binning_table.binSpectral(i_fov, spectra.signal[i_pol]);
        binning_table.binSpectral(i_fov, spectra.noise[i_pol]);
        binning_table.binSpectral(i_fov, spectra.mask[i_pol]);
    }
}

auto L1A::extractSpectra(const CKD& ckd,
                         const int i_fov,
                         Spectra& spectra) const -> void
{
    if (ckd.level >= ProcLevel::fov) {
        if (meta->l1x_level < ProcLevel::fov) {
            timers[static_cast<int>(ProcLevel::fov)]->start();
            spectra.extractFOV(ckd, image, noise, pixel_mask, i_fov);
            binSpectra(*binning_table, i_fov, spectra);
            timers[static_cast<int>(ProcLevel::fov)]->stop();
            if (l1x[static_cast<int>(ProcLevel::fov)]) {
                l1x[static_cast<int>(ProcLevel::fov)]->writeSpectrum(
                  spectra, l1x_image_offset, i_fov);
            }
        } else if (meta->l1x_level == ProcLevel::fov) {
            spectra = l1x_spectra[i_fov];
        }
    }
    if (ckd.level >= ProcLevel::rad) {
        if (meta->l1x_level < ProcLevel::rad) {
            timers[static_cast<int>(ProcLevel::rad)]->start();
            if (!ckd.rad_spectra.front().empty()) {
                spectra.calibrate(ckd, exposure_time, i_fov);
            }
            spectra.interpolateToCommon(ckd, i_fov);
            timers[static_cast<int>(ProcLevel::rad)]->stop();
            if (l1x[static_cast<int>(ProcLevel::rad)]) {
                l1x[static_cast<int>(ProcLevel::rad)]->writeSpectrum(
                  spectra, l1x_image_offset, i_fov);
            }
        } else if (meta->l1x_level == ProcLevel::rad) {
            spectra = l1x_spectra[i_fov];
        }
    }
}

auto moveL1ASlice(std::vector<L1A>::iterator src_beg,
                  std::vector<L1A>::iterator src_end,
                  std::vector<L1A>& dest) -> void
{
    dest.resize(std::ranges::distance(src_beg, src_end), { false });
    auto dest_beg { dest.begin() };
    for (auto src { src_beg }; src != src_end; ++src) {
        std::swap(*dest_beg, *src);
        ++dest_beg;
    }
}

} // namespace spex
