// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for reading and storing the contents of one L1A product and
// for performing operations on it such as calibrating.

#pragma once

#include "binning_table.h"
#include "quaternion.h"
#include "spectra.h"

namespace spex {

class L1AMetadata;
class L1X;
class Timer;

class L1A
{
private:
    // Whether this is a background image
    bool is_background {};
    // Apply row-dependent offset to an image
    auto rowDependentOffset(const SPEXSettings& settings,
                            const ProcLevel proc_level,
                            const CKD& ckd) -> void;
    // Correct image for stray light using the Van Cittert
    // deconvolution algorithm
    auto calibrateStrayLight(const CKD& ckd, const int n_van_cittert) -> void;

public:
    L1A(const bool is_background);
    ~L1A() = default;

    // Main data

    // This variable starts with a raw detector image and is updated
    // at each calibration step. After the spectra have been extracted
    // we don't need the detector image anymore (steps FOV and
    // higher).
    std::vector<double> image {};
    // Raw detector frames either read from file or a previously
    // allocated buffer. If [l1b][l1a_in_memory] is true then this
    // variable is allocated and filled by the L1A manager for all L1A
    // products. Otherwise it is empty and reading from a NetCDF
    // source takes place before calibrating each image.
    std::vector<uint16_t> frames_u16 {};
    // Same thing but for the detector image noise
    std::vector<double> noise {};
    // Background image associated with this L1A product (default is
    // no background image)
    std::shared_ptr<std::vector<double>> bgr_image {};
    // Individual frames that make up an image. Typically these are
    // discarded right after reading in and coadding. However, for
    // some operations such as noise CKD generation these need to be
    // saved for later use.
    std::vector<std::vector<double>> frames {};
    // Some operations require the original image to be saved and
    // overwritten later.
    std::vector<double> image_buf {};
    // Pixel mask is initially copied from the CKD but may be updated
    // during one or more calibrations steps. Only used for detector
    // steps, hence called pixel mask. From FOV onward we use the
    // spectral mask (part of Spectra).
    std::vector<bool> pixel_mask {};
    // All rows up to but not including jump_row use
    // ckd.dark.offset_short for the dark offset and
    // ckd.dark.offset_long otherwise.
    int jump_row {};

    // Meta data

    // Some of the meta data is in an L1AMetadata instance and the
    // rest is held by the L1A class directly.
    std::shared_ptr<L1AMetadata> meta {};
    // Image number where the current observation starts. In
    // diagnostic mode it is typically 0 unless some images need to be
    // skipped. In science mode, where a single L1A file contains all
    // the images which are not coadded, image_offset is the index of
    // the current image to be processed.
    int image_offset {};
    // Seconds since the start of the day
    double image_time {};
    // Absolute time for the solar model
    uint32_t seconds {};
    // Additional fraction of a second to be added
    double second_fraction {};
    // Binning table
    uint8_t binning_table_id {};
    std::shared_ptr<BinningTable> binning_table {};
    // Stray light might require special treatment for unbinning so
    // keep a separate binning_table for that.
    std::shared_ptr<BinningTable> binning_table_stray {};
    // Coadded digital offset. It is a float because the coadding
    // correction c_coadd (float) will be added to it.
    double digital_offset {};
    // Uncoadded original digital offset
    int16_t digital_offset_single {};
    // Number of coadditions. In diagnostic mode, this is summed over
    // images in the L1A product.
    int nr_coadditions {};
    // Coadded and averaged exposure time
    double exposure_time {};
    // Equals either pol_minus or pol_plus
    int polarization_state {};

    // Naviagion data

    // Rotation quaternion for attitude
    Quaternion att_quat {};
    // Interpolated satellite position
    std::array<double, dims::vec> satellite_position {};

    // L1X related

    // L1X instances for each step (null if not requested)
    std::vector<std::shared_ptr<L1X>> l1x {};
    // Which image in the L1X is owned by this L1A instance
    size_t l1x_image_offset {};
    // When spectra are read they are stashed into this array. This is
    // needed because the L1A usually has no spectra as members. When
    // a spectrum is requested it is retrieved from this stash rather
    // than acquired from the image which does not exist for data
    // ProcLevel::fov or higher.
    std::vector<Spectra> l1x_spectra {};

    // Timers for most steps in the process chain. All L1A instances
    // within a process (e.g. fov, rad, l1b) share the same timers so
    // that cumulative times can be reported in the end. The timers
    // are created and associated with L1A instances at the beginning
    // of each process after the L1A manager has been initialized.
    std::array<std::shared_ptr<Timer>, static_cast<int>(ProcLevel::n_levels)>
      timers {};

    auto isBackground() const -> bool { return is_background; }
    // Process data from the image_attributes NetCDF group
    auto initImageAttributes(const CKD& ckd) -> void;
    // Process data that was already read from the navigation_data NetCDF group
    auto processNavigationData() -> void;
    // Read an image from file or a previously allocated buffer and
    // perform basic image processing such as coadding of frames. If
    // reading from file, this becomes a high-data-load function to
    // read the image. The NetCDF source file (meta->nc) is assumed to
    // be opened. This function recognizes both L1A and L1X data
    // levels.
    auto readImage(const CKD& ckd) -> void;
    // Calibrate an image up to and including stray light
    auto calibrateImage(const SPEXSettings& settings,
                        const ProcLevel proc_level,
                        const CKD& ckd) -> void;
    // Extract a pair of radiometrically calibrated S+/S- spectra from
    // a detector image
    auto extractSpectra(const CKD& ckd, const int i_fov, Spectra& spectra) const
      -> void;
};

// Move a slice from an L1A vector to another vector. Since we can't
// actually move portions of a vector, the source array will be filled
// with dummy L1A instances as a result of swapping. It should be
// deallocated as soon as possible after calls to this function.
auto moveL1ASlice(std::vector<L1A>::iterator src_beg,
                  std::vector<L1A>::iterator src_end,
                  std::vector<L1A>& dest) -> void;

} // namespace spex
