// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_diagnostic.h"

#include "l1a_metadata.h"
#include "spex_settings.h"

namespace spex {

auto L1ADiagnostic::getNumberOfImages(
  const SPEXSettings& /*settings*/,
  const std::shared_ptr<L1AMetadata>& /*meta*/,
  const size_t n_files) const -> int
{
    // Each file is one image
    return static_cast<int>(n_files);
}

// Whether the number of discarded frames should be reduced. If a
// non-zero number of frames were requested to be discarded from the
// front and/or back of the frame stack, but less than 10 frames in
// total would remain when doing so, then do not discard that many
// frames.
static auto limitDiscardedFrame(const SPEXSettings& settings,
                                const ProcLevel proc_level,
                                const std::shared_ptr<L1AMetadata>& meta)
  -> bool
{
    const int discard_front { settings.discard_frames_front.at(
      static_cast<int>(proc_level)) };
    const int discard_back { settings.discard_frames_back.at(
      static_cast<int>(proc_level)) };
    if (discard_front == 0 && discard_back == 0) {
        return false;
    }
    meta->nc->reOpen();
    const int n_remaining { meta->number_of_images - discard_front
                            - discard_back };
    // Extract the coaddition factors from the L1A file and add them
    // up to get the total number of coadditions. Only count those
    // frames that remain after the discarding of frames from front
    // and/or back.
    std::vector<uint16_t> nr_coadditions_values(n_remaining);
    const std::vector<size_t> start { static_cast<size_t>(discard_front) };
    const std::vector<size_t> count {
        static_cast<size_t>(n_remaining),
    };
    meta->nc->getGroup("image_attributes")
      .getVar("nr_coadditions")
      .getVar(start, count, nr_coadditions_values.data());
    // Return whether the total number of remaining frames would be
    // less than a threshold
    constexpr int min_number_of_total_frames { 10 };
    return std::accumulate(
             nr_coadditions_values.begin(), nr_coadditions_values.end(), 0)
           < min_number_of_total_frames;
}

auto L1ADiagnostic::organize(const SPEXSettings& settings,
                             const ProcLevel proc_level,
                             std::vector<std::shared_ptr<L1AMetadata>>& meta)
  -> void
{
    // Each file is one L1A image
    for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size()); ++i_l1a) {
        l1a_products.at(i_l1a).meta = meta.at(i_l1a);
        // If some frames need to be discarded for the current
        // process, adjust the number of images (frames) that will be
        // read from the L1A file.
        if (limitDiscardedFrame(settings, proc_level, meta.at(i_l1a))) {
            // If less than a certain number of frames would remain
            // after the discarding of frames, then discard at most 1
            // frame from the front and no frames from the back of the
            // frame stack.
            const int discard_front { std::min(
              1,
              static_cast<int>(settings.discard_frames_front.at(
                static_cast<int>(proc_level)))) };
            meta.at(i_l1a)->number_of_images -= discard_front;
            l1a_products.at(i_l1a).image_offset = discard_front;
        } else {
            meta.at(i_l1a)->number_of_images -=
              settings.discard_frames_front.at(static_cast<int>(proc_level))
              + settings.discard_frames_back.at(static_cast<int>(proc_level));
            l1a_products.at(i_l1a).image_offset =
              settings.discard_frames_front.at(static_cast<int>(proc_level));
        }
    }
}

} // namespace spex
