// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_manager.h"

#include "l1a_metadata.h"
#include "l1x.h"
#include "spex_settings.h"

namespace spex {

// Must use an initialization function because it calls some pure
// virtual functions.
auto L1AManager::init(const SPEXSettings& settings,
                      const ProcLevel proc_level,
                      const CKD& ckd) -> void
{
    auto l1a_files { is_background
                       ? settings.l1a_bgr_files.at(static_cast<int>(proc_level))
                       : settings.l1a_files.at(static_cast<int>(proc_level)) };
    // If this is the stray light CKD process and we are reading from
    // a checkpoint file, the number of files becomes a dummy
    // variable. To save time set it to 1.
    if (proc_level == ProcLevel::stray && settings.stray.checkpoint.read
        || proc_level == ProcLevel::swath && settings.swath.checkpoint.read) {
        l1a_files = { l1a_files.front() };
    }
    // First read all the L1A meta data (low data load). That
    // determines how many L1A products should be allocated and how
    // they are treated.
    // If there are many L1A files then show the progress of reading
    // their metadata in increments of perc_inc %.
    constexpr size_t l1a_meta_verbose_thres { 300 };
    constexpr int perc_inc { 10 };
    const int perc_factor { static_cast<int>(l1a_files.size() / perc_inc) };
    const bool meta_progress_verbose { l1a_files.size()
                                       >= l1a_meta_verbose_thres };
    std::vector<std::shared_ptr<L1AMetadata>> l1a_meta(l1a_files.size());
    if (!is_background) {
        spdlog::info("Reading L1A meta data");
    } else if (meta_progress_verbose) {
        spdlog::info("Reading L1A background meta data");
    }
    for (int i_file {}; i_file < static_cast<int>(l1a_files.size()); ++i_file) {
        if (meta_progress_verbose && i_file % perc_factor == 0) {
            spdlog::info("  {:2}%", perc_inc * i_file / perc_factor);
        }
        l1a_meta.at(i_file) = std::make_shared<L1AMetadata>(
          l1a_files.at(i_file), settings, proc_level);
    }
    l1a_products = std::vector<L1A>(
      getNumberOfImages(settings, l1a_meta.front(), l1a_files.size()),
      { is_background });
    organize(settings, proc_level, l1a_meta);
    for (auto& l1a : l1a_products) {
        l1a.initImageAttributes(ckd);
        if (l1a.meta->sc_records > 0) {
            l1a.processNavigationData();
        }
    }
}

auto L1AManager::initL1X(const SPEXSettings& settings,
                         const ProcLevel proc_level,
                         const CKD& ckd) -> void
{
    std::vector<std::shared_ptr<L1X>> l1x(static_cast<int>(ProcLevel::n_levels),
                                          nullptr);
    for (int i_l1x {}; i_l1x < static_cast<int>(ProcLevel::n_levels); ++i_l1x) {
        const ProcLevel l1x_level { static_cast<ProcLevel>(i_l1x) };
        std::string l1x_filename {};
        if (l1x_level == ProcLevel::raw) {
            l1x_filename = settings.l1x_raw.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::dark) {
            l1x_filename = settings.l1x_dark.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::nonlin) {
            l1x_filename = settings.l1x_nonlin.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::prnu) {
            l1x_filename = settings.l1x_prnu.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::stray) {
            l1x_filename = settings.l1x_stray.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::fov) {
            l1x_filename = settings.l1x_fov.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::rad) {
            l1x_filename = settings.l1x_rad.at(static_cast<int>(proc_level));
        }
        if (!l1x_filename.empty()) {
            l1x.at(i_l1x) = std::make_shared<L1X>(
              ckd, settings, l1a_products, l1x_level, l1x_filename);
        }
    }
    for (int i {}; i < static_cast<int>(l1a_products.size()); ++i) {
        l1a_products.at(i).l1x = l1x;
        l1a_products.at(i).l1x_image_offset = i;
    }
}

} // namespace spex
