// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_metadata.h"

#include "geometry.h"
#include "spex_settings.h"

namespace spex {

L1AMetadata::L1AMetadata(const std::string& filename,
                         const SPEXSettings& settings,
                         const ProcLevel proc_level)
  : filename { filename }, keep_frames { proc_level == ProcLevel::noise }
{
    nc = std::make_shared<NcObject>(netCDF::NcFile::read, filename);
    for (const auto& [name, att] : nc->getAtts()) {
        if (name == "l1x_level") {
            att.getValues(&l1x_level);
        }
    }
    if (settings.l1a_type.at(static_cast<int>(proc_level))
        == L1AType::science) {
        n_ch = detector::n_channels_binning_mode;
    } else {
        n_ch = detector::n_channels_full_frame;
    }
    number_of_images =
      static_cast<int>(nc->getDim("number_of_images").getSize());
    if (l1x_level < ProcLevel::fov) {
        npix = static_cast<int>(nc->getDim("samples_per_image").getSize());
    }

    const auto image_attributes { nc->getGroup("image_attributes") };
    icu_times.resize(number_of_images);
    icu_times_subsec.resize(number_of_images);
    image_times.resize(number_of_images);
    binning_tables.resize(number_of_images);
    digital_offsets.resize(number_of_images);
    nr_coadditions.resize(number_of_images);
    exposure_times.resize(number_of_images);
    image_attributes.getVar("icu_time_sec").getVar(icu_times.data());
    image_attributes.getVar("icu_time_subsec").getVar(icu_times_subsec.data());
    image_attributes.getVar("image_time").getVar(image_times.data());
    image_attributes.getVar("binning_table").getVar(binning_tables.data());
    image_attributes.getVar("digital_offset").getVar(digital_offsets.data());
    image_attributes.getVar("nr_coadditions").getVar(nr_coadditions.data());
    image_attributes.getVar("exposure_time").getVar(exposure_times.data());

    if (const auto navigation_data { nc->getGroup("navigation_data") };
        !navigation_data.isNull()) {
        if (const auto dim { nc->getDim("SC_records") }; dim.isNull()) {
            sc_records =
              static_cast<int>(navigation_data.getDim("att_time").getSize());
        } else {
            sc_records = static_cast<int>(nc->getDim("SC_records").getSize());
        }
        nav_att_time.resize(sc_records);
        nav_att_quat.resize(sc_records);
        std::vector<double> nav_att_quat_raw(sc_records * dims::quat);
        nav_orb_time.resize(sc_records);
        nav_orb_pos.resize(sc_records);
        std::vector<double> nav_orb_pos_raw(sc_records * dims::vec);
        navigation_data.getVar("att_time").getVar(nav_att_time.data());
        navigation_data.getVar("att_quat").getVar(nav_att_quat_raw.data());
        for (int i {}; i < sc_records; ++i) {
            nav_att_quat[i] = { nav_att_quat_raw[i * dims::quat + 0],
                                nav_att_quat_raw[i * dims::quat + 1],
                                nav_att_quat_raw[i * dims::quat + 2],
                                nav_att_quat_raw[i * dims::quat + 3] };
            // Normalize the quaternion. It comes from independent
            // input, possibly polluted by ASCII roundings before it
            // ended up in NetCDF.
            nav_att_quat[i].normalize();
        }
        removeQuaternionDiscontinuities(nav_att_quat);
        navigation_data.getVar("orb_time").getVar(nav_orb_time.data());
        navigation_data.getVar("orb_pos").getVar(nav_orb_pos_raw.data());
        for (int i {}; i < sc_records; ++i) {
            nav_orb_pos[i] = { nav_orb_pos_raw[i * dims::vec + 0],
                               nav_orb_pos_raw[i * dims::vec + 1],
                               nav_orb_pos_raw[i * dims::vec + 2] };
        }
        std::vector<double> pos_values {};
        for (int i_dir {}; i_dir < dims::vec; ++i_dir) {
            pos_values.resize(sc_records);
            for (int i_pos {}; i_pos < sc_records; ++i_pos) {
                pos_values[i_pos] = nav_orb_pos[i_pos][i_dir];
            }
            pos_splines[i_dir] =
              make_unique<CubicSpline>(nav_orb_time, pos_values);
        }
    } else if (proc_level == ProcLevel::l1b && settings.l1b.geolocation) {
        throw std::runtime_error { "cannot do geolocation without navigation "
                                   "data; set [l1b][geolocation] to false" };
    }
    if (const auto gse_data { nc->getGroup("gse_data") }; !gse_data.isNull()) {
        for (const auto& [name, dim] : gse_data.getDims()) {
            if (name == "wavelength") {
                dim_refspec = static_cast<int>(dim.getSize());
            }
        }
        refspec_wavelength.resize(dim_refspec, NC_FILL_DOUBLE);
        refspec_radiance.resize(dim_refspec, NC_FILL_DOUBLE);
        for (const auto& [name, var] : gse_data.getVars()) {
            if (name == "wavelength") {
                var.getVar(refspec_wavelength.data());
            } else if (name == "signal") {
                var.getVar(refspec_radiance.data());
            } else if (name == "viewport") {
                var.getVar(&viewport);
            }
        }
        if (const auto wave_monitor { gse_data.getGroup("WaveMonitor") };
            !wave_monitor.isNull()) {
            dim_refspec =
              static_cast<int>(wave_monitor.getDim("wavelength").getSize());
            refspec_wavelength.resize(dim_refspec);
            wave_monitor.getVar("wavelength").getVar(refspec_wavelength.data());
            // Reference spectra will be taken as the sum over different times
            const auto dim_time { wave_monitor.getDim("time").getSize() };
            std::vector<double> refspec_radiance_raw(dim_refspec * dim_time);
            wave_monitor.getVar("spectrum").getVar(refspec_radiance_raw.data());
            refspec_radiance.resize(dim_refspec);
            std::ranges::fill(refspec_radiance, 0.0);
            int start_idx {};
            for (int i_time {}; i_time < static_cast<int>(dim_time); ++i_time) {
                for (int i {}; i < dim_refspec; ++i) {
                    refspec_radiance[i] += refspec_radiance_raw[i + start_idx];
                }
                start_idx += dim_refspec;
            }
        }
        if (const auto ref_signal_var { gse_data.getVar("reference_signal") };
            !ref_signal_var.isNull()) {
            ref_signal_var.getVar(&reference_signal);
        }
        if (const auto egse_var { gse_data.getVar("egse") };
            !egse_var.isNull()) {
            egse.resize(egse_var.getDim(0).getSize());
            egse_var.getVar(egse.data());
        }
        // Degree of linear polarization constant over the spectrum
        // (might or might not be needed)
        double dolp_value {};
        for (const auto& [name, att] : gse_data.getAtts()) {
            if (name == "Illumination_level") {
                att.getValues(&illumination_level);
            } else if (name == "DoLP") {
                att.getValues(&dolp_value);
            } else if (name == "ACT_rotationAngle") {
                att.getValues(&act_angle);
                act_angle *= math::deg_to_rad;
            } else if (name == "ALT_rotationAngle") {
                // If ALT angle is present override the previously set
                // viewport value.
                att.getValues(&alt_angle);
                constexpr double angle_comp_margin { 10.0 };
                for (const auto& [vp_idx, angle] : viewport_idx_to_angle) {
                    if (std::abs(angle - alt_angle) < angle_comp_margin) {
                        viewport = vp_idx;
                        break;
                    }
                }
                alt_angle *= math::deg_to_rad;
            }
        }
        if (const auto group { gse_data.getGroup("SpectralDolP") };
            !group.isNull()) {
            const auto n_wavelengths { group.getDim("wavelength").getSize() };
            std::vector<double> wavelengths(n_wavelengths);
            group.getVar("wavelength").getVar(wavelengths.data());
            std::vector<double> dolp(n_wavelengths);
            group.getVar("DoLP").getVar(dolp.data());
            dolp_spline = std::make_unique<CubicSpline>(wavelengths, dolp);
        } else {
            dolp_spline = std::make_unique<CubicSpline>(
              std::vector<double> { 0.0, NC_FILL_FLOAT },
              std::vector<double> { dolp_value, dolp_value });
        }
    }
    // Close the file because there may be an onslaught of L1A files
    nc->close();
}

} // namespace spex
