// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// L1A file meta data. It is specific for one L1A file but does not
// need to be associated with exactly one image. For instance,
// engineering data has to be interpolated for each image. If multiple
// images exist in one file (such as in science L1A) we would not like
// to repeat reading the engineering data for each image.

#pragma once

#include "cubic_spline.h"
#include "egse.h"
#include "nc_object.h"
#include "quaternion.h"

#include <memory>

namespace spex {

class SPEXSettings;

class L1AMetadata
{
public:
    // Read meta data of one file. This routine does not know which
    // image is interesting but reads important file-specific
    // information such as the number of images as well as meta data
    // with different time axes.
    L1AMetadata(const std::string& filename,
                const SPEXSettings& settings,
                const ProcLevel proc_level);

    // NetCDF object of the L1A file
    std::shared_ptr<NcObject> nc {};
    std::string filename {};
    // Number of images to be coadded. In diagnostic mode, typically
    // (but not always) all images found in an L1A file are coadded
    // and number_of_images is set equal to the number_of_images
    // dimension. In science mode, there is no coadding and
    // number_of_images should equal 1.
    int number_of_images {};
    // Number of pixels in an image (maximum)
    int npix {};
    // Whether to keep all frames of the raw L1A image or discard them
    // after coadding. Typically they are not needed after coadding
    // and thus the default is to discard them in order to save
    // memory.
    bool keep_frames { false };
    // Number of data channels
    int n_ch {};

    // Image attributes
    std::vector<uint32_t> icu_times {};
    std::vector<uint16_t> icu_times_subsec {};
    std::vector<double> image_times {};
    std::vector<uint8_t> binning_tables {};
    std::vector<int16_t> digital_offsets {};
    std::vector<uint16_t> nr_coadditions {};
    std::vector<double> exposure_times {};

    // Navigation data if requested
    int sc_records {};
    std::vector<double> nav_att_time {};
    std::vector<Quaternion> nav_att_quat {};
    std::vector<double> nav_orb_time {};
    std::vector<std::array<double, dims::vec>> nav_orb_pos {};
    // Cubic splines for interpolating the satellite position. These
    // are required because the navigation and science data have
    // different time axes.
    std::array<std::unique_ptr<CubicSpline>, dims::vec> pos_splines {};
    // An added challenge is that these do not interpolate straight
    // forward. The position and velocity use an interpolation where
    // the velocity is the derivative of the position, with Earth
    // rotation correction. And Quaternions are interpolated
    // symmetrical and naturally normalized (see eventual issues in
    // the appendix in the ATBD.)

    // GSE data (optional and totally absent for science L1A). This
    // group exists per file.
    // Number of wavelength of a reference spectrum (only for optics
    // measurements)
    int dim_refspec {};
    // Scalar illumination level for detector measurements
    double illumination_level {};
    // Biweight median of reference diode signal
    double reference_signal {};
    // Illuminated viewport
    uint8_t viewport {};
    // A reference spectrum
    std::vector<double> refspec_wavelength {};
    // Wavelengths for the reference spectrum
    std::vector<double> refspec_radiance {};
    // Degree of linear polarization as a function of wavelength. If
    // not present, the constant refspec_dolp is used.
    std::unique_ptr<CubicSpline> dolp_spline {};
    // Across and along track rotation angle of rotation stage
    double act_angle {};
    double alt_angle {};
    std::vector<EGSE> egse {};

    // L1X maturity level (global attribute only for L1X input)
    ProcLevel l1x_level { ProcLevel::l1a };
};

} // namespace spex
