// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_science.h"

#include "l1a_metadata.h"
#include "spex_settings.h"

namespace spex {

auto L1AScience::getNumberOfImages(const SPEXSettings& settings,
                                   const std::shared_ptr<L1AMetadata>& meta,
                                   const size_t n_files) const -> int
{
    if (n_files == 0) {
        return 0;
    }
    if (n_files != 1) {
        throw std::runtime_error { "number of L1A files should be exactly 1 in "
                                   "science mode" };
    }
    return settings.l1b.image_end.value_or(meta->number_of_images - 1)
           - settings.l1b.image_start + 1;
}

auto L1AScience::organize(const SPEXSettings& settings,
                          const ProcLevel proc_level,
                          std::vector<std::shared_ptr<L1AMetadata>>& meta)
  -> void
{
    // There is only one file so only one NetCDF object
    for (auto& l1a : l1a_products) {
        l1a.meta = meta.front();
    }
    for (int i {}; i < static_cast<int>(l1a_products.size()); ++i) {
        l1a_products.at(i).image_offset = i + settings.l1b.image_start;
        // All images cover only one image by definition
        meta.front()->number_of_images = 1;
    }
    // If requested, load all raw detector images into memory now as
    // opposed to reading them one by one during calibration.
    if (proc_level == ProcLevel::l1b && settings.l1b.l1a_in_memory) {
        const size_t npix { static_cast<size_t>(meta.front()->npix) };
        const std::vector<size_t> starts {
            static_cast<size_t>(settings.l1b.image_start), 0
        };
        const std::vector<size_t> counts { l1a_products.size(), npix };
        std::vector<uint16_t> detector_images(l1a_products.size() * npix);
        meta.front()->nc->reOpen();
        meta.front()
          ->nc->getGroup("science_data")
          .getVar("detector_images")
          .getVar(starts, counts, detector_images.data());
        for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size());
             ++i_l1a) {
            l1a_products[i_l1a].frames_u16 = {
                detector_images.begin() + i_l1a * npix,
                detector_images.begin() + (i_l1a + 1) * npix
            };
        }
        meta.front()->nc->close();
    }
}

} // namespace spex
