// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1b.h"

#include "geometry.h"
#include "io.h"
#include "l1a_metadata.h"
#include "linalg.h"
#include "linear_spline.h"
#include "read_solar_spectrum.h"
#include "solar_model.h"

namespace spex {

L1B::L1B(const SPEXSettings& settings, CKD& ckd)
  : Processor { settings, ProcLevel::l1b, ckd }
{
    if (settings.l1b.last_calibration_step
        && settings.l1b.last_calibration_step < ckd.level) {
        ckd.level = static_cast<ProcLevel>(
          static_cast<int>(settings.l1b.last_calibration_step.value()) + 1);
    }
}

auto L1B::processInit() -> void
{
    if (settings.l1b.geolocation
        && (settings.skip.at(static_cast<int>(ProcLevel::swath))
            || ckd.level <= ProcLevel::swath)) {
        throw std::runtime_error {
            "geolocation cannot be performed because the CKD does "
            "not contain swath vectors; turn off the geolocation flag to "
            "proceed"
        };
    }
    // We need L1A when writing output. Save it before it becomes a
    // null pointer.
    l1a_meta = l1a_manager->l1a_products.front().meta;
    // Allocate the main variables that constitute the L1B product
    const auto intensity_size { nL1A() * ckd.fov.n_act
                                * settings.l1b.intensity_wavelength.size() };
    const auto polarization_size { nL1A() * ckd.fov.n_act
                                   * settings.l1b.l1b_wavelength.size() };
    const auto geometry_size { nL1A() * ckd.fov.n_act };
    prod.polarization_bandpasses.resize(settings.l1b.l1b_wavelength.size());
    prod.image_time.resize(nL1A());
    prod.I.resize(intensity_size);
    prod.I_noise.resize(intensity_size);
    prod.I_polsample.resize(polarization_size);
    prod.I_polsample_noise.resize(polarization_size);
    prod.q.resize(polarization_size);
    prod.q_noise.resize(polarization_size);
    prod.u.resize(polarization_size);
    prod.u_noise.resize(polarization_size);
    prod.dolp.resize(polarization_size);
    prod.dolp_noise.resize(polarization_size);
    prod.aolp.resize(polarization_size);
    prod.aolp_noise.resize(polarization_size);
    prod.lat.resize(geometry_size);
    prod.lon.resize(geometry_size);
    prod.alt.resize(geometry_size);
    prod.vza.resize(geometry_size);
    prod.vaa.resize(geometry_size);
    prod.sza.resize(geometry_size);
    prod.saa.resize(geometry_size);
    // Find out where the retrieval windows are and how big they are
    fov_start_bin.resize(ckd.fov.n_act);
    fov_nmeas.resize(ckd.fov.n_act);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        int i_bin {};
        if (settings.l1b.wave_min) {
            while (ckd.wave_common[i_fov][i_bin] < settings.l1b.wave_min) {
                ++i_bin;
            }
        }
        fov_start_bin[i_fov] = i_bin;
        i_bin = ckd.fov.dims_spec[i_fov] - 1;
        if (settings.l1b.wave_max) {
            while (ckd.wave_common[i_fov][i_bin] > settings.l1b.wave_max) {
                --i_bin;
            }
        }
        fov_nmeas[i_fov] = i_bin + 1 - fov_start_bin[i_fov];
    }
    // We need to add (order-1) because the last order knots and the
    // endpoint are not included.
    n_demod_states = static_cast<int>(
      2 * (settings.l1b.knots.size() + settings.l1b.order - 1));
    n_demod_states_half = n_demod_states / 2;
    demod_size_full = n_demod_states * n_demod_states;
    demod_size_half = n_demod_states_half * n_demod_states_half;
    BSpline demod_b_spline { settings.l1b.order, settings.l1b.knots };
    // Initialize demodulation splines and the Gaussian kernel
    demod_spline_matrix.resize(ckd.fov.n_act);
    gauss_kernel.resize(ckd.fov.n_act);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        // This is the pure spline matrix used in demodulation with
        // the dimension n_demod_states_half because q and u are not
        // included yet.
        demod_spline_matrix[i_fov] =
          std::make_unique<PackedMatrix>(fov_nmeas[i_fov], n_demod_states_half);
        for (int i {}; i < demod_spline_matrix[i_fov]->nRows(); ++i) {
            demod_b_spline.evalBasis(
              ckd.wave_common[i_fov][fov_start_bin[i_fov] + i],
              i,
              *demod_spline_matrix[i_fov]);
        }
        // Gauss convolution
        const int dim_pol_wave { static_cast<int>(
          settings.l1b.l1b_wavelength.size()) };
        gauss_kernel[i_fov] =
          std::make_unique<PackedMatrix>(dim_pol_wave, fov_nmeas[i_fov]);
        for (int i {}; i < dim_pol_wave; ++i) {
            gauss_kernel[i_fov]->row_locations[i] =
              static_cast<int>(gauss_kernel[i_fov]->data.size());
            const double fwhm { settings.l1b.l1b_wavelength[i]
                                / settings.l1b.resolving_power };
            prod.polarization_bandpasses[i] = static_cast<float>(fwhm);
            const double sigma { fwhm
                                 / (2.0 * std::sqrt(2.0 * std::log(2.0))) };
            const double norm_inv {
                1.0 / (sigma * std::sqrt(2 * std::numbers::pi))
            };
            for (int j {}; j < fov_nmeas[i_fov]; ++j) {
                const int j0 { fov_start_bin[i_fov] + j };
                double res { ckd.wave_common[i_fov][j0]
                             - settings.l1b.l1b_wavelength[i] };
                res *= res;
                res /= 2 * sigma * sigma;
                res = norm_inv * std::exp(-res);
                double delta_lambda {};
                if (j == 0) {
                    delta_lambda = ckd.wave_common[i_fov][j0 + 1]
                                   - ckd.wave_common[i_fov][j0];
                } else if (j == fov_nmeas[i_fov] - 1) {
                    delta_lambda = ckd.wave_common[i_fov][j0]
                                   - ckd.wave_common[i_fov][j0 - 1];
                } else {
                    delta_lambda = 0.5
                                   * (ckd.wave_common[i_fov][j0 + 1]
                                      - ckd.wave_common[i_fov][j0 - 1]);
                }
                const double result { res * delta_lambda };
                if (result > settings.l1b.gauss_range_tol) {
                    gauss_kernel[i_fov]->data.push_back(result);
                    ++gauss_kernel[i_fov]->row_sizes[i];
                    if (gauss_kernel[i_fov]->global_col_indices[i]
                        == def::idx) {
                        gauss_kernel[i_fov]->global_col_indices[i] = j;
                    }
                }
            }
        }
    }
    spdlog::debug("B-spline knots: ");
    for (const double& knot : settings.l1b.knots) {
        spdlog::debug("  {}", knot);
    }
    if (!settings.l1b.dem_file.empty()) {
        constexpr double spexone_res_in_m { 2.6e3 };
        dem = std::make_unique<DEM>(settings.l1b.dem_file, spexone_res_in_m);
        if (settings.l1b.dem_in_memory) {
            dem->loadToMemory();
        }
    }
    // The solar spectrum could also be read at the end of the
    // calculation (when the whole orbit has been processed) but it's
    // better to do it early in case there are errors.
    intensity_F0.resize(settings.l1b.intensity_wavelength.size(), 0.0);
    polarization_F0.resize(settings.l1b.l1b_wavelength.size(), 0.0);
    if (settings.main.solar_spectrum_file.empty()) {
        return;
    }
    std::vector<double> solar_wavelengths {};
    std::vector<double> solar_spectrum {};
    readSolarSpectrum(
      settings.main.solar_spectrum_file, solar_wavelengths, solar_spectrum);
    const CubicSpline solar_spline { solar_wavelengths, solar_spectrum };
    // Solar spectrum on the intensity grid is convolved with the ISRF
    const auto isrf { [](const double fwhm, const double x) {
        const double sigma { fwhm / (2.0 * std::sqrt(2.0 * std::log(2.0))) };
        const double norm { sigma * std::sqrt(2.0 * std::numbers::pi) };
        return exp(-0.5 * x * x / (sigma * sigma)) / norm;
    } };
    for (int i {};
         i < static_cast<int>(settings.l1b.intensity_wavelength.size());
         ++i) {
        const double& wave_out { settings.l1b.intensity_wavelength[i] };
        constexpr double dw { 0.01 };
        constexpr double int_range { 10.0 };
        for (int j {}; j < static_cast<int>(int_range / dw) + 1; ++j) {
            const double wave_in { wave_out - 0.5 * int_range + j * dw };
            constexpr double fwhm { 2.0 };
            intensity_F0[i] +=
              dw * isrf(fwhm, wave_out - wave_in) * solar_spline.eval(wave_in);
        }
    }
    // Solar spectrum on the polarization grid is convolved with a
    // Gaussian of varying FWHM.
    for (int i {}; i < static_cast<int>(settings.l1b.l1b_wavelength.size());
         ++i) {
        const double& wave_out { settings.l1b.l1b_wavelength[i] };
        constexpr double dw { 0.1 };
        constexpr double int_range { 100.0 };
        for (int j {}; j < static_cast<int>(int_range / dw) + 1; ++j) {
            const double wave_in { wave_out - 0.5 * int_range + j * dw };
            polarization_F0[i] +=
              dw * isrf(prod.polarization_bandpasses[i], wave_out - wave_in)
              * solar_spline.eval(wave_in);
        }
    }
}

auto L1B::prepareBatches() -> void
{
    batchOne();
}

auto L1B::geolocate(const int i_img,
                    const spex::L1A& l1a,
                    std::vector<double>& aolp_shift) -> void
{
    // The solar model provides the sun vector in ECR coordinates and
    // the ECR quaternion.
    std::array<double, dims::vec> sun {};
    Quaternion q_ecr {};
    solarModel(l1a.seconds, l1a.second_fraction, sun, q_ecr);
    // Get the swath vectors in ECR coordinates by applying the
    // rotation from CKD to J2000 and from J2000 to ECR.
    Quaternion q_sc { q_ecr };
    q_sc.multiply(l1a.att_quat);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        std::array<double, dims::vec> pointing {
            ckd.swath.pointings[i_fov * dims::vec + 0],
            ckd.swath.pointings[i_fov * dims::vec + 1],
            ckd.swath.pointings[i_fov * dims::vec + 2]
        };
        q_sc.rotate(pointing);
        std::array<double, dims::vec> pos {};
        intersectEllipsoid(pointing, l1a.satellite_position, pos);
        double lat {};
        double lon {};
        double alt {};
        intersectTerrain(*dem, pointing, pos, lat, lon, alt);
        prod.lat[i_img * ckd.fov.n_act + i_fov] = static_cast<float>(lat);
        prod.lon[i_img * ckd.fov.n_act + i_fov] = static_cast<float>(lon);
        prod.alt[i_img * ckd.fov.n_act + i_fov] = static_cast<float>(alt);
        solarAndViewingGeometry(sun,
                                pointing,
                                lat,
                                lon,
                                prod.sza[i_img * ckd.fov.n_act + i_fov],
                                prod.saa[i_img * ckd.fov.n_act + i_fov],
                                prod.vza[i_img * ckd.fov.n_act + i_fov],
                                prod.vaa[i_img * ckd.fov.n_act + i_fov]);
        // Normal vector of the target point
        const std::array<double, dims::vec> zenith {
            std::cos(lat) * std::cos(lon),
            std::cos(lat) * std::sin(lon),
            std::sin(lat)
        };
        // Determine current viewport based on the viewport FOV starting indices
        const auto i_vp { std::distance(
                            ckd.fov.start_vp.begin(),
                            std::ranges::find_if(
                              ckd.fov.start_vp,
                              [i_fov](const auto x) { return x > i_fov; }))
                          - 1 };
        // Instrument reference plane normal
        const std::array<double, dims::vec> inst_plane_normal {
            ckd.swath.normals[i_vp * dims::vec + 0],
            ckd.swath.normals[i_vp * dims::vec + 1],
            ckd.swath.normals[i_vp * dims::vec + 2]
        };
        aolp_shift[i_fov] =
          aolpShift(q_sc, pointing, zenith, inst_plane_normal);
    }
}

auto L1B::demodulateSpectrum(const int i_img,
                             const int i_fov,
                             const Spectra& spectra,
                             DemodWork& wrk,
                             CovMatrices& S) -> void
{
    const int nmeas { fov_nmeas[i_fov] };
    for (int imeas {}; imeas < nmeas; ++imeas) {
        const int i { fov_start_bin[i_fov] + imeas };
        const double m { spectra.signal[pol_minus][i] };
        const double p { spectra.signal[pol_plus][i] };
        wrk.modulation[imeas] = (p - m) / (p + m);
        const double ns_m { spectra.noise[pol_minus][i] };
        const double ns_p { spectra.noise[pol_plus][i] };
        const double p_m { p + m };
        const double p_m2 { p_m * p_m };
        S.modulation[imeas] =
          4 * (m * m * ns_p * ns_p + p * p * ns_m * ns_m) / (p_m2 * p_m2);
        S.I_var[imeas] = ns_p * ns_p + ns_m * ns_m;
    }
    PackedMatrix spline_matrix_q { *demod_spline_matrix[i_fov] };
    PackedMatrix spline_matrix_u { *demod_spline_matrix[i_fov] };
    for (int i {}; i < spline_matrix_q.nRows(); ++i) {
        for (int j {}; j < spline_matrix_q.nCols(i); ++j) {
            spline_matrix_q(i, j) *=
              ckd.pol.m_q[i_fov][fov_start_bin[i_fov] + i]
              / std::sqrt(S.modulation[i]);
            spline_matrix_u(i, j) *=
              ckd.pol.m_u[i_fov][fov_start_bin[i_fov] + i]
              / std::sqrt(S.modulation[i]);
        }
    }
    if (settings.l1b.do_fraunhofer) {
        for (int i {}; i < spline_matrix_q.nRows(); ++i) {
            for (int j {}; j < spline_matrix_q.nCols(i); ++j) {
                spline_matrix_q(i, j) *=
                  ckd.pol.m_q_solar[i_fov][fov_start_bin[i_fov] + i];
                spline_matrix_u(i, j) *=
                  ckd.pol.m_u_solar[i_fov][fov_start_bin[i_fov] + i];
            }
        }
    }
    PackedMatrix spline_matrix_q_T { n_demod_states_half, nmeas };
    PackedMatrix spline_matrix_u_T { n_demod_states_half, nmeas };
    spline_matrix_q_T.transpose(spline_matrix_q);
    spline_matrix_u_T.transpose(spline_matrix_u);
    // BT*B is the linear inversion matrix. It is constructed
    // from four different segments.
    std::vector<double> BTB_11(demod_size_half);
    spline_matrix_q_T.multRightTransposed(spline_matrix_q_T, BTB_11);
    std::vector<double> BTB_12(demod_size_half);
    spline_matrix_q_T.multRightTransposed(spline_matrix_u_T, BTB_12);
    std::vector<double> BTB_22(demod_size_half);
    spline_matrix_u_T.multRightTransposed(spline_matrix_u_T, BTB_22);
    std::vector<double> BTB(demod_size_full);
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[i * n_demod_states + j] = BTB_11[i * n_demod_states_half + j];
        }
    }
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[i * n_demod_states + j + n_demod_states_half] =
              BTB_12[i * n_demod_states_half + j];
        }
    }
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[(i + n_demod_states_half) * n_demod_states + j] =
              BTB_12[i * n_demod_states_half + j];
        }
    }
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[(i + n_demod_states_half) * n_demod_states + j
                + n_demod_states_half] = BTB_22[i * n_demod_states_half + j];
        }
    }
    // BT*Y is the B-spline matrix multiplied with the modulation data
    std::vector<double> BTY(n_demod_states, 0.0);
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < spline_matrix_q_T.row_sizes[i]; ++j) {
            const int col_idx { spline_matrix_q_T.globalCol(i, j) };
            BTY[i] += spline_matrix_q_T(i, j) * wrk.modulation[col_idx]
                      / std::sqrt(S.modulation[col_idx]);
            BTY[i + n_demod_states_half] += spline_matrix_u_T(i, j)
                                            * wrk.modulation[col_idx]
                                            / std::sqrt(S.modulation[col_idx]);
        }
    }
    // Convert BTB into an upper triangular form
    std::vector<double> BTB_tri(n_demod_states * (n_demod_states + 1) / 2);
    for (int j {}; j < n_demod_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j + 1) * j / 2] = BTB[i * n_demod_states + j];
        }
    }
    // Do the linear inversion. x is the resulting q/u state
    // vector and Sx the corresponding covariance matrix.
    std::ranges::fill(S.x, 0.0);
    for (int i {}; i < n_demod_states; ++i) {
        S.x[i * n_demod_states + i] = 1.0;
    }
    std::vector<int> IPIV(n_demod_states);
    const char uplo { 'u' };
    int info {};
    dppsv_(&uplo,
           &n_demod_states,
           &n_demod_states,
           BTB_tri.data(),
           S.x.data(),
           &n_demod_states,
           &info);
    std::vector<double> x_result(n_demod_states, 0.0);
    for (int i {}; i < n_demod_states; ++i) {
        for (int k {}; k < n_demod_states; ++k) {
            x_result[i] += S.x[i * n_demod_states + k] * BTY[k];
        }
    }
    // Divide Sx into sectors
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            S.xq[i * n_demod_states_half + j] = S.x[i * n_demod_states + j];
            S.xu[i * n_demod_states_half + j] =
              S.x[(i + n_demod_states_half) * n_demod_states + j
                  + n_demod_states_half];
            S.xqu[i * n_demod_states_half + j] =
              S.x[i * n_demod_states + j + n_demod_states_half];
        }
    }
    // q and u values on the fine grid
    std::vector<double> q_fine(nmeas, 0.0);
    std::vector<double> u_fine(nmeas, 0.0);
    for (int i {}; i < nmeas; ++i) {
        for (int j {}; j < demod_spline_matrix[i_fov]->nCols(i); ++j) {
            const int x_idx { demod_spline_matrix[i_fov]->globalCol(i, j) };
            q_fine[i] +=
              demod_spline_matrix[i_fov]->operator()(i, j) * x_result[x_idx];
            u_fine[i] += demod_spline_matrix[i_fov]->operator()(i, j)
                         * x_result[n_demod_states_half + x_idx];
        }
    }
    // q and u variances on the fine grid
    timer_Sx_BT.start();
    if (settings.l1b.noise_propagation) {
        demod_spline_matrix[i_fov]->multLeftTransposedT(
          n_demod_states_half, S.xq, S.xq_BT_T);
        demod_spline_matrix[i_fov]->multLeftTransposedT(
          n_demod_states_half, S.xu, S.xu_BT_T);
        demod_spline_matrix[i_fov]->multLeftTransposedT(
          n_demod_states_half, S.xqu, S.xqu_BT_T);
    }
    timer_Sx_BT.stop();
    timer_Squ.start();
    if (settings.l1b.noise_propagation) {
        demod_spline_matrix[i_fov]->multRightTransposed(nmeas, S.xq_BT_T, S.q);
        demod_spline_matrix[i_fov]->multRightTransposed(nmeas, S.xu_BT_T, S.u);
        demod_spline_matrix[i_fov]->multRightTransposed(
          nmeas, S.xqu_BT_T, S.qu);
    }
    timer_Squ.stop();
    // Q and U
    std::vector<double> Q_fine { q_fine };
    std::vector<double> U_fine { u_fine };
    for (int i {}; i < nmeas; ++i) {
        Q_fine[i] *= wrk.rad[fov_start_bin[i_fov] + i];
        U_fine[i] *= wrk.rad[fov_start_bin[i_fov] + i];
    }
    timer_SQU.start();
    if (settings.l1b.noise_propagation) {
        for (int i {}; i < nmeas; ++i) {
            for (int j {}; j < nmeas; ++j) {
                const double rad_prod { wrk.rad[fov_start_bin[i_fov] + i]
                                        * wrk.rad[fov_start_bin[i_fov] + j] };
                const int idx { i * nmeas + j };
                S.Q[idx] = S.q[idx] * rad_prod;
                S.U[idx] = S.u[idx] * rad_prod;
                S.QU[idx] = S.qu[idx] * rad_prod;
            }
            S.Q[i * nmeas + i] += q_fine[i] * S.I_var[i] * q_fine[i];
            S.U[i * nmeas + i] += u_fine[i] * S.I_var[i] * u_fine[i];
            S.IQ[i] = q_fine[i] * S.I_var[i];
            S.IU[i] = u_fine[i] * S.I_var[i];
            S.QU[i * nmeas + i] += q_fine[i] * S.I_var[i] * u_fine[i];
        }
    }
    timer_SQU.stop();
    // Apply telescope polarization (Mueller element mt) to I and Q
    std::vector<double> mt2(nmeas);
    std::vector<double> mt22(nmeas);
    std::vector<double> mt22_inv(nmeas);
    std::vector<double> mt21_1_inv(nmeas);
    for (int i {}; i < nmeas; ++i) {
        const double mt { ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i] };
        mt2[i] = mt * mt;
        mt22[i] = (1 - mt2[i]) * (1 - mt2[i]);
        mt22_inv[i] = 1.0 / mt22[i];
        mt21_1_inv[i] = 1.0 / (1.0 - mt2[i]);
    }
    std::vector<double> I_1(nmeas);
    for (int i {}; i < nmeas; ++i) {
        const double mt { ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i] };
        const double I_1 { 1 / (1 - mt2[i]) * wrk.rad[fov_start_bin[i_fov] + i]
                           - ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i]
                               / (1 - mt2[i]) * Q_fine[i] };
        const double Q_fine_1 { 1 / (1 - mt2[i]) * Q_fine[i]
                                - mt / (1 - mt2[i])
                                    * wrk.rad[fov_start_bin[i_fov] + i] };
        wrk.rad[fov_start_bin[i_fov] + i] = I_1;
        Q_fine[i] = Q_fine_1;
    }
    if (settings.l1b.noise_propagation_telescope) {
        timer_Smt.start();
        for (int i {}; i < nmeas; ++i) {
            const int row_start { i * nmeas };
            for (int j {}; j < nmeas; ++j) {
                const int idx { row_start + j };
                S.Q_1[idx] = S.Q[idx] * mt22_inv[i];
                S.QU_1[idx] = S.QU[idx] * mt21_1_inv[i];
            }
        }
        for (int i {}; i < nmeas; ++i) {
            const double mt { ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i] };
            S.IQ_1[i] = -mt * S.Q[i * nmeas + i] / mt22[i];
            S.IU_1[i] = (S.IU[i] - S.QU[i * nmeas + i] * mt) / (1 - mt2[i]);
            S.Q_1[i * nmeas + i] += mt2[i] / mt22[i] * S.I_var[i];
            S.Q_1[i * nmeas + i] += -2 * mt * S.IQ[i] / mt22[i];
            S.I_var_1[i] =
              (S.I_var[i] + S.Q[i * nmeas + i] * mt2[i] - 2 * S.IQ[i] * mt)
              / mt22[i];
            S.IQ_1[i] +=
              (S.IQ[i] - mt * S.I_var[i] + mt2[i] * S.IQ[i]) / mt22[i];
            S.QU_1[i * nmeas + i] -= S.IU[i] * mt * mt21_1_inv[i];
        }
        std::swap(S.Q, S.Q_1);
        std::swap(S.I_var, S.I_var_1);
        std::swap(S.IQ, S.IQ_1);
        std::swap(S.IU, S.IU_1);
        std::swap(S.QU, S.QU_1);
        timer_Smt.stop();
    }
    // I Q U convolved with Gaussian
    const int dim_pol_wave { static_cast<int>(
      settings.l1b.l1b_wavelength.size()) };
    std::vector<double> Q_conv(dim_pol_wave, 0.0);
    std::vector<double> U_conv(dim_pol_wave, 0.0);
    std::vector<double> I_conv(dim_pol_wave, 0.0);
    std::vector<double> SI_conv(dim_pol_wave, 0.0);
    for (int i {}; i < dim_pol_wave; ++i) {
        for (int k {}; k < gauss_kernel[i_fov]->nCols(i); ++k) {
            const double& gauss { (*gauss_kernel[i_fov])(i, k) };
            const int x_idx { gauss_kernel[i_fov]->globalCol(i, k) };
            Q_conv[i] += gauss * Q_fine[x_idx];
            U_conv[i] += gauss * U_fine[x_idx];
            I_conv[i] += gauss * wrk.rad[fov_start_bin[i_fov] + x_idx];
            SI_conv[i] += gauss * gauss * S.I_var[x_idx];
        }
    }
    timer_SXT.start();
    std::vector<double> SQ_KT(nmeas * dim_pol_wave);
    std::vector<double> SU_KT(nmeas * dim_pol_wave);
    std::vector<double> SIQ_KT(nmeas * dim_pol_wave);
    std::vector<double> SIU_KT(nmeas * dim_pol_wave);
    std::vector<double> SQU_KT(nmeas * dim_pol_wave);
    if (settings.l1b.noise_propagation) {
        gauss_kernel[i_fov]->multLeftTransposed(nmeas, S.Q, SQ_KT);
        gauss_kernel[i_fov]->multLeftTransposed(nmeas, S.U, SU_KT);
        gauss_kernel[i_fov]->multLeftTransposed(nmeas, S.QU, SQU_KT);
    }
    timer_SXT.stop();
    timer_S_QUI_conv.start();
    std::vector<double> SQ_conv(dim_pol_wave, 0.0);
    std::vector<double> SU_conv(dim_pol_wave, 0.0);
    std::vector<double> SIQ_conv(dim_pol_wave, 0.0);
    std::vector<double> SIU_conv(dim_pol_wave, 0.0);
    std::vector<double> SQU_conv(dim_pol_wave, 0.0);
    for (int i {}; i < dim_pol_wave; ++i) {
        for (int k {}; k < gauss_kernel[i_fov]->nCols(i); ++k) {
            const double& gauss { (*gauss_kernel[i_fov])(i, k) };
            const int x_idx { gauss_kernel[i_fov]->globalCol(i, k) };
            SQ_conv[i] += gauss * SQ_KT[x_idx * dim_pol_wave + i];
            SU_conv[i] += gauss * SU_KT[x_idx * dim_pol_wave + i];
            SQU_conv[i] += gauss * SQU_KT[x_idx * dim_pol_wave + i];
        }
    }
    for (int i {}; i < dim_pol_wave; ++i) {
        for (int k {}; k < gauss_kernel[i_fov]->row_sizes[i]; ++k) {
            const double& gauss { (*gauss_kernel[i_fov])(i, k) };
            const int idx { gauss_kernel[i_fov]->globalCol(i, k) };
            SIQ_conv[i] += gauss * gauss * S.IQ[idx];
            SIU_conv[i] += gauss * gauss * S.IU[idx];
        }
    }
    timer_S_QUI_conv.stop();
    std::vector<double> q_conv(dim_pol_wave);
    std::vector<double> u_conv(dim_pol_wave);
    for (int i {}; i < dim_pol_wave; ++i) {
        q_conv[i] = Q_conv[i] / I_conv[i];
        u_conv[i] = U_conv[i] / I_conv[i];
    }
    std::vector<double> Sq_conv(dim_pol_wave);
    std::vector<double> Su_conv(dim_pol_wave);
    std::vector<double> SIq_conv(dim_pol_wave);
    std::vector<double> SIu_conv(dim_pol_wave);
    std::vector<double> Squ_conv(dim_pol_wave);
    for (int i {}; i < dim_pol_wave; ++i) {
        const double I_conv2 { I_conv[i] * I_conv[i] };
        const double q_conv2 { q_conv[i] * q_conv[i] };
        const double u_conv2 { u_conv[i] * u_conv[i] };
        Sq_conv[i] =
          (SQ_conv[i] + SI_conv[i] * q_conv2 - 2 * SIQ_conv[i] * q_conv[i])
          / I_conv2;
        Su_conv[i] =
          (SU_conv[i] + SI_conv[i] * u_conv2 - 2 * SIU_conv[i] * u_conv[i])
          / I_conv2;
        SIq_conv[i] = (SIQ_conv[i] - SI_conv[i] * q_conv[i]) / I_conv[i];
        SIu_conv[i] = (SIU_conv[i] - SI_conv[i] * u_conv[i]) / I_conv[i];
        Squ_conv[i] = (SQU_conv[i] + SI_conv[i] * q_conv[i] * u_conv[i]
                       - SIU_conv[i] * q_conv[i] - SIQ_conv[i] * u_conv[i])
                      / I_conv2;
    }
    std::vector<double> Sdolp(dim_pol_wave);
    std::vector<double> Saolp(dim_pol_wave);
    std::vector<double> Saolp_dolp(dim_pol_wave);
    const int pol_wave_idx_start { i_fov * dim_pol_wave };
    const int i_fov_start { (i_img * ckd.fov.n_act + i_fov) * dim_pol_wave };
    for (int i {}; i < dim_pol_wave; ++i) {
        wrk.dolp[pol_wave_idx_start + i] =
          std::sqrt(q_conv[i] * q_conv[i] + u_conv[i] * u_conv[i]);
        wrk.aolp[pol_wave_idx_start + i] =
          0.5 * std::atan2(u_conv[i], q_conv[i]);
        const double p2 { wrk.dolp[pol_wave_idx_start + i]
                          * wrk.dolp[pol_wave_idx_start + i] };
        const double q_conv2 { q_conv[i] * q_conv[i] };
        const double u_conv2 { u_conv[i] * u_conv[i] };
        Sdolp[i] = (Sq_conv[i] * q_conv2 + Su_conv[i] * u_conv2
                    + 2 * Squ_conv[i] * q_conv[i] * u_conv[i])
                   / p2;
        Saolp[i] = (Su_conv[i] * q_conv2 + Sq_conv[i] * u_conv2
                    - 2 * Squ_conv[i] * q_conv[i] * u_conv[i])
                   / (4 * p2 * p2);
        Saolp_dolp[i] = ((Su_conv[i] - Sq_conv[i]) * q_conv[i] * u_conv[i]
                         + Squ_conv[i] * (q_conv2 - u_conv2))
                        / (2 * p2 * wrk.dolp[pol_wave_idx_start + i]);
    }
    // Rotate q and u to the local meridian
    std::vector<double> q_conv_rot(dim_pol_wave);
    std::vector<double> u_conv_rot(dim_pol_wave);
    std::vector<double> Sq_conv_rot(dim_pol_wave);
    std::vector<double> Su_conv_rot(dim_pol_wave);
    std::vector<double> Squ_conv_rot(dim_pol_wave);
    for (int i {}; i < dim_pol_wave; ++i) {
        wrk.aolp[pol_wave_idx_start + i] += wrk.aolp_shift[i_fov];
        if (wrk.aolp[pol_wave_idx_start + i] > std::numbers::pi / 2) {
            wrk.aolp[pol_wave_idx_start + i] -= std::numbers::pi;
        } else if (wrk.aolp[pol_wave_idx_start + i] < -std::numbers::pi / 2) {
            wrk.aolp[pol_wave_idx_start + i] += std::numbers::pi;
        }
        q_conv_rot[i] = wrk.dolp[pol_wave_idx_start + i]
                        * std::cos(2.0 * wrk.aolp[pol_wave_idx_start + i]);
        u_conv_rot[i] = wrk.dolp[pol_wave_idx_start + i]
                        * std::sin(2.0 * wrk.aolp[pol_wave_idx_start + i]);
        const double p2 { wrk.dolp[pol_wave_idx_start + i]
                          * wrk.dolp[pol_wave_idx_start + i] };
        const double q_conv_rot2 { q_conv_rot[i] * q_conv_rot[i] };
        const double u_conv_rot2 { u_conv_rot[i] * u_conv_rot[i] };
        Sq_conv_rot[i] = 4 * Saolp[i] * u_conv_rot2
                         + Sdolp[i] * q_conv_rot2 / p2
                         - 4 * Saolp_dolp[i] * q_conv_rot[i] * u_conv_rot[i]
                             / wrk.dolp[pol_wave_idx_start + i];
        Su_conv_rot[i] = 4 * Saolp[i] * q_conv_rot2
                         + Sdolp[i] * u_conv_rot2 / p2
                         + 4 * Saolp_dolp[i] * q_conv_rot[i] * u_conv_rot[i]
                             / wrk.dolp[pol_wave_idx_start + i];
        Squ_conv_rot[i] = -4 * Saolp[i] * q_conv_rot[i] * u_conv_rot[i]
                          + Sdolp[i] * q_conv_rot[i] * u_conv_rot[i] / p2
                          + 2 * Saolp_dolp[i] * (q_conv_rot2 - u_conv_rot2)
                              / wrk.dolp[pol_wave_idx_start + i];
    }
    for (int i {}; i < dim_pol_wave; ++i) {
        wrk.aolp[pol_wave_idx_start + i] /= math::deg_to_rad;
        Saolp[i] /= math::deg_to_rad * math::deg_to_rad;
    }
    for (int i {}; i < dim_pol_wave; ++i) {
        prod.I_polsample[i_fov_start + i] = static_cast<float>(I_conv[i]);
        prod.I_polsample_noise[i_fov_start + i] =
          static_cast<float>(SI_conv[i]);
        prod.q[i_fov_start + i] = static_cast<float>(q_conv_rot[i]);
        prod.q_noise[i_fov_start + i] =
          static_cast<float>(std::sqrt(Sq_conv_rot[i]));
        prod.u[i_fov_start + i] = static_cast<float>(u_conv_rot[i]);
        prod.u_noise[i_fov_start + i] =
          static_cast<float>(std::sqrt(Su_conv_rot[i]));
        prod.dolp[i_fov_start + i] =
          static_cast<float>(wrk.dolp[pol_wave_idx_start + i]);
        prod.dolp_noise[i_fov_start + i] =
          static_cast<float>(std::sqrt(Sdolp[i]));
        prod.aolp[i_fov_start + i] =
          static_cast<float>(wrk.aolp[pol_wave_idx_start + i]);
        prod.aolp_noise[i_fov_start + i] =
          static_cast<float>(std::sqrt(Saolp[i]));
    }
}

// Process one image
auto L1B::processBatch(const int i_img) -> void
{
    if (const double p { getPercentage(i_img, batches.size()) }; p >= 0.0) {
        spdlog::info("Processing images {:6.2f}%", p);
    }
    const auto& l1a { batches.at(i_img).l1a_products.front() };
    DemodWork wrk {};
    wrk.aolp_shift.resize(ckd.fov.n_act);
    if (settings.l1b.geolocation) {
        geolocate(i_img, l1a, wrk.aolp_shift);
    }
    // Perform demodulation
    l1a.timers[static_cast<int>(ProcLevel::pol)]->start();
    const int fov_dim_max { ckd.level >= ProcLevel::fov
                              ? *std::max_element(ckd.fov.dims_spec.cbegin(),
                                                  ckd.fov.dims_spec.cend())
                              : 0 };
    const int nmeas_max { ckd.level >= ProcLevel::rad ? *std::max_element(
                            fov_nmeas.cbegin(), fov_nmeas.cend())
                                                      : 0 };
    const auto polarization_size { ckd.fov.n_act
                                   * settings.l1b.l1b_wavelength.size() };
    wrk.rad.resize(fov_dim_max);
    wrk.dolp.resize(polarization_size);
    wrk.aolp.resize(polarization_size);
    wrk.modulation.resize(nmeas_max);
    // Noise of radiance on the fine wavelength grid
    std::vector<double> rad_noise(fov_dim_max);
    CovMatrices S {};
    S.x.resize(demod_size_full);
    S.xq.resize(n_demod_states_half * n_demod_states_half);
    S.xu.resize(n_demod_states_half * n_demod_states_half);
    S.xqu.resize(n_demod_states_half * n_demod_states_half);
    S.xq_BT_T.resize(nmeas_max * n_demod_states_half);
    S.xu_BT_T.resize(nmeas_max * n_demod_states_half);
    S.xqu_BT_T.resize(nmeas_max * n_demod_states_half);
    S.q.resize(nmeas_max * nmeas_max);
    S.u.resize(nmeas_max * nmeas_max);
    S.qu.resize(nmeas_max * nmeas_max);
    S.Q.resize(nmeas_max * nmeas_max);
    S.U.resize(nmeas_max * nmeas_max);
    S.QU.resize(nmeas_max * nmeas_max);
    S.IQ.resize(nmeas_max);
    S.IU.resize(nmeas_max);
    S.Q_1.resize(nmeas_max * nmeas_max);
    S.I_var_1.resize(nmeas_max);
    S.IQ_1.resize(nmeas_max);
    S.IU_1.resize(nmeas_max);
    S.QU_1.resize(nmeas_max * nmeas_max);
    S.modulation.resize(nmeas_max);
    S.I_var.resize(nmeas_max);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        Spectra spectra {};
        l1a.extractSpectra(ckd, i_fov, spectra);
        if (ckd.level >= ProcLevel::rad) {
            for (int i {}; i < ckd.fov.dims_spec[i_fov]; ++i) {
                wrk.rad[i] =
                  spectra.signal[pol_plus][i] + spectra.signal[pol_minus][i];
                rad_noise[i] = sqrt(
                  spectra.noise[pol_minus][i] * spectra.noise[pol_minus][i]
                  + spectra.noise[pol_plus][i] * spectra.noise[pol_plus][i]);
            }
            const LinearSpline radiance_spline { ckd.wave_common[i_fov],
                                                 wrk.rad };
            const LinearSpline radiance_noise_spline { ckd.wave_common[i_fov],
                                                       rad_noise };
            const int n_waves { static_cast<int>(
              settings.l1b.intensity_wavelength.size()) };
            for (int i {}; i < n_waves; ++i) {
                prod.I[(i_img * ckd.fov.n_act + i_fov) * n_waves + i] =
                  static_cast<float>(
                    radiance_spline.eval(settings.l1b.intensity_wavelength[i]));
                prod.I_noise[(i_img * ckd.fov.n_act + i_fov) * n_waves + i] =
                  static_cast<float>(radiance_noise_spline.eval(
                    settings.l1b.intensity_wavelength[i]));
            }
        }
        if (ckd.level >= ProcLevel::pol) {
            demodulateSpectrum(i_img, i_fov, spectra, wrk, S);
        }
    }
    prod.image_time[i_img] = l1a.image_time;
    l1a.timers[static_cast<int>(ProcLevel::pol)]->stop();
}

auto L1B::output() -> void
{
    // L1B product
    NcObject nc { netCDF::NcFile::replace, settings.l1b.outputfile };

    // Global attributes
    nc.putAtt("cdm_data_type", "swath");
    nc.putAtt("processing_level", "L1B");
    nc.putAtt("title", "PACE SPEXone Level-1B data");

    // Dimensions
    const auto nc_bins_along_track { nc.addDim("bins_along_track", nL1A()) };
    const auto nc_number_of_views { nc.addDim("number_of_views", ckd.n_vp) };
    const auto nc_spatial_samples_per_image { nc.addDim(
      "spatial_samples_per_image", ckd.fov.n_act) };
    const auto nc_intensity_bands_per_view { nc.addDim(
      "intensity_bands_per_view", settings.l1b.intensity_wavelength.size()) };
    const auto nc_polarization_bands_per_view { nc.addDim(
      "polarization_bands_per_view", settings.l1b.l1b_wavelength.size()) };
    const std::vector<netCDF::NcDim> intensity_shape {
        nc_bins_along_track,
        nc_spatial_samples_per_image,
        nc_intensity_bands_per_view
    };
    const std::vector<netCDF::NcDim> polarization_shape {
        nc_bins_along_track,
        nc_spatial_samples_per_image,
        nc_polarization_bands_per_view
    };
    const std::vector<netCDF::NcDim> intensity_view_bands_shape {
        nc_number_of_views, nc_intensity_bands_per_view
    };
    const std::vector<netCDF::NcDim> polarization_view_bands_shape {
        nc_number_of_views, nc_polarization_bands_per_view
    };
    const std::vector<netCDF::NcDim> geometry_shape {
        nc_bins_along_track, nc_spatial_samples_per_image
    };

    nc.writeReducedConfig(settings);

    // Work variables for NetCDF groups and variables
    netCDF::NcGroup grp {};
    netCDF::NcVar var {};

    grp = nc.addGroup("SENSOR_VIEW_BANDS");

    var = grp.addVar(
      "viewport_index", netCDF::ncUbyte, nc_spatial_samples_per_image);
    var.putAtt("long_name",
               "zero-based viewport index for each spatial sample");
    var.putAtt("_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
    var.putAtt("valid_min", netCDF::ncUbyte, 0);
    var.putAtt(
      "valid_max", netCDF::ncUbyte, static_cast<uint8_t>(ckd.n_vp - 1));
    std::vector<double> vp_angles(ckd.n_vp);
    std::transform(viewport_idx_to_angle.cbegin(),
                   viewport_idx_to_angle.cend(),
                   vp_angles.begin(),
                   [](const auto& pair) { return pair.second; });
    var.putAtt("mapping", netCDF::ncDouble, ckd.n_vp, vp_angles.data());
    std::vector<uint8_t> viewport_indices(ckd.fov.n_act);
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        const int i_end { i_vp + 1 < ckd.n_vp ? ckd.fov.start_vp[i_vp + 1]
                                              : ckd.fov.n_act };
        for (int i { ckd.fov.start_vp[i_vp] }; i < i_end; ++i) {
            viewport_indices[i] = static_cast<uint8_t>(i_vp);
        }
    }
    var.putVar(viewport_indices.data());

    var = grp.addVar(
      "intensity_wavelengths", netCDF::ncFloat, intensity_view_bands_shape);
    var.putAtt("long_name", "intensity field center wavelengths at each view");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    // The official spectral coverage of SPEXone is 385–770 nm but
    // here we have taken the values by inspecting a CKD.
    static constexpr float wavelengths_valid_min { 340.0f };
    static constexpr float wavelengths_valid_max { 800.0f };
    var.putAtt("valid_min", netCDF::ncFloat, wavelengths_valid_min);
    var.putAtt("valid_max", netCDF::ncFloat, wavelengths_valid_max);
    var.putAtt("units", "nm");
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        var.putVar({ static_cast<size_t>(i_vp), 0 },
                   { 1, settings.l1b.intensity_wavelength.size() },
                   settings.l1b.intensity_wavelength.data());
    }

    var = grp.addVar(
      "intensity_bandpasses", netCDF::ncFloat, intensity_view_bands_shape);
    var.putAtt("long_name", "intensity field bandpasses at each view");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_max", netCDF::ncFloat, 100.0f);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 100.0f);
    var.putAtt("units", "nm");
    const std::vector<float> intensity_bandpasses(
      ckd.n_vp * settings.l1b.intensity_wavelength.size(), 2.0f);
    var.putVar(intensity_bandpasses.data());

    var =
      grp.addVar("intensity_F0", netCDF::ncFloat, intensity_view_bands_shape);
    var.putAtt("long_name", "intensity band solar irradiance");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 5000.0f);
    var.putAtt("units", "W m-2 um-1");
    std::vector<float> intensity_F0_f(ckd.n_vp * intensity_F0.size());
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        for (int j {}; j < static_cast<int>(intensity_F0.size()); ++j) {
            intensity_F0_f[i_vp * intensity_F0.size() + j] =
              static_cast<float>(intensity_F0[j]);
        }
    }
    var.putVar(intensity_F0_f.data());

    var = grp.addVar("polarization_wavelengths",
                     netCDF::ncFloat,
                     polarization_view_bands_shape);
    var.putAtt("long_name", "polarization field wavelengths at each view");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, wavelengths_valid_min);
    var.putAtt("valid_max", netCDF::ncFloat, wavelengths_valid_max);
    var.putAtt("units", "nm");
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        var.putVar({ static_cast<size_t>(i_vp), 0 },
                   { 1, settings.l1b.l1b_wavelength.size() },
                   settings.l1b.l1b_wavelength.data());
    }

    var = grp.addVar("polarization_bandpasses",
                     netCDF::ncFloat,
                     polarization_view_bands_shape);
    var.putAtt("long_name", "polarization field bandpasses at each view");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 2.5f);
    var.putAtt("valid_max", netCDF::ncFloat, 100.0f);
    var.putAtt("units", "nm");
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        var.putVar({ static_cast<size_t>(i_vp), 0 },
                   { 1, settings.l1b.l1b_wavelength.size() },
                   prod.polarization_bandpasses.data());
    }

    var = grp.addVar(
      "polarization_F0", netCDF::ncFloat, polarization_view_bands_shape);
    var.putAtt("long_name", "polarization band solar irradiance");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 5000.0f);
    var.putAtt("units", "W m-2 um-1");
    std::vector<float> polarization_F0_f(ckd.n_vp * polarization_F0.size());
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        for (int j {}; j < static_cast<int>(polarization_F0.size()); ++j) {
            polarization_F0_f[i_vp * polarization_F0.size() + j] =
              static_cast<float>(polarization_F0[j]);
        }
    }
    var.putVar(polarization_F0_f.data());

    grp = nc.addGroup("BIN_ATTRIBUTES");

    var = grp.addVar("image_time", netCDF::ncDouble, nc_bins_along_track);
    var.putAtt("long_name", "image time (seconds of day)");
    var.putAtt("_FillValue", netCDF::ncDouble, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncDouble, 0.0);
    var.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
    var.putAtt("units", "seconds");
    var.putVar(prod.image_time.data());

    grp = nc.addGroup("GEOLOCATION_DATA");

    var = grp.addVar("latitude", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "latitudes of bin locations");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    var.putAtt("units", "degrees_north");
    moduloAndConvertAngles(prod.lat);
    var.putVar(prod.lat.data());

    var = grp.addVar("longitude", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "longitudes of bin locations");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -180.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 180.0f);
    var.putAtt("units", "degrees_east");
    moduloAndConvertAngles(prod.lon);
    var.putVar(prod.lon.data());

    var = grp.addVar("altitude", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "altitude at bin locations");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -1000.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 10000.0f);
    var.putAtt("units", "m");
    var.putVar(prod.alt.data());

    var = grp.addVar("sensor_zenith", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "sensor zenith angle");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    var.putAtt("units", "degrees");
    moduloAndConvertAngles(prod.vza);
    var.putVar(prod.vza.data());

    var = grp.addVar("sensor_azimuth", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "sensor azimuth angle");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -180.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 180.0f);
    var.putAtt("units", "degrees");
    moduloAndConvertAngles(prod.vaa);
    var.putVar(prod.vaa.data());

    var = grp.addVar("solar_zenith", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "solar zenith angle at bin locations");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    var.putAtt("units", "degrees");
    moduloAndConvertAngles(prod.sza);
    var.putVar(prod.sza.data());

    var = grp.addVar("solar_azimuth", netCDF::ncFloat, geometry_shape);
    var.putAtt("long_name", "solar azimuth angle at bin locations");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -180.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 180.0f);
    var.putAtt("units", "degrees");
    moduloAndConvertAngles(prod.saa);
    var.putVar(prod.saa.data());

    grp = nc.addGroup("OBSERVATION_DATA");

    var = grp.addVar("I", netCDF::ncFloat, intensity_shape);
    var.putAtt("long_name", "I Stokes vector component");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "W m-2 sr-1 um-1");
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 999.0f);
    var.putVar(prod.I.data());

    var = grp.addVar("I_noise", netCDF::ncFloat, intensity_shape);
    var.putAtt("long_name", "random noise of I in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "W m-2 sr-1 um-1");
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 800.0f);
    var.putVar(prod.I_noise.data());

    var = grp.addVar("I_polsample", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name",
               "I Stokes vector component at polarimeter spectral sampling");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 999.0f);
    var.putAtt("units", "W m-2 sr-1 um-1");
    var.putVar(prod.I_polsample.data());

    var = grp.addVar("I_polsample_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name",
               "noise (standard deviation) of I at polarimeter sampling");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 800.0f);
    var.putAtt("units", "W m-2 sr-1 um-1");
    var.putVar(prod.I_polsample_noise.data());

    var = grp.addVar("q", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "Q over I (little q) Stokes vector component");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -1.5f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.5f);
    var.putAtt("units", "1");
    var.putVar(prod.q.data());

    var = grp.addVar("q_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "random noise of q in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "1");
    var.putVar(prod.q_noise.data());

    var = grp.addVar("u", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "U over I (little u) Stokes vector component");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -1.5f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.5f);
    var.putAtt("units", "1");
    var.putVar(prod.u.data());

    var = grp.addVar("u_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "random noise of u in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "1");
    var.putVar(prod.u_noise.data());

    var = grp.addVar("DoLP", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "degree of linear polarization");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    var.putAtt("units", "1");
    var.putVar(prod.dolp.data());

    var = grp.addVar("DoLP_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "noise (standard deviation) of DoLP");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "1");
    var.putVar(prod.dolp_noise.data());

    var = grp.addVar("AoLP", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "angle of linear polarization");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    var.putAtt("units", "degrees");
    var.putVar(prod.aolp.data());

    var = grp.addVar("AoLP_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "noise (standard deviation) of AoLP");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "degrees");
    var.putVar(prod.aolp_noise.data());

    // Copy navigation data and selected image attributes from L1A to L1B
    if (settings.l1b.geolocation) {
        copyNavigationData(*l1a_meta->nc, nc);
        copyImageAttributes(settings.l1b.image_start,
                            settings.l1b.image_end.value_or(
                              settings.l1b.image_start + nL1A() - 1),
                            *l1a_meta->nc,
                            nc);
    }
}

auto L1B::processFinalize() -> void
{
    Timer timer_io {};
    timer_io.start();
    output();
    timer_io.stop();
    spdlog::info("Time for writing output:  {:8.3f}", timer_io.time());
    spdlog::debug("Demodulation error propagation timings:");
    spdlog::debug("           Sx_BT: {:8.3f} s ({} calls)",
                  timer_Sx_BT.time(),
                  timer_Sx_BT.calls());
    spdlog::debug("             Squ: {:8.3f} s ({} calls)",
                  timer_Squ.time(),
                  timer_Squ.calls());
    spdlog::debug("             SQU: {:8.3f} s ({} calls)",
                  timer_SQU.time(),
                  timer_SQU.calls());
    spdlog::debug("             Smt: {:8.3f} s ({} calls)",
                  timer_Smt.time(),
                  timer_Smt.calls());
    spdlog::debug("             SXT: {:8.3f} s ({} calls)",
                  timer_SXT.time(),
                  timer_SXT.calls());
    spdlog::debug("      S_QUI_conv: {:8.3f} s ({} calls)",
                  timer_S_QUI_conv.time(),
                  timer_S_QUI_conv.calls());
}

} // namespace spex
