// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for generating a level 1B product

#pragma once

#include "dem.h"
#include "packed_matrix.h"
#include "processor.h"
#include "timer.h"

namespace spex {

class L1B : public Processor
{
public:
    L1B(const SPEXSettings& settings, CKD& ckd);
    ~L1B() = default;

protected:
    auto processInit() -> void override;
    // Each image is in its own batch. Since they're read from the
    // same file, open the NetCDF file and leave it open for the
    // entire process. It will be automatically closed in the end.
    auto prepareBatches() -> void override;
    auto processBatch(const int i_img) -> void override;
    // Write output and show some timings
    auto processFinalize() -> void override;

private:
    // Demodulation work variables
    struct DemodWork
    {
        // Radiance on the fine wavelength grid
        std::vector<double> rad {};
        std::vector<double> dolp {};
        std::vector<double> aolp {};
        // Help array from geolocation that is required for the
        // demodulation, for transferring polarization to local meridian
        // plane.
        std::vector<double> aolp_shift {};
        // Modulation (G) values
        std::vector<double> modulation {};
    };

    // Covariance matrices and related quantities
    struct CovMatrices
    {
        // Linear inversion covariance matrix Sx
        std::vector<double> x {};
        // The q-quarter of Sx
        std::vector<double> xq {};
        // The u-quarter of Sx
        std::vector<double> xu {};
        // The qu-quarter of Sx
        std::vector<double> xqu {};
        // (Sxq*B^T)^T where B is the B-spline matrix
        std::vector<double> xq_BT_T {};
        // (Sxu*B^T)^T
        std::vector<double> xu_BT_T {};
        // (Sxu*B^T)^T
        std::vector<double> xqu_BT_T {};
        // Covariance matrices of q, u, and their cross-term on the fine
        // wavelength grid
        std::vector<double> q {};
        std::vector<double> u {};
        std::vector<double> qu {};
        // Covariance matrices of Q, U, and their cross-terms, including
        // with I, on the fine wavelength grid
        std::vector<double> Q {};
        std::vector<double> U {};
        std::vector<double> QU {};
        std::vector<double> IQ {};
        std::vector<double> IU {};
        // Covariance matrices related to application of telescope
        std::vector<double> Q_1 {};
        std::vector<double> I_var_1 {};
        std::vector<double> IQ_1 {};
        std::vector<double> IU_1 {};
        std::vector<double> QU_1 {};
        // Modulation (G) variances
        std::vector<double> modulation {};
        // Intensity variances
        std::vector<double> I_var {};
    };

    // Binned spectral index marking the beginning of the retrieval
    // window per FOV (determined by [l1b][l1b_wavelength])
    std::vector<int> fov_start_bin {};
    // Number of measurements in the retrieval window per FOV
    std::vector<int> fov_nmeas {};
    // Number of linear inversion states when doing a B-spline fit is
    // close to twice the number of knots because both q and u are
    // included.
    int n_demod_states {};
    int n_demod_states_half {};
    // Size of matrices with dimensions (n_demod_states, n_demod_states)
    int demod_size_full {};
    int demod_size_half {};
    std::vector<std::unique_ptr<PackedMatrix>> demod_spline_matrix {};
    // Demodulation Gaussian kernel matrix
    std::vector<std::unique_ptr<PackedMatrix>> gauss_kernel {};

    // Geolocation related
    std::unique_ptr<DEM> dem {};

    // All variables that constitute the L1B Product (except those
    // that are easy enough to comput on the fly)
    struct
    {
        // Sensor view bands
        std::vector<float> polarization_bandpasses {};
        // Bin attributes
        std::vector<double> image_time {};
        // Geolocation data
        std::vector<float> lat {};
        std::vector<float> lon {};
        std::vector<float> alt {};
        std::vector<float> vza {};
        std::vector<float> vaa {};
        std::vector<float> sza {};
        std::vector<float> saa {};
        // Observation data
        std::vector<float> I {};
        std::vector<float> I_noise {};
        std::vector<float> I_polsample {};
        std::vector<float> I_polsample_noise {};
        std::vector<float> q {};
        std::vector<float> q_noise {};
        std::vector<float> u {};
        std::vector<float> u_noise {};
        std::vector<float> dolp {};
        std::vector<float> dolp_noise {};
        std::vector<float> aolp {};
        std::vector<float> aolp_noise {};
    } prod;

    // Solar spectrum related
    std::vector<double> intensity_F0 {};
    std::vector<double> polarization_F0 {};

    // We need an extra handle on L1A meta because otherwise it would
    // become a null pointer by the time we need to use when writing
    // output.
    std::shared_ptr<L1AMetadata> l1a_meta {};

    // Timers for the most expensive demodulation steps (only shown in
    // debug mode)
    Timer timer_Sx_BT {};
    Timer timer_Squ {};
    Timer timer_SQU {};
    Timer timer_Smt {};
    Timer timer_SXT {};
    Timer timer_S_QUI_conv {};

    // Compute the geolocation information for all ACT angles
    // corresponding to image (ALT position) i_img. Also generate the
    // AoLP shifts.
    auto geolocate(const int i_img,
                   const spex::L1A& l1a,
                   std::vector<double>& aolp_shift) -> void;

    // Compute the Stokes parameters and related quantities along with
    // their noise estimate for ALT position i_img and ACT angle
    // i_fov.
    auto demodulateSpectrum(const int i_img,
                            const int i_fov,
                            const Spectra& spectra,
                            DemodWork& wrk,
                            CovMatrices& S) -> void;

    // Create a level L1B product file
    auto output() -> void;
};

} // namespace spex
