// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1c.h"

#include "cubic_spline.h"
#include "dem.h"
#include "geometry.h"
#include "timer.h"

#include <span>

namespace spex {

// Scale factor for some of the geolocation data variables
constexpr float geo_scale_factor { 100.0f };

// Main dimensions used in this module
struct Dims
{
    // Number of rows in the list of L1C latitudes (nc variable
    // bins_across_track)
    int n_l1c_rows {};
    // Number of columns in the list of L1C latitudes (nc variable
    // bins_across_track)
    int n_l1c_cols {};
    // Number of rows and columns in the dense L1C grid
    int n_l1c_rows_dbl {};
    int n_l1c_cols_dbl {};
    // Number L1B rows and columns
    int n_l1b_rows {};
    int n_l1b_cols {};
    // Lowest row in the list of L1B latitudes used for interpolation
    int min_l1b_row { def::idx };
    // Highest L1B row used for interpolation
    int max_l1b_row { def::idx };
    // Lowest and highest L1C rows that should contain data. Normally
    // these values are 0 and the number of L1C rows
    // respectively. However, if the L1C grid is partially outside the
    // L1B grid, anything that is outside the min_l1c_row:max_l1c_row
    // range for a given viewport is set to a fill value. These are
    // computed separately for each viewport.
    int min_l1c_row {};
    int max_l1c_row {};
    // L1C row indices of L1B points
    std::vector<double> row_indices {};
    // L1C column indices of L1B points
    std::vector<double> col_indices {};
    // L1B columns indices interpolated along the L1C rows
    std::vector<double> interp_cols {};
    // For a L1C given row, number of columns on the left that should
    // be set to fill values because they are outside the L1B grid.
    std::vector<int> fill_cols_left {};
    // Index of the first L1C column on the right that is outside the
    // L1B swath and should set to fill values.
    std::vector<int> fill_cols_right {};
    // When dealing with L1C grids that cross the -180/180 deg
    // longitude boundary, bring all values to the 0 .. 360 range by
    // adding 2*pi to the negative values.
    bool add_2pi { false };
};

// Unbin and interpolate data from an MxN grid to a 2Mx2N grid
static auto expandGrid(const int n_rows,
                       const int n_cols,
                       std::vector<float>& data) -> void
{
    constexpr int bin { 2 };
    // First interpolate across rows
    std::vector<double> data_semi_dense(bin * data.size());
    std::vector<double> x_values(n_rows);
    std::vector<double> y_values(n_rows);
    for (int j {}; j < n_cols; ++j) {
        for (int i {}; i < n_rows; ++i) {
            x_values[i] = static_cast<double>(i);
            y_values[i] = static_cast<double>(data[i * n_cols + j]);
        }
        const CubicSpline spline { x_values, y_values };
        for (int i {}; i < bin * n_rows; ++i) {
            // Index 0 becomes -0.25, 0.25 after unbinning. When
            // working with the dense we move in steps of 0.5 in terms
            // of the original grid.
            const double arg_x { -0.25 + i * 0.5 };
            data_semi_dense[i * n_cols + j] = spline.eval(arg_x);
        }
    }
    // Next interpolate across columns
    std::vector<float> data_dense(bin * bin * data.size());
    x_values.resize(n_cols);
    y_values.resize(n_cols);
    for (int i {}; i < bin * n_rows; ++i) {
        for (int j {}; j < n_cols; ++j) {
            x_values[j] = j;
            y_values[j] = data_semi_dense[i * n_cols + j];
        }
        const CubicSpline spline { x_values, y_values };
        for (int j {}; j < bin * n_cols; ++j) {
            const double arg_y { -0.25 + j * 0.5 };
            data_dense[i * bin * n_cols + j] =
              static_cast<float>(spline.eval(arg_y));
        }
    }
    data = std::move(data_dense);
}

// Read in the L1C latitudes/longitudes, create a dense L1C, and
// compute the L1C normal vectors.
static auto prepareL1CGrid(const SPEXSettings& settings,
                           Dims& dims,
                           std::vector<float>& l1c_latitudes,
                           std::vector<float>& l1c_longitudes,
                           std::vector<float>& l1c_latitudes_dense,
                           std::vector<float>& l1c_longitudes_dense,
                           std::vector<double>& l1c_mid_vectors,
                           std::vector<double>& normal_vectors_r,
                           std::vector<double>& normal_vectors_c,
                           std::vector<double>& deltas_r,
                           std::vector<double>& deltas_c) -> void
{
    const NcObject nc_l1c { netCDF::NcFile::read, settings.l1c.l1c_in };
    dims.n_l1c_rows =
      static_cast<int>(nc_l1c.getDim("bins_along_track").getSize());
    dims.n_l1c_cols =
      static_cast<int>(nc_l1c.getDim("bins_across_track").getSize());
    dims.n_l1c_rows_dbl = 2 * dims.n_l1c_rows;
    dims.n_l1c_cols_dbl = 2 * dims.n_l1c_cols;
    l1c_latitudes_dense.resize(dims.n_l1c_rows * dims.n_l1c_cols);
    l1c_longitudes_dense.resize(dims.n_l1c_rows * dims.n_l1c_cols);
    const auto geolocation_data { nc_l1c.getGroup("geolocation_data") };
    geolocation_data.getVar("latitude").getVar(l1c_latitudes_dense.data());
    geolocation_data.getVar("longitude").getVar(l1c_longitudes_dense.data());
    for (auto& val : l1c_latitudes_dense) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    for (auto& val : l1c_longitudes_dense) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    l1c_latitudes = l1c_latitudes_dense;
    l1c_longitudes = l1c_longitudes_dense;
    // If the first and last longitude have opposite signs and one of
    // the values (test the middle one) is at least 90 deg we are
    // crossing the -180/180 deg boundary. This signals that we should
    // add 2*pi to all negative values to bring longitudes into the
    // range 0...360.
    dims.add_2pi =
      (l1c_longitudes_dense.front() < 0.0 ^ l1c_longitudes_dense.back() < 0.0)
      && std::abs(l1c_longitudes_dense[l1c_longitudes_dense.size() / 2])
           > std::numbers::pi / 2;
    if (dims.add_2pi) {
        for (auto& lon : l1c_longitudes_dense) {
            if (lon < 0.0) {
                lon += 2 * std::numbers::pi;
            }
        }
    }
    expandGrid(dims.n_l1c_rows, dims.n_l1c_cols, l1c_latitudes_dense);
    expandGrid(dims.n_l1c_rows, dims.n_l1c_cols, l1c_longitudes_dense);
    // Next compute the L1C unit vectors from latitude/longitude
    //     ( cos(lon)cos(lat) )
    // G = ( sin(lon)cos(lat) )
    //     ( sin(lat)         )
    std::vector<double> l1c_unit_vectors(dims.n_l1c_rows_dbl
                                         * dims.n_l1c_cols_dbl * dims::vec);
    for (int i {}; i < dims.n_l1c_rows_dbl * dims.n_l1c_cols_dbl; ++i) {
        l1c_unit_vectors[i * dims::vec] =
          std::cos(l1c_longitudes_dense[i]) * std::cos(l1c_latitudes_dense[i]);
        l1c_unit_vectors[i * dims::vec + 1] =
          std::sin(l1c_longitudes_dense[i]) * std::cos(l1c_latitudes_dense[i]);
        l1c_unit_vectors[i * dims::vec + 2] = std::sin(l1c_latitudes_dense[i]);
    }

    // Compute the L1C row and column normal vectors
    l1c_mid_vectors.resize(dims.n_l1c_rows_dbl * dims::vec);
    normal_vectors_r.resize(dims.n_l1c_rows_dbl * dims::vec);
    normal_vectors_c.resize(dims.n_l1c_rows_dbl * dims::vec);
    std::array<double, dims::vec> cross {}; // cross product
    const int mid_col { dims.n_l1c_cols_dbl / 2 };
    for (int i_row {}; i_row < dims.n_l1c_rows_dbl; ++i_row) {
        crossProduct(
          // Last column of the row
          std::span(&l1c_unit_vectors[((i_row + 1) * dims.n_l1c_cols_dbl - 1)
                                      * dims::vec],
                    dims::vec),
          // First column of the row
          std::span(&l1c_unit_vectors[i_row * dims.n_l1c_cols_dbl * dims::vec],
                    dims::vec),
          cross);
        const double norm { vectorNorm(cross) };
        for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
            normal_vectors_r[i_row * dims::vec + i_vec] = cross[i_vec] / norm;
        }
        // Because there is an even number of L1C columns, the middle
        // vector is defined as the average of the middle two.
        std::span<double> mid_vector {
            l1c_mid_vectors.begin() + i_row * dims::vec, dims::vec
        };
        for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
            mid_vector[i_vec] =
              (l1c_unit_vectors[(i_row * dims.n_l1c_cols_dbl + mid_col)
                                  * dims::vec
                                + i_vec]
               + l1c_unit_vectors[(i_row * dims.n_l1c_cols_dbl + mid_col + 1)
                                    * dims::vec
                                  + i_vec])
              / 2;
        }
        const double norm_mid { vectorNorm(mid_vector) };
        for (auto& val : mid_vector) {
            val /= norm_mid;
        }
        crossProduct(std::span(mid_vector.data(), dims::vec),
                     std::span(&normal_vectors_r[i_row * dims::vec], dims::vec),
                     cross);
        for (int i_vec {}; i_vec < dims::vec; ++i_vec) {
            normal_vectors_c[i_row * dims::vec + i_vec] = cross[i_vec];
        }
    }
    // Compute grid resolutions across rows and columns for this viewport
    deltas_r.resize(dims.n_l1c_rows_dbl);
    deltas_c.resize(dims.n_l1c_rows_dbl, 0.0);
    for (int i_row {}; i_row < dims.n_l1c_rows_dbl; ++i_row) {
        // Across rows, the resolution (half of the angular grid
        // spacing) is determined by the dot product between the unit
        // vectors of this and the next row. If looking at the last
        // row then compute the dot product with the previous row
        // instead.
        const int sign { i_row < dims.n_l1c_rows_dbl - 1 ? 1 : -1 };
        deltas_r[i_row] =
          std::acos(
            dotProduct(l1c_mid_vectors.begin() + i_row * dims::vec,
                       l1c_mid_vectors.begin() + (i_row + sign) * dims::vec))
          / 2;
        // Across columns, compute the average resolution of the given
        // row. Since we are looking at pairs, the iteration limit is
        // one less than the number of columns.
        for (int i_col {}; i_col < dims.n_l1c_cols_dbl - 1; ++i_col) {
            const int idx { i_row * dims.n_l1c_cols_dbl + i_col };
            const double norm1 { vectorNorm(std::span<double> {
              l1c_unit_vectors.begin() + idx * dims::vec, dims::vec }) };
            const double norm2 { vectorNorm(std::span<double> {
              l1c_unit_vectors.begin() + (idx + 1) * dims::vec, dims::vec }) };
            deltas_c[i_row] +=
              std::acos(
                dotProduct(l1c_unit_vectors.begin() + idx * dims::vec,
                           l1c_unit_vectors.begin() + (idx + 1) * dims::vec)
                / norm1 / norm2)
              / 2;
        }
        deltas_c[i_row] /= dims.n_l1c_cols_dbl - 1;
    }
}

// Assign a level 1C row and column to each L1B grid point
static auto genL1CCoordinates(const SPEXSettings& settings,
                              const std::vector<double>& l1b_lat_vp,
                              const std::vector<double>& l1b_lon_vp,
                              const std::vector<float>& l1c_latitudes,
                              const std::vector<float>& l1c_longitudes,
                              std::vector<double>& l1c_mid_vectors,
                              std::vector<double>& normal_vectors_r,
                              std::vector<double>& normal_vectors_c,
                              const std::vector<double>& deltas_r,
                              const std::vector<double>& deltas_c,
                              Dims& dims) -> void
{
    // Compute the L1B unit vectors (same formula as above)
    std::vector<double> l1b_unit_vectors(dims.n_l1b_rows * dims.n_l1b_cols
                                         * dims::vec);
    for (int i {}; i < dims.n_l1b_rows * dims.n_l1b_cols; ++i) {
        l1b_unit_vectors[i * dims::vec] =
          std::cos(l1b_lon_vp[i]) * std::cos(l1b_lat_vp[i]);
        l1b_unit_vectors[i * dims::vec + 1] =
          std::sin(l1b_lon_vp[i]) * std::cos(l1b_lat_vp[i]);
        l1b_unit_vectors[i * dims::vec + 2] = std::sin(l1b_lat_vp[i]);
    }
    // Compute the target L1C row and column indices
    std::vector<int> l1c_rows(dims.n_l1b_rows * dims.n_l1b_cols, def::idx);
    std::vector<int> l1c_cols(dims.n_l1b_rows * dims.n_l1b_cols, def::idx);
    const int mid_col { dims.n_l1c_cols_dbl / 2 };
#pragma omp parallel for
    for (int i_l1b = 0; i_l1b < dims.n_l1b_rows * dims.n_l1b_cols; ++i_l1b) {
        for (int i_l1c {}; i_l1c < dims.n_l1c_rows_dbl; ++i_l1c) {
            const double angle_r { dotProduct(
              l1b_unit_vectors.begin() + i_l1b * dims::vec,
              normal_vectors_r.begin() + i_l1c * dims::vec) };
            // Include a small margin to account for numerical
            // precision. It could be that if a point sits exactly in
            // between two L1C rows it is not associated with either
            // of them.
            constexpr double prec_margin { 1e-6 };
            if (std::abs(angle_r) <= deltas_r[i_l1c] + prec_margin) {
                // Guard against the case where the angle is small but
                // the L1B and L1C grid points are on the opposite
                // side of the Earth.
                if (dotProduct(l1c_mid_vectors.begin() + i_l1c * dims::vec,
                               l1b_unit_vectors.begin() + i_l1b * dims::vec)
                    < 0.0) {
                    continue;
                }
                l1c_rows[i_l1b] = (angle_r < 0.0) ? i_l1c : i_l1c - 1;
                if (l1c_rows[i_l1b] >= dims.n_l1c_rows_dbl - 1) {
                    l1c_rows[i_l1b] = def::idx;
                }
                const double angle_c { dotProduct(
                  l1b_unit_vectors.begin() + i_l1b * dims::vec,
                  normal_vectors_c.begin() + i_l1c * dims::vec) };
                l1c_cols[i_l1b] = static_cast<int>(
                  mid_col + 0.5 + angle_c / (2 * deltas_c[i_l1c]));
                // If the L1B and L1C grids have no overlap it could
                // still be that the angle between the L1B unit vector
                // and the L1C row normal vector is small. If so, the
                // assigned L1C column index is out of range. Compare
                // that to the L1C swath width plus a small margin for
                // numerical inaccuracies.
                constexpr int col_margin { 1 };
                if (l1c_cols[i_l1b] < -col_margin
                    || l1c_cols[i_l1b] > dims.n_l1c_cols_dbl + col_margin) {
                    l1c_rows[i_l1b] = def::idx;
                    l1c_cols[i_l1b] = def::idx;
                } else {
                    break;
                }
            }
        }
    }
    // Check if the L1C grid is partially outside the L1B grid of this
    // viewport.
    dims.min_l1c_row = 0;
    dims.max_l1c_row = dims.n_l1c_rows_dbl;
    for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
        dims.min_l1c_row = std::max(dims.min_l1c_row, l1c_rows[i_col]);
        const int l1c_row_upper { *(l1c_rows.crbegin() + i_col) };
        if (l1c_row_upper != def::idx) {
            dims.max_l1c_row = std::min(dims.max_l1c_row, l1c_row_upper + 1);
        }
    }

    // Find usable L1B range
    for (int i_row {}; i_row < dims.n_l1b_rows; ++i_row) {
        bool row_clean { true };
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            if (l1c_rows[i_row * dims.n_l1b_cols + i_col] == def::idx) {
                row_clean = false;
                break;
            }
        }
        if (row_clean) {
            if (dims.min_l1b_row == def::idx) {
                dims.min_l1b_row = i_row;
            }
            dims.max_l1b_row = i_row + 1;
        }
    }
    if (dims.min_l1b_row == def::idx && dims.max_l1b_row == def::idx) {
        throw std::runtime_error { "L1B and L1C grids have no overlap" };
    }
    // Extend the range a little bit in order to make the
    // interpolation more accurate. For this we copy the L1C row and
    // column indices to the rows that outside the L1C range. That is
    // strictly speaking not correct but then the relative coordinates
    // would assume large negative or positive values values which
    // makes it okay. Basically, we are linearly interpolating the L1C
    // row indices outside its range.
    constexpr int n_extra_l1b_rows { 2 };
    for (int i_row { dims.min_l1b_row - n_extra_l1b_rows };
         i_row < dims.min_l1b_row;
         ++i_row) {
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            if (i_row >= 0) {
                l1c_rows[i_row * dims.n_l1b_cols + i_col] =
                  l1c_rows[dims.min_l1b_row * dims.n_l1b_cols + i_col];
                l1c_cols[i_row * dims.n_l1b_cols + i_col] =
                  l1c_cols[dims.min_l1b_row * dims.n_l1b_cols + i_col];
            }
        }
    }
    dims.min_l1b_row = std::max(0, dims.min_l1b_row - n_extra_l1b_rows);
    for (int i_row { dims.max_l1b_row };
         i_row < dims.max_l1b_row + n_extra_l1b_rows;
         ++i_row) {
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            if (i_row < dims.n_l1b_rows) {
                l1c_rows[i_row * dims.n_l1b_cols + i_col] =
                  l1c_rows[(dims.max_l1b_row - 1) * dims.n_l1b_cols + i_col];
                l1c_cols[i_row * dims.n_l1b_cols + i_col] =
                  l1c_cols[(dims.max_l1b_row - 1) * dims.n_l1b_cols + i_col];
            }
        }
    }
    dims.max_l1b_row =
      std::min(dims.n_l1b_rows, dims.max_l1b_row + n_extra_l1b_rows);
    spdlog::debug("Using L1B rows {}...{} for the interpolation",
                  dims.min_l1b_row,
                  dims.max_l1b_row - 1);

    // Make sure there are no bad L1C row or column indices within the
    // usable L1B row range.
    for (int i_row { dims.min_l1b_row }; i_row < dims.max_l1b_row; ++i_row) {
        for (int i_col {}; i_col < dims.n_l1b_cols; ++i_col) {
            const int idx { i_row * dims.n_l1b_cols + i_col };
            if (l1c_rows[idx] == def::idx || l1c_cols[idx] == def::idx) {
                throw std::runtime_error {
                    "L1B point (" + std::to_string(i_row) + ", "
                    + std::to_string(i_col)
                    + ") has a bad L1C row/column index: "
                    + std::to_string(l1c_rows[idx]) + ", "
                    + std::to_string(l1c_cols[idx])
                };
            }
        }
    }

    // Compute the relative coordinates of an input point within an
    // L1C cell and then the full coordinates in terms of fractional
    // L1C row and column indices.
    dims.row_indices.resize(dims.n_l1b_rows * dims.n_l1b_cols);
    dims.col_indices.resize(dims.n_l1b_rows * dims.n_l1b_cols);
#pragma omp parallel for
    for (int i_row_l1b = dims.min_l1b_row; i_row_l1b < dims.max_l1b_row;
         ++i_row_l1b) {
        for (int i_col_l1b {}; i_col_l1b < dims.n_l1b_cols; ++i_col_l1b) {
            const int l1c_row {
                l1c_rows[i_row_l1b * dims.n_l1b_cols + i_col_l1b]
            };
            const int l1c_col {
                l1c_cols[i_row_l1b * dims.n_l1b_cols + i_col_l1b]
            };
            // Input location
            std::array<double, 2> p_x {
                l1b_lon_vp[i_row_l1b * dims.n_l1b_cols + i_col_l1b],
                l1b_lat_vp[i_row_l1b * dims.n_l1b_cols + i_col_l1b]
            };
            // Identify the three corners of the target L1C cell. The
            // fourth corner p11 is not needed for anything and is
            // thus omitted here.
            std::array<double, 2> p_00 {
                l1c_longitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col],
                l1c_latitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col]
            };
            std::array<double, 2> p_01 {
                l1c_longitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col + 1],
                l1c_latitudes[l1c_row * dims.n_l1c_cols_dbl + l1c_col + 1]
            };
            std::array<double, 2> p_10 {
                l1c_longitudes[(l1c_row + 1) * dims.n_l1c_cols_dbl + l1c_col],
                l1c_latitudes[(l1c_row + 1) * dims.n_l1c_cols_dbl + l1c_col]
            };

            // Shift the local coordinate system origin to zero. The
            // origin is defined to be p_00.
            p_x[0] -= p_00[0];
            p_x[1] -= p_00[1];
            p_01[0] -= p_00[0];
            p_01[1] -= p_00[1];
            p_10[0] -= p_00[0];
            p_10[1] -= p_00[1];
            // p_00 becomes 0 and is thus omitted wherever it needs to be used

            // Rotate along the x-axis such that the latitude of the
            // right bottom cell corner is zero.
            const double theta { -std::atan(p_01[1] / p_01[0]) };
            std::array<double, 4> rot_mat { std::cos(theta),
                                            -std::sin(theta),
                                            std::sin(theta),
                                            std::cos(theta) };
            // Return Rx where R is the rotation matrix
            const auto multRotMat { [&rot_mat](const std::array<double, 2>& x) {
                return std::array<double, 2> {
                    rot_mat[0] * x[0] + rot_mat[1] * x[1],
                    rot_mat[2] * x[0] + rot_mat[3] * x[1]
                };
            } };
            p_x = multRotMat(p_x);
            // p_00 is unaffected by this rotation
            p_01 = multRotMat(p_01);
            p_10 = multRotMat(p_10);
            // If the new longitude of the lower right corner of the
            // cell is negative then the orbit is descending. If so,
            // all cell coordinates should be flipped, otherwise, the
            // direction of the relative coordinates is incorrect
            // (they should be increasing in flight direction).
            if (p_01[0] < 0.0) {
                p_x = { -p_x[0], -p_x[1] };
                p_01 = { -p_01[0], -p_01[1] };
                p_10 = { -p_10[0], -p_10[1] };
            }

            // Undo skew
            const double alpha { 90.0 * math::deg_to_rad
                                 - std::atan(p_10[0] / p_10[1]) };
            rot_mat = { 1.0,
                        -1.0 / std::tan(alpha),
                        0.0,
                        std::abs(1.0 / std::sin(alpha)) };
            p_x = multRotMat(p_x);
            // p_00 is unaffected by this rotation
            p_01 = multRotMat(p_01);
            p_10 = multRotMat(p_10);

            // Normalize the relative coordinates to be in the range 0..1
            const double Dx { std::abs(p_01[0]) };
            const double Dy { std::abs(p_10[1]) };
            p_x = { p_x[0] / Dx, p_x[1] / Dy };

            // Add the relative cell coordinates to the row and cell
            // numbers to get the final L1B coordinates in terms of
            // the fractional L1C row and column numbers.
            dims.row_indices[i_row_l1b * dims.n_l1b_cols + i_col_l1b] =
              l1c_rows[i_row_l1b * dims.n_l1b_cols + i_col_l1b] + p_x[1];
            dims.col_indices[i_row_l1b * dims.n_l1b_cols + i_col_l1b] =
              l1c_cols[i_row_l1b * dims.n_l1b_cols + i_col_l1b] + p_x[0];

            // Notify if one or both of the relative coordinates are outside the
            // range [0, 1] (exclude the extra rows)
            if (i_row_l1b >= dims.min_l1b_row + n_extra_l1b_rows
                && i_row_l1b < dims.max_l1b_row - n_extra_l1b_rows
                && (p_x[0] < 0.0 || p_x[0] > 1.0 || p_x[1] < 0.0
                    || p_x[1] > 1.0)) {
                spdlog::debug("Relative coordinates of L1B point ({}, {}) are "
                              "{:6.3f}, {:6.3f}",
                              i_row_l1b,
                              i_col_l1b,
                              p_x[0],
                              p_x[1]);
            }
        }
    }
    // See if there is a jump in the list of L1C row indices assigned
    // to the middle L1B column.
    for (int i { dims.min_l1b_row + 1 }; i < dims.max_l1b_row; ++i) {
        const double this_row {
            dims.row_indices[i * dims.n_l1b_cols + mid_col]
        };
        const double prev_row {
            dims.row_indices[(i - 1) * dims.n_l1b_cols + mid_col]
        };
        if (std::abs(this_row - prev_row) > settings.l1c.acceptable_gap_size) {
            throw std::runtime_error { "L1B rows " + std::to_string(i - 1)
                                       + " and " + std::to_string(i)
                                       + " are assigned L1C rows "
                                       + std::to_string(prev_row) + " and "
                                       + std::to_string(this_row)
                                       + ". The gap is larger than allowed." };
        }
    }
}

// 2D interpolate a quantity stored in l1b_buf and 2x2 bin the result
// and store it in l1c_buf.
static auto interpolateAndBin(Dims& dims,
                              const std::vector<double>& l1b_buf,
                              std::vector<double>& l1c_buf) -> void
{
    // First interpolate across the L1B rows
    const int n_rows { dims.max_l1b_row - dims.min_l1b_row };
    // L1B data after interpolation across rows
    std::vector<double> data_semi(dims.n_l1c_rows_dbl * dims.n_l1b_cols);
    // x-values for the first interpolation are the fractional L1C row indices
    std::vector<double> x_values(n_rows);
    // y-values are the L1B data or the fractional L1C column indices
    // if the latter has not been computed yet.
    std::vector<double> y_values(n_rows);
    // The new column indices need to be computed once per viewport
    if (dims.interp_cols.empty()) {
        dims.interp_cols.resize(dims.n_l1c_rows_dbl * dims.n_l1b_cols);
        for (int j {}; j < dims.n_l1b_cols; ++j) {
            for (int i {}; i < n_rows; ++i) {
                const int idx { (dims.min_l1b_row + i) * dims.n_l1b_cols + j };
                x_values[i] = dims.row_indices[idx];
                y_values[i] = dims.col_indices[idx];
            }
            const CubicSpline spline { x_values, y_values };
            for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
                dims.interp_cols[i * dims.n_l1b_cols + j] =
                  spline.eval(static_cast<double>(i));
            }
        }
    }
    // L1B columns indices interpolated along the L1C rows
    for (int j {}; j < dims.n_l1b_cols; ++j) {
        for (int i {}; i < n_rows; ++i) {
            const int idx { (dims.min_l1b_row + i) * dims.n_l1b_cols + j };
            x_values[i] = dims.row_indices[idx];
            y_values[i] = l1b_buf[idx];
        }
        const CubicSpline spline { x_values, y_values };
        for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
            data_semi[i * dims.n_l1b_cols + j] =
              spline.eval(static_cast<double>(i));
        }
    }
    // Determine which L1C columns are outside the L1B swath. This
    // needs to be done once per viewport
    if (dims.fill_cols_left.empty()) {
        dims.fill_cols_left.resize(dims.n_l1c_rows_dbl, 0);
        dims.fill_cols_right.resize(dims.n_l1c_rows_dbl, dims.n_l1c_cols_dbl);
        for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
            const double l1b_col_first {
                dims.interp_cols[i * dims.n_l1b_cols]
            };
            const double l1b_col_last {
                dims.interp_cols[(i + 1) * dims.n_l1b_cols - 1]
            };
            for (int j {}; j < dims.n_l1c_cols_dbl; ++j) {
                if (j < l1b_col_first) {
                    dims.fill_cols_left[i] = j + 1;
                }
                if (j > l1b_col_last) {
                    dims.fill_cols_right[i] = j;
                    // Capture the first column on the right that is
                    // outside the L1B range. Do not process further.
                    break;
                }
            }
        }
    }
    // Next interpolate along the L1C rows. The x-values are now the
    // interpolated L1C column indices and y-values are the L1B data
    // interpolated onto the semi-dense grid.
    x_values.resize(dims.n_l1b_cols);
    y_values.resize(dims.n_l1b_cols);
    std::vector<double> l1c_buf_unbin(dims.n_l1c_rows_dbl
                                      * dims.n_l1c_cols_dbl);
    constexpr double buf_fill_value { 1e100 };
    for (int i {}; i < dims.n_l1c_rows_dbl; ++i) {
        for (int j {}; j < dims.n_l1b_cols; ++j) {
            x_values[j] = dims.interp_cols[i * dims.n_l1b_cols + j];
            y_values[j] = data_semi[i * dims.n_l1b_cols + j];
        }
        const CubicSpline spline { x_values, y_values };
        for (int j {}; j < dims.fill_cols_left[i]; ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] = buf_fill_value;
        }
        for (int j { dims.fill_cols_left[i] }; j < dims.fill_cols_right[i];
             ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] =
              spline.eval(static_cast<double>(j));
        }
        for (int j { dims.fill_cols_right[i] }; j < dims.n_l1c_cols_dbl; ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] = buf_fill_value;
        }
    }
    // Use fill values for parts of the L1C grid that are outside the L1B grid
    for (int i {}; i < dims.min_l1c_row; ++i) {
        for (int j {}; j < dims.n_l1c_cols_dbl; ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] = buf_fill_value;
        }
    }
    for (int i { dims.max_l1c_row }; i < dims.n_l1c_rows_dbl; ++i) {
        for (int j {}; j < dims.n_l1c_cols_dbl; ++j) {
            l1c_buf_unbin[i * dims.n_l1c_cols_dbl + j] = buf_fill_value;
        }
    }
    // Finally, bin the data and store in the L1C buffer
    constexpr int bin { 2 };
    std::fill(l1c_buf.begin(), l1c_buf.end(), 0.0);
    for (int i {}; i < dims.n_l1c_rows; ++i) {
        for (int j {}; j < dims.n_l1c_cols; ++j) {
            for (int m {}; m < bin; ++m) {
                for (int n {}; n < bin; ++n) {
                    l1c_buf[i * dims.n_l1c_cols + j] +=
                      l1c_buf_unbin[(bin * i + m) * dims.n_l1c_cols_dbl
                                    + bin * j + n];
                }
            }
        }
    }
    const double norm { 1.0 / bin / bin };
    for (auto& val : l1c_buf) {
        val *= norm;
        if (std::abs(val) > std::sqrt(buf_fill_value)) {
            val = def::l1bc_fill;
        }
    }
}

// Convert a float array to an integer array while multiplying each
// element with a scale factor to preserve accuracy.
static auto floatToShort(const std::vector<float>& data_in,
                         std::vector<int16_t>& data_out) -> void
{
    data_out.resize(data_in.size());
    for (int i {}; i < static_cast<int>(data_in.size()); ++i) {
        if (data_in[i] < 0.8 * def::l1bc_fill) {
            data_out[i] = static_cast<int16_t>(def::l1bc_fill);
        } else {
            data_out[i] = static_cast<int16_t>(geo_scale_factor * data_in[i]);
        }
    }
}

auto L1C::processInit() -> void
{
    // Read in and expand the L1C target grid
    Dims dims {};
    // Original L1C grid
    std::vector<float> l1c_latitudes {};
    std::vector<float> l1c_longitudes {};
    // Dense L1C grid
    std::vector<float> l1c_latitudes_dense {};
    std::vector<float> l1c_longitudes_dense {};
    // Unit vectors of the L1C row centers
    std::vector<double> l1c_mid_vectors {};
    // L1C row normal vectors (constructed from the middle vectors)
    std::vector<double> normal_vectors_r {};
    // L1C column normal vectors
    std::vector<double> normal_vectors_c {};
    // L1C grid resolution across rows as a function of the row index
    std::vector<double> deltas_r {};
    // L1C grid resolution across columns as a function of the row index
    std::vector<double> deltas_c {};
    prepareL1CGrid(settings,
                   dims,
                   l1c_latitudes,
                   l1c_longitudes,
                   l1c_latitudes_dense,
                   l1c_longitudes_dense,
                   l1c_mid_vectors,
                   normal_vectors_r,
                   normal_vectors_c,
                   deltas_r,
                   deltas_c);

    // L1B dimensions
    const NcObject nc_l1b { netCDF::NcFile::read, settings.l1c.l1b_in };
    const int number_of_views {
        static_cast<int>(nc_l1b.getDim("number_of_views").getSize()),
    };
    dims.n_l1b_rows =
      static_cast<int>(nc_l1b.getDim("bins_along_track").getSize());
    const int spatial_samples_per_image {
        static_cast<int>(nc_l1b.getDim("spatial_samples_per_image").getSize()),
    };
    const int intensity_bands_per_view {
        static_cast<int>(nc_l1b.getDim("intensity_bands_per_view").getSize()),
    };
    const int polarization_bands_per_view {
        static_cast<int>(
          nc_l1b.getDim("polarization_bands_per_view").getSize()),
    };

    // L1B sensor view bands
    const auto l1b_sensor_view_bands { nc_l1b.getGroup("SENSOR_VIEW_BANDS") };
    std::vector<uint8_t> viewport_index(spatial_samples_per_image);
    l1b_sensor_view_bands.getVar("viewport_index")
      .getVar(viewport_index.data());

    // Image time is given per along track position. We duplicate
    // values for each field of view and then interpolate each
    // viewport separately.
    std::vector<double> l1b_image_time_raw(dims.n_l1b_rows);
    nc_l1b.getGroup("BIN_ATTRIBUTES")
      .getVar("image_time")
      .getVar(l1b_image_time_raw.data());
    std::vector<float> l1b_image_time(dims.n_l1b_rows
                                      * spatial_samples_per_image);
    for (int i {}; i < dims.n_l1b_rows; ++i) {
        for (int j {}; j < spatial_samples_per_image; ++j) {
            l1b_image_time[i * spatial_samples_per_image + j] =
              l1b_image_time_raw[i];
        }
    }

    // L1B geolocation data
    const auto l1b_geolocation_data { nc_l1b.getGroup("GEOLOCATION_DATA") };
    const int l1b_latlon_size { dims.n_l1b_rows * spatial_samples_per_image };
    std::vector<float> l1b_latitudes(l1b_latlon_size);
    l1b_geolocation_data.getVar("latitude").getVar(l1b_latitudes.data());
    for (auto& val : l1b_latitudes) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    std::vector<float> l1b_longitudes(l1b_latlon_size);
    l1b_geolocation_data.getVar("longitude").getVar(l1b_longitudes.data());
    for (auto& val : l1b_longitudes) {
        val = static_cast<float>(static_cast<double>(val) * math::deg_to_rad);
    }
    std::vector<float> l1b_altitude(l1b_latlon_size);
    std::vector<float> l1b_sensor_zenith(l1b_latlon_size);
    std::vector<float> l1b_sensor_azimuth(l1b_latlon_size);
    std::vector<float> l1b_solar_zenith(l1b_latlon_size);
    std::vector<float> l1b_solar_azimuth(l1b_latlon_size);
    l1b_geolocation_data.getVar("altitude").getVar(l1b_altitude.data());
    l1b_geolocation_data.getVar("sensor_zenith")
      .getVar(l1b_sensor_zenith.data());
    l1b_geolocation_data.getVar("sensor_azimuth")
      .getVar(l1b_sensor_azimuth.data());
    l1b_geolocation_data.getVar("solar_zenith").getVar(l1b_solar_zenith.data());
    l1b_geolocation_data.getVar("solar_azimuth")
      .getVar(l1b_solar_azimuth.data());
    for (int i {}; i < l1b_latlon_size; ++i) {
        l1b_sensor_zenith[i] *= math::deg_to_rad;
        l1b_sensor_azimuth[i] *= math::deg_to_rad;
        l1b_solar_zenith[i] *= math::deg_to_rad;
        l1b_solar_azimuth[i] *= math::deg_to_rad;
    }

    // L1B observation variables
    const auto l1b_observation_data { nc_l1b.getGroup("OBSERVATION_DATA") };
    const int l1b_intensity_size { dims.n_l1b_rows * spatial_samples_per_image
                                   * intensity_bands_per_view };
    const int l1b_pol_size { dims.n_l1b_rows * spatial_samples_per_image
                             * polarization_bands_per_view };
    std::vector<float> l1b_I(l1b_intensity_size);
    std::vector<float> l1b_I_noise(l1b_intensity_size);
    std::vector<float> l1b_I_polsample(l1b_pol_size);
    std::vector<float> l1b_I_polsample_noise(l1b_pol_size);
    std::vector<float> l1b_q_noise(l1b_pol_size);
    std::vector<float> l1b_u_noise(l1b_pol_size);
    // Polarized radiance
    std::vector<float> l1b_pol_I(l1b_pol_size);
    std::vector<float> l1b_dolp_noise(l1b_pol_size);
    std::vector<float> l1b_aolp(l1b_pol_size);
    std::vector<float> l1b_aolp_noise(l1b_pol_size);
    l1b_observation_data.getVar("I").getVar(l1b_I.data());
    l1b_observation_data.getVar("I_noise").getVar(l1b_I_noise.data());
    l1b_observation_data.getVar("I_polsample").getVar(l1b_I_polsample.data());
    l1b_observation_data.getVar("I_polsample_noise")
      .getVar(l1b_I_polsample_noise.data());
    l1b_observation_data.getVar("q_noise").getVar(l1b_q_noise.data());
    l1b_observation_data.getVar("u_noise").getVar(l1b_u_noise.data());
    l1b_observation_data.getVar("DoLP").getVar(l1b_pol_I.data());
    for (int i {}; i < l1b_pol_size; ++i) {
        l1b_pol_I[i] *= l1b_I_polsample[i];
    }
    l1b_observation_data.getVar("DoLP_noise").getVar(l1b_dolp_noise.data());
    l1b_observation_data.getVar("AoLP").getVar(l1b_aolp.data());
    l1b_observation_data.getVar("AoLP_noise").getVar(l1b_aolp_noise.data());

    // L1C variables
    const int l1c_intensity_size { dims.n_l1c_rows * dims.n_l1c_cols
                                   * number_of_views
                                   * intensity_bands_per_view };
    const int l1c_pol_size { dims.n_l1c_rows * dims.n_l1c_cols * number_of_views
                             * polarization_bands_per_view };
    // Like lat/lon but with angular dependence
    const int l1c_ang_size { dims.n_l1c_rows * dims.n_l1c_cols
                             * number_of_views };
    std::vector<float> l1c_I(l1c_intensity_size);
    std::vector<float> l1c_I_noise(l1c_intensity_size);
    std::vector<float> l1c_I_polsample(l1c_pol_size);
    std::vector<float> l1c_I_polsample_noise(l1c_pol_size);
    std::vector<float> l1c_q(l1c_pol_size);
    std::vector<float> l1c_q_noise(l1c_pol_size);
    std::vector<float> l1c_u(l1c_pol_size);
    std::vector<float> l1c_u_noise(l1c_pol_size);
    // Polarized radiance
    std::vector<float> l1c_pol_I(l1c_pol_size);
    std::vector<float> l1c_dolp_noise(l1c_pol_size);
    std::vector<float> l1c_aolp(l1c_pol_size);
    std::vector<float> l1c_aolp_noise(l1c_pol_size);
    std::vector<float> l1c_sza(l1c_ang_size);
    std::vector<float> l1c_saa(l1c_ang_size);
    std::vector<float> l1c_vza(l1c_ang_size);
    std::vector<float> l1c_vaa(l1c_ang_size);
    std::vector<float> l1c_image_time(l1c_ang_size);

    // Main interpolation loop
    for (int i_vp {}; i_vp < number_of_views; ++i_vp) {
        spdlog::info("Processing viewport {}", i_vp);
        // Quantities that need to be reset and recomputed for each viewport
        dims.min_l1b_row = def::idx;
        dims.max_l1b_row = def::idx;
        dims.fill_cols_left.clear();
        dims.fill_cols_right.clear();
        dims.interp_cols.clear();
        // Number of across track bins of the current viewport
        dims.n_l1b_cols =
          static_cast<int>(std::ranges::count(viewport_index, i_vp));
        std::vector<int> fov_indices(dims.n_l1b_cols);
        std::iota(fov_indices.begin(),
                  fov_indices.end(),
                  std::distance(viewport_index.begin(),
                                std::ranges::find(viewport_index, i_vp)));
        std::vector<double> l1b_lat_vp(dims.n_l1b_rows * dims.n_l1b_cols);
        std::vector<double> l1b_lon_vp(dims.n_l1b_rows * dims.n_l1b_cols);
        for (int i {}; i < dims.n_l1b_rows; ++i) {
            for (int j {}; j < dims.n_l1b_cols; ++j) {
                const int idx { i * spatial_samples_per_image
                                + fov_indices[j] };
                const int idx_vp { i * dims.n_l1b_cols + j };
                l1b_lat_vp[idx_vp] = l1b_latitudes[idx];
                l1b_lon_vp[idx_vp] = l1b_longitudes[idx];
                if (dims.add_2pi && l1b_lon_vp[idx_vp] < 0.0) {
                    l1b_lon_vp[idx_vp] += 2 * std::numbers::pi;
                }
            }
        }

        // Generate the fractional L1C row and column indices. This
        // needs to be done only once per viewport.
        genL1CCoordinates(settings,
                          l1b_lat_vp,
                          l1b_lon_vp,
                          l1c_latitudes_dense,
                          l1c_longitudes_dense,
                          l1c_mid_vectors,
                          normal_vectors_r,
                          normal_vectors_c,
                          deltas_r,
                          deltas_c,
                          dims);
        fixAzimuthValues(dims.min_l1b_row,
                         dims.max_l1b_row,
                         spatial_samples_per_image,
                         l1b_solar_azimuth);
        fixAzimuthValues(dims.min_l1b_row,
                         dims.max_l1b_row,
                         spatial_samples_per_image,
                         l1b_sensor_azimuth);

        // Do the interpolation
#pragma omp parallel firstprivate(dims)
        {
            std::vector<double> l1b_buf(dims.n_l1b_rows * dims.n_l1b_cols);
            std::vector<double> l1c_buf(dims.n_l1c_rows * dims.n_l1c_cols);
            // L1B data to L1B buffer ready for interpolation
            const auto loadToBuf { [&](const int n_bands,
                                       const int i_wave,
                                       const std::vector<float>& l1b_in) {
                for (int i {}; i < dims.n_l1b_rows; ++i) {
                    for (int j {}; j < dims.n_l1b_cols; ++j) {
                        l1b_buf[i * dims.n_l1b_cols + j] =
                          l1b_in[(i * spatial_samples_per_image
                                  + fov_indices[j])
                                   * n_bands
                                 + i_wave];
                    }
                }
            } };
            // L1C buffer to L1C data after interpolation
            const auto loadFromBuf { [&](const int n_bands,
                                         const int i_wave,
                                         std::vector<float>& l1c_out) {
                for (int i {}; i < dims.n_l1c_rows * dims.n_l1c_cols; ++i) {
                    l1c_out[(i * number_of_views + i_vp) * n_bands + i_wave] =
                      static_cast<float>(l1c_buf[i]);
                }
            } };
#pragma omp for
            for (int i_wave = 0; i_wave < intensity_bands_per_view; ++i_wave) {
                loadToBuf(intensity_bands_per_view, i_wave, l1b_I);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(intensity_bands_per_view, i_wave, l1c_I);
                loadToBuf(intensity_bands_per_view, i_wave, l1b_I_noise);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(intensity_bands_per_view, i_wave, l1c_I_noise);
            }
#pragma omp for
            for (int i_wave = 0; i_wave < polarization_bands_per_view;
                 ++i_wave) {
                loadToBuf(polarization_bands_per_view, i_wave, l1b_I_polsample);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(
                  polarization_bands_per_view, i_wave, l1c_I_polsample);
                loadToBuf(
                  polarization_bands_per_view, i_wave, l1b_I_polsample_noise);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(
                  polarization_bands_per_view, i_wave, l1c_I_polsample_noise);
                loadToBuf(polarization_bands_per_view, i_wave, l1b_q_noise);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(polarization_bands_per_view, i_wave, l1c_q_noise);
                loadToBuf(polarization_bands_per_view, i_wave, l1b_u_noise);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(polarization_bands_per_view, i_wave, l1c_u_noise);
                loadToBuf(polarization_bands_per_view, i_wave, l1b_pol_I);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(polarization_bands_per_view, i_wave, l1c_pol_I);
                loadToBuf(polarization_bands_per_view, i_wave, l1b_dolp_noise);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(
                  polarization_bands_per_view, i_wave, l1c_dolp_noise);
                loadToBuf(polarization_bands_per_view, i_wave, l1b_aolp);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(polarization_bands_per_view, i_wave, l1c_aolp);
                loadToBuf(polarization_bands_per_view, i_wave, l1b_aolp_noise);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(
                  polarization_bands_per_view, i_wave, l1c_aolp_noise);
                // L1C q and u are computed from the interpolated DoLP and AoLP
                for (int i {}; i < dims.n_l1c_rows * dims.n_l1c_cols; ++i) {
                    const int idx { (i * number_of_views + i_vp)
                                      * polarization_bands_per_view
                                    + i_wave };
                    l1c_q[idx] = static_cast<float>(
                      l1c_pol_I[idx] * std::cos(2.0 * l1c_aolp[idx])
                      / l1c_I_polsample[idx]);
                    l1c_u[idx] = static_cast<float>(
                      l1c_pol_I[idx] * std::sin(2.0 * l1c_aolp[idx])
                      / l1c_I_polsample[idx]);
                }
            }
#pragma omp master
            {
                auto loadFromBuf { [&](std::vector<float>& l1c_out) {
                    for (int i {}; i < dims.n_l1c_rows * dims.n_l1c_cols; ++i) {
                        l1c_out[i * number_of_views + i_vp] =
                          static_cast<float>(l1c_buf[i]);
                    }
                } };
                loadToBuf(1, 0, l1b_solar_zenith);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(l1c_sza);
                loadToBuf(1, 0, l1b_solar_azimuth);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(l1c_saa);
                loadToBuf(1, 0, l1b_sensor_zenith);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(l1c_vza);
                loadToBuf(1, 0, l1b_sensor_azimuth);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(l1c_vaa);
                loadToBuf(1, 0, l1b_image_time);
                interpolateAndBin(dims, l1b_buf, l1c_buf);
                loadFromBuf(l1c_image_time);
            }
        }
    }
    // Finally, compute the altitude and its RMS. These are viewport
    // independent.
    const int l1c_latlon_size { dims.n_l1c_rows * dims.n_l1c_cols };
    std::vector<double> l1c_altitude(l1c_latlon_size);
    std::vector<double> l1c_altitude_variability(l1c_latlon_size);
    constexpr double l1c_grid_resolution { 5.2e3 };
    DEM dem { settings.l1c.dem_file, l1c_grid_resolution };
    constexpr double lat_margin { 0.1 };
    constexpr double lon_margin { 0.5 };
    dem.loadToMemory(*std::ranges::min_element(l1c_latitudes) - lat_margin,
                     *std::ranges::max_element(l1c_latitudes) + lat_margin,
                     *std::ranges::min_element(l1c_longitudes) - lon_margin,
                     *std::ranges::max_element(l1c_longitudes) + lon_margin);
#pragma omp parallel for
    for (int i = 0; i < l1c_latlon_size; ++i) {
        const auto [alt,
                    rms] { dem.getHeight(l1c_latitudes[i], l1c_longitudes[i]) };
        l1c_altitude[i] = alt;
        l1c_altitude_variability[i] = rms;
    }

    // Write L1C product to NetCDF file
    NcObject nc_l1c { netCDF::NcFile::replace, settings.l1c.l1c_out };
    // Also reopen the grid file for copying some attributes and variables
    const NcObject nc_l1c_grid { netCDF::NcFile::read, settings.l1c.l1c_in };
    constexpr int compression_level { 5 };
    // Global attributes
    nc_l1c.putAtt("title", "PACE SPEXone Level-1C data");
    nc_l1c.putAtt("instrument", "SPEXone");
    int nadir_bin {};
    for (const auto& [name, att] : nc_l1c_grid.getAtts()) {
        std::string buffer {};
        if (name == "project") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "keywords_vocabulary") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "naming_authority") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "processing_level") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "cdm_data_type") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "bin_size_at_nadir") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "startDirection") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "endDirection") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "time_coverage_start") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "time_coverage_end") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "license") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "nadir_bin") {
            att.getValues(&nadir_bin);
            nc_l1c.putAtt(name, netCDF::ncInt, nadir_bin);
        } else if (name == "orbit_number") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "spectral_response_function") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "sun_earth_distance") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        } else if (name == "systematic_uncertainty_model") {
            att.getValues(buffer);
            nc_l1c.putAtt(name, buffer);
        }
    }
    nc_l1c.putAtt("software_name", "SPEXone L1A-L1C processor");
    nc_l1c.putAtt("software_version", SPEXONE_PROJECT_VERSION);
    nc_l1c.putAtt("terrain_data_source",
                  "GEBCO 15-arcsec Global Relief Map and Land/Water Mask");
    nc_l1c.putAtt("history", "");

    // Work variable for NetCDF variables
    netCDF::NcVar var {};

    // Dimensions
    const auto nc_bins_along_track { nc_l1c.addDim("bins_along_track",
                                                   dims.n_l1c_rows) };
    const auto nc_bins_across_track { nc_l1c.addDim("bins_across_track",
                                                    dims.n_l1c_cols) };
    const auto nc_number_of_views { nc_l1c.addDim("number_of_views",
                                                  number_of_views) };
    const auto nc_intensity_bands_per_view { nc_l1c.addDim(
      "intensity_bands_per_view", intensity_bands_per_view) };
    const auto nc_polarization_bands_per_view { nc_l1c.addDim(
      "polarization_bands_per_view", polarization_bands_per_view) };
    const std::vector<netCDF::NcDim> intensity_shape {
        nc_bins_along_track,
        nc_bins_across_track,
        nc_number_of_views,
        nc_intensity_bands_per_view
    };
    std::vector<size_t> intensity_chunks { 2,
                                           static_cast<size_t>(dims.n_l1c_cols),
                                           static_cast<size_t>(number_of_views),
                                           static_cast<size_t>(
                                             intensity_bands_per_view) };
    const std::vector<netCDF::NcDim> polarization_shape {
        nc_bins_along_track,
        nc_bins_across_track,
        nc_number_of_views,
        nc_polarization_bands_per_view
    };
    std::vector<size_t> polarization_chunks {
        2,
        static_cast<size_t>(dims.n_l1c_cols),
        static_cast<size_t>(number_of_views),
        static_cast<size_t>(polarization_bands_per_view)
    };
    const std::vector<netCDF::NcDim> nc_latlon_dims {
        nc_bins_along_track,
        nc_bins_across_track,
    };
    std::vector<size_t> latlon_chunks { 2,
                                        static_cast<size_t>(dims.n_l1c_cols) };
    // Like lat/lon but with angular dependence
    const std::vector<netCDF::NcDim> nc_ang_dims { nc_bins_along_track,
                                                   nc_bins_across_track,
                                                   nc_number_of_views };
    std::vector<size_t> ang_chunks { 2,
                                     static_cast<size_t>(dims.n_l1c_cols),
                                     static_cast<size_t>(number_of_views) };

    // Sensor view bands
    auto l1c_sensor_views_bands { nc_l1c.addGroup("sensor_views_bands") };
    var = l1c_sensor_views_bands.addVar(
      "view_angles", netCDF::ncFloat, nc_number_of_views);
    var.putAtt("long_name", "along-track view angles for sensor");
    var.putAtt("units", "degrees");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -89.0);
    var.putAtt("valid_max", netCDF::ncFloat, 89.0);
    std::vector<float> vp_angles(number_of_views, 0);
    std::transform(viewport_idx_to_angle.cbegin(),
                   viewport_idx_to_angle.cend(),
                   vp_angles.begin(),
                   [](const auto& pair) { return pair.second; });
    var.putVar(vp_angles.data());
    NcObject::copyVar(l1b_sensor_view_bands.getVar("intensity_wavelengths"),
                      l1c_sensor_views_bands);
    NcObject::copyVar(l1b_sensor_view_bands.getVar("polarization_wavelengths"),
                      l1c_sensor_views_bands);
    NcObject::copyVar(l1b_sensor_view_bands.getVar("intensity_bandpasses"),
                      l1c_sensor_views_bands);
    NcObject::copyVar(l1b_sensor_view_bands.getVar("polarization_bandpasses"),
                      l1c_sensor_views_bands);
    NcObject::copyVar(l1b_sensor_view_bands.getVar("intensity_F0"),
                      l1c_sensor_views_bands);
    NcObject::copyVar(l1b_sensor_view_bands.getVar("polarization_F0"),
                      l1c_sensor_views_bands);

    // Bin attributes
    auto l1c_bin_attributes { nc_l1c.addGroup("bin_attributes") };
    var = l1c_bin_attributes.addVar(
      "nadir_view_time", netCDF::ncDouble, nc_bins_along_track);
    var.putAtt("long_name",
               "time bin " + std::to_string(nadir_bin)
                 + " was viewed at nadir view");
    var.putAtt("_FillValue", netCDF::ncDouble, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncDouble, 0.0);
    var.putAtt("valid_max", netCDF::ncDouble, 86400.999);
    var.putAtt("units", "seconds");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, latlon_chunks);
    std::vector<double> nadir_view_time(dims.n_l1c_rows);
    for (int i {}; i < dims.n_l1c_rows; ++i) {
        constexpr int nadir_viewport { 2 };
        nadir_view_time[i] =
          l1c_image_time[(i * dims.n_l1c_cols + nadir_bin) * number_of_views
                         + nadir_viewport];
    }
    var.putVar(nadir_view_time.data());
    var = l1c_bin_attributes.addVar(
      "view_time_offsets", netCDF::ncDouble, nc_ang_dims);
    var.putAtt("long_name", "time offsets of views from nadir view");
    var.putAtt("_FillValue", netCDF::ncDouble, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncDouble, -200.0);
    var.putAtt("valid_max", netCDF::ncDouble, 200.0);
    var.putAtt("units", "seconds");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, ang_chunks);
    std::vector<double> view_time_offsets(dims.n_l1c_rows * dims.n_l1c_cols
                                          * number_of_views);
    for (int i_row {}; i_row < dims.n_l1c_rows; ++i_row) {
        for (int i_col {}; i_col < dims.n_l1c_cols; ++i_col) {
            for (int i_vp {}; i_vp < number_of_views; ++i_vp) {
                const int idx { i_row * dims.n_l1c_cols + i_col };
                const int idx_ang { idx * number_of_views + i_vp };
                if (l1c_image_time[idx_ang] < 0.8 * def::l1bc_fill
                    || nadir_view_time[idx] < 0.8 * def::l1bc_fill) {
                    view_time_offsets[idx_ang] = def::l1bc_fill;
                } else {
                    view_time_offsets[idx_ang] =
                      l1c_image_time[idx_ang] - nadir_view_time[i_row];
                }
            }
        }
    }
    var.putVar(view_time_offsets.data());

    // Geolocation data
    auto l1c_geolocation_data { nc_l1c.addGroup("geolocation_data") };
    // For converting float arrays to integer arrays and scaling
    std::vector<int16_t> buf_16 {};
    // To avoid loss of accuracy, reread the latitudes and longitudes
    // from the original grid file.
    var = NcObject::copyVar(l1b_geolocation_data.getVar("latitude"),
                            l1c_geolocation_data,
                            nc_latlon_dims);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, latlon_chunks);
    nc_l1c_grid.getGroup("geolocation_data")
      .getVar("latitude")
      .getVar(l1c_latitudes.data());
    var.putVar(l1c_latitudes.data());
    var = NcObject::copyVar(l1b_geolocation_data.getVar("longitude"),
                            l1c_geolocation_data,
                            nc_latlon_dims);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, latlon_chunks);
    nc_l1c_grid.getGroup("geolocation_data")
      .getVar("longitude")
      .getVar(l1c_longitudes.data());
    var.putVar(l1c_longitudes.data());
    // Cannot copy altitudes because of different variable type
    var =
      l1c_geolocation_data.addVar("altitude", netCDF::ncShort, nc_latlon_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "altitude at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, -1000);
    var.putAtt("valid_max", netCDF::ncShort, 10000);
    var.putAtt("units", "meters");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, latlon_chunks);
    buf_16.resize(l1c_latitudes.size());
    for (int i {}; i < static_cast<int>(buf_16.size()); ++i) {
        buf_16[i] = static_cast<int16_t>(l1c_altitude[i]);
    }
    var.putVar(buf_16.data());
    var = l1c_geolocation_data.addVar(
      "altitude_variability", netCDF::ncShort, nc_latlon_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "RMS variability of altitude at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, 0);
    var.putAtt("valid_max", netCDF::ncShort, 1000.0);
    var.putAtt("units", "meters");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, latlon_chunks);
    for (int i {}; i < static_cast<int>(buf_16.size()); ++i) {
        buf_16[i] = static_cast<int16_t>(l1c_altitude_variability[i]);
    }
    var.putVar(buf_16.data());
    var =
      l1c_geolocation_data.addVar("solar_zenith", netCDF::ncShort, nc_ang_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "solar zenith angle at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, 0);
    var.putAtt("valid_max", netCDF::ncShort, 18000);
    var.putAtt("units", "degrees");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, ang_chunks);
    moduloAndConvertAngles(l1c_sza);
    floatToShort(l1c_sza, buf_16);
    var.putVar(buf_16.data());
    var = l1c_geolocation_data.addVar(
      "solar_azimuth", netCDF::ncShort, nc_ang_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "solar azimuth angle at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, -18000);
    var.putAtt("valid_max", netCDF::ncShort, 18000);
    var.putAtt("units", "degrees");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, ang_chunks);
    moduloAndConvertAngles(l1c_saa);
    floatToShort(l1c_saa, buf_16);
    var.putVar(buf_16.data());
    var = l1c_geolocation_data.addVar(
      "sensor_zenith", netCDF::ncShort, nc_ang_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "sensor zenith angle at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, 0);
    var.putAtt("valid_max", netCDF::ncShort, 18000);
    var.putAtt("units", "degrees");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, ang_chunks);
    moduloAndConvertAngles(l1c_vza);
    floatToShort(l1c_vza, buf_16);
    var.putVar(buf_16.data());
    var = l1c_geolocation_data.addVar(
      "sensor_azimuth", netCDF::ncShort, nc_ang_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "sensor azimuth angle at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, -18000);
    var.putAtt("valid_max", netCDF::ncShort, 18000);
    var.putAtt("units", "degrees");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, ang_chunks);
    moduloAndConvertAngles(l1c_vaa);
    floatToShort(l1c_vaa, buf_16);
    var.putVar(buf_16.data());
    // The scattering angle is computed from the solar and sensor
    // viewing angles.
    var = l1c_geolocation_data.addVar(
      "scattering_angle", netCDF::ncShort, nc_ang_dims);
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("long_name", "scattering angle at bin locations");
    var.putAtt("valid_min", netCDF::ncShort, 0);
    var.putAtt("valid_max", netCDF::ncShort, 18000);
    var.putAtt("units", "degrees");
    var.putAtt("scale_factor", netCDF::ncFloat, 1.0f / geo_scale_factor);
    var.putAtt("add_offset", netCDF::ncFloat, 0.0f);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, ang_chunks);
    std::vector<float> scattering_angle(dims.n_l1c_rows * dims.n_l1c_cols
                                        * number_of_views);
    for (int i {}; i < static_cast<int>(scattering_angle.size()); ++i) {
        if (l1c_vza[i] < 0.8 * def::l1bc_fill) {
            scattering_angle[i] = def::l1bc_fill;
            continue;
        }
        const double theta { math::deg_to_rad * l1c_vza[i] };
        const double theta_s { math::deg_to_rad * l1c_sza[i] };
        const double phi { math::deg_to_rad * l1c_vaa[i] };
        const double phi_s { math::deg_to_rad * l1c_saa[i] };
        const double u { std::cos(theta + std::numbers::pi) };
        const double u_s { std::cos(theta_s) };
        const double cos_alpha {
            u * u_s
            + std::sqrt(1 - u * u) * std::sqrt(1 - u_s * u_s)
                * std::cos(phi + std::numbers::pi - phi_s)
        };
        scattering_angle[i] = static_cast<float>(std::acos(cos_alpha));
    }
    moduloAndConvertAngles(scattering_angle);
    floatToShort(scattering_angle, buf_16);
    var.putVar(buf_16.data());

    // Observation data
    Timer timer_obs {};
    timer_obs.start();
    auto l1c_observation_data { nc_l1c.addGroup("observation_data") };
    var = l1c_geolocation_data.addVar(
      "obs_per_view", netCDF::ncShort, nc_latlon_dims);
    var.putAtt("long_name", "observations contributing to bin from each view");
    var.putAtt("_FillValue", netCDF::ncShort, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncShort, 0);
    var.putAtt("valid_max", netCDF::ncShort, 999);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, latlon_chunks);
    std::vector<int16_t> obs_per_view(dims.n_l1c_rows * dims.n_l1c_cols, 0);
    for (int i_row {}; i_row < dims.n_l1c_rows; ++i_row) {
        for (int i_col {}; i_col < dims.n_l1c_cols; ++i_col) {
            for (int i_vp {}; i_vp < number_of_views; ++i_vp) {
                const int idx { i_row * dims.n_l1c_cols + i_col };
                const int idx_ang { idx * number_of_views + i_vp };
                if (l1c_image_time[idx_ang] > def::l1bc_fill) {
                    ++obs_per_view[i_row * dims.n_l1c_cols + i_col];
                }
            }
        }
    }
    var.putVar(obs_per_view.data());
    var = l1c_observation_data.addVar(
      "QC_bitwise", netCDF::ncFloat, intensity_shape);
    var.putAtt("long_name", "bitwise quality indicator");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 65536.0f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, intensity_chunks);
    var = l1c_observation_data.addVar("QC", netCDF::ncFloat, intensity_shape);
    var.putAtt("long_name", "quality indicator");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncShort, 0.0f);
    var.putAtt("valid_max", netCDF::ncShort, 10.0f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, intensity_chunks);
    var = NcObject::copyVar(
      l1b_observation_data.getVar("I"), l1c_observation_data, intensity_shape);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, intensity_chunks);
    var.putVar(l1c_I.data());
    var = NcObject::copyVar(l1b_observation_data.getVar("I_noise"),
                            l1c_observation_data,
                            intensity_shape);
    var.setCompression(false, true, compression_level);
    var.putVar(l1c_I_noise.data());
    var = l1c_observation_data.addVar(
      "QC_polsample_bitwise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "bitwise quality indicator");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 65536.0f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var = l1c_observation_data.addVar(
      "QC_polsample", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "quality indicator");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncShort, 0.0f);
    var.putAtt("valid_max", netCDF::ncShort, 10.0f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var = NcObject::copyVar(l1b_observation_data.getVar("I_polsample"),
                            l1c_observation_data,
                            polarization_shape);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_I_polsample.data());
    var = NcObject::copyVar(l1b_observation_data.getVar("I_polsample_noise"),
                            l1c_observation_data,
                            polarization_shape);
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_I_polsample_noise.data());
    var = l1c_observation_data.addVar(
      "Q_over_I", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "Q over I (little q) Stokes vector component");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -1.5f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.5f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_q.data());
    var = l1c_observation_data.addVar(
      "Q_over_I_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "random noise of Q over I in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_q_noise.data());
    var = l1c_observation_data.addVar(
      "U_over_I", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "U over I (little u) Stokes vector component");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -1.5f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.5f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_u.data());
    var = l1c_observation_data.addVar(
      "U_over_I_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "random noise of U over I in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_u_noise.data());
    var =
      l1c_observation_data.addVar("DOLP", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "degree of linear polarization");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    // Convert polarized radiance to DoLP
    for (int i {}; i < l1c_pol_size; ++i) {
        l1c_pol_I[i] /= l1c_I_polsample[i];
    }
    var.putVar(l1c_pol_I.data());
    var = l1c_observation_data.addVar(
      "DOLP_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "random noise of DOLP in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, 0.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    var.putAtt("units", "1");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_dolp_noise.data());
    var =
      l1c_observation_data.addVar("AOLP", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "angle of linear polarization");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("valid_min", netCDF::ncFloat, -90.0f);
    var.putAtt("valid_max", netCDF::ncFloat, 90.0f);
    var.putAtt("units", "degrees");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_aolp.data());
    var = l1c_observation_data.addVar(
      "AOLP_noise", netCDF::ncFloat, polarization_shape);
    var.putAtt("long_name", "random noise of AOLP in bin");
    var.putAtt("_FillValue", netCDF::ncFloat, def::l1bc_fill);
    var.putAtt("units", "degrees");
    var.setCompression(false, true, compression_level);
    var.setChunking(netCDF::NcVar::nc_CHUNKED, polarization_chunks);
    var.putVar(l1c_aolp_noise.data());
    timer_obs.stop();
    spdlog::info("L1C observation data writing time: {:6.3f} s",
                 timer_obs.time());
}

} // namespace spex
