// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1x.h"

#include "io.h"
#include "l1a.h"
#include "l1a_metadata.h"
#include "nc_object.h"
#include "spectra.h"

#include <omp.h>

namespace spex {

auto L1X::createVariables(const ProcLevel level,
                          const CKD& ckd,
                          const size_t number_of_images,
                          const size_t samples_per_image,
                          NcObject& nc) -> void
{
    // Global attributes

    if (level == ProcLevel::l1a) {
        nc.putAtt("processing_level", "L1A");
    } else if (level == ProcLevel::l1b) {
        nc.putAtt("processing_level", "L1A");
    } else if (level == ProcLevel::l1c) {
        nc.putAtt("processing_level", "L1C");
    } else {
        nc.putAtt("processing_level", "L1X");
        nc.putAtt("l1x_level", netCDF::ncInt, static_cast<int>(level));
    }

    // Dimensions

    const auto number_of_images_nc {
        nc.addDim("number_of_images", number_of_images),
    };
    netCDF::NcDim samples_per_image_nc {};
    netCDF::NcDim spatial_detector_pixels {};
    netCDF::NcDim spatial_samples_per_image {};
    netCDF::NcDim polarization_directions {};
    if (level <= ProcLevel::stray) {
        samples_per_image_nc =
          nc.addDim("samples_per_image", samples_per_image);
        spatial_detector_pixels =
          nc.addDim("spatial_detector_pixels", ckd.n_detector_rows);
    } else if (level < ProcLevel::l1b) {
        spatial_samples_per_image =
          nc.addDim("spatial_samples_per_image", ckd.fov.n_act);
        polarization_directions =
          nc.addDim("polarization_directions", dims::pol);
    }
    // Spectral dimension is required for both the detector image and
    // when saving spectra. The spectra will have different sizes
    // depending on the viewport and ACT angle (FOV index) but to keep
    // storage simple we use the original number of columns (i.e. the
    // spectral dimension) when saving them.
    const auto spectral_detector_pixels {
        nc.addDim("spectral_detector_pixels", ckd.n_detector_cols),
    };

    // Image attributes

    auto image_attributes { nc.addGroup("image_attributes") };
    auto image_time {
        image_attributes.addVar(
          "image_time", netCDF::ncDouble, number_of_images_nc),
    };
    image_time.putAtt("long_name", "image time (seconds of day)");
    image_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
    image_time.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
    image_time.putAtt("units", "seconds");

    auto icu_time_sec {
        image_attributes.addVar(
          "icu_time_sec", netCDF::ncUint, number_of_images_nc),
    };
    icu_time_sec.putAtt("long_name", "ICU time stamp (seconds)");
    icu_time_sec.putAtt("description", "Science TM parameter ICU_TIME_SEC");
    icu_time_sec.putAtt("valid_min", netCDF::ncUint, 1577800000);
    icu_time_sec.putAtt("valid_max", netCDF::ncUint, 1735700000);
    icu_time_sec.putAtt("units", "seconds since 1970-01-01 00:00:00");

    auto icu_time_subsec {
        image_attributes.addVar(
          "icu_time_subsec", netCDF::ncUshort, number_of_images_nc),
    };
    icu_time_subsec.putAtt("long_name", "ICU time stamp (sub-seconds)");
    icu_time_subsec.putAtt("description",
                           "Science TM parameter ICU_TIME_SUBSEC");
    icu_time_subsec.putAtt("valid_min", netCDF::ncUshort, 0);
    icu_time_subsec.putAtt("valid_max", netCDF::ncUshort, 65535);
    icu_time_subsec.putAtt("units", "1/65536 s");

    auto binning_table {
        image_attributes.addVar(
          "binning_table", netCDF::ncUbyte, number_of_images_nc),
    };
    binning_table.putAtt("long_name", "binning-table ID");
    binning_table.putAtt("valid_min", netCDF::ncUbyte, 0);
    binning_table.putAtt("valid_max", netCDF::ncUbyte, 255);

    auto digital_offset {
        image_attributes.addVar(
          "digital_offset", netCDF::ncShort, number_of_images_nc),
    };
    digital_offset.putAtt("long_name", "digital offset");
    digital_offset.putAtt("units", "1");

    auto nr_coadditions {
        image_attributes.addVar(
          "nr_coadditions", netCDF::ncUshort, number_of_images_nc),
    };
    nr_coadditions.putAtt("_FillValue", netCDF::ncUshort, NC_FILL_USHORT);
    nr_coadditions.putAtt("long_name", "number of coadditions");
    nr_coadditions.putAtt("units", "1");

    auto exposure_time {
        image_attributes.addVar(
          "exposure_time", netCDF::ncDouble, number_of_images_nc),
    };
    exposure_time.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
    exposure_time.putAtt("long_name", "exposure time");
    exposure_time.putAtt("units", "s");

    // Science data

    auto science_data { nc.addGroup("science_data") };
    constexpr std::array<uint8_t, 2> mask_values { 0, 1 };
    if (level == ProcLevel::l1a) {
        auto detector_images {
            science_data.addVar("detector_images",
                                netCDF::ncUshort,
                                { number_of_images_nc, samples_per_image_nc }),
        };
        detector_images.putAtt("_FillValue", netCDF::ncUshort, 65535);
        detector_images.putAtt("long_name", "detector pixel values");
        detector_images.putAtt("units", "counts");
        detector_images.putAtt("valid_min", netCDF::ncUshort, 0);
        detector_images.putAtt("valid_max", netCDF::ncUshort, 65534);
        std::vector<std::size_t> chunksize { 1, samples_per_image };
        detector_images.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
    } else if (level <= ProcLevel::stray) {
        std::vector<std::size_t> chunksize_detector { 1, samples_per_image };
        std::vector<std::size_t> chunksize_detector_unbinned {
            1,
            static_cast<size_t>(ckd.n_detector_rows),
            static_cast<size_t>(ckd.n_detector_cols)
        };
        // Binned and unbinned detector images
        auto detector_images {
            science_data.addVar("detector_images",
                                netCDF::ncDouble,
                                { number_of_images_nc, samples_per_image_nc }),
        };
        detector_images.setChunking(netCDF::NcVar::nc_CHUNKED,
                                    chunksize_detector);
        detector_images.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_images.putAtt("long_name",
                               "(binned) detector images at current stage");
        detector_images.putAtt("units", "counts");
        auto detector_images_unbinned {
            science_data.addVar("detector_images_unbinned",
                                netCDF::ncDouble,
                                { number_of_images_nc,
                                  spatial_detector_pixels,
                                  spectral_detector_pixels }),
        };
        detector_images_unbinned.setChunking(netCDF::NcVar::nc_CHUNKED,
                                             chunksize_detector_unbinned);
        detector_images_unbinned.putAtt(
          "_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_images_unbinned.putAtt(
          "long_name", "unbinned detector images at current stage");
        detector_images_unbinned.putAtt("units", "counts");
        // Binned and unbinned detector image masks
        auto pixel_mask {
            science_data.addVar("pixel_mask",
                                netCDF::ncUbyte,
                                { number_of_images_nc, samples_per_image_nc }),
        };
        pixel_mask.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_detector);
        pixel_mask.putAtt("_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        pixel_mask.putAtt("long_name", "(binned) detector pixel mask");
        pixel_mask.putAtt("flag_values",
                          netCDF::ncUbyte,
                          mask_values.size(),
                          mask_values.data());
        pixel_mask.putAtt("flag_meanings", "good bad");
        auto pixel_mask_unbinned {
            science_data.addVar("pixel_mask_unbinned",
                                netCDF::ncUbyte,
                                { number_of_images_nc,
                                  spatial_detector_pixels,
                                  spectral_detector_pixels }),
        };
        pixel_mask_unbinned.setChunking(netCDF::NcVar::nc_CHUNKED,
                                        chunksize_detector_unbinned);
        pixel_mask_unbinned.putAtt(
          "_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        pixel_mask_unbinned.putAtt("long_name", "unbinned detector pixel mask");
        pixel_mask_unbinned.putAtt("flag_values",
                                   netCDF::ncUbyte,
                                   mask_values.size(),
                                   mask_values.data());
        pixel_mask_unbinned.putAtt("flag_meanings", "good bad");
        // Binned and unbinned detector image noise
        auto detector_noise {
            science_data.addVar("detector_noise",
                                netCDF::ncDouble,
                                { number_of_images_nc, samples_per_image_nc }),
        };
        detector_noise.setChunking(netCDF::NcVar::nc_CHUNKED,
                                   chunksize_detector);
        detector_noise.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_noise.putAtt("long_name",
                              "(binned) noise of the detector images");
        detector_noise.putAtt("units", "counts");
        auto detector_noise_unbinned {
            science_data.addVar("detector_noise_unbinned",
                                netCDF::ncDouble,
                                { number_of_images_nc,
                                  spatial_detector_pixels,
                                  spectral_detector_pixels }),
        };
        detector_noise_unbinned.setChunking(netCDF::NcVar::nc_CHUNKED,
                                            chunksize_detector_unbinned);
        detector_noise_unbinned.putAtt(
          "_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_noise_unbinned.putAtt("long_name",
                                       "unbinned noise of the detector images");
        detector_noise_unbinned.putAtt("units", "counts");
    } else if (level < ProcLevel::l1b) {
        std::vector<std::size_t> chunksize {
            1,
            static_cast<size_t>(ckd.fov.n_act),
            static_cast<size_t>(dims::pol),
            static_cast<size_t>(ckd.n_detector_cols),
        };
        const std::string unit { level == ProcLevel::fov ? "counts"
                                                         : "W.m-2.sr-1.um-1" };
        // Spectra can have different sizes which need to be recorded separately
        auto spectra_sizes {
            science_data.addVar(
              "spectra_sizes", netCDF::ncInt, spatial_samples_per_image),
        };
        spectra_sizes.putAtt("long_name",
                             "size (number of wavelengths) of each spectrum");
        spectra_sizes.putAtt("units", "1");
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            const std::vector<size_t> start { static_cast<size_t>(i_fov) };
            const std::vector<size_t> count { 1 };
            const auto size { ckd.wave_full.front().at(i_fov).size() };
            spectra_sizes.putVar(start, count, &size);
        }
        // Need to save wavelengths of all spectra because they differ
        // from those in the CKD which are unbinned.
        auto wavelengths {
            science_data.addVar("wavelengths",
                                netCDF::ncDouble,
                                { spatial_samples_per_image,
                                  polarization_directions,
                                  spectral_detector_pixels }),
        };
        wavelengths.putAtt("long_name",
                           "wavelengths of extracted spectra at current stage");
        wavelengths.putAtt("units", "nm");
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                const std::vector<size_t> start { static_cast<size_t>(i_fov),
                                                  static_cast<size_t>(i_pol),
                                                  0 };
                const std::vector<size_t> count {
                    1, 1, ckd.wave_full.at(i_pol).at(i_fov).size()
                };
                if (level < ProcLevel::rad) {
                    wavelengths.putVar(
                      start, count, ckd.wave_full.at(i_pol).at(i_fov).data());
                } else {
                    wavelengths.putVar(
                      start, count, ckd.wave_common.at(i_fov).data());
                }
            }
        }
        // All spectra
        auto spectra {
            science_data.addVar("spectra",
                                netCDF::ncDouble,
                                { number_of_images_nc,
                                  spatial_samples_per_image,
                                  polarization_directions,
                                  spectral_detector_pixels }),
        };
        spectra.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
        spectra.putAtt("long_name", "extracted spectra at current stage");
        spectra.putAtt("units", unit);
        // All spectral noise
        auto spectra_noise { science_data.addVar(
          "spectra_noise",
          netCDF::ncDouble,
          { number_of_images_nc,
            spatial_samples_per_image,
            polarization_directions,
            spectral_detector_pixels }) };
        spectra_noise.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
        spectra_noise.putAtt(
          "long_name", "noise estimate on extracted spectra at current stage");
        spectra_noise.putAtt("units", unit);
        // All spectral masks
        auto spectra_mask {
            science_data.addVar("spectra_mask",
                                netCDF::ncUbyte,
                                { number_of_images_nc,
                                  spatial_samples_per_image,
                                  polarization_directions,
                                  spectral_detector_pixels }),
        };
        spectra_mask.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
        spectra_mask.putAtt("_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        spectra_mask.putAtt("long_name", "spectral mask at current stage");
        spectra_mask.putAtt("flag_values",
                            netCDF::ncUbyte,
                            mask_values.size(),
                            mask_values.data());
        spectra_mask.putAtt("flag_meanings", "good bad");
    }
}

L1X::L1X(const CKD& ckd,
         const SPEXSettings& settings,
         const std::vector<L1A>& l1a_products,
         const ProcLevel level,
         const std::string& filename)
  : ckd { ckd }, level { level }, filename { filename }
{
    NcObject nc { netCDF::NcFile::replace, filename };

    nc.writeReducedConfig(settings);
    createVariables(
      level,
      ckd,
      l1a_products.size(),
      static_cast<int>(l1a_products.front().binning_table->nBins()),
      nc);

    // Image attributes

    for (size_t i_l1a {}; i_l1a < l1a_products.size(); ++i_l1a) {
        auto grp { nc.getGroup("image_attributes") };
        const std::vector<size_t> start { i_l1a };
        const std::vector<size_t> count { 1 };
        grp.getVar("image_time")
          .putVar(start, count, &l1a_products.at(i_l1a).image_time);
        grp.getVar("icu_time_sec")
          .putVar(start, count, &l1a_products.at(i_l1a).seconds);
        const uint16_t subseconds {
            static_cast<uint16_t>(
              std::round(65535 * l1a_products.at(i_l1a).second_fraction)),
        };
        grp.getVar("icu_time_subsec").putVar(start, count, &subseconds);
        grp.getVar("nr_coadditions")
          .putVar(start, count, &l1a_products.at(i_l1a).nr_coadditions);
        grp.getVar("binning_table")
          .putVar(start, count, &l1a_products.at(i_l1a).binning_table_id);
        const int16_t digital_offset { static_cast<int16_t>(
          l1a_products.at(i_l1a).digital_offset) };
        grp.getVar("digital_offset").putVar(start, count, &digital_offset);
        grp.getVar("exposure_time")
          .putVar(start, count, &l1a_products.at(i_l1a).exposure_time);
    }

    // Navigation data

    if (!l1a_products.front().meta->nav_att_time.empty()) {
        copyNavigationData(*l1a_products.front().meta->nc, nc);
    }
}

auto L1X::writeImage(const L1A& l1a) -> void
{
    if (omp_get_num_threads() > 1) {
        throw std::runtime_error {
            "running L1X with multiple threads currently not supported"
        };
    }
    NcObject nc { netCDF::NcFile::write, filename };
    auto science_data { nc.getGroup("science_data") };
    const std::vector<size_t> start { l1a.l1x_image_offset, 0 };
    const std::vector<size_t> count { 1, l1a.image.size() };
    const std::vector<size_t> start_unbinned { l1a.l1x_image_offset, 0, 0 };
    const std::vector<size_t> count_unbinned {
        1,
        static_cast<size_t>(ckd.n_detector_rows),
        static_cast<size_t>(ckd.n_detector_cols)
    };
    // Write current detector image in binned and unbinned formats
    science_data.getVar("detector_images")
      .putVar(start, count, l1a.image.data());
    std::vector<double> image_unbinned(ckd.npix);
    l1a.binning_table->unbin(l1a.image, image_unbinned);
    science_data.getVar("detector_images_unbinned")
      .putVar(start_unbinned, count_unbinned, image_unbinned.data());
    // Write current detector image mask
    std::vector<uint8_t> pixel_mask_u8(l1a.pixel_mask.size());
    for (int i {}; i < static_cast<int>(pixel_mask_u8.size()); ++i) {
        pixel_mask_u8[i] = static_cast<uint8_t>(l1a.pixel_mask[i]);
    }
    science_data.getVar("pixel_mask")
      .putVar(start, count, pixel_mask_u8.data());
    std::vector<uint8_t> pixel_mask_unbinned_u8(ckd.npix);
    l1a.binning_table->unbin(pixel_mask_u8, pixel_mask_unbinned_u8);
    science_data.getVar("pixel_mask_unbinned")
      .putVar(start_unbinned, count_unbinned, pixel_mask_unbinned_u8.data());
    science_data.getVar("detector_noise")
      .putVar(start, count, l1a.noise.data());
    std::vector<double> noise_unbinned(ckd.npix);
    l1a.binning_table->unbin(l1a.noise, noise_unbinned);
    science_data.getVar("detector_noise_unbinned")
      .putVar(start_unbinned, count_unbinned, noise_unbinned.data());
}

auto L1X::writeSpectrum(const Spectra& spectra,
                        const size_t l1x_image_offset,
                        const int i_fov) -> void
{
    if (omp_get_num_threads() > 1) {
        throw std::runtime_error {
            "running L1X with multiple threads currently not supported"
        };
    }
    NcObject nc { netCDF::NcFile::write, filename };
    auto science_data { nc.getGroup("science_data") };
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        const std::vector<size_t> start { static_cast<size_t>(l1x_image_offset),
                                          static_cast<size_t>(i_fov),
                                          static_cast<size_t>(i_pol),
                                          0 };
        const std::vector<size_t> count {
            1, 1, 1, spectra.signal.front().size()
        };
        science_data.getVar("spectra").putVar(
          start, count, spectra.signal.at(i_pol).data());
        science_data.getVar("spectra_noise")
          .putVar(start, count, spectra.noise.at(i_pol).data());
        std::vector<uint8_t> mask_u8(spectra.signal.front().size());
        for (int i {}; i < static_cast<int>(mask_u8.size()); ++i) {
            mask_u8[i] = static_cast<uint8_t>(spectra.mask[i_pol][i]);
        }
        science_data.getVar("spectra_mask")
          .putVar(start, count, mask_u8.data());
    }
}

} // namespace spex
