// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "linalg.h"

namespace spex {

auto dppsv(const std::vector<double>& data,
           const std::vector<double>& std_devs,
           std::vector<double>& B,
           std::vector<double>& state_vector) -> void
{
    // Number of state parameters
    const int n_states { static_cast<int>(B.size() / data.size()) };
    // Scale each row of data by noise (standard deviations)
    for (int i {}; i < static_cast<int>(data.size()); ++i) {
        for (int j {}; j < n_states; ++j) {
            B[i * n_states + j] /= std_devs[i];
        }
    }
    // B^T x B
    std::vector<double> BTB(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int j {}; j < n_states; ++j) {
            for (int k {}; k < static_cast<int>(data.size()); ++k) {
                BTB[i * n_states + j] +=
                  B[k * n_states + i] * B[k * n_states + j];
            }
        }
    }
    // B^T x Y where Y is the data vector
    std::vector<double> BTY(static_cast<int>(data.size()), 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < static_cast<int>(data.size()); ++k) {
            BTY[i] += B[k * n_states + i] * data[k] / std_devs[k];
        }
    }
    // Upper triangle of B^T x B
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j + 1) * j / 2] = BTB[i * n_states + j];
        }
    }
    // Covariance matrix of linear inversion initialized to unity
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    constexpr char uplo { 'u' };
    int info {};
    dppsv_(
      &uplo, &n_states, &n_states, BTB_tri.data(), Sx.data(), &n_states, &info);
    state_vector.assign(n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            state_vector[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
}

auto dppsv(const std::vector<double>& data,
           const std::vector<double>& std_devs,
           PackedMatrix& B,
           std::vector<double>& state_vector) -> void
{
    // Scale each row of data by noise (standard deviations)
    for (int i {}; i < B.nRows(); ++i) {
        for (int j {}; j < B.nCols(i); ++j) {
            B(i, j) /= std_devs[i];
        }
    }
    // B^T x B
    PackedMatrix B_T { B.nCols(), B.nRows() };
    B_T.transpose(B);
    std::vector<double> BTB(B.nCols() * B.nCols(), 0.0);
    B_T.multRightTransposed(B_T, BTB);
    // B^T x Y where Y is the data vector
    std::vector<double> BTY(B_T.nRows(), 0.0);
    for (int i {}; i < B_T.nRows(); ++i) {
        for (int j {}; j < B_T.nCols(i); ++j) {
            const int col_idx { B_T.globalCol(i, j) };
            BTY[i] += B_T(i, j) * data.at(col_idx) / std_devs.at(col_idx);
        }
    }
    // Number of state parameters
    const int n_states { B.nCols() };
    // Upper triangle of B^T x B
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j + 1) * j / 2] = BTB[i * n_states + j];
        }
    }
    // Covariance matrix of linear inversion initialized to unity
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    constexpr char uplo { 'u' };
    int info {};
    dppsv_(
      &uplo, &n_states, &n_states, BTB_tri.data(), Sx.data(), &n_states, &info);
    state_vector.assign(n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            state_vector[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
}

} // namespace spex
