// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "linear_spline_multi.h"

#include "algorithm.h"

namespace spex {

LinearSplineMulti::LinearSplineMulti(const int n_curves, const int n_knots)
  : n_curves { n_curves }, n_knots { n_knots }
{
    knots.resize(n_knots);
    y.resize(n_curves * n_knots);
}

auto LinearSplineMulti::setLookupIdxScale() -> void
{
    lookup_idx_scale =
      static_cast<int>(knots.size() - 1) / (knots.back() - knots.front());
}

auto LinearSplineMulti::generateSpline(const int i_curve,
                                       const std::vector<double>& values)
  -> void
{
    setLookupIdxScale();
    for (int i_knot {}; i_knot < n_knots; ++i_knot) {
        const int idx { i_curve * n_knots + i_knot };
        y[idx] = values[i_knot];
    }
}

auto LinearSplineMulti::lookupIdx(const double x) const -> int
{
    return static_cast<int>(lookup_idx_scale * (x - knots.front()));
}

[[nodiscard]] auto LinearSplineMulti::eval(const int i_curve,
                                           const double x) const -> double
{
    const int idx { equal_spacing ? lookupIdx(x) : binaryFindIdx(knots, x) };
    // Duplicate first or last knots outside the interpolation range
    if (x <= knots.front()) {
        const double t { (x - knots.front()) / (knots[1] - knots.front()) };
        return std::lerp(y[i_curve * n_knots], y[i_curve * n_knots + 1], t);
    }
    if (x >= knots.back()) {
        const double& knots_2 { knots[n_knots - 2] };
        const double t { (x - knots_2) / (knots.back() - knots_2) };
        return std::lerp(
          y[(i_curve + 1) * n_knots - 2], y[(i_curve + 1) * n_knots - 1], t);
    }
    const double t { (x - knots[idx]) / (knots[idx + 1] - knots[idx]) };
    return std::lerp(
      y[i_curve * n_knots + idx], y[i_curve * n_knots + idx + 1], t);
}

[[nodiscard]] auto LinearSplineMulti::deriv(const int i_curve,
                                            const double x) const -> double
{
    const int idx { equal_spacing ? lookupIdx(x) : binaryFindIdx(knots, x) };
    // Duplicate first or last derivatives outside the interpolation range
    if (x <= knots.front()) {
        return (y[i_curve * n_knots + 1] - y[i_curve * n_knots])
               / (knots[1] - knots.front());
    }
    if (x >= knots.back()) {
        return (y[(i_curve + 1) * n_knots - 1] - y[(i_curve + 1) * n_knots - 2])
               / (knots.back() - knots[n_knots - 2]);
    }
    return (y[i_curve * n_knots + idx + 1] - y[i_curve * n_knots + idx])
           / (knots[idx + 1] - knots[idx]);
}

} // namespace spex
