// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for generating and evaluating linear multiple splines. The
// curves share the same knot vector but have different
// coefficients. The coefficients are stored in large shared arrays
// with the dimensions (n_curves, n_knots) or (n_curves, n_knots - 1)
// depending on the parameter. These are accessible publicly so they
// can be set or read directly (for instance when reading/writing to a
// CKD file) but are meant to be generated with calls to
// generateSpline.

#pragma once

#include "constants.h"

#include <vector>

namespace spex {

class LinearSplineMulti
{
private:
    bool equal_spacing { true };
    int n_curves {};
    int n_knots {};
    // Scaling factor for the index lookup
    double lookup_idx_scale {};
    auto lookupIdx(const double x) const -> int;

public:
    std::vector<double> knots {};
    std::vector<double> y {};

    LinearSplineMulti() = default;
    LinearSplineMulti(const int n_curves, const int n_knots);
    // Set the scaling factor using the current knot vector. Without
    // setting this the index lookup won't work. It is called
    // automatically by generateSpline but if the coefficients are set
    // manually then this also needs to be called manually.
    auto setLookupIdxScale() -> void;
    // Using the function values generate the spline coefficients for
    // curve i_curve. The values are supposed to be given for points
    // of the knot vector which was set previously.
    auto generateSpline(const int i_curve, const std::vector<double>& values)
      -> void;
    [[nodiscard]] auto eval(const int i_curve, const double x) const -> double;
    // First derivative at x
    [[nodiscard]] auto deriv(const int i_curve, const double x) const -> double;
};

} // namespace spex
