// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "nc_object.h"

#include "date.h"
#include "io.h"
#include "spex_settings.h"

#include <filesystem>
#include <numeric>

namespace spex {

NcObject::NcObject(const netCDF::NcFile::FileMode fmode,
                   const std::string& filename)
  : fmode { fmode }
{
    if (!filename.empty()) {
        open(filename);
    }
}

auto NcObject::open(const std::string& filename) -> void
{
    netCDF::NcFile::close();
    this->filename = filename;
    netCDF::NcFile::open(filename, fmode);
    if (fmode != netCDF::NcFile::read) {
        putAtt("Conventions", "CF-1.6");
        putAtt("institution", "SRON Netherlands Institute for Space Research");
        putAtt("instrument", "SPEXone");
        putAtt("processor_version", SPEXONE_PROJECT_VERSION);
        putAtt("product_name", std::filesystem::path { filename }.filename());
        putAtt("project", "PACE Project");
        putAtt("creator_email", "SPEXone-MPC@sron.nl");
        putAtt("creator_name", "SRON/Earth Science");
        putAtt("creator_url",
               "https://www.sron.nl/missions-earth/pace-spexone");
        putAtt("publisher_email", "SPEXone-MPC@sron.nl");
        putAtt("publisher_name", "SRON/Earth Science");
        putAtt("publisher_url",
               "https://www.sron.nl/missions-earth/pace-spexone");
        putAtt("date_created", getDateAndTime());
        putAtt("git_commit", SPEXONE_GIT_COMMIT_ABBREV);
    }
    is_open = true;
}

auto NcObject::close() -> void
{
    netCDF::NcFile::close();
    is_open = false;
}

auto NcObject::reOpen() -> void
{
    if (filename.empty()) {
        throw std::runtime_error {
            "cannot reopen a NetCDF file that has not been opened yet"
        };
    }
    netCDF::NcFile::open(filename, fmode);
    is_open = true;
}

auto NcObject::writeReducedConfig(const SPEXSettings& settings) -> void
{
    auto var { addVar("processor_configuration", netCDF::ncString) };
    var.putAtt("comment",
               "configuration parameters used during the processor run that "
               "produced this file");
    std::string conf_str { settings.getReducedConfig().c_str() };
    const char* conf_char { conf_str.c_str() };
    var.putVar(&conf_char);
}

auto NcObject::getGroup(const std::string& path) const -> netCDF::NcGroup
{
    const std::vector<std::string> name_list { splitString(path, '/') };
    auto name_it { name_list.cbegin() };
    // The first group is taken from NcFile and subsequent groups, if
    // any, are taken as subgroups of the current group.
    netCDF::NcGroup group { this->NcFile::getGroup(*name_it) };
    while (++name_it != name_list.cend() && !group.isNull()) {
        group = group.getGroup(*name_it);
    }
    return group;
}

auto NcObject::copyAtts(const netCDF::NcVar& in, netCDF::NcVar& out) -> void
{
    void* value_buf[32768] {};
    for (const auto& [name, att] : in.getAtts()) {
        att.getValues(value_buf);
        out.putAtt(name, att.getType(), att.getAttLength(), value_buf);
    }
}

auto NcObject::copyVar(const netCDF::NcVar& var,
                       netCDF::NcGroup& group,
                       const std::vector<netCDF::NcDim>& dims) -> netCDF::NcVar
{
    auto var_out { group.addVar(
      var.getName(), var.getType(), dims.empty() ? var.getDims() : dims) };
    copyAtts(var, var_out);
    // Only copy the values over if dimensions were not given
    if (dims.empty()) {
        std::vector<size_t> _dims {};
        for (const auto& dim : var.getDims()) {
            _dims.push_back(dim.getSize());
        }
        size_t total_size { std::accumulate(_dims.cbegin(),
                                            _dims.cend(),
                                            static_cast<size_t>(1),
                                            std::multiplies<size_t>()) };
        void* values[var.getType().getSize() * total_size];
        var.getVar(values);
        var_out.putVar(values);
    }
    return var_out;
}

NcObject::~NcObject()
{
#pragma omp critical
    {
        if (!filename.empty()) {
            if (fmode != netCDF::NcFile::read) {
                putAtt("date_finalized", getDateAndTime());
            }
        }
        close();
    }
}

} // namespace spex
