// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "noise_gen.h"

#include "io.h"
#include "l1a_metadata.h"

#include <gadfit/lm_solver.h>
#include <numbers>

namespace spex {

auto NoiseGen::processInit() -> void
{
    // First dimension is the number of unbinned pixels, second
    // dimension is the total number of L1A files (calibration
    // measurements).
    averages.resize(ckd.npix * nL1A(), 0.0);
    sigmas.resize(ckd.npix * nL1A(), 0.0);
}

auto NoiseGen::prepareBatches() -> void
{
    batchOne();
}

auto NoiseGen::processBatch(const int i_batch) -> void
{
    if (const double p { getPercentage(i_batch, batches.size()) }; p >= 0.0) {
        spdlog::info("Computing averages and standard deviations: {:5.1f}%", p);
    }
    // Take the one and only L1A product from the batch
    const auto& l1a { batches.at(i_batch).l1a_products.front() };
    // Define a local array in order to avoid non-contiguous access of
    // averages.
    std::vector<double> averages_slice(ckd.npix, 0.0);
    for (auto& frame : l1a.frames) {
        std::ranges::transform(averages_slice,
                               frame,
                               averages_slice.begin(),
                               [](const double cur_sum, const double new_val) {
                                   return cur_sum + new_val;
                               });
    }
    for (auto& val : averages_slice) {
        val /= l1a.meta->number_of_images;
    }
    for (int i {}; i < ckd.npix; ++i) {
        averages[i * nL1A() + i_batch] = averages_slice[i];
    }
    // Same as with averages_slice.
    std::vector<double> sigmas_slice(ckd.npix, 0.0);
    for (auto& frame : l1a.frames) {
        for (int i {}; i < ckd.npix; ++i) {
            const double diff { frame[i] - averages_slice[i] };
            sigmas_slice[i] += diff * diff;
        }
    }
    // Unbiased standard deviation is
    //   sigma = f_sigma * sigma1
    // where the prefactor is
    //   f_sigma = sqrt[(n-1)/(2pi)] * B[(n-1)/2, 1/2]
    // and sigma1 is computed by the normal standard deviation formula.
    const double f_sigma {
        std::sqrt((l1a.meta->number_of_images - 1) / (2.0 * std::numbers::pi))
        * std::beta((l1a.meta->number_of_images - 1) / 2.0, 0.5)
    };
    for (auto& val : sigmas_slice) {
        val = f_sigma * std::sqrt(val / (l1a.meta->number_of_images - 1));
    }
    for (int i {}; i < ckd.npix; ++i) {
        sigmas[i * nL1A() + i_batch] = sigmas_slice[i];
    }
}

auto NoiseGen::processFinalize() -> void
{
    // Allocate normal and debug variables
    ckd.noise.g.resize(ckd.npix);
    ckd.noise.n.resize(ckd.npix);
    if (settings.debug.at(static_cast<int>(ProcLevel::noise)).value()) {
        ckd.noise.std_dev_g.resize(ckd.npix);
        ckd.noise.std_dev_n.resize(ckd.npix);
    }
    // Work arrays for storing the averaged intensities and standard deviations
    std::vector<std::array<double, 2>> values(nL1A());
    std::vector<double> x_values {}; // Intensities for a given pixel
    std::vector<double> y_values {}; // sigmas for a given pixel
#pragma omp parallel for firstprivate(values) private(x_values, y_values)
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        if (const double p { getPercentage(i_pix, ckd.npix, 1.0) }; p >= 0.0) {
            spdlog::info("Fitting noise g and n parameters: {:5.1f}%", p);
        }
        // Sort intensities and standard deviations by intensity values
        for (int i {}; i < nL1A(); ++i) {
            const int idx { i_pix * nL1A() + i };
            values[i] = { averages[idx], sigmas[idx] };
        }
        std::ranges::sort(values, [](const auto& a, const auto& b) {
            return a.front() < b.front();
        });
        // Find the first positive intensity value
        auto values_it { std::ranges::find_if(
          values, [](const auto& x) { return x.front() > 0.0; }) };
        x_values.clear();
        y_values.clear();
        // Only include positive intensities and the corresponding
        // standard deviations up to the intensity threshold.
        while (values_it != values.cend()) {
            if (values_it->front() < settings.noise.intensity_threshold) {
                x_values.push_back(values_it->front());
                y_values.push_back(values_it->back());
                ++values_it;
            } else {
                break;
            }
        }
        gadfit::LMsolver solver {
            [](const std::vector<gadfit::AdVar>& parameters,
               const double x) -> gadfit::AdVar {
                return sqrt(parameters[0] * x + parameters[1]);
            }
        };
        solver.addDataset(x_values, y_values);
        solver.setPar(0, settings.noise.fitting.initial_g, true);
        solver.setPar(1, settings.noise.fitting.initial_n, true);
        solver.settings.iteration_limit =
          settings.noise.fitting.iteration_limit;
        solver.settings.verbosity = gadfit::io::hide_all;
        solver.fit();
        ckd.noise.g[i_pix] = solver.getParValue(0);
        ckd.noise.n[i_pix] = solver.getParValue(1);
        if (settings.debug.at(static_cast<int>(ProcLevel::noise)).value()) {
            const double chi2 { solver.chi2() / solver.degreesOfFreedom() };
            auto& inv_JTJ { solver.getInvJTJ() };
            ckd.noise.std_dev_g[i_pix] = std::sqrt(chi2 * inv_JTJ[0]);
            ckd.noise.std_dev_n[i_pix] = std::sqrt(chi2 * inv_JTJ[3]);
        }
    }
    // Update pixel mask
    if (settings.noise.mask_g_min) {
        updateMaskLt(ckd.noise.g,
                     settings.noise.mask_g_min.value(),
                     "pixels masked bad because their noise inverse conversion "
                     "gain (g) is less than",
                     ckd.pixel_mask);
    }
    if (settings.noise.mask_g_max) {
        updateMaskGt(ckd.noise.g,
                     settings.noise.mask_g_max.value(),
                     "pixels masked bad because their noise inverse conversion "
                     "gain (g) is greater than",
                     ckd.pixel_mask);
    }
    if (settings.noise.mask_n_min) {
        updateMaskLt(ckd.noise.n,
                     settings.noise.mask_n_min.value(),
                     "pixels masked bad because their dark noise (n) is less "
                     "than",
                     ckd.pixel_mask);
    }
    if (settings.noise.mask_n_max) {
        updateMaskGt(ckd.noise.n,
                     settings.noise.mask_n_max.value(),
                     "pixels masked bad because their dark noise (n) is "
                     "greater than",
                     ckd.pixel_mask);
    }
}

} // namespace spex
