// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "nonlin_gen.h"

#include "functions.h"
#include "io.h"
#include "l1a_metadata.h"
#include "linalg.h"
#include "linear_spline.h"
#include "lininv.h"

#include <gadfit/lm_solver.h>

namespace spex {

auto NonlinGen::processInit() -> void
{
    const int pre_bin_spat { settings.main.detector_spat_full.value()
                             / settings.main.detector_spat };
    const int pre_bin_spec { settings.main.detector_spec_full.value()
                             / settings.main.detector_spec };
    pre_bin = pre_bin_spat * pre_bin_spec;
    ckd.nonlin.model = settings.nonlin.model;
    ckd.nonlin.debug_pixels = settings.nonlin.debug_pixels;
    ckd.nonlin.order = settings.nonlin.order;
    ckd.dim_nonlin_knot = settings.nonlin.knots.size();
    if (settings.nonlin.knots.empty()) {
        ckd.nonlin.b_spline_knots.resize(settings.nonlin.n_knots);
        const int beg { settings.nonlin.target_counts_min };
        const int end { settings.nonlin.target_counts_max };
        for (int i {}; i < settings.nonlin.n_knots; ++i) {
            ckd.nonlin.b_spline_knots[i] =
              beg + i * (end - beg) / (settings.nonlin.n_knots - 1);
        }
    } else {
        ckd.nonlin.b_spline_knots = settings.nonlin.knots;
    }
    if (ckd.nonlin.model != NonlinModel::single) {
        return;
    }
    b = std::make_unique<Bspline>(ckd.nonlin.order + 1,
                                  ckd.dim_nonlin_knot,
                                  ckd.nonlin.b_spline_knots.data());
    ckd.dim_nonlin_spline = b->nspline;
    abscissa.resize(nL1A());
    for (int i_l1a {}; i_l1a < nL1A(); ++i_l1a) {
        abscissa[i_l1a] =
          l1a_manager->l1a_products[i_l1a].exposure_time
          * l1a_manager->l1a_products[i_l1a].meta->illumination_level;
    }
    ckd.nonlin_signal_scale.resize(ckd.npix);
    nonlin_lin_slope.resize(ckd.npix);
    nonlin_lin_chi2.resize(ckd.npix);
    ckd.nonlin_fit.resize(ckd.npix * ckd.dim_nonlin_spline);
    nonlin_chi2.resize(ckd.npix);
}

// Modify signal by adding signal-induced offset to it
// [Eq. (80)]. After this the image contains the dark current. If
// !active do not add signal induced offset, only add dark current.
static auto addSignalInduced(const bool active,
                             CKD& ckd,
                             std::vector<L1A>& l1a_products) -> void
{
    spdlog::info("Applying signal induced offset");
    if (ckd.dark.current.empty()) {
        // This module relies on parts of the dark CKD. If not
        // present, use dummy values.
        ckd.dark.current.resize(ckd.npix, 0.0);
        ckd.dark.offset_long.resize(ckd.npix, 0.0);
    }
#pragma omp parallel for
    for (int i_l1a = 0; i_l1a < static_cast<int>(l1a_products.size());
         ++i_l1a) {
        for (int i_row {}; i_row < ckd.n_detector_rows; ++i_row) {
            // Average across columns for this row
            double average {};
            // Number of values to be averaged can be less than the
            // detector dimension if there are bad pixels.
            int n_values {};
            for (int i_col {}; i_col < ckd.n_detector_cols; ++i_col) {
                const int idx { i_row * ckd.n_detector_cols + i_col };
                if (ckd.pixel_mask[idx]) {
                    continue;
                }
                average += l1a_products[i_l1a].image[idx];
                ++n_values;
            }
            constexpr double rel_factor { 0.17 };
            const double fac { (active && average >= 0.0 && n_values > 0)
                                 ? rel_factor * std::sqrt(average / n_values)
                                 : 0.0 };
            for (int i_col {}; i_col < ckd.n_detector_cols; ++i_col) {
                const int idx { i_row * ckd.n_detector_cols + i_col };
                if (ckd.pixel_mask[idx]) {
                    continue;
                }
                l1a_products[i_l1a].image[idx] +=
                  ckd.dark.current[idx] * l1a_products[i_l1a].exposure_time
                  - fac;
            }
        }
    }
}

// Sort measurements into batches such that within each batch the
// reference diode value is more or less constant. Preliminary to step
// 4.
static auto groupByReferenceSignal(const std::vector<L1A>& l1a_products)
  -> std::vector<std::vector<std::reference_wrapper<const L1A>>>
{
    std::vector<std::reference_wrapper<const L1A>> l1a_refs {
        l1a_products.begin(), l1a_products.end()
    };
    std::ranges::sort(l1a_refs, [](const auto& l1a_1, const auto& l1a_2) {
        return l1a_1.get().meta->reference_signal
               < l1a_2.get().meta->reference_signal;
    });
    // Group by light intensity
    std::vector<std::vector<std::reference_wrapper<const L1A>>>
      l1a_grouped_by_phi {};
    auto l1a_first { l1a_refs.begin() };
    auto l1a_last { l1a_refs.begin() };
    while (l1a_last++ != l1a_refs.end()) {
        const auto& meta_first { l1a_first->get().meta };
        const auto& meta_last { l1a_last != l1a_refs.end()
                                  ? l1a_last->get().meta
                                  : meta_first };
        const auto rel_ref_signal_diff { std::abs(
          (meta_first->reference_signal - meta_last->reference_signal)
          / meta_first->reference_signal) };
        constexpr double rel_ref_signal_diff_thres { 0.01 };
        if (l1a_last == l1a_refs.end()
            || meta_first->filename.find("_lin_") == std::string::npos
            || meta_last->filename.find("_lin_") == std::string::npos
            || rel_ref_signal_diff > rel_ref_signal_diff_thres) {
            // Do not create a group if there are 3 or less
            // measurements (exposure times)
            if (std::ranges::distance(l1a_first, l1a_last) > 3) {
                l1a_grouped_by_phi.push_back({ l1a_first, l1a_last });
            }
            l1a_first = l1a_last;
        }
    }
    for (auto& group : l1a_grouped_by_phi) {
        std::ranges::sort(group, [](const auto& l1a_1, const auto& l1a_2) {
            return l1a_1.get().exposure_time < l1a_2.get().exposure_time;
        });
    }
    // Now all measurements are grouped by light level and exposure time
    return l1a_grouped_by_phi;
}

auto NonlinGen::computeSignalRates(const bool include_f2,
                                   std::vector<L1A>& l1a_products) const -> void
{
    // Number of parameters for the Ic fit, Eq. (81). If 1 then only
    // f1 is fitted, else both f1 and f2.
    const int n_parameters { include_f2 ? 2 : 1 };
    spdlog::info("Computing signal rates");
    // L1A products grouped by light level (reference diode value)
    const auto l1a_grouped_by_phi { groupByReferenceSignal(l1a_products) };
    // Debug information
    const int npix_d { static_cast<int>(ckd.nonlin.debug_pixels.size()) };
    ckd.nonlin.signal_rates.resize(
      npix_d, std::vector<double>(l1a_products.size(), 0.0));
    ckd.nonlin.reference_rates.resize(
      npix_d, std::vector<double>(l1a_products.size(), 0.0));
    ckd.nonlin.f1.resize(ckd.npix, 0.0);
    ckd.nonlin.f2.resize(ckd.npix, 0.0);
    // These are the actual work arrays for signal and reference rates
    std::vector<double> signal_rates(l1a_grouped_by_phi.size(), 0.0);
    std::vector<double> reference_rates(l1a_grouped_by_phi.size(), 0.0);
    std::vector<double> exposure_times {};
    std::vector<double> weights {};
    std::vector<double> jacobian {};
#pragma omp parallel for firstprivate(signal_rates, reference_rates) private(  \
  exposure_times, weights, jacobian)
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        if (ckd.pixel_mask[i_pix]) {
            continue;
        }
        // If using all signal rates over all intensities, start with
        // zero length and build up the arrays for each
        // pixel. Otherwise use the default sizes.
        if (settings.nonlin.use_all_signal_rates) {
            signal_rates.clear();
            reference_rates.clear();
        }
        for (int i_group {};
             i_group < static_cast<int>(l1a_grouped_by_phi.size());
             ++i_group) {
            const int n_l1a { static_cast<int>(
              l1a_grouped_by_phi.at(i_group).size()) };
            const int n_l1a_min { static_cast<int>(0.5 * n_l1a) };
            // Compute signal rate per pixel with relative least squares
            // regression
            double Q_00 {};
            double Q_11 {};
            double Q_22 {};
            exposure_times.resize(n_l1a);
            for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                exposure_times[i_l1a] =
                  l1a_grouped_by_phi[i_group][i_l1a].get().exposure_time;
            }
            double sigma_min { 1e100 };
            int data_point_count {};
            for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                double w;
                if (i_l1a == 0) {
                    w = 2 * (exposure_times[1] - exposure_times.front());
                } else if (i_l1a == n_l1a - 1) {
                    w = 2 * (exposure_times.back() - exposure_times[n_l1a - 2]);
                } else {
                    w = exposure_times[i_l1a + 1] - exposure_times[i_l1a - 1];
                }
                const double w_over_exp { w / exposure_times[i_l1a] };
                const double w_over_exp2 { w_over_exp / exposure_times[i_l1a] };
                Q_00 += w;
                const double& counts {
                    l1a_grouped_by_phi[i_group][i_l1a].get().image[i_pix]
                };
                Q_11 += w_over_exp * counts;
                Q_22 += w_over_exp2 * counts * counts;
                const double signal_rate { Q_11 / Q_00 };
                const int degrees_of_freedom { i_l1a - 1 };
                const double chi2 { Q_22 - Q_00 * signal_rate * signal_rate };
                const double sigma { chi2 / degrees_of_freedom / Q_00 };
                if (i_l1a > n_l1a_min && sigma < sigma_min) {
                    sigma_min = sigma;
                    const double& dc { ckd.dark.current[i_pix] };
                    if (settings.nonlin.use_all_signal_rates) {
                        signal_rates.push_back(counts / exposure_times[i_l1a]
                                               - dc);
                        reference_rates.push_back(
                          l1a_grouped_by_phi[i_group][i_l1a]
                            .get()
                            .meta->reference_signal);
                    } else {
                        signal_rates[i_group] = signal_rate - dc;
                        data_point_count = i_l1a + 1;
                    }
                }
            }
            // Only average the reference rates within a group if not
            // using all signal rates in the fit.
            if (!settings.nonlin.use_all_signal_rates) {
                reference_rates[i_group] = 0.0;
                for (int i_l1a {}; i_l1a < data_point_count; ++i_l1a) {
                    reference_rates[i_group] +=
                      l1a_grouped_by_phi[i_group][i_l1a]
                        .get()
                        .meta->reference_signal;
                }
                reference_rates[i_group] /= data_point_count;
            }
        }
        // Determine average reference current (step 5)
        weights.resize(signal_rates.size(), 1.0);
        jacobian.resize(signal_rates.size() * n_parameters);
        // Loop over light intensities
        for (int i {}; i < static_cast<int>(signal_rates.size()); ++i) {
            jacobian[i * n_parameters + 0] = reference_rates[i];
        }
        if (n_parameters == 2) {
            for (int i {}; i < static_cast<int>(signal_rates.size()); ++i) {
                jacobian[i * n_parameters + 1] =
                  reference_rates[i] * std::sqrt(reference_rates[i]);
            }
        }
        // Eq. (81)
        std::vector<double> control_points {};
        dppsv(signal_rates, weights, jacobian, control_points);
        ckd.nonlin.f1[i_pix] = control_points[0];
        if (n_parameters == 2) {
            ckd.nonlin.f2[i_pix] =
              control_points[1]
              / (ckd.nonlin.f1[i_pix] * std::sqrt(ckd.nonlin.f1[i_pix]));
        } else {
            ckd.nonlin.f2[i_pix] = 0.0;
        }
        // Save debug info
        if (settings.debug.at(static_cast<int>(ProcLevel::nonlin)).value()
            && !ckd.nonlin.debug_pixels.empty()) {
            for (int i_pix_d {}; i_pix_d < npix_d; ++i_pix_d) {
                if (ckd.nonlin.debug_pixels[i_pix_d] == i_pix) {
                    for (int i_val {};
                         i_val < static_cast<int>(signal_rates.size());
                         ++i_val) {
                        ckd.nonlin.signal_rates[i_pix_d][i_val] =
                          signal_rates[i_val];
                        ckd.nonlin.reference_rates[i_pix_d][i_val] =
                          reference_rates[i_val];
                    }
                }
            }
        }
    }
}

auto NonlinGen::computeMaximumSignals(
  const std::vector<L1A>& l1a_products) const -> void
{
    spdlog::info("Computing saturated signal parameters (m0, m1, m2)");
    std::vector<double> values_for_median(ckd.npix, 0.0);
    for (int i {}; i < ckd.npix; ++i) {
        if (!ckd.pixel_mask[i]) {
            values_for_median[i] = ckd.dark.offset_long[i];
        }
    }
    const double dark_offset_median { median(values_for_median) };
    const double clipping_threshold {
        pre_bin * (1000.0 - l1a_products.front().digital_offset_single)
        - dark_offset_median
    };
    std::vector<int> good_measurements {};
    // Photodiode reference values
    std::vector<double> I_ref {};
    for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size()); ++i_l1a) {
        // Only consider long exposure times where data is saturated
        constexpr double exposure_thres { 2.9 };
        if (l1a_products[i_l1a].exposure_time < exposure_thres) {
            continue;
        }
        for (int i {}; i < ckd.npix; ++i) {
            if (!ckd.pixel_mask[i]) {
                values_for_median[i] = l1a_products[i_l1a].image[i];
            }
        }
        const double med { median(values_for_median) };
        if (med < clipping_threshold) {
            I_ref.push_back(l1a_products[i_l1a].meta->reference_signal);
            ckd.nonlin.S_max_data.push_back(med);
            good_measurements.push_back(i_l1a);
        }
    }
    // Medians of linear detector currents with dark current
    ckd.nonlin.I_c_med.resize(I_ref.size());
#pragma omp parallel for firstprivate(values_for_median)
    for (int i_lev = 0; i_lev < static_cast<int>(I_ref.size()); ++i_lev) {
        for (int i {}; i < ckd.npix; ++i) {
            if (!ckd.pixel_mask[i]) {
                values_for_median[i] =
                  ckd.dark.current[i] + ckd.nonlin.f1[i] * I_ref[i_lev]
                  + ckd.nonlin.f2[i]
                      * std::pow(ckd.nonlin.f1[i] * I_ref[i_lev], 1.5);
            } else {
                values_for_median[i] = 0.0;
            }
        }
        ckd.nonlin.I_c_med[i_lev] = median(values_for_median);
    }
    gadfit::LMsolver solver { [](const std::vector<gadfit::AdVar>& parameters,
                                 const double x) -> gadfit::AdVar {
        return parameters[0] + parameters[1] * std::log(x)
               + parameters[2] / std::sqrt(x);
    } };
    solver.addDataset(ckd.nonlin.I_c_med, ckd.nonlin.S_max_data);
    solver.setPar(0, 1e3, true);
    solver.setPar(1, 1e1, true);
    solver.setPar(2, -1e1, true);
    solver.settings.verbosity = gadfit::io::final_only;
    solver.fit();
    ckd.nonlin.m0.resize(ckd.npix);
    ckd.nonlin.m1 = solver.getParValue(1);
    ckd.nonlin.m2 = solver.getParValue(2);
    // Refit m0
    std::vector<double> I_lin(good_measurements.size());
    std::vector<double> m0_data(good_measurements.size());
#pragma omp parallel for firstprivate(I_lin, m0_data)
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        for (int i_meas {}; i_meas < static_cast<int>(good_measurements.size());
             ++i_meas) {
            I_lin[i_meas] =
              ckd.dark.current[i_pix] + ckd.nonlin.f1[i_pix] * I_ref[i_meas]
              + ckd.nonlin.f2[i_pix]
                  * std::pow(ckd.nonlin.f1[i_pix] * I_ref[i_meas], 1.5);
            m0_data[i_meas] =
              l1a_products[good_measurements[i_meas]].image[i_pix]
              - ckd.nonlin.m1 * std::log(I_lin[i_meas])
              - ckd.nonlin.m2 / std::sqrt(I_lin[i_meas]);
        }
        ckd.nonlin.m0[i_pix] =
          std::accumulate(m0_data.cbegin(), m0_data.cend(), 0.0)
          / static_cast<int>(m0_data.size());
    }
    // Sort for easier visualization (these are not used anymore)
    std::ranges::sort(ckd.nonlin.I_c_med);
    std::ranges::sort(ckd.nonlin.S_max_data);
    for (const auto& I_c : ckd.nonlin.I_c_med) {
        ckd.nonlin.S_max_fit.push_back(solver.getValue(I_c));
    }
}

auto NonlinGen::fitNonlinearity(const std::vector<L1A>& l1a_products) const
  -> void
{
    // Arrays to hold the linear and measured signals. Number of
    // points can be different for each pixel.
    std::vector<double> S_lin {};
    std::vector<double> S_meas {};
    // Like S_max but without m0: S_max_partial = m1 * ln(Ic) + m2 / sqrt(Ic)
    std::vector<double> S_max_partial {};
    // The index array serves as x-values for the solver. Its values
    // refers to elements in the S_lin and S_max_partial arrays inside
    // the solver.
    std::vector<double> S_idx {};
    // Linear current at the original exposure time of the calibration
    // measurement
    std::vector<double> I_lin_tau0 {};
    // Linear current at the target exposure time
    std::vector<double> I_lin_tau {};
    // B-spline matrix of basis functions
    std::vector<double> B_mat {};
    // Work array for the B-spline control points
    std::vector<double> control_points_wrk {};
    // Noise for the B-spline fit (all ones)
    std::vector<double> std_devs {};
    // S_meas and S_lin sorted by ascending S_meas
    std::vector<double> S_meas_sorted {};
    std::vector<double> S_lin_sorted {};
    // Perumation array for sortin S_meas
    std::vector<int> indices {};
    // Parameters for the sigmoidal function (if using that nonlinearity model)
    if (ckd.nonlin.model == NonlinModel::sigmoidal) {
        ckd.nonlin.r.resize(ckd.nonlin_n_r, std::vector<double>(ckd.npix));
        ckd.nonlin.c.resize(ckd.npix);
    } else {
        ckd.nonlin.b_spline = std::make_unique<BSpline>(
          ckd.nonlin.order, ckd.nonlin.b_spline_knots);
        ckd.nonlin.b_spline_deriv = std::make_unique<BSpline>(
          ckd.nonlin.order, ckd.nonlin.b_spline_knots);
        ckd.nonlin.control_points.resize(ckd.npix
                                         * ckd.nonlin.b_spline->nStates());
    }
    // If debug pixels were requested, set up additional arrays to be
    // saved in the CKD
    const int npix_d { static_cast<int>(ckd.nonlin.debug_pixels.size()) };
    if (settings.debug.at(static_cast<int>(ProcLevel::nonlin)).value()
        && !ckd.nonlin.debug_pixels.empty()) {
        const int n_max_signals { static_cast<int>(l1a_products.size()) };
        ckd.nonlin.S_meas.resize(npix_d,
                                 std::vector<double>(n_max_signals, 0.0));
        ckd.nonlin.S_lin.resize(npix_d,
                                std::vector<double>(n_max_signals, 0.0));
        ckd.nonlin.nonlinearity_fits.resize(
          npix_d, std::vector<double>(n_max_signals, 0.0));
    }
    // Main loop that is parallelized over pixels
#pragma omp parallel for private(                                              \
  S_meas, S_lin, I_lin_tau0, I_lin_tau, S_max_partial, S_idx)
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        if (ckd.pixel_mask[i_pix]) {
            continue;
        }
        if (const double p { getPercentage(i_pix, ckd.npix, 0.1) }; p >= 0.0) {
            spdlog::info("Fitting the nonlinearity model {:5.1f}%", p);
        }
        // These arrays can have different lengths for each pixel
        S_meas.clear();
        S_lin.clear();
        I_lin_tau0.clear();
        I_lin_tau.clear();
        S_max_partial.clear();
        S_idx.clear();
        const double clipping_threshold {
            pre_bin * (1020.0 - l1a_products.front().digital_offset_single)
            - ckd.dark.offset_long[i_pix]
        };
        constexpr double exposure_thres { 1700.0 };
        constexpr double current_thres { 50.0e3 }; // per ms
        for (const auto& l1a : l1a_products) {
            const double I_lin { ckd.dark.current[i_pix]
                                 + ckd.nonlin.f1[i_pix]
                                     * l1a.meta->reference_signal };
            const double I_c_nlin { std::max(I_lin - ckd.dark.current[i_pix],
                                             0.0) };
            const double I_c {
                I_lin + ckd.nonlin.f2[i_pix] * I_c_nlin * std::sqrt(I_c_nlin)
            };
            if (l1a.image[i_pix] < clipping_threshold
                && I_lin * l1a.exposure_time < pre_bin * exposure_thres
                && I_lin < pre_bin * current_thres) {
                S_lin.push_back(I_c * l1a.exposure_time);
                S_meas.push_back(l1a.image[i_pix]);
                S_max_partial.push_back(ckd.nonlin.m1 * std::log(I_c)
                                        + ckd.nonlin.m2 / std::sqrt(I_c));
                I_lin_tau0.push_back(I_c);
                I_lin_tau.push_back(I_c * l1a.exposure_time
                                    / settings.nonlin.target_exposure_time);
            }
        }
        if (ckd.nonlin.model == NonlinModel::sigmoidal) {
            S_idx.resize(S_lin.size());
            std::iota(S_idx.begin(), S_idx.end(), 0.0);
            // Define the fitting function
            const auto sigmoidal { [&S_lin, &S_max_partial](
                                     const std::vector<gadfit::AdVar>& pars,
                                     const double idx_d) -> gadfit::AdVar {
                using gadfit::AdVar;
                const int idx { static_cast<int>(idx_d) };
                const AdVar S_max { pars[6] + S_max_partial[idx] };
                const AdVar x { S_lin[idx] / S_max };
                const AdVar x2 { x * x };
                constexpr double eps { 1e-10 };
                const AdVar ln_x { log(abs(x) + eps) };
                const AdVar r { pars[0] + pars[1] * ln_x + pars[2] * x2
                                + pars[3] * x2 * ln_x + pars[4] * x2 * abs(x) };
                const AdVar& c { pars[5] };
                const AdVar xrc { x * r * c / tanh(c) };
                return S_max / (2.0 * c) * log(cosh(c + xrc) / cosh(c - xrc));
            } };
            // Set up solver
            gadfit::LMsolver solver { sigmoidal };
            solver.addDataset(S_idx, S_meas);
            solver.settings.iteration_limit = 100;
            solver.settings.loss = gadfit::Loss::cauchy;
            solver.settings.verbosity = gadfit::io::hide_all;
            // Set initial parameters
            solver.setPar(0, 1.00, true);                 // r0
            solver.setPar(1, 0.020, true);                // r1
            solver.setPar(2, -0.01, true);                // r2
            solver.setPar(3, -0.01, true);                // r3
            solver.setPar(4, 0.01, true);                 // r4
            solver.setPar(5, 7.0, true);                  // c
            solver.setPar(6, ckd.nonlin.m0[i_pix], true); // m0
            // Perform fit
            solver.fit();
            // Retrieve parameters
            for (int i_r {}; i_r < ckd.nonlin_n_r; ++i_r) {
                ckd.nonlin.r[i_r][i_pix] = solver.getParValue(i_r);
            }
            ckd.nonlin.c[i_pix] = solver.getParValue(5);
            ckd.nonlin.m0[i_pix] = solver.getParValue(6);
            // Save debugging information
            if (!ckd.nonlin.nonlinearity_fits.empty()) {
                for (int i_pix_d {}; i_pix_d < npix_d; ++i_pix_d) {
                    if (ckd.nonlin.debug_pixels[i_pix_d] == i_pix) {
                        for (int i_lin {};
                             i_lin < static_cast<int>(S_lin.size());
                             ++i_lin) {
                            const double S_max_tau0 {
                                ckd.nonlin.m0[i_pix]
                                + ckd.nonlin.m1 * std::log(I_lin_tau0[i_lin])
                                + ckd.nonlin.m2 / std::sqrt(I_lin_tau0[i_lin])
                            };
                            const double S_max_tau {
                                ckd.nonlin.m0[i_pix]
                                + ckd.nonlin.m1 * std::log(I_lin_tau[i_lin])
                                + ckd.nonlin.m2 / std::sqrt(I_lin_tau[i_lin])
                            };
                            const double ratio { S_max_tau / S_max_tau0 };
                            ckd.nonlin.S_lin[i_pix_d][i_lin] =
                              ratio * S_lin[i_lin];
                            ckd.nonlin.S_meas[i_pix_d][i_lin] =
                              ratio * S_meas[i_lin];
                            ckd.nonlin.nonlinearity_fits[i_pix_d][i_lin] =
                              ratio
                              * solver.getValue(static_cast<double>(i_lin));
                        }
                        break;
                    }
                }
            }
        } else {
            for (int i {}; i < static_cast<int>(S_lin.size()); ++i) {
                const double S_max_tau0 {
                    ckd.nonlin.m0[i_pix]
                    + ckd.nonlin.m1 * std::log(I_lin_tau0[i])
                    + ckd.nonlin.m2 / std::sqrt(I_lin_tau0[i])
                };
                const double S_max_tau {
                    ckd.nonlin.m0[i_pix]
                    + ckd.nonlin.m1 * std::log(I_lin_tau[i])
                    + ckd.nonlin.m2 / std::sqrt(I_lin_tau[i])
                };
                const double ratio { S_max_tau / S_max_tau0 };
                S_lin[i] *= ratio;
                S_meas[i] *= ratio;
            }
            indices.resize(S_meas.size());
            std::iota(indices.begin(), indices.end(), 0);
            std::ranges::sort(indices, [&S_meas](const int a, const int b) {
                return S_meas[a] < S_meas[b];
            });
            S_meas_sorted.resize(S_meas.size());
            S_lin_sorted.resize(S_lin.size());
            for (int i {}; i < static_cast<int>(indices.size()); ++i) {
                S_meas_sorted[i] = S_meas[indices[i]];
                S_lin_sorted[i] = S_lin[indices[i]];
            }
            ckd.nonlin.b_spline->evalBasis(S_meas_sorted, B_mat);
            std_devs.resize(S_meas_sorted.size(), 1.0);
            dppsv(S_lin_sorted, std_devs, B_mat, control_points_wrk);
            std::copy(control_points_wrk.cbegin(),
                      control_points_wrk.cend(),
                      ckd.nonlin.control_points.begin()
                        + i_pix * ckd.nonlin.b_spline->nStates());
            // Save debugging information
            if (!ckd.nonlin.nonlinearity_fits.empty()) {
                for (int i_pix_d {}; i_pix_d < npix_d; ++i_pix_d) {
                    if (ckd.nonlin.debug_pixels[i_pix_d] == i_pix) {
                        for (int i {};
                             i < static_cast<int>(S_meas_sorted.size());
                             ++i) {
                            ckd.nonlin.S_meas[i_pix_d][i] = S_meas_sorted[i];
                            ckd.nonlin.S_lin[i_pix_d][i] = S_lin_sorted[i];
                            ckd.nonlin.nonlinearity_fits[i_pix_d][i] =
                              ckd.nonlin.b_spline->deBoor(control_points_wrk,
                                                          S_meas_sorted[i]);
                        }
                        break;
                    }
                }
            }
        }
    }
}

// Here is the sigmoidal function again but rewritten for faster
// evaluation. The one used in the solver is much slower.
[[nodiscard]] auto NonlinGen::sigmoidal(const int pre_bin,
                                        const CKD& ckd,
                                        const double S_lin,
                                        const double exposure_time,
                                        const int i_pix) -> double
{
    const double I_lin { S_lin / exposure_time };
    const double S_max_min { pre_bin * 200.0 };
    const double S_max { std::max(
      S_max_min,
      I_lin <= 0.0 ? S_max_min
                   : ckd.nonlin.m0[i_pix] + ckd.nonlin.m1 * std::log(I_lin)
                       + ckd.nonlin.m2 / std::sqrt(I_lin)) };
    const double x { (I_lin * exposure_time) / S_max };
    const double x2 { x * x };
    const double eps { pre_bin * 1e-10 };
    const double ln_x { std::log(std::abs(x) + eps) };
    const double r { ckd.nonlin.r[0][i_pix] + ckd.nonlin.r[1][i_pix] * ln_x
                     + ckd.nonlin.r[2][i_pix] * x2
                     + ckd.nonlin.r[3][i_pix] * x2 * ln_x
                     + ckd.nonlin.r[4][i_pix] * x2 * std::abs(x) };
    const double c { ckd.nonlin.c[i_pix] };
    const double xrc { x * r * c / std::tanh(c) };
    return S_max / (2.0 * c)
           * std::log(std::cosh(c + xrc) / std::cosh(c - xrc));
}

auto NonlinGen::parametrizeSigmoidal() const -> void
{
    // Interpolation is first done on a dense (S_lin, S_meas) grid and
    // then interpolated onto a sparser (S_meas, S_lin) grid.
    constexpr int knot_multiplier { 10 };
    const int n_S_lin_dense {
        knot_multiplier * static_cast<int>(ckd.nonlin.b_spline_knots.size())
    };
    std::vector<double> S_lin_dense(n_S_lin_dense);
    // Minimum and maximum counts of the linear signal
    const int c_min { settings.nonlin.target_counts_min };
    double c_max {};
    // Find the maximum linear signal by testing the values of the
    // sigmoidal function using the target max measured signal as
    // argument.
    const int center_pixel { ckd.npix / 2 + ckd.n_detector_cols / 2 };
    for (int counts {}; counts < pre_bin * 1000; ++counts) {
        const int S_lin { settings.nonlin.target_counts_max + counts };
        if (sigmoidal(pre_bin,
                      ckd,
                      static_cast<double>(S_lin),
                      settings.nonlin.target_exposure_time,
                      center_pixel)
            >= settings.nonlin.target_counts_max) {
            c_max = S_lin;
            break;
        }
    }
    for (int i {}; i < n_S_lin_dense; ++i) {
        S_lin_dense[i] = c_min + i * (c_max - c_min) / (n_S_lin_dense - 1.0);
    }
    // Define knots for the CKD nonlinearity spline. These are the
    // same for each pixel and won't change. Note that the dense knot
    // vector (S_lin_dense) refers to the linear signals. The CKD knot
    // vector refers to measured signals and its largest value is
    // determined by the sigmoidal function using the last value of
    // S_lin_dense.
    ckd.nonlin.linear_spline = {
        ckd.npix, static_cast<int>(ckd.nonlin.b_spline_knots.size())
    };
    const double c_max_meas { sigmoidal(pre_bin,
                                        ckd,
                                        S_lin_dense.back(),
                                        settings.nonlin.target_exposure_time,
                                        center_pixel) };
    for (int i {}; i < static_cast<int>(ckd.nonlin.b_spline_knots.size());
         ++i) {
        ckd.nonlin.linear_spline.knots[i] =
          c_min
          + i * (c_max_meas - c_min)
              / (static_cast<int>(ckd.nonlin.b_spline_knots.size()) - 1);
    }
    std::vector<double> S_meas_dense(n_S_lin_dense);
    std::vector<double> S_lin_sparse(
      static_cast<int>(ckd.nonlin.b_spline_knots.size()));
#pragma omp parallel for firstprivate(S_meas_dense, S_lin_sparse)
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        if (ckd.pixel_mask[i_pix]) {
            continue;
        }
        if (const double p { getPercentage(i_pix, ckd.npix, 1.0) }; p >= 0.0) {
            spdlog::info("Parameterizing the fit with splines {:5.1f}%", p);
        }
        for (int i_knot {}; i_knot < n_S_lin_dense; ++i_knot) {
            S_meas_dense[i_knot] =
              sigmoidal(pre_bin,
                        ckd,
                        S_lin_dense[i_knot],
                        settings.nonlin.target_exposure_time,
                        i_pix);
        }
        // Now that we have the dense mapping, interpolate the linear
        // signal values using the CKD knot vector (measured signals)
        // which is shared by all pixels.
        const LinearSpline spline { S_meas_dense, S_lin_dense };
        for (int i {}; i < static_cast<int>(ckd.nonlin.b_spline_knots.size());
             ++i) {
            S_lin_sparse[i] = spline.eval(ckd.nonlin.linear_spline.knots[i]);
        }
        ckd.nonlin.linear_spline.generateSpline(i_pix, S_lin_sparse);
    }
}

static auto readInputCKD(const std::string filename, CKD& ckd) -> void
{
    NcObject ckd_in { netCDF::NcFile::read, filename };
    const auto grp { ckd_in.getGroup("DEBUG").getGroup("NON_LINEARITY") };
    ckd.readNonlinDebugData(grp);
}

auto NonlinGen::processBatch() -> void
{
    if (settings.nonlin.restart_ckd.empty()) {
        // We start with dark offset and current corrected data. After
        // the first step the signal again includes dark current.
        addSignalInduced(settings.nonlin.signal_induced_offset,
                         ckd,
                         batches.front().l1a_products);
        computeSignalRates(settings.nonlin.include_f2,
                           batches.front().l1a_products);
        computeMaximumSignals(batches.front().l1a_products);
        fitNonlinearity(batches.front().l1a_products);
        // The L1A products are not needed anymore. Free memory before
        // starting with the parameterization.
        batches.front().l1a_products = std::vector<L1A> {};
    } else {
        batches.front().l1a_products = std::vector<L1A> {};
        readInputCKD(settings.nonlin.restart_ckd, ckd);
    }
    if (ckd.nonlin.model == NonlinModel::sigmoidal) {
        parametrizeSigmoidal();
    }
}

auto NonlinGen::processMedian() -> void
{
    const auto l1a { std::ranges::max(
      batches.front().l1a_products, [](const auto& a, const auto& b) {
          return a.meta->illumination_level < b.meta->illumination_level;
      }) };
    if (settings.nonlin.scale_median) {
        std::vector<double> buf { l1a.image };
        const double med { median(buf) };
        for (auto& val : ckd.nonlin_signal_scale) {
            val = med;
        }
    } else {
        for (int i {}; i < ckd.npix; ++i) {
            ckd.nonlin_signal_scale[i] = l1a.image[i];
        }
    }
}

auto NonlinGen::processLin() -> void
{
    const auto& l1a_products { batches.front().l1a_products };
    // Linear result pointers
    double* linslopeptr = &nonlin_lin_slope[0];
    double* linchi2ptr = &nonlin_lin_chi2[0];

    spdlog::info("Calculating linear fit per pixel");
    for (size_t ipix {}; ipix < ckd.npix; ipix++) {
        // Put fillvalues everywhere
        *linslopeptr = -999.0;
        *linchi2ptr = -999.0;

        // Breakable do-once loop
        for (bool ctd = true; ctd; ctd = false) {

            // Die if you are already dead
            if (ckd.pixel_mask[ipix])
                break;
            ckd.pixel_mask[ipix] =
              true; // Unless you survive until the end of the loop

            // Gather the measurements at this pixel index
            std::vector<double> signal(nL1A());
            std::vector<double> noise(nL1A());
            std::vector<bool> pixel_mask(nL1A());
            for (size_t il1a = 0; il1a < nL1A(); il1a++) {
                signal[il1a] = l1a_products[il1a].image[ipix];
                noise[il1a] = l1a_products[il1a].noise[ipix];
                pixel_mask[il1a] = l1a_products[il1a].pixel_mask[ipix];
            }
            // Filter for signals between the chosen range
            size_t nl1a_lin = 0;
            size_t nl1a_lin_healthy = 0;
            // Maximum allocation size
            std::vector<double> abscissa_lin(nL1A());
            std::vector<double> signal_lin(nL1A());
            std::vector<double> s_y_lin(nL1A());
            std::vector<bool> pixel_mask_lin(nL1A());
            for (size_t il1a = 0; il1a < nL1A(); il1a++) {
                double& sig = signal[il1a];
                if (sig > settings.nonlin.lmin && sig < settings.nonlin.lmax) {
                    abscissa_lin[nl1a_lin] = abscissa[il1a];
                    signal_lin[nl1a_lin] = sig;
                    s_y_lin[nl1a_lin] = pow(noise[il1a], 2.0);
                    pixel_mask_lin[nl1a_lin] = pixel_mask[il1a];
                    if (!pixel_mask_lin[nl1a_lin]) {
                        nl1a_lin_healthy++;
                    }
                    ++nl1a_lin;
                }
            }

            // Encapsulate vectors to their intended sizes
            abscissa_lin.resize(nl1a_lin);
            signal_lin.resize(nl1a_lin);
            s_y_lin.resize(nl1a_lin);
            pixel_mask_lin.resize(nl1a_lin);

            // Nothing is possible if there are no reliable measurments
            if (nl1a_lin_healthy < 1)
                break;

            // Perform the linear fit. Note that if it fails, no mess is created
            // in the result before returning.
            if (linear_invert(1,
                              nl1a_lin,
                              abscissa_lin.data(),
                              OPT_DIAG,
                              s_y_lin.data(),
                              &pixel_mask_lin,
                              signal_lin.data(),
                              linslopeptr)
                != 0)
                break;
            *linchi2ptr = 0.0;
            for (size_t il1a = 0; il1a < nl1a_lin; il1a++) {
                if (pixel_mask_lin[il1a])
                    continue;
                *linchi2ptr +=
                  pow((abscissa_lin[il1a] * *linslopeptr - signal_lin[il1a]),
                      2.0)
                  / s_y_lin[il1a] / (nl1a_lin_healthy - 1);
            }

            // Apply mask criterion: Linear chi squared
            if (settings.nonlin.mask_lin_chi2_max != -999.0
                && *linchi2ptr > settings.nonlin.mask_lin_chi2_max)
                break;

            // Survived to the end of the loop
            ckd.pixel_mask[ipix] =
              false; // Unless you survive until the end of the loop
        }

        // Progress pointers
        linslopeptr++;
        linchi2ptr++;
    }
}

auto NonlinGen::processNonlin() -> void
{
    const auto& l1a_products { batches.front().l1a_products };
    // Pointers to earlier calculated results
    double* linslopeptr = &nonlin_lin_slope[0];
    double* scaleptr = &ckd.nonlin_signal_scale[0];
    // Pointers to results
    double* resptr = &ckd.nonlin_fit[0];
    double* chi2ptr = &nonlin_chi2[0];

    spdlog::info("Calculating nonlinearity fit per pixel");
    for (size_t ipix {}; ipix < ckd.npix; ipix++) {
        // Fill fillvalues everywhere
        for (size_t ispline = 0; ispline < ckd.dim_nonlin_spline; ispline++)
            resptr[ispline] = -999.0;
        *chi2ptr = -999.0;

        for (bool ctd = true; ctd; ctd = false) {
            if (ckd.pixel_mask[ipix]) {
                break;
            }
            ckd.pixel_mask[ipix] =
              true; // Unless you survive until the end of the loop

            std::vector<double> signal_corrected_scaled(nL1A());
            std::vector<double> signal_measured(nL1A());
            std::vector<double> s_y_measured(nL1A());
            std::vector<bool> pixel_mask_measured(nL1A());
            size_t nl1a_healthy = 0;
            for (size_t il1a = 0; il1a < nL1A(); il1a++) {
                L1A* l1a = &l1a_manager->l1a_products[il1a];
                signal_corrected_scaled[il1a] =
                  *linslopeptr * l1a->exposure_time
                  * l1a->meta->illumination_level / (*scaleptr);
                signal_measured[il1a] = l1a->image[ipix];
                s_y_measured[il1a] = pow(l1a->noise[ipix], 2.0);
                pixel_mask_measured[il1a] = l1a->pixel_mask[ipix];
                if (!pixel_mask_measured[il1a])
                    nl1a_healthy++;
            }

            // Perform the fit
            std::vector<double> jac(ckd.dim_nonlin_spline * nL1A(), 0.0);
            b->jaccalc(nL1A(), signal_corrected_scaled.data(), jac.data());
            // Add hard constraint at zero. Zero signal should mean zero
            // mismatch.
            double zero = 0.0; // Used as abscissa value and as measurement
            std::vector<double> jac_zero(ckd.dim_nonlin_spline);
            b->jaccalc(1, &zero, jac_zero.data());
            if (constrained_linear_invert(ckd.dim_nonlin_spline,
                                          nL1A(),
                                          1,
                                          jac.data(),
                                          jac_zero.data(),
                                          OPT_DIAG,
                                          s_y_measured.data(),
                                          &pixel_mask_measured,
                                          signal_measured.data(),
                                          &zero,
                                          resptr)
                != 0)
                break;
            *chi2ptr = 0.0;
            std::vector<double> mod(nL1A());
            Matrix::matmul_rl_fold_slow(
              ckd.dim_nonlin_spline, nL1A(), jac.data(), resptr, mod.data());
            for (size_t il1a = 0; il1a < nL1A(); il1a++) {
                if (pixel_mask_measured[il1a])
                    continue;
                // The plus one is the hard constraint, reducing the
                // DFS by one (or adding a measurement with no bias).
                *chi2ptr += pow(mod[il1a] - signal_measured[il1a], 2.0)
                            / s_y_measured[il1a]
                            / (nl1a_healthy - ckd.dim_nonlin_spline + 1);
            }

            // Apply mask criterion: Nonlinear chi squared
            if (settings.nonlin.mask_chi2_max
                && *chi2ptr > settings.nonlin.mask_chi2_max)
                break;

            // Survived to the end of the loop
            ckd.pixel_mask[ipix] = false;
        }

        // Progress pointers
        linslopeptr++;
        scaleptr += 1;
        resptr += ckd.dim_nonlin_spline;
        chi2ptr += 1;
    }
}

auto NonlinGen::processBatchSingle() -> void
{
    // Apply signal-induced offset
    addSignalInduced(
      settings.nonlin.signal_induced_offset, ckd, batches.front().l1a_products);
    // Subtract dark current
    for (auto& l1a : batches.front().l1a_products) {
        for (int i {}; i < ckd.npix; ++i) {
            l1a.image[i] -= ckd.dark.current[i] * l1a.exposure_time;
        }
    }
    processMedian();
    processLin();
    processNonlin();
}

auto NonlinGen::processBatch(const int /* i_batch */) -> void
{
    if (settings.nonlin.model != NonlinModel::single) {
        processBatch();
    } else {
        processBatchSingle();
    }
}

} // namespace spex
