// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for generating the nonlinearity CKD. The equation numbers
// refer to Paul's tech note SRON-SPEX-TN-2020-001 issue 0.5.

#pragma once

#include "processor.h"

namespace spex {

class Bspline;

class NonlinGen : public Processor
{
private:
    //  If data was prebinned (detector_spat_reduced and/or
    // detector_spec_reduced) then use a scaling factor for some of
    // the operations.
    int pre_bin { 1 };
    // Used when parallelizing over pixels
    std::vector<int> pix_ranges {};
    // Slope fit during linear fit
    std::vector<double> nonlin_lin_slope {};
    // Residual chi squared on the linear fit
    std::vector<double> nonlin_lin_chi2 {};
    // Residual chi squared on non-linearity correction
    std::vector<double> nonlin_chi2 {};
    // Exposure time times illumination level
    std::vector<double> abscissa {};
    // Fitting spline
    std::unique_ptr<Bspline> b {};
    std::vector<int> nonlin_il1a_tracker {};

    auto processInit() -> void override;

    // Compute signal rates per light intensity (step 4)
    auto computeSignalRates(const bool include_f2,
                            std::vector<L1A>& l1a_products) const -> void;
    // Determine offset corrected maximum signals from detector medians (steps 7
    // and 8)
    auto computeMaximumSignals(const std::vector<L1A>& l1a_products) const
      -> void;
    // Fit the nonlinearity model (step 9, Eqs. (72) and (73))
    auto fitNonlinearity(const std::vector<L1A>& l1a_products) const -> void;
    // Compute the (S_meas, S_lin) curves for each pixel and
    // parametrize them with cubic splines
    auto parametrizeSigmoidal() const -> void;
    // Generate the nonlinearity CKD using the sigmoidal function or
    // the B-spline model
    auto processBatch() -> void;

    // Previous implementation based on B-splines
    auto processMedian() -> void;
    auto processLin() -> void;
    auto processNonlin() -> void;
    auto processBatchSingle() -> void;

    // Driver routine that selects the model to run
    auto processBatch(const int /* ibatch */) -> void override;

public:
    NonlinGen(const SPEXSettings& settings, CKD& ckd)
      : Processor { settings, ProcLevel::nonlin, ckd }
    {}
    // Sigmoidal function for evaluating the nonlinearity. We leave it
    // static so it can be called from other codes, most importantly
    // the instrument model.
    [[nodiscard]] static auto sigmoidal(const int pre_bin,
                                        const CKD& ckd,
                                        const double S_lin,
                                        const double exposure_time,
                                        const int i_pix) -> double;
    ~NonlinGen() = default;
};

} // namespace spex
