// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Implementation of matrices with packed storage where the non-zero
// elements of each row are assumed to be short contiguous slices.

#pragma once

#include "constants.h"

#include <vector>

namespace spex {

class PackedMatrix
{
private:
    // Number of rows in the global (unpacked) matrix
    int n_rows {};
    // Number of columns in the global matrix
    int n_cols {};

public:
    // If i * n_cols is the starting index of the ith row in the
    // global matrix, row_locations[i] is the starting index in packed
    // storage.
    std::vector<int> row_locations {};
    // The first element of the ith row in packed storage has the
    // global column index global_col_indices[i].
    std::vector<int> global_col_indices {};
    // Number of non-zero elements of each row
    std::vector<int> row_sizes {};
    // Matrix contents in packed storage
    std::vector<double> data {};
    // Initialize with global dimensions
    PackedMatrix(const int n_rows, const int n_cols);
    // Number of rows in the global matrix
    [[nodiscard]] auto nRows() const -> int { return n_rows; }
    // Number of columns in the global matrix
    [[nodiscard]] auto nCols() const -> int { return n_cols; }
    // Number of elements on row i in packed matrix
    [[nodiscard]] auto nCols(const int i) const -> int { return row_sizes[i]; }
    // Return data element corresponding to the global indices i and j
    auto operator()(const int i, const int j) -> double&;
    // Return global column index that corresponds to the jth element
    // of the ith row in the packed matrix.
    auto globalCol(const int i_row, const int j_packed) const -> int;
    // Y =  X A^T where X is self
    auto multRightTransposed(const PackedMatrix& A,
                             std::vector<double>& Y) const -> void;
    // Y =  X A^T where X is self
    auto multRightTransposed(const int A_rows,
                             const std::vector<double>& A,
                             std::vector<double>& Y) const -> void;
    // Y = A X^T where X is self
    auto multLeftTransposed(const int A_rows,
                            const std::vector<double>& A,
                            std::vector<double>& Y) const -> void;
    // Y^T = A X^T where X is self
    auto multLeftTransposedT(const int A_rows,
                             const std::vector<double>& A,
                             std::vector<double>& Y) const -> void;
    // Tranpose X and write the result to self
    auto transpose(const PackedMatrix& X) -> void;
    ~PackedMatrix() = default;
};

} // namespace spex
