// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "pol_gen.h"

#include "io.h"
#include "l1a_metadata.h"
#include "linalg.h"
#include "read_solar_spectrum.h"

#include <fstream>
#include <gadfit/lm_solver.h>
#include <numbers>

namespace spex {

auto PolGen::processInit() -> void
{
    // Radiometric CKD cannot be skipped for this process
    if (ckd.rad_spectra.front().empty()) {
        throw std::runtime_error {
            "Cannot generate polarimetric CKD without the radiometric CKD "
            "being generated first"
        };
    }
    // Allocate polarimetric CKD variables, including the debug ones
    // even if not used.
    ckd.pol.m_q.resize(ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
    ckd.pol.m_u.resize(ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
    ckd.pol.m_t.resize(ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
    ckd.pol.eps.resize(ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
    ckd.pol.delta.resize(ckd.fov.n_act,
                         std::vector<double>(ckd.n_detector_cols));
    // Read prior data for retardance (delta)
    netCDF::NcFile nc_delta_prior { settings.pol.delta_prior_file,
                                    netCDF::NcFile::read };
    const int n_delta_wavelengths { static_cast<int>(
      nc_delta_prior.getDim("wavelengths").getSize()) };
    std::vector<double> wavelengths(n_delta_wavelengths);
    nc_delta_prior.getVar("wavelengths").getVar(wavelengths.data());
    std::vector<double> deltas(n_delta_wavelengths);
    nc_delta_prior.getVar("retardance").getVar(deltas.data());
    retardance_prior = make_unique<CubicSpline>(wavelengths, deltas);
    // Read prior data for telescope polarization (mt)
    netCDF::NcFile nc_mt_prior { settings.pol.mt_prior_file,
                                 netCDF::NcFile::read };
    const int n_mt_viewports { static_cast<int>(
      nc_mt_prior.getDim("viewports").getSize()) };
    const int n_mt_fovs_per_viewport { static_cast<int>(
      nc_mt_prior.getDim("fovs_per_viewport").getSize()) };
    const int n_mt_wavelengths { static_cast<int>(
      nc_mt_prior.getDim("wavelengths").getSize()) };
    // mt prior depends on the viewport and FOV
    telescope_prior.resize(n_mt_viewports);
    for (auto& el : telescope_prior) {
        el.resize(n_mt_fovs_per_viewport);
    }
    wavelengths.resize(n_mt_wavelengths);
    nc_mt_prior.getVar("wavelengths").getVar(wavelengths.data());
    std::vector<double> mt_prior(n_mt_viewports * n_mt_fovs_per_viewport
                                 * n_mt_wavelengths);
    // Retrieve all prior data from the nc file
    nc_mt_prior.getVar("mt").getVar(mt_prior.data());
    for (int i_vp {}; i_vp < n_mt_viewports; ++i_vp) {
        for (int i_fov {}; i_fov < n_mt_fovs_per_viewport; ++i_fov) {
            // Define a cubic spline for each segment (viewport and
            // field of view) separately
            const auto mt_prior_beg { mt_prior.cbegin()
                                      + i_vp * n_mt_fovs_per_viewport
                                          * n_mt_wavelengths
                                      + i_fov * n_mt_wavelengths };
            std::vector<double> mt_prior_segment(
              mt_prior_beg, mt_prior_beg + n_mt_wavelengths);
            // If mt is not fit for, still evaluate the spline but use zeros
            if (!settings.pol.fitting.mt_order) {
                std::ranges::fill(mt_prior_segment, 0.0);
            }
            telescope_prior.at(i_vp).at(i_fov) =
              make_unique<CubicSpline>(wavelengths, mt_prior_segment);
        }
    }
    spdlog::info("Calculation details:");
    if (settings.pol.mueller.enabled) {
        spdlog::info("  Mueller elements computed per wavelength");
        if (settings.pol.mueller.knots) {
            spdlog::info("  Each spectrum first smoothed with B-spline of "
                         "order {} using {} knots",
                         settings.pol.mueller.b_spline_order.value,
                         settings.pol.mueller.knots.value());
        }
    } else {
        spdlog::info("  fitting each spectrum with a model function");
    }
    if (settings.main.isrf_file.empty()
        || settings.main.solar_spectrum_file.empty()) {
        spdlog::warn("Either the ISRF or solar spectrum file missing");
        spdlog::warn("which means the solar spectrum correction (Fraunhofer)");
        spdlog::warn("factors are not computed");
    } else {
        isrf = std::make_unique<ISRF>(settings.main.isrf_file);
        std::vector<double> solar_wavelengths {};
        std::vector<double> solar_spectrum {};
        readSolarSpectrum(
          settings.main.solar_spectrum_file, solar_wavelengths, solar_spectrum);
        solar_spline =
          std::make_unique<CubicSpline>(solar_wavelengths, solar_spectrum);
    }
    // The solar correction factors are always allocated even if not
    // populated with real data.
    ckd.pol.m_q_solar.resize(ckd.fov.n_act,
                             std::vector<double>(ckd.n_detector_cols));
    ckd.pol.m_u_solar.resize(ckd.fov.n_act,
                             std::vector<double>(ckd.n_detector_cols));
}

auto PolGen::prepareBatches() -> void
{
    batchViewport();
}

// Perform a B-spline fit of signal and overwrite it with the result
static auto smoothSpectrum(const int spline_order,
                           const int n_knots,
                           const std::vector<double>& wavelengths,
                           std::vector<double>& signal,
                           const std::vector<double>& noise) -> void
{
    // The ends of the knot vector should be slightly outside the
    // wavelength range.
    std::vector<double> knots(n_knots);
    for (int i {}; i < n_knots; ++i) {
        knots[i] = wavelengths.front() - 1e-10
                   + i * (wavelengths.back() - wavelengths.front())
                       / (n_knots - 1 - 1e-10);
    }
    BSpline b_spline { spline_order, knots };
    // B-spline basis matrix in packed format
    PackedMatrix B_packed { static_cast<int>(wavelengths.size()),
                            b_spline.nStates() };
    for (int i {}; i < B_packed.nRows(); ++i) {
        b_spline.evalBasis(wavelengths[i], i, B_packed);
    }
    std::vector<double> state_vector {};
    auto B_packed_copy { B_packed };
    dppsv(signal, noise, B_packed_copy, state_vector);
    std::ranges::fill(signal, 0.0);
    for (int i {}; i < B_packed.nRows(); ++i) {
        for (int j {}; j < B_packed.nCols(i); ++j) {
            signal[i] +=
              B_packed(i, j) * state_vector[B_packed.globalCol(i, j)];
        }
    }
}

// Fit G = mq * q + mu * u to find the Mueller elements mq and mu. The
// modulation values G are given per AoLP and correspond to a fixed
// wavelength.
static auto muellerFit(const double dolp,
                       const std::vector<double>& aolp,
                       const std::vector<double>& modulation,
                       double& mq,
                       double& mu,
                       double& mt,
                       double& eps) -> void
{
    gadfit::LMsolver solver { [&dolp](
                                const std::vector<gadfit::AdVar>& parameters,
                                const double aolp) {
        using gadfit::AdVar;
        const AdVar& mq { parameters[0] };
        const AdVar& mu { parameters[1] };
        const double q { dolp * std::cos(2 * aolp) };
        const double u { dolp * std::sin(2 * aolp) };
        return mq * q + mu * u;
    } };
    solver.addDataset(aolp, modulation);
    // Use precomputed mq and mu as initial guesses
    solver.setPar(0, mq, true);
    solver.setPar(1, mu, true);
    solver.settings.iteration_limit = 100;
    solver.settings.verbosity = gadfit::io::hide_all;
    solver.fit();
    mq = solver.getParValue(0);
    mu = solver.getParValue(1);
    // Telescope polarization not considered for now
    mt = 0.0;
    eps = std::sqrt(mq * mq + mu * mu);
}

auto PolGen::processBatch(const int i_vp) -> void
{
    spdlog::info("Generating polarimetric CKD for viewport: {} deg",
                 viewport_idx_to_angle.at(
                   batches.at(i_vp).l1a_products.front().meta->viewport));
    // Number of images in a batch (viewport) is the number of AoLP values
    const int n_l1a { static_cast<int>(batches.at(i_vp).l1a_products.size()) };
    // Angles of polarization
    std::vector<double> aolp(n_l1a);
    // Wavelengths used in the fitting procedure. If
    // fit_window_min/max is not used it is the same as
    // ckd.wave_common for a given FOV.
    std::vector<double> wavelengths {};
    wavelengths.reserve(ckd.n_detector_cols);
    // Modulation spectra corresponding to different AoLP values
    std::vector<std::vector<double>> modulation(
      n_l1a, std::vector<double>(ckd.n_detector_cols, 0.0));
    // Modulation standard deviations
    std::vector<std::vector<double>> modulation_stddev(
      n_l1a, std::vector<double>(ckd.n_detector_cols, 0.0));
    // Do not yet extract spectra, only extract reference spectrum metadata
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        const auto& egse { batches[i_vp].l1a_products[i_l1a].meta->egse };
        std::vector<float> aolp_values(egse.size());
        for (int i {}; i < static_cast<int>(egse.size()); ++i) {
            switch (settings.pol.aolp_from.value()) {
            case EGSEMember::POLARIZER:
                aolp_values[i] = egse[i].POLARIZER;
                break;
            case EGSEMember::ACT_ANGLE:
                aolp_values[i] = egse[i].ACT_ANGLE;
                break;
            }
        }
        aolp[i_l1a] = median(aolp_values) * math::deg_to_rad
                      + settings.pol.aolp_offset.value();
        if (settings.pol.aolp_sign_flip.value()) {
            aolp[i_l1a] *= -1;
        }
    }
    spdlog::debug("AoLP values for this viewport are:");
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        spdlog::debug("  {} deg", aolp[i_l1a] / math::deg_to_rad);
    }
    // Fit spectra per L1B FOV
#pragma omp parallel for firstprivate(                                         \
  wavelengths, modulation, modulation_stddev)
    for (int i_fov = ckd.fov.start_vp[i_vp];
         i_fov < ckd.fov.start_vp[i_vp] + ckd.fov.n_act_vp[i_vp];
         ++i_fov) {
        if (const double p { getPercentage(i_fov - ckd.fov.start_vp[i_vp],
                                           ckd.fov.n_act_vp[i_vp]) };
            p >= 0.0) {
            spdlog::info("  Fitting per FOV: {:2.0f}%", p);
        }
        // Ranges of the fitting window
        int fit_idx_beg {};
        int fit_idx_end { ckd.n_detector_cols };
        if (settings.pol.fit_window_min) {
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                if (ckd.wave_common[i_fov][i]
                    < settings.pol.fit_window_min.value()) {
                    fit_idx_beg = i + 1;
                }
            }
        }
        if (settings.pol.fit_window_max) {
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                if (ckd.wave_common[i_fov][i]
                    < settings.pol.fit_window_max.value()) {
                    fit_idx_end = i + 1;
                }
            }
        }
        wavelengths.resize(fit_idx_end - fit_idx_beg);
        for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
            modulation[i_l1a].resize(wavelengths.size());
            modulation_stddev[i_l1a].resize(wavelengths.size());
            Spectra spectra {};
            batches[i_vp].l1a_products[i_l1a].extractSpectra(
              ckd, i_fov, spectra);
            if (settings.pol.mueller.enabled && settings.pol.mueller.knots) {
                for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                    smoothSpectrum(settings.pol.mueller.b_spline_order,
                                   settings.pol.mueller.knots.value(),
                                   ckd.wave_full[i_pol][i_fov],
                                   spectra.signal[i_pol],
                                   spectra.noise[i_pol]);
                }
            }
            for (int i { fit_idx_beg }; i < fit_idx_end; ++i) {
                const double& m { spectra.signal[pol_minus][i] };
                const double& p { spectra.signal[pol_plus][i] };
                modulation[i_l1a][i - fit_idx_beg] = (p - m) / (p + m);
                const double& ns_m { spectra.noise[pol_minus][i] };
                const double& ns_p { spectra.noise[pol_plus][i] };
                const double p_m { p + m };
                const double p_m2 { p_m * p_m };
                modulation_stddev[i_l1a][i - fit_idx_beg] =
                  std::sqrt(4 * (m * m * ns_p * ns_p + p * p * ns_m * ns_m)
                            / (p_m2 * p_m2));
            }
        }
        const auto& dolp {
            batches[i_vp].l1a_products.front().meta->dolp_spline
        };
        // Telescope polarization prior was defined for a number of
        // FOVs different from the actual number of FOVs. Simply take
        // the mt prior cubic spline that's closest to the current
        // FOV.
        const int i_fov_for_mt { static_cast<int>(std::round(
          static_cast<double>(i_fov - ckd.fov.start_vp[i_vp])
          * telescope_prior.at(i_vp).size() / ckd.fov.n_act_vp[i_vp])) };
        const auto& mt_prior_cur { telescope_prior.at(i_vp).at(i_fov_for_mt) };
        // Define the fitting function
        const int delta_order { settings.pol.fitting.delta_order };
        const int eps_order { settings.pol.fitting.eps_order };
        const int mt_order { settings.pol.fitting.mt_order.value_or(-1) };
        const auto G_fit { [this,
                            delta_order,
                            eps_order,
                            mt_order,
                            &dolp,
                            &mt_prior_cur](
                             const std::vector<gadfit::AdVar>& parameters,
                             const double lambda) {
            using gadfit::AdVar;
            // Number of wavelength orders required (lambda^1, ..., lambda^n)
            const int n_lambdas { std::max(delta_order, eps_order) };
            std::vector<double> lambdas(n_lambdas);
            lambdas.front() = lambda;
            for (int i { 1 }; i < n_lambdas; ++i) {
                lambdas[i] = lambdas[i - 1] * lambda;
            }
            // Decipher the parameter list that was passed to this
            // fitting function.
            int idx { -1 };
            // Angle of linear polarization (AoLP) is constant
            // (wavelength independent) per spectrum
            const AdVar& aolp { parameters[++idx] };
            // Retardance,
            //   delta = sum_{i=0}^Nr delta^i x lambda^i,
            // where Nr is the retardance order
            // [pol][fitting][delta_order]. The next Nr+1 values
            // (including a constant) from the parameter list are the
            // retardance orders. The order is at least 0, meaning we
            // are always fitting for at least a constant on top of
            // the prior.
            AdVar delta { retardance_prior->eval(lambda) + parameters[++idx] };
            for (int i { 1 }; i <= delta_order; ++i) {
                delta += parameters[++idx] * lambdas[i - 1];
            }
            // Efficiency is built up similarly as retardance
            AdVar eps { parameters[++idx] };
            for (int i { 1 }; i <= eps_order; ++i) {
                eps += parameters[++idx] * lambdas[i - 1];
            }
            // Telescope polarization Mueller element is built up
            // similarly as retardance and efficiency. If order is
            // less than zero mt is not included in the fit.
            AdVar mt { mt_prior_cur->eval(lambda), gadfit::passive_idx };
            if (mt_order >= 0) {
                mt += parameters[++idx];
            }
            for (int i { 1 }; i <= mt_order; ++i) {
                mt += parameters[++idx] * lambdas[i - 1];
            }
            // Now the parameter list has been processed. The rest is
            // about building up the polarimetric CKD model as
            // described in the ATBD. The q and u values are close to
            // constant but in general vary if the DoLP varies with
            // wavelength.
            const AdVar q { dolp->eval(lambda) * cos(2 * aolp) };
            const AdVar u { dolp->eval(lambda) * sin(2 * aolp) };
            const AdVar mq { eps * cos(2 * std::numbers::pi * delta / lambda) };
            const AdVar mu { eps * sin(2 * std::numbers::pi * delta / lambda) };
            const AdVar mtq { mt * mq };
            const AdVar xi { (1 - mtq) / (1 + mtq) };
            const AdVar Sp { xi * (1.0 + mtq + (mq + mt) * q + mu * u)
                             / (1.0 + mtq) };
            const AdVar Sm { (1.0 - mtq + (-mq + mt) * q - mu * u)
                             / (1.0 - mtq) };
            return (Sp - Sm) / (Sp + Sm);
        } };
        gadfit::LMsolver solver { G_fit };
        for (int i { fit_idx_beg }; i < fit_idx_end; ++i) {
            wavelengths[i - fit_idx_beg] = ckd.wave_common[i_fov][i];
        }
        for (int i_aolp {}; i_aolp < n_l1a; ++i_aolp) {
            solver.addDataset(
              wavelengths, modulation[i_aolp], modulation_stddev[i_aolp]);
        }
        // AoLP is fixed for each calibration measurement
        for (int i_aolp {}; i_aolp < n_l1a; ++i_aolp) {
            solver.setPar(0, aolp.at(i_aolp), false, i_aolp);
        }
        // All other model parameters are free to vary
        int idx {};
        for (int i {}; i <= settings.pol.fitting.delta_order; ++i) {
            solver.setPar(++idx, 0.0, true, "delta_" + std::to_string(i));
        }
        // Unlike retardance, the first estimate of efficiency must be non-zero
        solver.setPar(++idx, 1.0, true, "eps_0");
        for (int i { 1 }; i <= settings.pol.fitting.eps_order; ++i) {
            solver.setPar(++idx, 0.0, true, "eps_" + std::to_string(i));
        }
        for (int i {}; i <= mt_order; ++i) {
            solver.setPar(++idx, 0.0, true, "mt_" + std::to_string(i));
        }
        solver.settings.iteration_limit = 100;
        solver.settings.verbosity = gadfit::io::hide_all;
        solver.fit();
        // Fitting procedure complete, now retrieve all parameters. By
        // setting the index variable to 0 and using ++idx when
        // extracting the first fitting parameter value we are
        // skipping the (fixed) AoLP value.
        idx = 0;
        // Number of retardance orders to be extracted is order+1
        // because we include the constant value.
        std::vector<double> deltas(settings.pol.fitting.delta_order + 1);
        for (auto& delta : deltas) {
            delta = solver.getParValue(++idx);
        }
        std::vector<double> efficiencies(settings.pol.fitting.eps_order + 1);
        for (auto& eps : efficiencies) {
            eps = solver.getParValue(++idx);
        }
        std::vector<double> mts(mt_order + 1);
        for (auto& mt : mts) {
            mt = solver.getParValue(++idx);
        }
        // Construct the Mueller elements and other quantities to be
        // saved in the CKD. For meaning of lambdas, delta, eps, and
        // others, see the comments in G_fit.
        std::vector<double> lambdas(std::max(settings.pol.fitting.delta_order,
                                             settings.pol.fitting.eps_order));
        for (int i_spec {}; i_spec < ckd.n_detector_cols; ++i_spec) {
            lambdas.front() = ckd.wave_common[i_fov][i_spec];
            for (int i { 1 }; i < static_cast<int>(lambdas.size()); ++i) {
                lambdas[i] = lambdas[i - 1] * lambdas.front();
            }
            double delta { retardance_prior->eval(lambdas.front())
                           + (deltas.empty() ? 0.0 : deltas.front()) };
            for (int i { 1 }; i <= settings.pol.fitting.delta_order; ++i) {
                delta += deltas[i] * lambdas[i - 1];
            }
            double eps { efficiencies.empty() ? 0.0 : efficiencies.front() };
            for (int i { 1 }; i <= settings.pol.fitting.eps_order; ++i) {
                eps += efficiencies[i] * lambdas[i - 1];
            }
            double mt { mt_prior_cur->eval(lambdas.front()) };
            if (mt_order >= 0) {
                mt += mts.front();
            }
            for (int i { 1 }; i <= mt_order; ++i) {
                eps += mts[i] * lambdas[i - 1];
            }
            double mq { eps
                        * cos(2 * std::numbers::pi * delta / lambdas.front()) };
            double mu { eps
                        * sin(2 * std::numbers::pi * delta / lambdas.front()) };
            if (settings.pol.mueller.enabled) {
                // Extract modulation values per wavelength
                std::vector<double> modulation_wave(n_l1a);
                for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                    modulation_wave[i_l1a] = modulation[i_l1a][i_spec];
                }
                muellerFit(dolp->eval(ckd.wave_common[i_fov][i_spec]),
                           aolp,
                           modulation_wave,
                           mq,
                           mu,
                           mt,
                           eps);
            }
            ckd.pol.m_q[i_fov][i_spec] = mq;
            ckd.pol.m_u[i_fov][i_spec] = mu;
            ckd.pol.m_t[i_fov][i_spec] = mt;
            ckd.pol.eps[i_fov][i_spec] = eps;
            ckd.pol.delta[i_fov][i_spec] = delta;
            // Update the radiometric CKD
            // C+ = C0+ / (1+mq*mt)
            ckd.rad_spectra[pol_plus][i_fov][i_spec] /= 1.0 + mt * mq;
            // C- = C0- / (1-mq*mt)
            ckd.rad_spectra[pol_minus][i_fov][i_spec] /= 1.0 - mt * mq;
        }
        // Solar spectrum correction
        if (!(settings.main.isrf_file.empty()
              || settings.main.solar_spectrum_file.empty())) {
            for (int i_col {}; i_col < ckd.n_detector_cols; ++i_col) {
                // Convolved values at the current column (wavelength index)
                double mq_I_conv {};
                double mu_I_conv {};
                double mq_conv {};
                double mu_conv {};
                double I_conv {};
                for (int k_col {}; k_col < ckd.n_detector_cols; ++k_col) {
                    if (std::abs(i_col - k_col) > settings.pol.isrf_cutoff) {
                        continue;
                    }
                    const double delta_arg { 2 * std::numbers::pi
                                             * ckd.pol.delta[i_fov][k_col]
                                             / ckd.wave_common[i_fov][k_col] };
                    const double isrf_v { isrf->eval(
                      i_fov, i_col, i_col - k_col) };
                    const double sol_spectrum { solar_spline->eval(
                      ckd.wave_common[i_fov][k_col]) };
                    mq_I_conv += isrf_v * sol_spectrum * std::cos(delta_arg);
                    mu_I_conv += isrf_v * sol_spectrum * std::sin(delta_arg);
                    mq_conv += isrf_v * std::cos(delta_arg);
                    mu_conv += isrf_v * std::sin(delta_arg);
                    I_conv += isrf_v * sol_spectrum;
                }
                ckd.pol.m_q_solar[i_fov][i_col] = mq_I_conv / I_conv / mq_conv;
                ckd.pol.m_u_solar[i_fov][i_col] = mu_I_conv / I_conv / mu_conv;
            }
        }
        if (settings.pol.print_debug.enabled
            && i_vp == settings.pol.print_debug.viewport
            && i_fov - ckd.fov.start_vp[i_vp] == settings.pol.print_debug.fov) {
            std::ofstream out { settings.pol.print_debug.output_folder
                                + "/polcal_vp_" + std::to_string(i_vp) + "_fov_"
                                + std::to_string(i_fov - ckd.fov.start_vp[i_vp])
                                + ".txt" };
            // Pivoting array for sorting AoLP values
            std::vector<int> indices(n_l1a);
            std::iota(indices.begin(), indices.end(), 0);
            std::ranges::sort(indices, [&aolp](const int i, const int j) {
                return aolp.at(i) < aolp.at(j);
            });
            for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                const int i_aolp { indices.at(i_l1a) };
                out << "# AoLP = " << aolp.at(i_aolp) / math::deg_to_rad
                    << " deg\n";
                // In the Mueller approach print the S/P spectra and
                // the B-spline fits not the individual fits by
                // wavelength.
                if (settings.pol.mueller.enabled) {
                    out << "# Wavelength (nm), S spectrum, B-spline fit, P "
                           "spectrum, B-spline fit\n";
                    // It it easier to recalculate the spectra and the
                    // fits than store them from earlier.
                    Spectra spectra {};
                    batches[i_vp].l1a_products[i_aolp].extractSpectra(
                      ckd, i_fov, spectra);
                    Spectra spectra_spline { spectra };
                    if (settings.pol.mueller.knots) {
                        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                            smoothSpectrum(settings.pol.mueller.b_spline_order,
                                           settings.pol.mueller.knots.value(),
                                           ckd.wave_full[i_pol][i_fov],
                                           spectra_spline.signal[i_pol],
                                           spectra_spline.noise[i_pol]);
                        }
                    }
                    for (int i {}; i < ckd.n_detector_cols; ++i) {
                        out << ckd.wave_common[i_fov][i] << ' '
                            << spectra.signal[pol_minus][i] << ' '
                            << spectra_spline.signal[pol_minus][i] << ' '
                            << spectra.signal[pol_plus][i] << ' '
                            << spectra_spline.signal[pol_plus][i] << ' '
                            << '\n';
                    }
                } else {
                    // Otherwise print the modulation spectra and the model fit
                    out << "# Wavelength (nm), modulation, fit\n";
                    for (int i {}; i < fit_idx_end - fit_idx_beg; ++i) {
                        out << wavelengths[i] << ' ' << modulation[i_aolp][i]
                            << ' ' << solver.getValue(wavelengths[i], i_aolp)
                            << '\n';
                    }
                }
                out << "\n\n";
            }
        }
    }
}

} // namespace spex
