// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "processor.h"

#include "io.h"
#include "l1a_metadata.h"
#include "l1a_science.h"
#include "spex_settings.h"
#include "timer.h"

#include <omp.h>

namespace spex {

Processor::Processor(const SPEXSettings& settings,
                     const ProcLevel proc_level,
                     CKD& ckd)
  : settings { settings }, proc_level { proc_level }, ckd { ckd }
{
    // Do not change the CKD level if this is L1B or higher. If the
    // CKD is incomplete, e.g. the polarimetric section is missing,
    // then its level should stay at ProcLevel::pol. Increasing it
    // here will mislead L1B and higher processes thinking the CKD is
    // more mature than it actually is.
    if (proc_level < ProcLevel::l1b) {
        ckd.level = static_cast<ProcLevel>(static_cast<int>(proc_level) - 1);
    }
}

auto Processor::execute() -> void
{
    // STEP 1 - Welcome message

    switch (proc_level) {
    case ProcLevel::dark:
        printHeading("Dark CKD generation");
        break;
    case ProcLevel::noise:
        printHeading("Noise CKD generation");
        break;
    case ProcLevel::nonlin:
        printHeading("Nonlinearity CKD generation");
        break;
    case ProcLevel::prnu:
        printHeading("PRNU CKD generation");
        break;
    case ProcLevel::stray:
        printHeading("Stray light CKD generation");
        break;
    case ProcLevel::fov:
        printHeading("Field of view CKD generation");
        break;
    case ProcLevel::swath:
        printHeading("Swath vector CKD generation");
        break;
    case ProcLevel::wave:
        printHeading("Spectral CKD generation");
        break;
    case ProcLevel::rad:
        printHeading("Radiometric CKD generation");
        break;
    case ProcLevel::pol:
        printHeading("Polarimetric CKD generation");
        break;
    case ProcLevel::l1b:
        printHeading("L1B process");
        break;
    case ProcLevel::l1c:
        printHeading("L1C process");
        break;
    default:
        printHeading("Unknown process");
    }
    setLoggerLevel(settings.debug.at(static_cast<int>(proc_level)).value());

    // STEP 2 - Return or continue

    if (settings.skip.at(static_cast<int>(proc_level))) {
        spdlog::info("Skipping process");
        return;
    }

    // STEP 3 - Timers

    // Start main timer for this process
    Timer timer {};
    timer.start();
    // Set up the rest of the timers
    std::array<std::shared_ptr<Timer>, static_cast<int>(ProcLevel::n_levels)>
      timers {};
    for (auto& timer : timers) {
        timer = std::make_shared<Timer>();
    }

    // STEP 4 - Organize L1A files

    // Construct an L1A manager which is a container for all L1A
    // products. After this, the L1A products are available as
    // l1a_manager.l1a_products but only their metadata is read. The
    // image contents will be read later during calibration.
    switch (settings.l1a_type.at(static_cast<int>(proc_level))) {
    case L1AType::science:
        l1a_manager = std::make_unique<L1AScience>(false);
        break;
    case L1AType::diagnostic:
        l1a_manager = std::make_unique<L1ADiagnostic>(false);
        break;
    default:
        throw std::runtime_error { "unknown L1A type" };
    }
    l1a_manager->init(settings, proc_level, ckd);
    countL1A();
    for (auto& l1a_product : l1a_manager->l1a_products) {
        l1a_product.timers = timers;
    }

    // STEP 5 - Binning table

    // Initialize the binning table, associate it with each L1A
    // object, and bin the CKD if running the L1B process.
    std::shared_ptr<BinningTable> binning_table {};
    std::shared_ptr<BinningTable> binning_table_stray {};
    if (!l1a_manager->l1a_products.empty()) {
        binning_table = std::make_shared<BinningTable>(
          settings.main.detector_spat,
          settings.main.detector_spec,
          settings.main.binningtable,
          l1a_manager->l1a_products.front().binning_table_id,
          settings.main.spectral_binning);
        // For stray light correction, the binning table dimensions
        // are either the same as the detector dimension or reduced by
        // a factor of two.
        if (ckd.stray.n_kernels > 0
            && settings.van_cittert_steps.at(static_cast<int>(proc_level))
                 > 0) {
            const int reduction { ckd.stray.reduced_kernels ? 2 : 1 };
            binning_table_stray = std::make_shared<BinningTable>(
              settings.main.detector_spat / reduction,
              settings.main.detector_spec / reduction,
              settings.main.binningtable,
              l1a_manager->l1a_products.front().binning_table_id,
              settings.main.spectral_binning);
            if (ckd.stray.n_spatial * ckd.stray.n_spectral
                < binning_table->nBins()) {
                throw std::runtime_error {
                    "stray light binning table is inconsistent with the main "
                    "binning table (perhaps you are trying to use reduced "
                    "kernels with unbinned images)"
                };
            }
        }
    }
    for (auto& l1a_product : l1a_manager->l1a_products) {
        l1a_product.binning_table = binning_table;
        l1a_product.binning_table_stray = binning_table_stray;
    }
    if (proc_level == ProcLevel::l1b) {
        binning_table->binCKD(ckd);
    }

    // STEP 6 - Initialize L1X instances (if any)

    l1a_manager->initL1X(settings, proc_level, ckd);

    // STEP 7 - Perform initialization, if any, of the current process

    processInit();
    prepareBatches();

    // STEP 8 - Background images

    // Separate L1A manager for background images
    L1ADiagnostic l1a_manager_bgr { true };
    l1a_manager_bgr.init(settings, proc_level, ckd);
    // Reduce the CKD level temporarily to dark, i.e. nonlinearity
    // or higher should not be applied to background images.
    const ProcLevel ckd_level_save { ckd.level };
    ckd.level = std::min(ckd_level_save, ProcLevel::dark);
    for (auto& l1a_product : l1a_manager_bgr.l1a_products) {
        l1a_product.timers = timers;
        l1a_product.binning_table = binning_table;
    }
    prepareBackground(l1a_manager_bgr);
    ckd.level = ckd_level_save;

    // STEP 9 - Print a summary of what will be done in this process

    if (proc_level != ProcLevel::l1c) {
        spdlog::info("Total number of L1A files: {}", nL1A());
        spdlog::info("Processing steps:");
        // Flag indicating if any calibration steps are performed for
        // this process
        bool no_steps { true };
        const auto& l1x_level {
            batches.front().l1a_products.front().meta->l1x_level
        };
        for (int i_proc {}; i_proc < static_cast<int>(ProcLevel::n_levels);
             ++i_proc) {
            const auto proc { static_cast<ProcLevel>(i_proc) };
            if (batches.front().l1a_products.front().meta->l1x_level < proc
                && ckd.level >= proc && !settings.skip.at(i_proc)) {
                switch (proc) {
                case ProcLevel::dark:
                    spdlog::info("  Dark correction");
                    break;
                case ProcLevel::noise:
                    spdlog::info("  Noise estimation");
                    break;
                case ProcLevel::nonlin:
                    spdlog::info("  Nonlinearity correction");
                    break;
                case ProcLevel::prnu:
                    spdlog::info("  PRNU correction");
                    break;
                case ProcLevel::stray:
                    spdlog::info("  Stray light correction");
                    break;
                case ProcLevel::fov:
                    spdlog::info("  Spectra extraction (FOV)");
                    break;
                case ProcLevel::rad:
                    spdlog::info("  Radiometric correction");
                    break;
                case ProcLevel::pol:
                    spdlog::info("  Demodulation");
                    break;
                default:
                    continue;
                }
                no_steps = false;
            }
        }
        if (batches.front().l1a_products.front().bgr_image) {
            spdlog::info("  Subtract background image");
            no_steps = false;
        }
        if (settings.dark_cols_left.at(static_cast<int>(proc_level)) > 0
            || settings.dark_cols_right.at(static_cast<int>(proc_level)) > 0) {
            spdlog::info("  Row-dependent offset");
            no_steps = false;
        }
        if (no_steps) {
            spdlog::info("  None");
        }
    }

    // STEP 10 - Perform the main loop over L1A batches

    // Run the core algorithm of the current process. Parallelization
    // might or might not be over batches because each process can set
    // its own value for n_threads_batches.
#pragma omp parallel for num_threads(n_threads_batches)
    for (int i_batch = 0; i_batch < static_cast<int>(batches.size());
         ++i_batch) {
        // Viewports may be skipped with [main][target_viewport]
        if (const auto viewport { static_cast<int>(
              batches.at(i_batch).l1a_products.front().meta->viewport) };
            settings.main.target_viewport_i && viewport > 0
            && settings.main.target_viewport_i.value() != viewport) {
            continue;
        }
        if (proc_level == ProcLevel::stray && settings.stray.checkpoint.read
            || proc_level == ProcLevel::swath
                 && settings.swath.checkpoint.read) {
            continue;
        }
        // Read L1A images (large data load) and calibrate them
        batches.at(i_batch).calibrate(settings, proc_level, ckd);
        // Execute the processor per batch. Batch number may indicate
        // the phase of the process.
        processBatch(i_batch);
        // Deallocate used images
        batches.at(i_batch).l1a_products = std::vector<L1A> {};
    }

    // STEP 11 - Do postprocessing, if any, for this process

    processFinalize();

    // STEP 12 - Print timings for processing step applied in this process

    timer.stop();
    spdlog::info("Timings:");
    spdlog::info("         Dark correction: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::dark))->time(),
                 timers.at(static_cast<int>(ProcLevel::dark))->calls());
    spdlog::info("        Noise estimation: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::noise))->time(),
                 timers.at(static_cast<int>(ProcLevel::noise))->calls());
    spdlog::info(" Nonlinearity correction: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::nonlin))->time(),
                 timers.at(static_cast<int>(ProcLevel::nonlin))->calls());
    spdlog::info("         PRNU correction: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::prnu))->time(),
                 timers.at(static_cast<int>(ProcLevel::prnu))->calls());
    spdlog::info("  Stray light correction: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::stray))->time(),
                 timers.at(static_cast<int>(ProcLevel::stray))->calls());
    spdlog::info("Spectra extraction (FOV): {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::fov))->time(),
                 timers.at(static_cast<int>(ProcLevel::fov))->calls());
    spdlog::info(" Radiometric calibration: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::rad))->time(),
                 timers.at(static_cast<int>(ProcLevel::rad))->calls());
    spdlog::info("            Demodulation: {:8.3f} s ({} calls)",
                 timers.at(static_cast<int>(ProcLevel::pol))->time(),
                 timers.at(static_cast<int>(ProcLevel::pol))->calls());
    spdlog::info("  Total time for process: {:8.3f} s", timer.time());
}

auto Processor::prepareBatches() -> void
{
    batches.resize(1);
    batches.front().l1a_products = std::move(l1a_manager->l1a_products);
    for (L1A& l1a : batches.front().l1a_products) {
        l1a.meta->nc->reOpen();
    }
}

auto Processor::prepareBackground(L1ADiagnostic& l1a_manager_bgr) -> void
{
    // Put all background images into one batch and calibrate them in
    // parallel. This is mostly to save memory.
    L1ABatch batch_bgr {};
    batch_bgr.l1a_products = std::move(l1a_manager_bgr.l1a_products);
    batch_bgr.calibrate(settings, proc_level, ckd);
    for (auto& l1a_bgr : batch_bgr.l1a_products) {
        auto bgr_image { std::make_shared<std::vector<double>>() };
        *bgr_image = l1a_bgr.image;
        for (auto& batch : batches) {
            for (auto& l1a : batch.l1a_products) {
                if (l1a_bgr.meta->viewport == l1a.meta->viewport
                    || l1a_bgr.meta->viewport == 0) {
                    l1a.bgr_image = bgr_image;
                }
            }
        }
    }
}

auto Processor::batchViewport() -> void
{
    std::ranges::sort(l1a_manager->l1a_products,
                      [](const auto& a, const auto& b) {
                          return a.meta->viewport < b.meta->viewport;
                      });
    batches.resize(ckd.n_vp);
    for (auto batch { batches.begin() };
         const auto& [vp_idx, angle] : viewport_idx_to_angle) {
        // Find the first L1A product that matches the given viewport
        auto l1a_beg { std::ranges::find_if(
          l1a_manager->l1a_products, [vp_idx = vp_idx](const auto& l1a) {
              // First test if L1A has its metadata. It might have
              // been reset with a previous call to moveL1ASlice.
              return l1a.meta && l1a.meta->viewport == vp_idx;
          }) };
        // Find last L1A product with the same viewport as l1a_beg
        auto l1a_end { std::ranges::find_if(l1a_beg,
                                            l1a_manager->l1a_products.cend(),
                                            [vp_idx = vp_idx](const auto& x) {
                                                return x.meta->viewport
                                                       != vp_idx;
                                            }) };
        // Assign this range of L1A products to the current batch
        moveL1ASlice(l1a_beg, l1a_end, batch->l1a_products);
        ++batch;
    }
    l1a_manager->l1a_products = {};
}

auto Processor::batchOne() -> void
{
    batches.resize(nL1A());
    for (int i {}; i < nL1A(); ++i) {
        moveL1ASlice(l1a_manager->l1a_products.begin() + i,
                     l1a_manager->l1a_products.begin() + i + 1,
                     batches.at(i).l1a_products);
    }
    l1a_manager->l1a_products = std::vector<L1A> {};
    // Parallelization is over batches where one batch is one image
    n_threads_batches = omp_get_max_threads();
    for (L1ABatch& batch : batches) {
        if (batch.l1a_products.front().frames_u16.empty()) {
            batch.l1a_products.front().meta->nc->reOpen();
        }
    }
}

Processor::~Processor()
{
    // Same comment as in the constructor. The CKD level should only
    // be increased here if we are generating the CKD.
    if (proc_level < ProcLevel::l1b) {
        ckd.level = static_cast<ProcLevel>(static_cast<int>(ckd.level) + 1);
    }
}

} // namespace spex
