// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "quaternion.h"

#include "algorithm.h"

namespace spex {

Quaternion::Quaternion(const std::array<double, dims::vec * dims::vec>& rot)
{
    if (const double trace { rot[0 * dims::vec + 0] + rot[1 * dims::vec + 1]
                             + rot[2 * dims::vec + 2] };
        trace >= 0.0) {
        const double s { 0.5 / std::sqrt(trace + 1.0) };
        v = { (rot[2 * dims::vec + 1] - rot[1 * dims::vec + 2]) * s,
              (rot[0 * dims::vec + 2] - rot[2 * dims::vec + 0]) * s,
              (rot[1 * dims::vec + 0] - rot[0 * dims::vec + 1]) * s };
        r = 0.25 / s;
    } else {
        int i {};
        if (rot[1 * dims::vec + 1] > rot[0 * dims::vec + 0]) {
            i = 1;
        }
        if (rot[1 * dims::vec + 1] > rot[i * dims::vec + i]) {
            i = 2;
        }
        const int j { (i + 1) % dims::vec };
        const int k { (j + 1) % dims::vec };
        double s { std::sqrt(rot[i * dims::vec + i] - rot[j * dims::vec + j]
                             - rot[k * dims::vec + k] + 1.0) };
        v[i] = 0.5 * s;
        s = 0.5 / s;
        r = (rot[k * dims::vec + j] - rot[j * dims::vec + k]) * s;
        v[j] = (rot[j * dims::vec + i] + rot[i * dims::vec + j]) * s;
        v[k] = (rot[k * dims::vec + i] + rot[i * dims::vec + k]) * s;
    }
}

auto Quaternion::normalize() -> void
{
    const double norm_inv { 1 / std::sqrt(r * r + dotProduct(v, v)) };
    for (double& val : v) {
        val *= norm_inv;
    }
    r *= norm_inv;
}

auto Quaternion::conjugate() -> void
{
    for (double& el : v) {
        el = -el;
    }
}

auto Quaternion::rotate(std::array<double, dims::vec>& a) const -> void
{
    const double vv { dotProduct(v, v) };
    const double av { dotProduct(a, v) };
    std::array<double, dims::vec> va {};
    crossProduct(v, a, va);
    for (int i {}; i < dims::vec; ++i) {
        a[i] = (r * r - vv) * a[i] + 2.0 * av * v[i] + 2.0 * r * va[i];
    }
}

auto Quaternion::multiply(const Quaternion& rhs) -> void
{
    double r_new { r * rhs.r - dotProduct(v, rhs.v) };
    std::array<double, dims::vec> cross {};
    crossProduct(v, rhs.v, cross);
    for (int i {}; i < dims::vec; ++i) {
        v[i] = r * rhs.v[i] + rhs.r * v[i] + cross[i];
    }
    r = std::move(r_new);
}

auto Quaternion::slerp(const Quaternion& q, const double t) -> void
{
    const double theta { std::acos(dotProduct(v, q.v) + r * q.r) };
    const double f1 { std::sin((1 - t) * theta) / sin(theta) };
    const double f2 { std::sin(t * theta) / sin(theta) };
    for (int i {}; i < dims::vec; ++i) {
        v[i] = f1 * v[i] + f2 * q.v[i];
    }
    r = f1 * r + f2 * q.r;
}

Vector::Vector()
{
    v.resize(dims::vec);
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        v[ivec] = 0.0;
}
Vector::Vector(double* input)
{
    v.resize(dims::vec);
    memcpy(v.data(), input, dims::vec * sizeof(double));
}
Vector::Vector(double x, double y, double z)
{
    v.resize(dims::vec);
    v[0] = x;
    v[1] = y;
    v[2] = z;
}
// Elements
double& Vector::operator[](int idx)
{
    return v[idx];
}
// Multiplication (vector and scalar, commutative)
Vector Vector::operator*(double rhs)
{
    Vector res = *this;
    res *= rhs;
    return res;
}
Vector operator*(double lhs, Vector rhs)
{
    return rhs * lhs;
}
Vector& Vector::operator*=(double rhs)
{
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        v[ivec] *= rhs;
    return *this;
}
// Division (vector divided by scalar)
Vector Vector::operator/(double rhs)
{
    Vector res = *this;
    res /= rhs;
    return res;
}
Vector& Vector::operator/=(double rhs)
{
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        v[ivec] /= rhs;
    return *this;
}
// Addition, subtraction and unary minus sign
Vector& Vector::operator+=(Vector rhs)
{
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        v[ivec] += rhs.v[ivec];
    return *this;
}
Vector Vector::operator+(Vector rhs)
{
    Vector res = *this;
    res += rhs;
    return res;
}
Vector Vector::operator-=(Vector rhs)
{
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        v[ivec] -= rhs.v[ivec];
    return *this;
}
Vector Vector::operator-(Vector rhs)
{
    Vector res = *this;
    res -= rhs;
    return res;
}
Vector Vector::operator-()
{
    Vector res = *this;
    res *= -1.0;
    return res;
}
// Dot and cross product
double Vector::dot(Vector other)
{
    double res = 0.0;
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        res += v[ivec] * other[ivec];
    return res;
}
Vector Vector::cross(Vector other)
{
    Vector res {};
    for (size_t ivec = 0; ivec < dims::vec; ivec++)
        res[ivec] = v[(ivec + 1) % dims::vec] * other[(ivec + 2) % dims::vec]
                    - v[(ivec + 2) % dims::vec] * other[(ivec + 1) % dims::vec];
    return res;
}
// Normalization
void Vector::normalize()
{
    double mag = sqrt(std::pow(v[0], 2.0) + pow(v[1], 2.0) + pow(v[2], 2.0));
    if (mag != 0.0)
        this->operator/=(mag);
}
// Transform to double array
void Vector::get(double* res)
{
    memcpy(res, v.data(), dims::vec * sizeof(double));
}

// Constructors
Quaternion0::Quaternion0()
{
    v = Vector();
    r = 0.0;
}
Quaternion0::Quaternion0(double* input)
{
    v = Vector(input);
    r = input[dims::vec];
}
Quaternion0::Quaternion0(Vector v, double r)
{
    this->v = Vector(v);
    this->r = r;
}
Quaternion0::Quaternion0(double x, double y, double z, double r)
{
    v = Vector(x, y, z);
    this->r = r;
}
// Elements
double& Quaternion0::operator[](int idx)
{
    return idx < (int)dims::vec ? v[idx] : r;
}
// Multiplication (quaternion and scalar, commutative)
Quaternion0 Quaternion0::operator*(double rhs)
{
    Quaternion0 res = *this;
    res *= rhs;
    return res;
}
Quaternion0 operator*(double lhs, Quaternion0 rhs)
{
    return rhs * lhs;
}
Quaternion0& Quaternion0::operator*=(double rhs)
{
    v *= rhs;
    r *= rhs;
    return *this;
}
// Division (quaternion divided by scalar)
Quaternion0 Quaternion0::operator/(double rhs)
{
    Quaternion0 res = *this;
    res /= rhs;
    return res;
}
Quaternion0& Quaternion0::operator/=(double rhs)
{
    v /= rhs;
    r /= rhs;
    return *this;
}
// Addition, subtraction and unary minus sign
Quaternion0& Quaternion0::operator+=(Quaternion0 rhs)
{
    v += rhs.v;
    r += rhs.r;
    return *this;
}
Quaternion0 Quaternion0::operator+(Quaternion0 rhs)
{
    Quaternion0 res = *this;
    res += rhs;
    return res;
}
Quaternion0 Quaternion0::operator-=(Quaternion0 rhs)
{
    v -= rhs.v;
    r -= rhs.r;
    return *this;
}
Quaternion0 Quaternion0::operator-(Quaternion0 rhs)
{
    Quaternion0 res = *this;
    res -= rhs;
    return res;
}
Quaternion0 Quaternion0::operator-()
{
    Quaternion0 res = *this;
    res *= -1.0;
    return res;
}
// Conjugate
Quaternion0 Quaternion0::conjug()
{
    return Quaternion0(-this->v, this->r);
}
// Rotation
Vector Quaternion0::quaternion_rotate(Vector orig)
{
    return (r * r - v.dot(v)) * orig + 2.0 * r * v.cross(orig)
           + 2.0 * v * v.dot(orig);
}
// Quaternion0 multiplication
Quaternion0 Quaternion0::operator*(Quaternion0 rhs)
{
    return Quaternion0(r * rhs.v + rhs.r * v + v.cross(rhs.v),
                       r * rhs.r - v.dot(rhs.v));
}
// Quaternion0 division
Quaternion0 Quaternion0::operator/(Quaternion0 rhs)
{
    return (*this) * rhs.conjug() / (pow(rhs.r, 2.0) + rhs.v.dot(rhs.v));
}
// Raising a power
Quaternion0 Quaternion0::power(double p)
{
    double norm = sqrt(pow(r, 2.0) + v.dot(v));
    double c = r / norm;
    Vector dir = v / norm;
    double s = sqrt(dir.dot(dir));
    dir /= s;
    double mag = pow(norm, p);
    double ang = atan2(s, c) * p;
    return mag * Quaternion0(dir * sin(ang), cos(ang));
}
// Normalization
void Quaternion0::normalize()
{
    double mag = sqrt(pow(r, 2.0) + v.dot(v));
    *this /= mag;
}
// Transform to double array
void Quaternion0::get(double* res)
{
    v.get(res);
    res[3] = r;
}

} // namespace spex
