// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Quaternion algebra

#pragma once

#include "constants.h"

#include <array>
#include <vector>

namespace spex {

class Quaternion
{
private:
    // Vector part
    std::array<double, dims::vec> v {};
    // Real part
    double r {};

public:
    Quaternion() = default;
    Quaternion(const double x, const double y, const double z, const double r)
      : v { x, y, z }, r { r }
    {}
    // Initialize from a rotation matrix using the algorithm of
    // "Quaternion Calculus and Fast Animation", Ken Shoemake, 1987 to
    // convert the rotation matrix to a quaternion. Also, see the
    // implementation in Eigen: src/Geometry/Quaternion.h .
    Quaternion(const std::array<double, dims::vec * dims::vec>& rot);

    auto operator[](const int i) const -> const double& { return v[i]; }
    auto operator[](const int i) -> double& { return v[i]; }
    auto real() const -> const double& { return r; }
    auto real() -> double& { return r; }

    auto normalize() -> void;
    // Reverse the vector component
    auto conjugate() -> void;
    // Apply Q a Q^-1 to vector 'a' where Q is this quaternion. The
    // formula is Q a Q^-1 = (r^2 - v^2)a + 2(av)v + 2r(v x a).
    auto rotate(std::array<double, dims::vec>& a) const -> void;
    // Multiply this quaternion with rhs from the right and store the
    // result in self
    auto multiply(const Quaternion& rhs) -> void;
    // Spherical linear interpolation (Ken Shoemake "Animating
    // rotation with quaternion curves", SIGGRAPH Computer Graphics,
    // 19:245 (1985).
    auto slerp(const Quaternion& q, const double t) -> void;
    ~Quaternion() = default;
};

struct Vector
{
public:
    // Member
    std::vector<double> v;

public:
    // Constructors
    Vector();
    Vector(double* input);
    Vector(double x, double y, double z);
    // Elements
    double& operator[](int idx);
    // Multiplication (vector and scalar, commutative)
    Vector operator*(double rhs);
    friend Vector operator*(double lhs, Vector rhs);
    Vector& operator*=(double rhs);
    // Division (vector divided by scalar)
    Vector operator/(double rhs);
    Vector& operator/=(double rhs);
    // Addition, subtraction and unary minus sign
    Vector& operator+=(Vector rhs);
    Vector operator+(Vector rhs);
    Vector operator-=(Vector rhs);
    Vector operator-(Vector rhs);
    Vector operator-();
    // Dot and cross product
    double dot(Vector other);
    Vector cross(Vector other);
    // Normalization
    void normalize();
    // Transform to double array
    void get(double* res);
};
struct Quaternion0
{
public:
    // Members
    double r {};
    Vector v {};

public:
    // Constructors
    Quaternion0();
    Quaternion0(double* input);
    Quaternion0(Vector v, double r);
    Quaternion0(double x, double y, double z, double r);
    // Elements
    double& operator[](int idx);
    // Multiplication (quaternion and scalar, commutative)
    Quaternion0 operator*(double rhs);
    friend Quaternion0 operator*(double lhs, Quaternion0 rhs);
    Quaternion0& operator*=(double rhs);
    // Division (quaternion divided by scalar)
    Quaternion0 operator/(double rhs);
    Quaternion0& operator/=(double rhs);
    // Addition, subtraction and unary minus sign
    Quaternion0& operator+=(Quaternion0 rhs);
    Quaternion0 operator+(Quaternion0 rhs);
    Quaternion0 operator-=(Quaternion0 rhs);
    Quaternion0 operator-(Quaternion0 rhs);
    Quaternion0 operator-();
    // Conjugate
    Quaternion0 conjug();
    // Rotation
    Vector quaternion_rotate(Vector orig);
    // Quaternion0 multiplication
    Quaternion0 operator*(Quaternion0 rhs);
    // Quaternion0 division
    Quaternion0 operator/(Quaternion0 rhs);
    // Raising a power
    Quaternion0 power(double p);
    // Normalization
    void normalize();
    // Transform to double array
    void get(double* res);
};

} // namespace spex
