// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "rad_gen.h"

#include "l1a_metadata.h"
#include "linear_spline.h"

namespace spex {

auto RadGen::processInit() -> void
{
    for (auto& rad : ckd.rad_spectra) {
        rad.resize(ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
    }
}

auto RadGen::prepareBatches() -> void
{
    batchViewport();
}

// NetCDF compound type for the reference diode
struct ref_dtype
{
    double Unix_Time {};
    float Excel_Time {};
    double Normalized_Time {};
    float amps {};
    float scaled {};
    double last_zero {};
    uint8_t lamp_on {};
    float voltage {};
    float current {};
};

auto RadGen::processBatch(const int i_vp) -> void
{
    spdlog::info("Generating radiometric CKD for viewport: {} deg",
                 viewport_idx_to_angle.at(
                   batches.at(i_vp).l1a_products.front().meta->viewport));
    const auto& l1a { batches.at(i_vp).l1a_products.front() };
    // Extract reference radiance corresponding to this viewport.
    const netCDF::NcFile nc { l1a.meta->filename, netCDF::NcFile::read };
    const auto reference_spectrum {
        nc.getGroup("gse_data").getGroup("ReferenceSpectrum")
    };
    const int n_wavelengths { static_cast<int>(
      reference_spectrum.getDim("wavelength").getSize()) };
    std::vector<double> reference_wavelengths(n_wavelengths);
    reference_spectrum.getVar("wavelength")
      .getVar(reference_wavelengths.data());
    std::vector<double> reference_radiance(n_wavelengths);
    reference_spectrum.getVar("spectral_radiance")
      .getVar(reference_radiance.data());
    // The reference radiance needs to be corrected by a scaling
    // factor. First retrieve the median current of the reference
    // diode in amps.
    const auto reference_diode {
        nc.getGroup("gse_data").getGroup("ReferenceDiode")
    };
    const int n_times { static_cast<int>(
      reference_diode.getDim("time").getSize()) };
    std::vector<float> diode_amps(n_times);
    reference_diode.getVar("amps").getVar(diode_amps.data());
    const double amps_median { median(diode_amps) };
    // Next retrieve the detector open current value. It's given as a
    // group attribute so need to do basic string manipulation to
    // extract the value.
    const auto detector_open { reference_spectrum.getAtt("detector.Open") };
    std::string detector_open_data(detector_open.getAttLength(), ' ');
    detector_open.getValues(detector_open_data.data());
    // Discard the unit (A) and convert the rest of the string to a
    // double.
    const double detector_open_value { std::atof(
      detector_open_data.substr(0, detector_open_data.find(' ')).data()) };
    // The scaling factor is the ratio of reference diode amps to the
    // open current. The factor 1e3 is for converting
    // spectral_radiance units from W/(m^2.sr.nm) to W/(m^2.sr.um).
    const double scaling_factor { 1e3 * amps_median / detector_open_value };
    // Finally, correct the reference radiance with the scaling factor
    for (auto& val : reference_radiance) {
        val *= scaling_factor;
    }
    // Because the reference and measurement wavelength grids are
    // different we interpolate the reference spectrum.
    const LinearSpline ref_spline { reference_wavelengths, reference_radiance };
    for (int i_fov { ckd.fov.start_vp.at(i_vp) };
         i_fov < ckd.fov.start_vp.at(i_vp) + ckd.fov.n_act_vp.at(i_vp);
         ++i_fov) {
        Spectra spectra {};
        l1a.extractSpectra(ckd, i_fov, spectra);
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            // The intended radiance is half the actual radiance
            // because this is unpolarized light in which case S+ and
            // S- are half the radiance.
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                ckd.rad_spectra[i_pol][i_fov][i] =
                  0.5 * ref_spline.eval(ckd.wave_full[i_pol][i_fov][i])
                  * l1a.exposure_time / spectra.signal[i_pol][i];
                // Filter out zeros, NaNs, and infinites
                if (!std::isnormal(ckd.rad_spectra[i_pol][i_fov][i])) {
                    ckd.rad_spectra[i_pol][i_fov][i] = NC_FILL_DOUBLE;
                }
            }
        }
    }
}

} // namespace spex
