// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "settings.h"

#include <iostream>

namespace spex {

auto Settings::init() -> void
{
    scanKeys();
    unrecognizedKeywordCheck();
    checkParameters();
}

// Extract the key of each node as a list of strings.
// NOLINTNEXTLINE(misc-no-recursion)
static auto extractYAMLKeys(const YAML::Node& node,
                            std::vector<std::vector<std::string>>& keys,
                            std::vector<std::string>& cur_key) -> void
{
    if (node.IsMap()) {
        for (YAML::const_iterator it { node.begin() }; it != node.end(); ++it) {
            cur_key.push_back(it->first.as<std::string>());
            extractYAMLKeys(node[it->first], keys, cur_key);
            cur_key.pop_back();
        }
    } else {
        // Anything that is not a map means that we reached the end of
        // the tree of a configuration parameter (node). Store the
        // current key in the list and continue the search for the
        // next key.
        keys.push_back(cur_key);
    }
}

auto Settings::unrecognizedKeywordCheck() const -> void
{
    std::vector<std::vector<std::string>> all_keys {};
    std::vector<std::string> cur_key {};
    extractYAMLKeys(YAML::Clone(config), all_keys, cur_key);
    // Test all keys from the input file against all valid keys.
    for (const auto& key : all_keys) {
        bool found { false };
        for (const auto& valid_key : all_valid_keys) {
            if (key == valid_key) {
                found = true;
                break;
            }
        }
        if (!found) {
            std::cerr << "Warning: unrecognized input parameter: "
                      << Setting<bool> { key, false, "" }.keyToStr() << '\n';
        }
    }
}

auto Settings::c_str() -> const char*
{
    do_dump = true;
    yaml_emitter.SetBoolFormat(YAML::YesNoBool);
    yaml_emitter.SetNullFormat(YAML::LowerNull);
    // Everything will be stored in a global map.
    yaml_emitter << YAML::BeginMap;
    // First print the non-verbose configuration. This can be used
    // directly as input for the processor.
    scanKeys();
    // Make sure all maps are properly closed.
    for (int i {}; i < static_cast<int>(cur_map_loc.size()); ++i) {
        yaml_emitter << YAML::EndMap;
    }
    // Next print the verbose version. This differs from the format of
    // the processor input but contains additional information about
    // each configuration parameter.
    yaml_emitter.verbose = true;
    cur_map_loc = std::vector<std::string> {};
    yaml_emitter << YAML::Key << "verbose";
    yaml_emitter << YAML::BeginMap;
    scanKeys();
    // We don't bother to close any maps here
    return yaml_emitter.c_str();
}

} // namespace spex
