// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "solar_model.h"

#include <cmath>
#include <stdexcept>

namespace spex {

auto solarModel(const uint32_t TAI_seconds,
                const double TAI_second_fraction,
                std::array<double, dims::vec>& sun,
                Quaternion& q_ecr) -> void
{
    // UT1 time in modified Julian day 2000 (MJD2000) format
    const auto [day_fract,
                t] { daysSinceJ2000(TAI_seconds, TAI_second_fraction) };

    // Compute solar ephemeris parameters. These are used to compute
    // the solar longitude and the nutation in longitude and
    // obliquity.
    const auto [sun_mean_lon, sma] { solarEphemeris(t) };
    const auto [moon_mean_lon, moon_asc_node] { lunarEphemeris(t) };

    // Compute nutation corrections for this day
    const auto [nutation_lon, mean_obliquity, nutation_obl] { nutate(
      t, sun_mean_lon, sma, moon_mean_lon, moon_asc_node) };

    // Greenwich hour angle in radians
    const double gha { getGHA2000(t, day_fract, nutation_lon, nutation_obl) };

    // Transform Greenwich hour angle to rotation Quaternion
    const Quaternion q_gha {
        0.0, 0.0, -std::sin(gha / 2.0), std::cos(gha / 2.0)
    };

    // Transform the sun and the satellite into ECR. The nutation and
    // such is already done. Only the Greenwich hour angle has to be
    // applied.
    getSun2000(
      t, sun_mean_lon, sma, moon_mean_lon, nutation_lon, nutation_obl, sun);
    q_gha.rotate(sun);

    // J2000 to ECR rotation of the Sun
    Quaternion q_j2000_to_mod {};
    j2000ToMOD(t, q_j2000_to_mod);

    // Nutation quaternion
    Quaternion q_nutation {};
    getNutationQuaternion(
      nutation_lon, nutation_obl, mean_obliquity, q_nutation);
    q_nutation.conjugate();

    // We can now compute the ECR transformation for the satellite
    // which is a combination of the Greenwich hour angle rotation,
    // nutation rotation, and the J2000 to MOD rotation.
    q_ecr = { q_gha };
    q_ecr.multiply(q_nutation);
    q_ecr.multiply(q_j2000_to_mod);
}

[[nodiscard]] auto julianDayNumber(const int year,
                                   const int month,
                                   const int day) -> int
{
    const int a { (month - 14) / 12 };
    const int y { year + 4800 + a };
    return (1461 * y) / 4 + 367 * (month - 2 - 12 * a) / 12
           - (3 * ((y + 100) / 100)) / 4 + day - 32075;
}

[[nodiscard]] auto deltaTerrestrialUT1(const double y) -> double
{
    // https://eclipse.gsfc.nasa.gov/LEcat5/time.html
    if (y < 1961) {
        throw std::runtime_error { "years before 1961 not supported" };
    }
    if (y < 1986) {
        const double t { y - 1975.0 };
        const double t2 { t * t };
        const double t3 { t2 * t };
        return 45.45 + 1.067 * t - t2 / 260 - t3 / 718;
    }
    if (y < 2005) {
        const double t { y - 2000.0 };
        const double t2 { t * t };
        const double t3 { t2 * t };
        const double t4 { t3 * t };
        const double t5 { t4 * t };
        return 63.86 + 0.3345 * t - 0.060374 * t2 + 0.0017275 * t3
               + 0.000651814 * t4 + 0.00002373599 * t5;
    }
    if (y < 2050.0) {
        const double t { y - 2000.0 };
        const double t2 { t * t };
        return 62.92 + 0.32217 * t + 0.005589 * t2;
    }
    if (y < 2150) {
        const double d { ((y - 1820) / 100) };
        return -20 + 32 * d * d - 0.5628 * (2150 - y);
    }
    throw std::runtime_error { "years after 2150 not supported" };
}

[[nodiscard]] auto daysSinceJ2000(const uint32_t icu_seconds,
                                  const double second_fraction)
  -> std::array<double, 2>
{
    // Current Julian day number. The ICU time stamp is given relative
    // to 1 January 1958.
    const int days { static_cast<int>(icu_seconds / 86400) };
    const int julian_day_number { julianDayNumber(1958, 1, 1) + days };
    // Difference between terrestrial time (TT) and international
    // atomic time (TAI): TT = TAI + 32 s.
    constexpr double TAI_TT_diff { 32.184 };
    // Number of years since 1958
    const double years { 1958 + icu_seconds / 86400 / 365.25 };
    // Difference between TAI and UT1
    const double delta_UT { deltaTerrestrialUT1(years) - TAI_TT_diff };
    // Day fraction in the UT1 time scale
    const double day_fract_UT1 {
        (icu_seconds % 86400 + second_fraction - delta_UT) / 86400.0
    };
    const double julian_day { julian_day_number - 0.5 + day_fract_UT1 };
    return { day_fract_UT1, julian_day - 2451544.5 };
}

[[nodiscard]] auto solarEphemeris(const double t) -> std::array<double, 2>
{
    return { // Sun mean longitude
             fmod(280.46592 + 0.9856473516 * t, 360.0) * math::deg_to_rad,
             // Sun mean anomaly
             fmod(357.52772 + 0.9856002831 * t, 360.0) * math::deg_to_rad
    };
}

[[nodiscard]] auto lunarEphemeris(const double t) -> std::array<double, 2>
{
    return { // Moon mean longitude
             fmod(218.31643 + 13.17639648 * t, 360.0) * math::deg_to_rad,
             // Ascending node of Moon's mean orbit
             fmod(125.04452 - 0.0529537648 * t, 360.0) * math::deg_to_rad
    };
}

[[nodiscard]] auto nutate(const double t,
                          const double sun_mean_lon,
                          const double sma,
                          const double moon_mean_lon,
                          const double moon_asc_node) -> std::array<double, 3>
{
    // Nutation in Longitude (conversion from arcseconds to radians at the end)
    const double nutation_lon { (-17.1996 * std::sin(moon_asc_node)
                                 + 0.2062 * std::sin(2.0 * moon_asc_node)
                                 - 1.3187 * std::sin(2.0 * sun_mean_lon)
                                 + 0.1426 * std::sin(sma)
                                 - 0.2274 * std::sin(2.0 * moon_mean_lon))
                                * math::deg_to_rad / 3600.0 };
    //  Mean Obliquity of the Ecliptic
    const double mean_obliquity { (23.439291 - 3.560e-7 * t)
                                  * math::deg_to_rad };
    //  Nutation in Obliquity
    const double nutation_obl { mean_obliquity
                                + (9.2025 * std::cos(moon_asc_node)
                                   + 0.5736 * std::cos(2.0 * sun_mean_lon))
                                    * math::deg_to_rad / 3600.0 };
    return { nutation_lon, mean_obliquity, nutation_obl };
}

auto getSun2000(const double t,
                const double sun_mean_lon,
                const double sma,
                const double moon_mean_lon,
                const double nutation_lon,
                const double nutation_obl,
                std::array<double, dims::vec>& sun) -> void
{
    // Compute planet mean anomalies
    const double anomaly_venus { fmod(50.40828 + 1.60213022 * t, 360.0)
                                 * math::deg_to_rad };
    const double anomaly_mars { fmod(19.38816 + 0.52402078 * t, 360.0)
                                * math::deg_to_rad };
    const double anomaly_jupiter { fmod(20.35116 + 0.08309121 * t, 360.0)
                                   * math::deg_to_rad };
    // Compute geometric solar longitude
    const double geom_sun_lon {
        sun_mean_lon
        + ((6893.0 - 4.6543463e-4 * t) * std::sin(sma)
           + 72.0 * std::sin(2.0 * sma) - 7.0 * std::cos(sma - anomaly_jupiter)
           + 6.0 * std::sin(moon_mean_lon - sun_mean_lon)
           + 5.0
               * std::sin(4.0 * sma - 8.0 * anomaly_mars
                          + 3.0 * anomaly_jupiter)
           - 5.0 * std::cos(2.0 * sma - 2.0 * anomaly_venus)
           - 4.0 * std::sin(sma - anomaly_venus)
           + 4.0
               * std::cos(4.0 * sma - 8.0 * anomaly_mars
                          + 3.0 * anomaly_jupiter)
           + 3.0 * std::sin(2.0 * sma - 2.0 * anomaly_venus)
           - 3.0 * std::sin(anomaly_jupiter)
           - 3.0 * std::sin(2.0 * sma - 2.0 * anomaly_jupiter))
            * math::deg_to_rad / 3600.0
    };
    // Solar distance in astronomical units (AU)
    const double sun_dist { 1.00014 - 0.01671 * std::cos(sma)
                            - 0.00014 * std::cos(2.0 * sma) };
    // Constant of aberration
    const double abberation { 0.0056932 * math::deg_to_rad };
    // Apparent solar longitude including corrections for nutation in
    // longitude and velocity aberration
    const double app_sun_lon { geom_sun_lon + nutation_lon
                               - abberation / sun_dist };
    sun[0] = std::cos(app_sun_lon);
    sun[1] = std::sin(app_sun_lon) * std::cos(nutation_obl);
    sun[2] = std::sin(app_sun_lon) * std::sin(nutation_obl);
}

[[nodiscard]] auto getGHA2000(const double t,
                              const double day_fract,
                              const double nutation_lon,
                              const double nutation_obl) -> double
{
    // Greenwich Mean Sidereal Time (radians)
    const double gmst { fmod(100.4606184 + 0.9856473663 * t + 2.908e-13 * t * t,
                             360.0)
                        * math::deg_to_rad };
    // Include apparent time correction and time-of-day
    return { gmst + nutation_lon * std::cos(nutation_obl)
             + day_fract * 360.0 * math::deg_to_rad };
}

auto j2000ToMOD(const double t, Quaternion& q_j2000_to_mod) -> void
{
    // UT MJD2000 time in J2000 centuries
    const double tc { (t - 0.5) / 36525.0 };
    const double tc2 { tc * tc };
    const double tc3 { tc2 * tc };
    const double zeta0 { (2306.2181 * tc + 0.302 * tc2 + 0.018 * tc3)
                         * math::deg_to_rad / 3600.0 };
    const double thetap { (2004.3109 * tc - 0.4266 * tc2 - 0.04160 * tc3)
                          * math::deg_to_rad / 3600.0 };
    const double xip { (2306.2181 * tc + 1.095 * tc2 + 0.018 * tc3)
                       * math::deg_to_rad / 3600.0 };
    const std::array<double, dims::vec * dims::vec> rot_mat {
        // 00
        -std::sin(zeta0) * std::sin(xip)
          + std::cos(zeta0) * std::cos(xip) * std::cos(thetap),
        // 01
        -std::cos(zeta0) * std::sin(xip)
          - std::sin(zeta0) * std::cos(xip) * std::cos(thetap),
        // 02
        -std::cos(xip) * std::sin(thetap),
        // 10
        std::sin(zeta0) * std::cos(xip)
          + std::cos(zeta0) * std::sin(xip) * std::cos(thetap),
        // 11
        std::cos(zeta0) * std::cos(xip)
          - std::sin(zeta0) * std::sin(xip) * std::cos(thetap),
        // 12
        -std::sin(xip) * std::sin(thetap),
        // 20
        std::cos(zeta0) * std::sin(thetap),
        // 21
        -std::sin(zeta0) * std::sin(thetap),
        // 22
        std::cos(thetap)
    };
    q_j2000_to_mod = { rot_mat };
}

auto getNutationQuaternion(const double nutation_lon,
                           const double nutation_obl,
                           const double mean_obliquity,
                           Quaternion& q_nutation) -> void
{
    // Nutation rotation matrix
    const std::array<double, dims::vec * dims::vec> rot_mat {
        // 00
        std::cos(nutation_lon),
        // 01
        std::sin(nutation_lon) * std::cos(nutation_obl),
        // 01
        std::sin(nutation_lon) * std::sin(nutation_obl),
        // 10
        -std::sin(nutation_lon) * std::cos(mean_obliquity),
        // 11
        std::cos(nutation_lon) * std::cos(nutation_obl)
            * std::cos(mean_obliquity)
          + std::sin(nutation_obl) * std::sin(mean_obliquity),
        // 12
        std::cos(nutation_lon) * std::sin(nutation_obl)
            * std::cos(mean_obliquity)
          - std::cos(nutation_obl) * std::sin(mean_obliquity),
        // 20
        -std::sin(nutation_lon) * std::sin(mean_obliquity),
        // 21
        std::cos(nutation_lon) * std::cos(nutation_obl)
            * std::sin(mean_obliquity)
          - std::sin(nutation_obl) * std::cos(mean_obliquity),
        // 22
        std::cos(nutation_lon) * std::sin(nutation_obl)
            * std::sin(mean_obliquity)
          + std::cos(nutation_obl) * std::cos(mean_obliquity),
    };
    q_nutation = { rot_mat };
}

} // namespace spex
