// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "spectra.h"

#include "ckd.h"
#include "cubic_spline.h"

namespace spex {

auto Spectra::extractFOV(const CKD& ckd,
                         const std::vector<double>& image,
                         const std::vector<double>& image_noise,
                         const std::vector<bool>& pixel_mask,
                         const int i_fov) -> void
{
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        signal[i_pol].resize(ckd.n_detector_cols, 0.0);
        noise[i_pol].resize(ckd.n_detector_cols, 0.0);
        mask[i_pol].resize(ckd.n_detector_cols, false);
    }
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i {}; i < ckd.n_detector_cols; ++i) {
            double total_weight {};
            for (int i_ind {}; i_ind < ckd.fov.n_indices; ++i_ind) {
                const int idx { i * ckd.fov.n_indices + i_ind };
                const int& ipix { ckd.fov.pix_indices[i_pol][i_fov][idx] };
                if (ipix != def::idx && !pixel_mask[ipix]) {
                    const double& w { ckd.fov.weights[i_pol][i_fov][idx] };
                    signal[i_pol][i] += image[ipix] * w;
                    const double n { image_noise[ipix] * w };
                    noise[i_pol][i] += n * n;
                    total_weight += w;
                }
            }
            if (total_weight > 0.0) {
                signal[i_pol][i] /= total_weight;
                noise[i_pol][i] =
                  std::sqrt(noise[i_pol][i] / (total_weight * total_weight));
            } else {
                signal[i_pol][i] = NC_FILL_DOUBLE;
                noise[i_pol][i] = NC_FILL_DOUBLE;
                mask[i_pol][i] = true;
            }
        }
    }
}

auto Spectra::calibrate(const CKD& ckd,
                        const double exposure_time,
                        const int i_fov) -> void
{
    const double exposure_time_inv { 1.0 / exposure_time };
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i {}; i < static_cast<int>(signal.front().size()); ++i) {
            if (mask[i_pol][i]) {
                continue;
            }
            constexpr double nc_fill_sqrt { std::sqrt(NC_FILL_DOUBLE) };
            if (ckd.rad_spectra[i_pol][i_fov][i] > nc_fill_sqrt) {
                mask[i_pol][i] = true;
            } else {
                signal[i_pol][i] *=
                  ckd.rad_spectra[i_pol][i_fov][i] * exposure_time_inv;
                noise[i_pol][i] *=
                  ckd.rad_spectra[i_pol][i_fov][i] * exposure_time_inv;
            }
        }
    }
}

auto Spectra::interpolateToCommon(const CKD& ckd, const int i_fov) -> void
{
    std::array<std::vector<double>, dims::pol> wavelengths {};
    std::array<std::vector<double>, dims::pol> signal_in {};
    std::array<std::vector<double>, dims::pol> noise_in {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        wavelengths[i_pol].reserve(signal.front().size());
        signal_in[i_pol].reserve(signal.front().size());
        noise_in[i_pol].reserve(signal.front().size());
        for (int i {}; i < static_cast<int>(signal.front().size()); ++i) {
            if (!mask[i_pol][i]) {
                wavelengths[i_pol].push_back(ckd.wave_full[i_pol][i_fov][i]);
                signal_in[i_pol].push_back(signal[i_pol][i]);
                noise_in[i_pol].push_back(noise[i_pol][i]);
            }
        }
    }
    const std::array<CubicSpline, dims::pol> signal_spline {
        CubicSpline { wavelengths[pol_minus], signal_in[pol_minus] },
        CubicSpline { wavelengths[pol_plus], signal_in[pol_plus] }
    };
    const std::array<CubicSpline, dims::pol> noise_spline {
        CubicSpline { wavelengths[pol_minus], noise_in[pol_minus] },
        CubicSpline { wavelengths[pol_plus], noise_in[pol_plus] }
    };
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i {}; i < static_cast<int>(signal.front().size()); ++i) {
            signal[i_pol][i] =
              signal_spline[i_pol].eval(ckd.wave_common[i_fov][i]);
            noise[i_pol][i] =
              noise_spline[i_pol].eval(ckd.wave_common[i_fov][i]);
        }
    }
}

} // namespace spex
