// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include <cstddef>
#include <vector>

namespace spex {

// This is an inversion routine that attempts to fit a set of Gaussians.
// This is used to recognize a number of peaks in a line.
class Splitgaussfit
{
public:
    Splitgaussfit(
      size_t a_ngauss, // Number of peaks (gausses)
      size_t a_npoly,  // Number of polynomial parameters for the background
      size_t a_nmeas); // Number of elements in the measurement vector
    ~Splitgaussfit() = default;

private:
    size_t ngauss {}; // Total number of peaks
    size_t npoly {}; // Number of polynomial terms, one higher than the order
    size_t nmeas {};
    std::vector<double> own_x {};
    double* x {};
    bool new_x {};
    std::vector<double> own_meas {};
    double* meas {};
    std::vector<double> s_y {};
    std::vector<bool> own_mask {}; // Pixel mask, created with copy or by default
    std::vector<bool> *mask {}; // Pixel mask, eventually a pointer to its own mask, or to an external one
    std::vector<double> mat {};

    // Execution of the inversion
    bool gaussfit = true;

    // Inversion properties
    size_t inv_maxiter {}; // Maximum number of iterations to reach convergence
    double inv_steptolerance {}; // Tolerance for a step to be recognized as successful
    double inv_initial_step_reducer = -999.0; // Starting value of the Levenberg-Marquardt step control parameter
    double inv_reducerfactor_fail = -999.0; // Multiplication of step reducer for a failed step
    double inv_reducerfactor_success = -999.0; // Multiplication of step reducer for a successful step
    double inv_reducer_limit = -999.0; // Maximum step control reducer for convergence

    // Inversion convergence tolerances
    double inv_magtol {}; // Tolerance in peak magnitude in units of a-priori magnitude
    double inv_postol {}; // Tolerance in peak position in X-coordinate units
    double inv_widthtol {}; // Tolerance in peak FWHM in X-coordinate units
    double inv_backgroundtol {}; // Tolerance in background in line medians

public:
    auto setX(double* a_x, // Independent parameter of the line
              bool copy=false) -> void; // Flag for copying the input data. Leave false to pass pointer
    auto setMeas(double* y, bool copy=false) -> void;
    // s is flag for providing S-elements (no need to square anymore)
    auto setNoise(double* y_noise, bool s=false) -> void;
    auto removeNoise() -> void;
    auto setMask(std::vector<bool> *a_mask, bool copy=false) -> void;
    auto removeMask() -> void;

    // Set flag for executing iterative fit of a Gaussian
    auto setGaussfit(bool a_gaussfit) -> void {gaussfit=a_gaussfit;}

    // Simple setters
    auto setMaxiter(size_t a_maxiter) -> void {inv_maxiter=a_maxiter;}
    auto setSteptolerance(double a_steptolerance) -> void {inv_steptolerance=a_steptolerance;}
    auto setInitialStepReducer(double a_initial_step_reducer) -> void {inv_initial_step_reducer = a_initial_step_reducer;}
    auto setReducerfactorFail(double a_reducerfactor_fail) -> void {inv_reducerfactor_fail = a_reducerfactor_fail;}
    auto setReducerfactorSuccess(double a_reducerfactor_success) -> void {inv_reducerfactor_success = a_reducerfactor_success;}
    auto setReducerLimit(double a_reducer_limit) -> void {inv_reducer_limit = a_reducer_limit;}

    // Tolerance setters
    auto setMagtol(double a_magtol) -> void {inv_magtol=a_magtol;}
    auto setPostol(double a_postol) -> void {inv_postol=a_postol;}
    auto setWidthtol(double a_widthtol) -> void {inv_widthtol=a_widthtol;}
    auto setBackgroundtol(double a_backgroundtol) -> void {inv_backgroundtol=a_backgroundtol;}

    auto solve(int insignificant_island, // Islands of this many pixels are not recognized
               double domain_range, // Size of a gauss tail to form domains
               int maxreso, // Highest denominator in setting thresholds
               double* res, // Retrieval result
               bool debug=false) -> int; // Logs debug information
};

} // namespace spex
