// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "stray_gen.h"

#include "fourier.h"
#include "l1a_metadata.h"

#include <fstream>

namespace spex {

// Threshold for comparing floating point numbers from L1A metadata
// such as the ACT angles. This parameter doesn't matter too much as
// long as the values of interest are distinguishable.
constexpr double comp_thres { 1e-10 };

// Extract additional L1A metadata for stray light calculations
static auto extraMeta(std::vector<L1A>& l1a_products) -> void
{
    for (auto& l1a : l1a_products) {
        // Determine polarization state by file name for now
        if (l1a.meta->filename.find("B-TNOM") == std::string::npos) {
            l1a.polarization_state = pol_plus;
        } else {
            l1a.polarization_state = pol_minus;
        }
        if (l1a.meta->egse.front().C_FLEX_405nm == 1) {
            l1a.meta->refspec_wavelength = { 405.0 };
        } else if (l1a.meta->egse.front().C_FLEX_457nm == 1) {
            l1a.meta->refspec_wavelength = { 457.0 };
        } else if (l1a.meta->egse.front().C_FLEX_515nm == 1) {
            l1a.meta->refspec_wavelength = { 515.0 };
        } else if (l1a.meta->egse.front().C_FLEX_561nm == 1) {
            l1a.meta->refspec_wavelength = { 561.0 };
        } else if (l1a.meta->egse.front().C_FLEX_660nm == 1) {
            l1a.meta->refspec_wavelength = { 660.0 };
        } else if (l1a.meta->egse.front().COBOLT_785nm == 1) {
            l1a.meta->refspec_wavelength = { 785.0 };
        } else if (l1a.meta->egse.front().CRYSTA_STATUS == 1) {
            l1a.meta->refspec_wavelength = { 732.0 };
        } else {
            l1a.meta->refspec_wavelength = { 762.0 };
        }
    }
}

auto StrayGen::prepareBatches() -> void
{
    extraMeta(l1a_manager->l1a_products);
    // Sort by:
    //   1 - Polarization state (S+/-)
    //   2 - Viewport
    //   3 - Wavelength
    //   4 - ACT angle
    //   5 - Exposure time
    std::ranges::sort(
      l1a_manager->l1a_products, [](const auto& l1a_1, const auto& l1a_2) {
          if (l1a_1.polarization_state != l1a_2.polarization_state) {
              return l1a_1.polarization_state < l1a_2.polarization_state;
          }
          if (l1a_1.meta->viewport != l1a_2.meta->viewport) {
              return l1a_1.meta->viewport < l1a_2.meta->viewport;
          }
          if (std::abs(l1a_1.meta->refspec_wavelength.front()
                       - l1a_2.meta->refspec_wavelength.front())
              > comp_thres) {
              return l1a_1.meta->refspec_wavelength.front()
                     < l1a_2.meta->refspec_wavelength.front();
          }
          if (std::abs(l1a_1.meta->act_angle - l1a_2.meta->act_angle)
              > comp_thres) {
              return l1a_1.meta->act_angle < l1a_2.meta->act_angle;
          }
          return l1a_1.exposure_time < l1a_2.exposure_time;
      });
    auto l1a_first { l1a_manager->l1a_products.begin() };
    auto l1a_last { l1a_manager->l1a_products.begin() };
    // Create batches where each batch corresponds to one polarization
    // state, viewport, and wavelength but different ACT angles and
    // exposure times. Given how the L1A instances are sorted we only
    // need to check if the wavelength changes when closing off a
    // batch and moving to the next one.
    while (l1a_last != l1a_manager->l1a_products.end()) {
        ++l1a_last;
        if (l1a_last == l1a_manager->l1a_products.end()
            || std::abs(l1a_first->meta->refspec_wavelength.front()
                        - l1a_last->meta->refspec_wavelength.front())
                 > comp_thres) {
            batches.push_back({});
            moveL1ASlice(l1a_first, l1a_last, batches.back().l1a_products);
            l1a_first = l1a_last;
        }
    }
    l1a_manager->l1a_products = {};
    // Now that we know how many batches there are, allocate the
    // kernel structures (each batch corresponds to one kernel).
    ckd.stray.n_kernels = static_cast<int>(batches.size());
    ckd.stray.kernels_real.resize(ckd.stray.n_kernels,
                                  std::vector<double>(ckd.npix, 0.0));
    centers.resize(ckd.stray.n_kernels);
}

auto StrayGen::prepareBackground(L1ADiagnostic& l1a_manager_bgr) -> void
{
    // Put all background images into one batch and calibrate them in parallel
    L1ABatch batch_bgr {};
    batch_bgr.l1a_products = std::move(l1a_manager_bgr.l1a_products);
    batch_bgr.calibrate(settings, proc_level, ckd);
    extraMeta(batch_bgr.l1a_products);
    for (auto& l1a_bgr : batch_bgr.l1a_products) {
        auto bgr_image { std::make_shared<std::vector<double>>() };
        *bgr_image = std::move(l1a_bgr.image);
        for (auto& batch : batches) {
            for (auto& l1a : batch.l1a_products) {
                if (l1a.polarization_state == l1a_bgr.polarization_state
                    && l1a.meta->viewport == l1a_bgr.meta->viewport
                    && std::abs(l1a.meta->refspec_wavelength.front()
                                - l1a_bgr.meta->refspec_wavelength.front())
                         < comp_thres
                    && std::abs(l1a.exposure_time - l1a_bgr.exposure_time)
                         < comp_thres) {
                    l1a.bgr_image = bgr_image;
                }
            }
        }
    }
}

// Generate a composite image from images of different exposure
// times. Here the L1A products correspond to the same polarization
// state, viewport, laser wavelength, and ACT angle but different
// exposure times. The order of exposure times is assumed to be
// ascending, i.e. l1a_products.front() has the lowest exposure time.
static auto genCompositeImage(const double intensity_threshold,
                              const std::vector<L1A>& l1a_products,
                              std::vector<double>& composite_image) -> void
{
    // Assume all images in l1a_products have the same binning and thus size
    composite_image.resize(l1a_products.front().image.size());
    for (int i {}; i < static_cast<int>(composite_image.size()); ++i) {
        for (auto l1a { l1a_products.crbegin() }; l1a != l1a_products.crend();
             ++l1a) {
            if (l1a->image[i] < intensity_threshold) {
                composite_image[i] = l1a->image[i] / l1a->exposure_time;
                break;
            }
        }
    }
}

// Compute coordinates of the center of mass of an image. n_rows and
// n_cols are the image dimensions and polarization_state determines
// which half of the image to consider.
static auto findCenterOfMass(const int n_rows,
                             const int n_cols,
                             const int polarization_state,
                             std::vector<double>& image)
  -> std::array<double, 2>
{
    std::array<int, 2> max_point { 0, 0 };
    std::array<size_t, 2> search_limits {};
    if (polarization_state == pol_minus) {
        search_limits = { 0, image.size() / 2 };
    } else {
        search_limits = { image.size() / 2, image.size() };
    }
    const auto max_idx {
        std::ranges::distance(
          image.cbegin(),
          std::ranges::max_element(image.cbegin() + search_limits.front(),
                                   image.cbegin() + search_limits.back())),
    };
    max_point.front() = static_cast<int>(max_idx / n_cols);
    max_point.back() = static_cast<int>(max_idx - n_cols * max_point.front());
    // Next find the center of maximum in a region close to the point
    // of maximum light level using the center of mass method.
    const int search_range { static_cast<int>(0.05 * n_rows) };
    double intensity_sum {};
    std::array<double, 2> center_of_mass { 0.0, 0.0 };
    for (int i { std::max(0, max_point[0] - search_range) };
         i < std::min(n_rows, max_point[0] + search_range);
         ++i) {
        for (int j { std::max(0, max_point[1] - search_range) };
             j < std::min(n_cols, max_point[1] + search_range);
             ++j) {
            const int idx { i * n_cols + j };
            intensity_sum += image[idx];
            center_of_mass[0] += i * image[idx];
            center_of_mass[1] += j * image[idx];
        }
    }
    for (auto& el : center_of_mass) {
        el /= intensity_sum;
    }
    return center_of_mass;
}

// Using bilinear interpolation center an image around the point "center"
static auto shiftToCenter(const int n_rows,
                          const int n_cols,
                          const std::array<double, 2>& center,
                          std::vector<double>& image) -> void
{
    const double row_midpoint { n_rows / 2.0 };
    const double col_midpoint { n_cols / 2.0 };
    std::vector<double> image_shifted(image.size(), 0.0);
    for (int i {}; i < n_rows; ++i) {
        for (int j {}; j < n_cols; ++j) {
            const double i_frac { i + center[0] - row_midpoint };
            const double j_frac { j + center[1] - col_midpoint };
            const int i0 { static_cast<int>(i_frac) };
            const int j0 { static_cast<int>(j_frac) };
            const int i1 { i0 + 1 };
            const int j1 { j0 + 1 };
            const double i_delta { i_frac - i0 };
            const double j_delta { j_frac - j0 };
            const double i_delta1 { 1.0 - i_delta };
            const double j_delta1 { 1.0 - j_delta };
            if (i_frac >= 0 && i_frac + 1 < n_rows && j_frac >= 0
                && j_frac + 1 < n_cols) {
                image_shifted[i * n_cols + j] =
                  i_delta1 * j_delta1 * image[i0 * n_cols + j0]
                  + i_delta * j_delta1 * image[i1 * n_cols + j0]
                  + i_delta1 * j_delta * image[i0 * n_cols + j1]
                  + i_delta * j_delta * image[i1 * n_cols + j1];
            }
        }
    }
    image = std::move(image_shifted);
}

auto StrayGen::processBatch(const int i_batch) -> void
{
    const auto& l1a_products { batches.at(i_batch).l1a_products };
    const std::string polarization_state {
        l1a_products.front().polarization_state == pol_plus ? "plus" : "minus"
    };
    spdlog::info(
      "Generating kernel for polarization state \"{}\", viewport {} deg, "
      "wavelength {} nm",
      polarization_state,
      viewport_idx_to_angle.at(l1a_products.front().meta->viewport),
      l1a_products.front().meta->refspec_wavelength.front());

    // STEP 1 - Generate composite images from multiple exposure times

    std::vector<std::vector<double>> composite_images {};
    // In the following search, we already know that images have been
    // sorted by ACT angles and exposure times.
    auto l1a_first { l1a_products.begin() };
    auto l1a_last { l1a_products.begin() };
    while (l1a_last != l1a_products.end()) {
        ++l1a_last;
        // If the next L1A image has a different ACT angle, close the
        // range of exposure times, generate a composite image, and
        // move on to the next ACT angle.
        if (l1a_last == l1a_products.end()
            || std::abs(l1a_first->meta->act_angle - l1a_last->meta->act_angle)
                 > comp_thres) {
            composite_images.push_back({});
            genCompositeImage(settings.stray.exposure_intensity_thres,
                              { l1a_first, l1a_last },
                              composite_images.back());
            l1a_first = l1a_last;
        }
    }

    // STEP 2 - Shift images to their centers

    // Find center of mass of each image
    std::vector<std::array<double, 2>> center_of_masses(
      composite_images.size());
    for (int i {}; i < static_cast<int>(composite_images.size()); ++i) {
        center_of_masses.at(i) =
          findCenterOfMass(ckd.n_detector_rows,
                           ckd.n_detector_cols,
                           l1a_products.front().polarization_state,
                           composite_images.at(i));
    }
    // Use the average of those for the kernel
    centers.at(i_batch) = { 0.0, 0.0 };
    for (int i {}; i < static_cast<int>(center_of_masses.size()); ++i) {
        centers.at(i_batch).at(0) += center_of_masses.at(i).at(0);
        centers.at(i_batch).at(1) += center_of_masses.at(i).at(1);
    }
    centers.at(i_batch).at(0) /= static_cast<double>(center_of_masses.size());
    centers.at(i_batch).at(1) /= static_cast<double>(center_of_masses.size());
    // Shift the center of mass of all image to their centers
    for (int i {}; i < static_cast<int>(composite_images.size()); ++i) {
        shiftToCenter(ckd.n_detector_rows,
                      ckd.n_detector_cols,
                      center_of_masses.at(i),
                      composite_images.at(i));
    }

    // STEP 3 - Construct kernel as the median of images

    std::vector<double> pixel_values(composite_images.size());
    for (int i {}; i < ckd.npix; ++i) {
        for (int i_img {}; i_img < static_cast<int>(composite_images.size());
             ++i_img) {
            pixel_values[i_img] = composite_images[i_img][i];
        }
        ckd.stray.kernels_real[i_batch][i] = median(pixel_values);
    }
    // Normalize kernel
    double sum {};
    for (auto& val : ckd.stray.kernels_real[i_batch]) {
        sum += val;
    }
    for (auto& val : ckd.stray.kernels_real[i_batch]) {
        val /= sum;
    }
}

// Set all values at the center of an image within a radius to zero
static auto carveHole(const int n_rows,
                      const int n_cols,
                      const double hole_radius,
                      std::vector<double>& kernel) -> void
{
    const double row_midpoint { n_rows / 2.0 };
    const double col_midpoint { n_cols / 2.0 };
    for (int i {}; i < n_rows; ++i) {
        for (int j {}; j < n_rows; ++j) {
            const double id { i - row_midpoint };
            const double jd { j - col_midpoint };
            if (std::sqrt(id * id + jd * jd) <= hole_radius) {
                kernel[i * n_cols + j] = 0.0;
            }
        }
    }
}

// Shift all kernels to their centers sum the up in a single
// image. This is useful for plotting but otherwise serves no purpose.
static auto shiftAndSumKernels(
  const int n_rows,
  const int n_cols,
  const std::vector<std::vector<double>>& kernels,
  const std::vector<std::array<double, 2>>& centers,
  std::vector<double>& all_kernels) -> void
{
    const int n_rows_half { n_rows / 2 };
    const int n_cols_half { n_cols / 2 };
    all_kernels.resize(kernels.front().size(), 0.0);
#pragma omp declare reduction(vplus : std::vector<double> : std::transform(    \
    omp_out.begin(),                                                           \
      omp_out.end(),                                                           \
      omp_in.begin(),                                                          \
      omp_out.begin(),                                                         \
      std::plus<double>())) initializer(omp_priv = omp_orig)
#pragma omp parallel for reduction(vplus : all_kernels)
    for (int i_kernel = 0; i_kernel < static_cast<int>(kernels.size());
         ++i_kernel) {
        for (int i {}; i < n_rows; ++i) {
            // ii and jj are i and j shifted by the kernel center
            const int ii { i - n_rows_half
                           + static_cast<int>(
                             std::round(centers[i_kernel][0])) };
            for (int j {}; j < n_cols; ++j) {
                const int jj { j - n_cols_half
                               + static_cast<int>(
                                 std::round(centers[i_kernel][1])) };
                if (ii >= 0 && ii < n_rows && jj >= 0 && jj < n_cols) {
                    all_kernels[ii * n_cols + jj] +=
                      kernels[i_kernel][i * n_cols + j];
                }
            }
        }
    }
}

// Shift an image cyclically such that its center moves to the first element
static auto roll(const int n_rows, const int n_cols, std::vector<double>& image)
  -> void
{
    const int n_rows_half { n_rows / 2 };
    const int n_cols_half { n_cols / 2 };
    const int n_rows_half_one { n_rows_half + n_rows % 2 };
    const int n_cols_half_one { n_cols_half + n_cols % 2 };
    std::vector<double> image_initial { image };
    for (int i {}; i < n_rows_half_one; ++i) {
        const int i_shifted { (i + n_rows_half) * n_cols };
        for (int j {}; j < n_cols_half_one; ++j) {
            const int idx { i_shifted + j + n_cols_half };
            image[i * n_cols + j] = image_initial[idx];
        }
    }
    for (int i { n_rows_half_one }; i < n_rows; ++i) {
        const int i_shifted { (i - n_rows_half_one) * n_cols };
        for (int j { n_cols_half_one }; j < n_cols; ++j) {
            const int idx { i_shifted + j - n_cols_half_one };
            image[i * n_cols + j] = image_initial[idx];
        }
    }
    for (int i { n_rows_half_one }; i < n_rows; ++i) {
        const int i_shifted { (i - n_rows_half_one) * n_cols };
        for (int j {}; j < n_cols_half_one; ++j) {
            const int idx { i_shifted + j + n_cols_half };
            image[i * n_cols + j] = image_initial[idx];
        }
    }
    for (int i {}; i < n_rows_half_one; ++i) {
        const int i_shifted { (i + n_rows_half) * n_cols };
        for (int j { n_cols_half_one }; j < n_cols; ++j) {
            const int idx { i_shifted + j - n_cols_half_one };
            image[i * n_cols + j] = image_initial[idx];
        }
    }
}

// Shift an image such that its center is at the first element and
// take its Fourier transform.
static auto rollAndFourier(const int n_rows,
                           const int n_cols,
                           const std::vector<double>& image,
                           std::vector<std::complex<double>>& image_fft) -> void
{
    std::vector<double> image_rolled { image };
    roll(n_rows, n_cols, image_rolled);
    fft_r2c(n_rows, n_cols, image_rolled, image_fft);
}

// Construct the kernel weight matrices
static auto constructKernelWeights(
  const int n_rows,
  const int n_cols,
  const std::vector<std::array<double, 2>>& centers,
  std::vector<std::vector<double>>& weights,
  std::vector<int>& edges) -> void
{
    const int weight_size { n_rows * n_cols };
    const int n_kernels { static_cast<int>(centers.size()) };
    weights.resize(n_kernels, std::vector<double>(weight_size, 0.0));
    constexpr int n_edges { 4 };
    edges.resize(n_edges * n_kernels);
    // Sort all centers by row numbers
    auto centers_sorted { centers };
    std::ranges::sort(centers_sorted);
    // Convert the list of centers into a matrix and additionally sort
    // by columns. For instance, centers_matrix[0][0] is the location
    // of the kernel center at the bottom left corner.
    std::vector<std::vector<std::array<double, 2>>> centers_matrix {};
    // If the row number of two centers differs by less than
    // next_group_distance then consider them to be in the same row of
    // kernels. The objective is to construct a sorted N x M matrix of
    // kernel centers. For instance, in the case of 10 bands on the
    // detector and 7 laser wavelengths, N = 10 and M = 7 accordingly.
    constexpr double rel_group_separation { 0.05 };
    const double next_group_distance { rel_group_separation * n_rows };
    auto range_beg { centers_sorted.begin() };
    auto range_end { range_beg };
    do {
        ++range_end;
        // If the row index of the next kernel center does not align
        // with the current row, close the group (one row of
        // centers_matrix) and sort by column indices.
        if (range_end == centers_sorted.cend()
            || std::abs((*range_end)[0] - (*range_beg)[0])
                 > next_group_distance) {
            std::ranges::sort(
              range_beg, range_end, [](const auto& a, const auto& b) {
                  // a[0] would correspond to a row position
                  return a[1] < b[1];
              });
            centers_matrix.push_back({});
            for (auto it { range_beg }; it != range_end; ++it) {
                centers_matrix.back().push_back(*it);
            }
            range_beg = range_end;
        }
    } while (range_end != centers_sorted.cend());
    for (const auto& row : centers_matrix) {
        if (row.size() != centers_matrix.front().size()) {
            throw std::runtime_error {
                "kernel centers do not form a coherent grid"
            };
        }
    }
    for (int i_kernel {}; i_kernel < n_kernels; ++i_kernel) {
        // Identify the location of the current kernel in centers_matrix
        std::array<int, 2> loc {};
        for (int i {}; i < static_cast<int>(centers_matrix.size()); ++i) {
            for (int j {}; j < static_cast<int>(centers_matrix.at(i).size());
                 ++j) {
                const double diff_i { centers.at(i_kernel)[0]
                                      - centers_matrix.at(i).at(j)[0] };
                const double diff_j { centers.at(i_kernel)[1]
                                      - centers_matrix.at(i).at(j)[1] };
                if (std::abs(diff_i * diff_i + diff_j * diff_j) < 1e-10) {
                    loc = { i, j };
                    break;
                }
            }
        }
        // Determine the four edges that define the weights of the
        // this kernel. The default box dimensions are the entire
        // detector.
        edges.at(i_kernel * box::n + box::b) = 0;
        edges.at(i_kernel * box::n + box::t) = n_rows;
        edges.at(i_kernel * box::n + box::l) = 0;
        edges.at(i_kernel * box::n + box::r) = n_cols;
        // The current kernel center is at C[i,j] where C is
        // centers_matrix. Its neighboring kernel centers are at
        // C[i-1,j], C[i,j-1], C[i+1,j], and C[i,j+1]. The box of the
        // current kernel is determined by the neighboring kernel
        // centers or one or more detector edges.
        if (loc.at(0) > 0) {
            edges.at(i_kernel * box::n + box::b) = static_cast<int>(
              centers_matrix.at(loc.at(0) - 1).at(loc.at(1))[0]);
        }
        if (loc.at(0) < static_cast<int>(centers_matrix.size()) - 1) {
            edges.at(i_kernel * box::n + box::t) = static_cast<int>(
              centers_matrix.at(loc.at(0) + 1).at(loc.at(1))[0]);
        }
        if (loc.at(1) > 0) {
            edges.at(i_kernel * box::n + box::l) = static_cast<int>(
              centers_matrix.at(loc.at(0)).at(loc.at(1) - 1)[1]);
        }
        if (loc.at(1)
            < static_cast<int>(centers_matrix.at(loc.at(0)).size()) - 1) {
            edges.at(i_kernel * box::n + box::r) = static_cast<int>(
              centers_matrix.at(loc.at(0)).at(loc.at(1) + 1)[1]);
        }
        // Now that the box has been drawn, compute the weight
        // functions in it.
        for (int i { edges.at(i_kernel * box::n + box::b) };
             i < edges.at(i_kernel * box::n + box::t);
             ++i) {
            // Make sure the weight is never exactly zero. This could
            // make it more difficult to normalize it later.
            constexpr double value_eps { 1e-10 };
            double x_value { value_eps };
            if (i < centers[i_kernel][0]) {
                x_value +=
                  static_cast<double>(i - edges.at(i_kernel * box::n + box::b))
                  / (centers[i_kernel][0]
                     - edges.at(i_kernel * box::n + box::b));
            } else {
                x_value +=
                  static_cast<double>(edges.at(i_kernel * box::n + box::t) - i)
                  / (edges.at(i_kernel * box::n + box::t)
                     - centers[i_kernel][0]);
            }
            for (int j { edges.at(i_kernel * box::n + box::l) };
                 j < edges.at(i_kernel * box::n + box::r);
                 ++j) {
                double y_value { value_eps };
                if (j < centers[i_kernel][1]) {
                    y_value += static_cast<double>(
                                 j - edges.at(i_kernel * box::n + box::l))
                               / (centers[i_kernel][1]
                                  - edges.at(i_kernel * box::n + box::l));
                } else {
                    y_value += static_cast<double>(
                                 edges.at(i_kernel * box::n + box::r) - j)
                               / (edges.at(i_kernel * box::n + box::r)
                                  - centers[i_kernel][1]);
                }
                weights[i_kernel][i * n_cols + j] = x_value * y_value;
            }
        }
    }
    // Normalize the kernel weights
    std::vector<double> weights_sum(weight_size);
    for (const auto& weight : weights) {
        for (int i {}; i < weight_size; ++i) {
            weights_sum[i] += weight[i];
        }
    }
    for (auto& weight : weights) {
        for (int i {}; i < weight_size; ++i) {
            weight[i] /= weights_sum[i];
        }
    }
}

// Enlarge an image with the new dimensions (n_rows + 2*extent_rows,
// n_cols + 2*extent_cols) and set the new elements to 0.
static auto enlargeAndPad(std::vector<double>& image,
                          const int n_rows,
                          const int n_cols,
                          const int extent_rows,
                          const int extent_cols) -> void
{
    const int n_rows_half { n_rows / 2 };
    const int n_cols_half { n_cols / 2 };
    const int n_rows_large { n_rows + 2 * extent_rows };
    const int n_cols_large { n_cols + 2 * extent_cols };
    const int n_rows_large_half { n_rows_large / 2 };
    const int n_cols_large_half { n_cols_large / 2 };
    std::vector<double> image_large(n_rows_large * n_cols_large, 0.0);
    for (int i { -extent_rows }; i < extent_rows; ++i) {
        for (int j { -extent_cols }; j < extent_cols; ++j) {
            image_large[(i + n_rows_large_half) * n_cols_large + j
                        + n_cols_large_half] =
              image[(i + n_rows_half) * n_cols + j + n_cols_half];
        }
    }
    image = std::move(image_large);
}

// Compute internal scattering factor from the kernel and the kernel weights
static auto computeEta(const CKD& ckd,
                       const std::vector<double>& weights,
                       const std::vector<double>& kernel,
                       std::vector<double>& eta) -> void
{
    // Double the dimension of weights and pad with zeros
    std::vector<double> weights_large(weights.size());
    for (int i {}; i < static_cast<int>(weights.size()); ++i) {
        weights_large[i] = weights[i];
    }
    enlargeAndPad(weights_large,
                  ckd.stray.kernel_n_rows_full,
                  ckd.stray.kernel_n_cols_full,
                  ckd.stray.kernel_n_rows_full / 2,
                  ckd.stray.kernel_n_cols_full / 2);
    const int kernel_n_rows_dbl { 2 * ckd.stray.kernel_n_rows_full };
    const int kernel_n_cols_dbl { 2 * ckd.stray.kernel_n_cols_full };
    std::vector<std::complex<double>> weights_fft(
      getFFTSize(kernel_n_rows_dbl, kernel_n_cols_dbl));
    rollAndFourier(
      kernel_n_rows_dbl, kernel_n_cols_dbl, weights_large, weights_fft);
    std::vector<double> eta_tmp(kernel.size());
    std::vector<double> d_kernel(kernel.size());
    for (int i {}; i < static_cast<int>(kernel.size()); ++i) {
        d_kernel[i] = kernel[i];
    }
    std::vector<std::complex<double>> d_kernel_fft(weights_fft.size());
    convolve(ckd.stray.kernel_n_rows_full,
             ckd.stray.kernel_n_cols_full,
             d_kernel,
             kernel_n_rows_dbl,
             kernel_n_cols_dbl,
             weights_fft,
             d_kernel_fft,
             eta_tmp);
    for (int i {}; i < static_cast<int>(eta.size()); ++i) {
        eta[i] = eta_tmp[i];
    }
}

// Reduce kernel to an optimal size. It remains at the center of the
// reduced image.
static auto reduceKernelSize(const int n_rows_in,
                             const int n_cols_in,
                             const std::vector<double>& kernel,
                             const int n_rows_out,
                             const int n_cols_out,
                             std::vector<double>& kernel_reduced) -> void
{
    assert(n_rows_out <= n_rows_in && n_cols_out <= n_cols_in
           && "output array should be smaller than input array");
    kernel_reduced.resize(n_rows_out * n_cols_out, 0.0);
    const int shift_row { (n_rows_in - n_rows_out) / 2 };
    const int shift_col { (n_cols_in - n_cols_out) / 2 };
    for (int i {}; i < n_rows_out; ++i) {
        for (int j {}; j < n_cols_out; ++j) {
            kernel_reduced[i * n_cols_out + j] =
              kernel[(i + shift_row) * n_cols_in + j + shift_col];
        }
    }
}

// First write the number of kernels, then all real-space kernels, and
// then all the kernel centers into a checkpoint file.
auto static writeCheckpoint(const SPEXSettings& settings,
                            const std::vector<std::vector<double>>& kernels,
                            const std::vector<std::array<double, 2>> centers)
  -> void
{
    std::ofstream out { settings.stray.checkpoint.filename, std::ios::binary };
    const int n_kernels { static_cast<int>(kernels.size()) };
    out.write((char*)&n_kernels, sizeof(int));
    for (const auto& kernel : kernels) {
        out.write((char*)kernel.data(), kernel.size() * sizeof(double));
    }
    for (const auto& center : centers) {
        out.write((char*)center.data(), center.size() * sizeof(double));
    }
}

// Basically like writeCheckpoint but in reverse
auto static readCheckpoint(const SPEXSettings& settings,
                           CKD& ckd,
                           std::vector<std::array<double, 2>>& centers) -> void
{
    try {
        std::ifstream in { settings.stray.checkpoint.filename,
                           std::ios::binary };
        in.exceptions(std::ofstream::failbit);
        in.read((char*)&ckd.stray.n_kernels, sizeof(int));
        ckd.stray.kernels_real.resize(ckd.stray.n_kernels,
                                      std::vector<double>(ckd.npix));
        centers.resize(ckd.stray.n_kernels);
        for (auto& kernel : ckd.stray.kernels_real) {
            in.read((char*)kernel.data(), kernel.size() * sizeof(double));
        }
        for (auto& center : centers) {
            in.read((char*)center.data(), center.size() * sizeof(double));
        }
    } catch (const std::ofstream::failure& e) {
        std::cerr << "error reading from " << settings.stray.checkpoint.filename
                  << '\n';
        throw;
    }
}

auto StrayGen::processFinalize() -> void
{
    if (settings.stray.checkpoint.write) {
        writeCheckpoint(settings, ckd.stray.kernels_real, centers);
    }
    if (settings.stray.checkpoint.read) {
        readCheckpoint(settings, ckd, centers);
    }
    spdlog::info("Computing internal scattering factors and weights");
    // Default is for the kernel (before enlarging that happens at the
    // end of this function) to have the same dimensions as the
    // detector. If reduced kernels were requested then those
    // dimensions are halved.
    ckd.stray.reduced_kernels = settings.stray.reduced_kernels;
    ckd.stray.kernel_n_rows_full = ckd.n_detector_rows;
    ckd.stray.kernel_n_cols_full = ckd.n_detector_cols;
    if (settings.stray.reduced_kernels) {
        ckd.stray.kernel_n_rows_full /= 2;
        ckd.stray.kernel_n_cols_full /= 2;
    }
    ckd.stray.n_spatial = ckd.stray.kernel_n_rows_full;
    ckd.stray.n_spectral = ckd.stray.kernel_n_cols_full;
    const int kernel_size { ckd.stray.n_spatial * ckd.stray.n_spectral };

    // Carve out a hole from the center
#pragma omp parallel for
    for (int i_kernel = 0; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        carveHole(ckd.n_detector_rows,
                  ckd.n_detector_cols,
                  settings.stray.kernel_hole_radius,
                  ckd.stray.kernels_real.at(i_kernel));
    }

    // Shift kernels to their centers and compute their sum. In debug
    // mode the result is stored in the CKD but is not used otherwise.
    shiftAndSumKernels(ckd.n_detector_rows,
                       ckd.n_detector_cols,
                       ckd.stray.kernels_real,
                       centers,
                       ckd.stray.all_kernels);

    // Reduce kernel dimensions if requested. If bin_size == 1 these
    // loops do nothing.
    const int bin_size { settings.stray.reduced_kernels ? 2 : 1 };
    std::vector<std::vector<double>> kernels_reduced(
      ckd.stray.n_kernels, std::vector<double>(kernel_size, 0.0));
    for (int i_kernel {}; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        for (int i {}; i < ckd.stray.n_spatial; ++i) {
            for (int j {}; j < ckd.stray.n_spectral; ++j) {
                for (int m {}; m < bin_size; ++m) {
                    for (int n {}; n < bin_size; ++n) {
                        kernels_reduced[i_kernel]
                                       [i * ckd.stray.n_spectral + j] +=
                          ckd.stray.kernels_real[i_kernel]
                                                [(i * bin_size + m)
                                                   * settings.main.detector_spec
                                                 + j * bin_size + n];
                    }
                }
            }
        }
    }
    ckd.stray.kernels_real = std::move(kernels_reduced);
    // Scale locations of kernel centers and kernel extent accordingly
    for (auto& center : centers) {
        center[0] /= bin_size;
        center[1] /= bin_size;
    }
    const int kernel_extent { settings.stray.kernel_extent / bin_size };

    // Construct the kernel weight matrices
    constructKernelWeights(ckd.stray.n_spatial,
                           ckd.stray.n_spectral,
                           centers,
                           ckd.stray.weights,
                           ckd.stray.edges);

    // Compute etas
    ckd.stray.etas.resize(ckd.stray.n_kernels,
                          std::vector<double>(kernel_size));
#pragma omp parallel for
    for (int i_kernel = 0; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        computeEta(ckd,
                   ckd.stray.weights.at(i_kernel),
                   ckd.stray.kernels_real.at(i_kernel),
                   ckd.stray.etas.at(i_kernel));
    }
    // The final eta used during deconvolution is a sum over
    // individual contributions.
    ckd.stray.eta.resize(kernel_size, 0.0);
    for (int i_kernel {}; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        for (int i {}; i < kernel_size; ++i) {
            ckd.stray.eta[i] += ckd.stray.etas[i_kernel][i];
        }
    }

    // Enlarge and pad all kernels with zeros
    std::vector<std::vector<double>> kernels_large {};
    // Update kernel dimensions
    ckd.stray.kernel_n_rows_full += 2 * kernel_extent;
    ckd.stray.kernel_n_cols_full += 2 * kernel_extent;
#pragma omp parallel for
    for (int i_kernel = 0; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        enlargeAndPad(ckd.stray.kernels_real[i_kernel],
                      ckd.stray.n_spatial,
                      ckd.stray.n_spectral,
                      kernel_extent,
                      kernel_extent);
    }

    // Number of rows and columns of each subimage
    std::vector<int> image_n_rows(ckd.stray.n_kernels, 0);
    std::vector<int> image_n_cols(ckd.stray.n_kernels, 0);
    // Until now 'edges' described the box size and location of each
    // kernel weight. We now expand it by kernel_extent which is a
    // distance in pixels. Then edges[i] describes a subimage that
    // must be extracted from the full image in order to convolve it
    // with kernel i.
    for (int i_kernel {}; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        ckd.stray.edges[i_kernel * box::n + box::b] = std::max(
          0, ckd.stray.edges[i_kernel * box::n + box::b] - kernel_extent);
        ckd.stray.edges[i_kernel * box::n + box::t] =
          std::min(ckd.stray.n_spatial,
                   ckd.stray.edges[i_kernel * box::n + box::t] + kernel_extent);
        ckd.stray.edges[i_kernel * box::n + box::l] = std::max(
          0, ckd.stray.edges[i_kernel * box::n + box::l] - kernel_extent);
        ckd.stray.edges[i_kernel * box::n + box::r] =
          std::min(ckd.stray.n_spectral,
                   ckd.stray.edges[i_kernel * box::n + box::r] + kernel_extent);
        image_n_rows[i_kernel] = ckd.stray.edges[i_kernel * box::n + box::t]
                                 - ckd.stray.edges[i_kernel * box::n + box::b];
        image_n_cols[i_kernel] = ckd.stray.edges[i_kernel * box::n + box::r]
                                 - ckd.stray.edges[i_kernel * box::n + box::l];
    }
    // Next generate the dimensions for all kernels. Each kernel is
    // defined by a box that covers a subimage plus at least
    // 2*kernel_extent. That is necessary to avoid periodic effect
    // when taking fast Fourier transforms.
    ckd.stray.kernel_rows.resize(ckd.stray.n_kernels);
    ckd.stray.kernel_cols.resize(ckd.stray.n_kernels);
    ckd.stray.kernel_fft_sizes.resize(ckd.stray.n_kernels);
    for (int i_kernel {}; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        ckd.stray.kernel_rows[i_kernel] =
          image_n_rows[i_kernel] + 2 * kernel_extent;
        ckd.stray.kernel_cols[i_kernel] =
          image_n_cols[i_kernel] + 2 * kernel_extent;
        // Extend kernel size to the next multiple of mult which is a
        // power of 2.
        constexpr int mult { 128 };
        ckd.stray.kernel_rows[i_kernel] =
          mult * (ckd.stray.kernel_rows[i_kernel] / mult + 1);
        ckd.stray.kernel_cols[i_kernel] =
          mult * (ckd.stray.kernel_cols[i_kernel] / mult + 1);
        ckd.stray.kernel_fft_sizes[i_kernel] = getFFTSize(
          ckd.stray.kernel_rows[i_kernel], ckd.stray.kernel_cols[i_kernel]);
    }
    // Finally, the most important part, which is taking FFTs of the kernels
    ckd.stray.kernels_fft.resize(ckd.stray.n_kernels);
    for (int i {}; i < ckd.stray.n_kernels; ++i) {
        ckd.stray.kernels_fft[i].resize(ckd.stray.kernel_fft_sizes[i]);
    }
#pragma omp parallel for
    for (int i_kernel = 0; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
        std::vector<double> kernel_reduced {};
        reduceKernelSize(ckd.stray.kernel_n_rows_full,
                         ckd.stray.kernel_n_cols_full,
                         ckd.stray.kernels_real[i_kernel],
                         ckd.stray.kernel_rows[i_kernel],
                         ckd.stray.kernel_cols[i_kernel],
                         kernel_reduced);
        rollAndFourier(ckd.stray.kernel_rows[i_kernel],
                       ckd.stray.kernel_cols[i_kernel],
                       kernel_reduced,
                       ckd.stray.kernels_fft[i_kernel]);
    }
}

} // namespace spex
