// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "swath_gen.h"

#include "l1a_metadata.h"
#include "lininv.h"

#include <fstream>
#include <gadfit/lm_solver.h>

namespace spex {

auto SwathGen::processInit() -> void
{
    act_angles.resize(ckd.n_vp);
    alt_angles.resize(ckd.n_vp);
    peak_medians.resize(ckd.n_vp);
    // Work arrays for constructing the FOV row indices spline
    std::vector<double> y_data {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        fov_splines_m[i_pol].resize(ckd.n_vp);
        for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
            fov_splines_m[i_pol][i_vp] =
              LinearSplineMulti { ckd.n_detector_cols, ckd.fov.n_act_vp[i_vp] };
            for (int i_fov {}; i_fov < ckd.fov.n_act_vp[i_vp]; ++i_fov) {
                fov_splines_m[i_pol][i_vp].knots[i_fov] =
                  ckd.fov.act_angles[ckd.fov.start_vp[i_vp] + i_fov];
            }
            y_data.resize(ckd.fov.n_act_vp[i_vp]);
            for (int i_col {}; i_col < ckd.n_detector_cols; ++i_col) {
                for (int i_fov {}; i_fov < ckd.fov.n_act_vp[i_vp]; ++i_fov) {
                    y_data[i_fov] =
                      ckd.fov.row_indices[i_pol][ckd.fov.start_vp[i_vp] + i_fov]
                                         [i_col];
                }
                fov_splines_m[i_pol][i_vp].generateSpline(i_col, y_data);
            }
        }
    }
}

auto SwathGen::prepareBatches() -> void
{
    std::ranges::sort(
      l1a_manager->l1a_products, [](const auto& l1a_1, const auto& l1a_2) {
          if (l1a_1.meta->viewport != l1a_2.meta->viewport) {
              return l1a_1.meta->viewport < l1a_2.meta->viewport;
          }
          if (l1a_1.meta->act_angle != l1a_2.meta->act_angle) {
              return l1a_1.meta->act_angle < l1a_2.meta->act_angle;
          }
          return l1a_1.meta->alt_angle < l1a_2.meta->alt_angle;
      });
    auto l1a_first { l1a_manager->l1a_products.begin() };
    auto l1a_last { l1a_manager->l1a_products.begin() };
    while (l1a_last != l1a_manager->l1a_products.end()) {
        ++l1a_last;
        constexpr double comp_thres { 1e-10 };
        if (l1a_last == l1a_manager->l1a_products.end()
            || std::abs(l1a_first->meta->act_angle - l1a_last->meta->act_angle)
                 > comp_thres) {
            batches.push_back({});
            moveL1ASlice(l1a_first, l1a_last, batches.back().l1a_products);
            l1a_first = l1a_last;
        }
    }
    // Batches have been generated. The L1A array of L1A manager is no
    // longer needed.
    l1a_manager->l1a_products = std::vector<L1A> {};
}

auto SwathGen::processBatch(const int i_batch) -> void
{
    const auto& l1as { batches[i_batch].l1a_products };
    // Determine current viewport (0..4)
    const int i_vp { static_cast<int>(
      std::distance(viewport_idx_to_angle.cbegin(),
                    viewport_idx_to_angle.find(l1as.front().meta->viewport))) };
    // ACT angles are all the same inside the batch
    act_angles[i_vp].push_back(
      0.5 * (l1as.front().meta->act_angle + l1as.back().meta->act_angle));
    spdlog::info("Processing viewport {}, ACT angle {} deg",
                 viewport_idx_to_angle.at(l1as.front().meta->viewport),
                 act_angles[i_vp].back() / math::deg_to_rad);
    // Use spectra medians as a measure of how strong the signal with
    // a given ALT angle is.
    std::vector<double> medians(l1as.size());
    std::vector<double> spectra(dims::pol * ckd.n_detector_cols);
    for (int i_alt {}; i_alt < l1as.size(); ++i_alt) {
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            for (int i_col {}; i_col < ckd.n_detector_cols; ++i_col) {
                const int idx { i_pol * ckd.n_detector_cols + i_col };
                const double row_index { fov_splines_m[i_pol][i_vp].eval(
                  i_col, act_angles[i_vp].back()) };
                const int row_dn { static_cast<int>(row_index) };
                const double weight { row_dn + 1.0 - row_index };
                const int pix_dn { row_dn * ckd.n_detector_cols + i_col };
                const int pix_up { (row_dn + 1) * ckd.n_detector_cols + i_col };
                if (l1as[i_alt].pixel_mask[pix_dn]
                    && l1as[i_alt].pixel_mask[pix_up]) {
                    spectra[idx] = 0.0;
                } else if (l1as[i_alt].pixel_mask[pix_dn]) {
                    spectra[idx] = l1as[i_alt].image[pix_up];
                } else if (l1as[i_alt].pixel_mask[pix_up]) {
                    spectra[idx] = l1as[i_alt].image[pix_dn];
                } else {
                    spectra[idx] = weight * l1as[i_alt].image[pix_dn]
                                   + (1.0 - weight) * l1as[i_alt].image[pix_up];
                }
            }
        }
        medians[i_alt] = median(spectra);
    }
    // Find peak ALT angle and the corresponding median
    std::vector<double> alt_measurements(l1as.size());
    for (int i_alt {}; i_alt < l1as.size(); ++i_alt) {
        alt_measurements[i_alt] = l1as[i_alt].meta->alt_angle;
    }
    gadfit::LMsolver solver { [](const std::vector<gadfit::AdVar>& parameters,
                                 const double x) {
        // a * exp(-(x0-x)^2/sigma) + bgr
        const gadfit::AdVar diff { parameters[0] - x };
        return parameters[2] * exp(-diff * diff / parameters[1])
               + parameters[3];
    } };
    solver.addDataset(alt_measurements, medians);
    // The highest median is set as the initial guess for the
    // Levenberg-Marquardt solver.
    const auto max_loc { std::ranges::distance(
      medians.cbegin(), std::ranges::max_element(medians)) };
    solver.setPar(0, alt_measurements[max_loc], true, "ALT angle peak");
    solver.setPar(1, 1e-7, true, "sigma");
    solver.setPar(2, medians[max_loc], true, "signal strength");
    solver.setPar(3, 0.0, true, "background");
    solver.settings.iteration_limit = 100;
    solver.settings.verbosity = gadfit::io::final_only;
    solver.fit();
    alt_angles[i_vp].push_back(solver.getParValue(0));
    peak_medians[i_vp].push_back(solver.getParValue(2));
}

// Write the ACT and ALT angles and the spectra medians to file. Keep
// track of how many angles (measurements) correspond to each
// viewport.
auto static writeCheckpoint(
  const SPEXSettings& settings,
  const CKD& ckd,
  const std::vector<std::vector<double>>& act_angles,
  const std::vector<std::vector<double>>& alt_angles,
  const std::vector<std::vector<double>>& peak_medians) -> void
{
    std::ofstream out { settings.swath.checkpoint.filename, std::ios::binary };
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        const int batch_size { static_cast<int>(act_angles[i_vp].size()) };
        out.write((char*)&batch_size, sizeof(int));
        out.write((char*)act_angles[i_vp].data(), batch_size * sizeof(double));
        out.write((char*)alt_angles[i_vp].data(), batch_size * sizeof(double));
        out.write((char*)peak_medians[i_vp].data(),
                  batch_size * sizeof(double));
    }
}

// Basically like writeCheckpoint but in reverse
auto static readCheckpoint(const SPEXSettings& settings,
                           CKD& ckd,
                           std::vector<std::vector<double>>& act_angles,
                           std::vector<std::vector<double>>& alt_angles,
                           std::vector<std::vector<double>>& peak_medians)
  -> void
{
    try {
        std::ifstream in { settings.swath.checkpoint.filename,
                           std::ios::binary };
        in.exceptions(std::ofstream::failbit);
        for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
            int batch_size {};
            in.read((char*)&batch_size, sizeof(int));
            act_angles[i_vp].resize(batch_size);
            alt_angles[i_vp].resize(batch_size);
            peak_medians[i_vp].resize(batch_size);
            in.read((char*)act_angles[i_vp].data(),
                    batch_size * sizeof(double));
            in.read((char*)alt_angles[i_vp].data(),
                    batch_size * sizeof(double));
            in.read((char*)peak_medians[i_vp].data(),
                    batch_size * sizeof(double));
        }
    } catch (const std::ofstream::failure& e) {
        std::cerr << "error reading from " << settings.stray.checkpoint.filename
                  << '\n';
        throw;
    }
}

auto SwathGen::processFinalize() -> void
{
    spdlog::info("Evaluating ACT and ALT angles on L1B FOVs");
    if (settings.swath.checkpoint.write) {
        writeCheckpoint(settings, ckd, act_angles, alt_angles, peak_medians);
    }
    if (settings.swath.checkpoint.read) {
        readCheckpoint(settings, ckd, act_angles, alt_angles, peak_medians);
    }
    // Allocate swath CKD
    ckd.swath.pointings.resize(ckd.fov.n_act * dims::vec);
    ckd.swath.normals.resize(ckd.n_vp * dims::vec);
    // Instrument rotation quaternion is conjugated. The input is
    // spacecraft to SPEXone matrix. We use a quaternion to rotate
    // SPEXone to the spacecraft.
    Quaternion q_inst { settings.swath.rotmat };
    q_inst.conjugate();
    q_inst.normalize(); // Getting rid of rounding errors

    Theodolite theo {};
    auto viewport_alt_angles { settings.swath.viewport_alt_angles };
    theo.calculate_reference(ckd.n_vp,
                             settings.swath.theodolite_reference_filename,
                             viewport_alt_angles.data(),
                             settings.swath.theodolite_alt_angle_tol,
                             settings.swath.theo_altmodel_maxiter,
                             settings.swath.theo_altmodel_tol,
                             settings.swath.theo_altmodel_steptolerance,
                             settings.swath.theo_altmodel_initial_step_reducer,
                             settings.swath.theo_altmodel_reducerfactor_fail,
                             settings.swath.theo_altmodel_reducerfactor_success,
                             settings.swath.theo_altmodel_reducer_limit);
    theo.calculate_spotcheck(settings.swath.theodolite_spotcheck_filename);
    for (int i_vp {}; i_vp < ckd.n_vp; i_vp++) {
        gadfit::LMsolver solver {
            [order = settings.swath.order](
              const std::vector<gadfit::AdVar>& parameters, const double x) {
                std::vector<double> args(order + 1);
                args[0] = 1.0;
                for (int i { 1 }; i <= order; ++i) {
                    args[i] = args[i - 1] * x;
                }
                gadfit::AdVar result { parameters[0] };
                for (int i {}; i < order; ++i) {
                    result += parameters[i + 1] * args[i + 1];
                }
                return result;
            }
        };
        std::vector<double> weights_inv(peak_medians[i_vp].size());
        for (int i {}; i < static_cast<int>(weights_inv.size()); ++i) {
            weights_inv[i] = 1 / peak_medians[i_vp][i];
        }
        solver.addDataset(act_angles[i_vp], alt_angles[i_vp], weights_inv);
        for (int i_order {}; i_order <= settings.swath.order; ++i_order) {
            solver.setPar(i_order, 0.0, true);
        }
        solver.settings.iteration_limit = 100;
        solver.settings.verbosity = gadfit::io::final_only;
        solver.fit();
        std::vector<double> alt_angles_interp(ckd.fov.n_act_vp[i_vp]);
        for (int i_fov {}; i_fov < ckd.fov.n_act_vp[i_vp]; ++i_fov) {
            alt_angles_interp[i_fov] = solver.getValue(
              ckd.fov.act_angles[ckd.fov.start_vp[i_vp] + i_fov]);
        }
        // Turn these angles into swath vectors
        for (int i_fov {}; i_fov < ckd.fov.n_act_vp[i_vp]; i_fov++) {
            // Instrument reference frame
            Vector res_irf {};
            theo.calculate_swath_vector(
              ckd.fov.act_angles[ckd.fov.start_vp[i_vp] + i_fov],
              alt_angles_interp[i_fov],
              res_irf);
            // Rotate to satellite reference frame
            std::array<double, dims::vec> res { res_irf[0],
                                                res_irf[1],
                                                res_irf[2] };
            q_inst.rotate(res);
            for (int i_dir {}; i_dir < dims::vec; ++i_dir) {
                ckd.swath.pointings[(ckd.fov.start_vp[i_vp] + i_fov) * dims::vec
                                    + i_dir] = res[i_dir];
            }
        }

        // Fit a plane through all the swath vectors.
        // It is a plane that goes through (0,0,0).
        // In the ideal case, the equation is ax + cz = 0 and
        // y does not matter. In reality, y may matter a bit.
        // Let us use the model x + by + cz = 0, fixing the x
        // contribution to one so that the solution becomes unique.
        // This reasoning depends on the fact that the swath vectors
        // are already in satellite coordinates and that x is the
        // flight direction, z is nadir and y is to the right.
        // With just the instrument reference frame, this had not
        // been safe.

        // There are two state parameters and nfov_vp measurements
        std::vector<double> jac(2 * ckd.fov.n_act_vp[i_vp], 0.0);
        std::vector<double> meas(ckd.fov.n_act_vp[i_vp], 0.0);
        std::vector<double> res_bc(2, 0.0);
        // The linear model will be: by + cz = -x with state vector
        // [b,c] and known parameters x,y,z.
        for (int i_fov {}; i_fov < ckd.fov.n_act_vp[i_vp]; i_fov++) {
            // Swath vector as double pointer
            double* xyz =
              &ckd.swath
                 .pointings[(ckd.fov.start_vp[i_vp] + i_fov) * dims::vec];
            // d/db = y
            jac[i_fov] = xyz[1];
            // d/dc = z
            jac[ckd.fov.n_act_vp[i_vp] + i_fov] = xyz[2];
            // Measurement is -x
            meas[i_fov] = -xyz[0];
        }
        linear_invert(2,
                      ckd.fov.n_act_vp[i_vp],
                      jac.data(),
                      OPT_NONE,
                      NULL,
                      NULL,
                      meas.data(),
                      res_bc.data());

        // Normalize, using the Vector class
        Vector normal_cur(1.0, res_bc[0], res_bc[1]);
        normal_cur.normalize();
        normal_cur.get(&ckd.swath.normals[i_vp * dims::vec]);
    }
}

} // namespace spex
