// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for generating the swath vector CKD

#pragma once

#include "linear_spline_multi.h"
#include "processor.h"
#include "quaternion.h"
#include "theodolite.h"

namespace spex {

class SwathGen : public Processor
{
public:
    SwathGen(const SPEXSettings& settings, CKD& ckd)
      : Processor { settings, ProcLevel::swath, ckd }
    {}
    ~SwathGen() = default;

protected:
    // Construct splines for FOV row indices so that we can extract
    // spectra at arbitrary ACT angles from detector images.
    auto processInit() -> void override;
    // Sort L1A products by viewports, ACT angles, and ALT angles and
    // make batches per viewport per ACT angle.
    auto prepareBatches() -> void;
    auto processBatch(const int i_batch) -> void override;
    auto processFinalize() -> void override;

private:
    // ACT angles and peak ALT angles grouped by viewports
    std::vector<std::vector<double>> act_angles {};
    std::vector<std::vector<double>> alt_angles {};
    // Spectra medians corresponding to ALT angles that yield maximum signals
    std::vector<std::vector<double>> peak_medians {};
    // Spline representation of the FOV CKD. This is used for
    // extracting spectra at arbitrary ACT angles. The dimensions are
    // (polarization states, viewports). When evaluated at a column
    // index and ACT angle, returns the corresponding fractional row
    // index.
    std::array<std::vector<LinearSplineMulti>, dims::pol> fov_splines_m {};
};

} // namespace spex
