// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "inversion.h"
#include "quaternion.h"

#include <memory>

namespace spex {

// Some extra constants for logging
const std::vector<std::string> targetnames = {
    "-X",
    "+X",
    "-Y",
    "+Y",
    "-Z",
    "+Z"
};
const std::vector<std::string> hvnames = {"H","V"};
// Static dimension sizes. It is nice to see if a '2' is
// a number of targets or a number of H/V theodolite angles.
const size_t ntar = 2;
const size_t nhv = 2;

class Theodolite
{
public:
    Theodolite() = default;
    ~Theodolite() = default;

    // Do what you can do with the reference file
    auto calculate_reference(size_t dim_vp,
                             std::string reference_filename,
                             double* viewport_alt_angles,
                             double alt_angle_tol,
                             size_t altinv_maxiter,
                             double altinv_tol,
                             double altinv_steptolerance,
                             double altinv_initial_step_reducer,
                             double altinv_reducerfactor_fail,
                             double altinv_reducerfactor_success,
                             double altinv_reducer_limit) -> void;

    // Add information from spot check file
    auto calculate_spotcheck(std::string spotcheck_filename) -> void;

    // Get swath vector from stage angle using the theodolite measurements
    auto calculate_swath_vector(
      double act, double alt, Vector& swathhvector) -> void;

    // Getters for private data members
    size_t getDimVp();

    // Intermediate results are stored into member variables
private:
    size_t dim_vp {}; // Number of viewports
    std::vector<int> tar {}; // First and second target. For now, this must be consistent all over the files
    // Results from reference calculation
    std::vector<double> actual_alt_angles {}; // ALT-angles per viewport, to recognize viewport
    std::vector<double> diff_theo_angs_act {}; // Derivatives of theodolite angles with respect to ACT stage angle, for each target and viewport
    std::unique_ptr<Quaternion0> altmodel_base {}; // Base quaternion for the ALT angle dependence model
    std::unique_ptr<Vector> altmodel_rotax {}; // Rotation axis for ALT angle dependence model
    // Results from spotcheck calculation
    std::vector<double> offset {};
    std::unique_ptr<Vector> star_stimulus {};

    // Read a theodolite file
    auto read(std::string filename,
              size_t& nmeas,
              std::vector<double> &act,
              std::vector<double> &alt,
              std::vector<int> &tar,
              std::vector<double> &ang,
              std::vector<double> &stst) -> void;

    // Turn theodolite angles into orientation quaternion
    auto theo_to_quat(double* theo_angs, // Input theodolite angles
                      Quaternion0* quat) -> void; // Output orientation quaternion

    // Do the exact opposite: Turn orientation quaternion into theodolite angles
    auto quat_to_theo(Quaternion0* quat, // Input orientation quaternion
                      double* theo_angs) -> void; // Output theodolite angles

    // Get a unit vector from the matrix that belongs to a quaternion
    auto quat_to_unitvec(Quaternion0 quat, int ax, Vector& vec) -> void;

    // Acquire thedolite angles that correspond to a certain stage position
    auto stage_to_theo(double act, // Across-track stage angle
                       double alt, // Along-track stage angle
                       double* theo_angs) -> void; // Output theodolite angles
};

// The non-linear inversion for the ALT-stage model
class Altmodel_Inversion : public Inversion
{
public:
    Altmodel_Inversion(Theodolite* creator, // Creating theodolite instance
                       size_t a_ivec_fixed, // Dimension of rotation axis to set to one (before normalization)
                       bool neg, // True if it is minus one instead of plus one
                       std::vector<double> a_alt_angles); // ALT stage angles at measurement points
    ~Altmodel_Inversion();

    // Overwrite forward function
    auto fwd(double* state, // State vector
             double* signal, // Modelled signal (outout)
             double* jacobian) -> void override; // Modelled jacobian (outout)

    // Translate state function into information for the theodolite
    auto translate(double* state, // State vector
                   Quaternion0* base, // Base (ALT=0) orientation quaternion
                   Vector* rotax) -> void; // Rotation axis

private:
    // Mapping parameters for fourth and fifth state parameters
    size_t ivec_fixed {};
    size_t ivec_rotax_first {};
    size_t ivec_rotax_second {};
    double val_fixed {};
    // Abscissa
    std::vector<double> alt_angles {};
    // Copy of number of viewports
    size_t dim_vp {};
};

} // namespace spex
