// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "wave_gen.h"

#include "functions.h"
#include "l1a_metadata.h"
#include "lininv.h"

namespace spex {

// 1. Put all the fitted peaks in a map
// 2. Calibrate wavelengths in extracted spectrum perspective
// 3. Define target wavelength grid
auto WaveGen::processInit() -> void
{
    for (auto& wavelengths : ckd.wave_full) {
        wavelengths.resize(ckd.fov.n_act,
                           std::vector<double>(ckd.n_detector_cols));
    }
    ckd.wave_common.resize(ckd.fov.n_act,
                           std::vector<double>(ckd.n_detector_cols));

    // Peak fitter for spectra
    spl = std::make_unique<Splitgaussfit>(settings.wave.npeak,settings.wave.peak_backgroundorder,ckd.n_detector_cols);
    spl->setGaussfit(settings.wave.peak_gaussfit); // Flag for performing gauss fit
    spl->setMaxiter(settings.wave.peak_maxiter); // Number of iterations for fitting Gauss peak shape
    spl->setMagtol(settings.wave.peak_magtol); // Tolerance of peak magnitude (in units of a-priori magnitude)
    spl->setPostol(settings.wave.peak_postol); // Tolerance of peak positions (in indices)
    spl->setWidthtol(settings.wave.peak_widthtol); // Tolerance of peak FWHM (in indices)
    spl->setBackgroundtol(settings.wave.peak_backgroundtol); // Tolerance of background B-spline coefficients (in median signals)
    spl->setSteptolerance(settings.wave.peak_steptolerance); // Inversion step acceptance criterion for fitting Gauss peak shape
    spl->setInitialStepReducer(settings.wave.peak_initial_step_reducer); // Starting value of the Levenberg-Marquardt step control parameter
    spl->setReducerfactorFail(settings.wave.peak_reducerfactor_fail); // Multiplication of step reducer for a failed step
    spl->setReducerfactorSuccess(settings.wave.peak_reducerfactor_success); // Multiplication of step reducer for a successful step
    spl->setReducerLimit(settings.wave.peak_reducer_limit); // Maximum step control reducer for convergence

    // Function fit of wavelenght as function of spectral index
    std::vector<double> knots = {0.0,(double)ckd.n_detector_cols-1};
    b = std::make_unique<Bspline>(settings.wave.order,2,knots.data());
    evalmat.resize(ckd.n_detector_cols*b->nspline);
    std::vector<double> evalispec(ckd.n_detector_cols);
    for (size_t ispec=0 ; ispec<ckd.n_detector_cols ; ispec++) evalispec[ispec] = (double)ispec;
    b->jaccalc(ckd.n_detector_cols,evalispec.data(),evalmat.data());
}

auto WaveGen::prepareBatches() -> void
{
    batchViewport();
}

auto WaveGen::processBatch(const int i_vp) -> void
{
    // 1. Put all the fitted peaks in a map
    spdlog::info("Generating spectral CKD for viewport: {} deg",
                 viewport_idx_to_angle.at(
                   batches.at(i_vp).l1a_products.front().meta->viewport));
    size_t npeak = settings.wave.npeak;
    std::vector<double> map_wavelength(ckd.fov.n_act_vp[i_vp]*dims::pol*npeak);
    std::vector<double> map_ispec(ckd.fov.n_act_vp[i_vp]*dims::pol*npeak);

    // Running pointers
    double* ispec_cur = map_ispec.data(); // These are the positions that are fit
    double* wave_cur = map_wavelength.data();

    auto& l1a { batches.at(i_vp).l1a_products.front() };
    // Recognize peaks from reference spectrum.
    // In principle, any L1A file can have a differently-shaped reference spectrum.
    // So, we do not recycle any of the properties of this fitter.
    Splitgaussfit refspl(npeak,settings.wave.refspec_backgroundorder,l1a.meta->dim_refspec);
    refspl.setX(l1a.meta->refspec_wavelength.data());
    refspl.setGaussfit(settings.wave.refspec_gaussfit); // Flag for performing gauss fit
    refspl.setMaxiter(settings.wave.refspec_maxiter); // Number of iterations for fitting Gauss peak shape
    refspl.setMagtol(settings.wave.refspec_magtol); // Tolerance of peak magnitude (in units of a-priori magnitude)
    refspl.setPostol(settings.wave.refspec_postol); // Tolerance of peak positions (in indices)
    refspl.setWidthtol(settings.wave.refspec_widthtol); // Tolerance of peak FWHM (in indices)
    refspl.setBackgroundtol(settings.wave.refspec_backgroundtol); // Tolerance of background B-spline coefficients (in median signals)
    refspl.setSteptolerance(settings.wave.refspec_steptolerance); // Inversion step acceptance criterion for fitting Gauss peak shape
    refspl.setInitialStepReducer(settings.wave.refspec_initial_step_reducer); // Starting value of the Levenberg-Marquardt step control parameter
    refspl.setReducerfactorFail(settings.wave.refspec_reducerfactor_fail); // Multiplication of step reducer for a failed step
    refspl.setReducerfactorSuccess(settings.wave.refspec_reducerfactor_success); // Multiplication of step reducer for a successful step
    refspl.setReducerLimit(settings.wave.refspec_reducer_limit); // Maximum step control reducer for convergence
    refspl.setMeas(l1a.meta->refspec_radiance.data());
    // Reference spectrum has no noise or pixel mask
    std::vector<double> res(3*npeak, 0.0);
    refspl.solve(settings.wave.refspec_insignificant_island,
                 settings.wave.refspec_domain_range,
                 settings.wave.refspec_maxreso,
                 res.data());
    for (int i {}; i < npeak; ++i) {
        res[i * 3 + 1] += settings.wave.peak_offset;
    }
    if (!settings.wave.reference_wavelengths.empty()) {
        for (int i {}; i < npeak; ++i) {
            res[i * 3 + 1] = settings.wave.reference_wavelengths[i];
        }
    }
    for (int i {}; i < npeak; ++i) {
        spdlog::debug("  Registered peak at {} nm", res[i * 3 + 1]);
    }
    std::vector<double> peaks(npeak, 0.0);
    for (size_t ipeak=0 ; ipeak<npeak ; ipeak++) {
        peaks[ipeak] = res[3*ipeak+1];
    }
    for (size_t ifov=0 ; ifov<ckd.fov.n_act_vp[i_vp] ; ifov++) {
        Spectra specs {};
        l1a.extractSpectra(ckd, ckd.fov.start_vp[i_vp] + ifov, specs);

        for (size_t ipol=0 ; ipol<dims::pol ; ipol++) {
            // Extract mask, because vector<bool> does not convert
            // properly to a bool pointer.
            std::vector<bool> mask_extracted { specs.mask[ipol] };
            fill_holes(ckd.n_detector_cols,mask_extracted,specs.signal[ipol].data());
            spl->setMeas(specs.signal[ipol].data());
            spl->setNoise(specs.noise[ipol].data());
            spl->setMask(&mask_extracted);
            std::vector<double> res(3*npeak);
            spl->solve(settings.wave.peak_insignificant_island,settings.wave.peak_domain_range,settings.wave.peak_maxreso,res.data());
            for (size_t ipeak=0 ; ipeak<npeak ; ipeak++) {
                ispec_cur[ipeak] = res[3*ipeak+1];
                wave_cur[ipeak] = peaks[ipeak];
            }
            // Move the pointers
            ispec_cur += npeak;
            wave_cur += npeak;
        }
    }

    // 2. Calibrate wavelengths in extracted spectrum perspective. {{{
    spdlog::info("Fit wavelengths per L1B field of view");
    // For this, we fit a polynomial in ispec through all the fitted (ispec,wave)
    // points.
    // Reset moving pointers
    ispec_cur = map_ispec.data();
    wave_cur = map_wavelength.data();
    // CKD pointer
    for (int ifov { ckd.fov.start_vp[i_vp] };
         ifov < ckd.fov.n_act_vp[i_vp] + ckd.fov.start_vp[i_vp];
         ++ifov) {
        for (size_t ipol=0 ; ipol<dims::pol ; ipol++) {
            std::vector<double> mat(npeak*b->nspline, 0.0);
            b->jaccalc(npeak,ispec_cur,mat.data());
            std::vector<double> fit(b->nspline, 0.0);
            linear_invert(b->nspline,npeak,mat.data(),OPT_NONE,NULL,NULL,wave_cur,fit.data());
            // Manual matrix multiplication. The evaluation matrix is always the same,
            // so it is saved from viewport to viewport.
            for (int ispec {}; ispec < ckd.n_detector_cols; ++ispec) {
                ckd.wave_full[ipol][ifov][ispec] = 0.0;
                for (int ispline {}; ispline < b->nspline; ++ispline) {
                    ckd.wave_full[ipol][ifov][ispec] +=
                      fit[ispline]*evalmat[ispline*ckd.n_detector_cols+ispec];
                }
            }
            ispec_cur += npeak;
            wave_cur += npeak;
        }
    }

    // 3. Define target wavelengths for co-sampling S+ and S-
    for (int ifov { ckd.fov.start_vp[i_vp] };
         ifov < ckd.fov.start_vp[i_vp] + ckd.fov.n_act_vp[i_vp];
         ++ifov) {
        for (int i {}; i < ckd.n_detector_cols; ++i) {
            ckd.wave_common[ifov][i] =
              0.5 * (ckd.wave_full[pol_plus][ifov][i]
                     + ckd.wave_full[pol_minus][ifov][i]);
        }
    }
}

} // namespace spex
