// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "yaml.h"

namespace YAML {

auto convert<spex::L1AType>::decode(const Node& node, spex::L1AType& rhs)
  -> bool
{
    const std::string l1a_type_str { node.as<std::string>() };
    if (l1a_type_str == "science") {
        rhs = spex::L1AType::science;
    } else if (l1a_type_str == "diagnostic") {
        rhs = spex::L1AType::diagnostic;
    } else {
        throw std::invalid_argument { "unknown L1A type: " + l1a_type_str };
    }
    return true;
}

auto convert<spex::NonlinModel>::decode(const Node& node,
                                        spex::NonlinModel& rhs) -> bool
{
    const std::string model_str { node.as<std::string>() };
    if (model_str == "bspline") {
        rhs = spex::NonlinModel::bspline;
    } else if (model_str == "sigmoidal") {
        rhs = spex::NonlinModel::sigmoidal;
    } else if (model_str == "single") {
        rhs = spex::NonlinModel::single;
    } else {
        throw std::invalid_argument { "unknown nonlinearity model: "
                                      + model_str };
    }
    return true;
}

auto convert<spex::EGSEMember>::decode(const Node& node, spex::EGSEMember& rhs)
  -> bool
{
    const std::string egse_member_str { node.as<std::string>() };
    if (egse_member_str == "POLARIZER") {
        rhs = spex::EGSEMember::POLARIZER;
    } else if (egse_member_str == "ACT_ANGLE") {
        rhs = spex::EGSEMember::ACT_ANGLE;
    } else {
        throw std::invalid_argument { "unknown EGSE member: "
                                      + egse_member_str };
    }
    return true;
}

auto convert<spex::ProcLevel>::decode(const Node& node, spex::ProcLevel& rhs)
  -> bool
{
    try {
        rhs = spex::proc_level_str_to_enum.at(node.as<std::string>());
    } catch (const std::out_of_range&) {
        throw std::invalid_argument { "unknown process: "
                                      + node.as<std::string>() };
    }
    return true;
}

} // namespace YAML

namespace spex {

auto operator<<(YAML::Emitter& out, const ProcLevel proc_level)
  -> YAML::Emitter&
{
    const auto it { std::ranges::find_if(
      proc_level_str_to_enum,
      [proc_level](const auto& item) { return item.second == proc_level; }) };
    out << it->first;
    return out;
}

auto operator<<(YAML::Emitter& out, const L1AType l1a_type) -> YAML::Emitter&
{
    switch (l1a_type) {
    case L1AType::science:
        out << "science";
        break;
    case L1AType::diagnostic:
        out << "diagnostic";
        break;
    default:
        out << "unknown";
    }
    return out;
}

auto operator<<(YAML::Emitter& out, const NonlinModel model) -> YAML::Emitter&
{
    switch (model) {
    case NonlinModel::bspline:
        out << "bspline";
        break;
    case NonlinModel::sigmoidal:
        out << "sigmoidal";
        break;
    case NonlinModel::single:
        out << "single";
        break;
    default:
        out << "unknown";
    }
    return out;
}

auto operator<<(YAML::Emitter& out, const EGSEMember egse_member)
  -> YAML::Emitter&
{
    switch (egse_member) {
    case EGSEMember::POLARIZER:
        out << "POLARIZER";
        break;
    case EGSEMember::ACT_ANGLE:
        out << "ACT_ANGLE";
        break;
    default:
        out << "unknown";
    }
    return out;
}

auto operator<<(YAML::Emitter& out, const SettingLevel setting_level)
  -> YAML::Emitter&
{
    switch (setting_level) {
    case SettingLevel::normal:
        out << "normal";
        break;
    case SettingLevel::debug:
        out << "debug";
        break;
    default:
        out << "unknown";
    }
    return out;
}

} // namespace spex
