// Test B-splines

#include "bspline_data.h"

#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_floating_point.hpp>
#include <spexone_cal/bspline.h>
#include <spexone_cal/linalg.h>

using Catch::Matchers::WithinRel;

TEST_CASE("Least squares")
{
    // Define number of initial knots. The knot vector is padded
    // later.
    constexpr int n_knots { 15 };
    // The knot range must be less than the data range but not by too
    // much.
    constexpr double d_beg { 454.0 };
    constexpr double d_end { 465.0 };
    // B-spline order
    constexpr int order { 3 };
    std::vector<double> knots(n_knots, 0.0);
    for (int i {}; i < n_knots; ++i) {
        knots[i] = d_beg + i * (d_end - d_beg) / (n_knots - 1);
    }
    spex::BSpline b_spline { order, knots };
    const int n_points { static_cast<int>(x_data.size()) };
    const int n_states { b_spline.nStates() };
    // Construct the B-spline matrix
    std::vector<double> B {};
    b_spline.evalBasis(x_data, B);
    // Each row is scaled by noise
    for (int i {}; i < n_points; ++i) {
        for (int j {}; j < n_states; ++j) {
            B[i * n_states + j] /= std::sqrt(noise[i]);
        }
    }
    std::vector<double> BTB(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int j {}; j < n_states; ++j) {
            for (int k {}; k < n_points; ++k) {
                BTB[i * n_states + j] +=
                  B[k * n_states + i] * B[k * n_states + j];
            }
        }
    }
    std::vector<double> BTY(n_points, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_points; ++k) {
            BTY[i] += B[k * n_states + i] * y_data[k] / std::sqrt(noise[k]);
        }
    }
    // Upper triangle of B^T*B
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j+1)*j/2] = BTB[i * n_states + j];
        }
    }
    // Covariance matrix of linear inversion initialized to unity
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    const char uplo { 'u' };
    int info {};
    dppsv_(&uplo,
           &n_states,
           &n_states,
           BTB_tri.data(),
           Sx.data(),
           &n_states,
           &info);
    // The resulting state vector
    std::vector<double> x_result(n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            x_result[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
    constexpr std::array ref {
        97.3564075280751,
        99.3872290797305,
        98.1036228704817,
        107.3865396455,
        110.358715544159,
        117.060445583357,
        118.3129794002,
        117.893750905433,
        120.711696780604,
        122.159064328912,
        122.272854054485,
        120.610347238412,
        118.534057858669,
        112.036272414326,
        107.469363015539,
        103.322570875858,
        99.5583898586142
    };
    for (int i {}; i < n_states; ++i) {
        CHECK_THAT(x_result.at(i), WithinRel(ref.at(i), 1e-14));
    }
}
